# Portal::Accounting::DB::Ledger.pm
# Created by JT Moree, Xperience, Inc.
# Thu Jan  9 13:45:24 UTC 2003

#$Id: Ledger.pm,v 1.42 2004/10/13 21:49:56 moreejt Exp $
#This code is licensed under the GPL

=head1 NAME

Ledger - Object used to interact with Database dealing with Ledger functions.

=head1 SYNOPSIS

  use Portal::Accounting::DB::Ledger;
  my $obj = Portal::Accounting::DB::Ledger->new(dbHandle => , langObj => );
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }

=head1 DESCRIPTION

Ledger contains methods for interchanging ledger data with the database.

=cut
package Portal::Accounting::DB::Ledger;
use strict;
use Portal::Accounting::Objects::Ledger;
use Portal::Accounting::Objects::SubEntities;
use Portal::Accounting::Objects::SubEntityObject;
use Portal::Accounting::DB::Main;
use Date::Calc qw (Today Today_and_Now);
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::DB::Main Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '.5';

=head1 Exported FUNCTIONS

=item scalar new()

    Creates a new instance of the Portal::Accounting::DB::Ledger
    object.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.
  my $sth = $self->{dbHandle}->write(sql => "SET datestyle TO 'ISO'");

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (not $self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }
  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item bool updateLedger(ledgerCode, color, businessType, cashOrAccrual)

    Updates a database ledger entry when given valid data.
    Only certain items can be changed on a ledger:
      color, businessType, cashOrAccrual

=cut
sub updateLedger
{
  my $self = shift;
  my %args = ( ledgerCode => "", color => "", businessType => "", cashOrAccrual => "", @_ ) ;
  my $color = "";
  my $businessType = "";
  my $cashOrAccrual = "";
  my $sql = "update ledger_tb set ";

  if (defined $args{ledgerObject})
  {
    if (! $args{ledgerObject}->isValid)
    {
      $self->prefixError("LedgerObject is invalid.<br>\n".$args{ledgerObject}->errorMessage);
      return 0;
    }
    $color = $args{ledgerObject}->{color};
    $cashOrAccrual = $args{ledgerObject}->{cashOrAccrual};
    $businessType = $args{ledgerObject}->{businessType};
  }
  elsif ($args{ledgerCode} ne "")
  {
    $color = $args{color};
    $cashOrAccrual = $args{cashOrAccrual};
    $businessType = $args{businessType};
  }
  else
  {
    $self->prefixError("LedgerCode or LedgerObject is required.<br>\n");
    return 0;
  }
  my $ledgerOrig = ($self->getLedgers(ledgerCode => $args{ledgerCode}))[0];
  if ($self->didErrorOccur)
  {
    $self->prefixError("Error while retrieving original LedgerObject for $args{ledgerCode}.<br>\n");
    return 0;
  }
  if ($color eq $ledgerOrig->{color})
  {  $color = ""; }
  if ($cashOrAccrual eq $ledgerOrig->{cashOrAccrual})
  {  $cashOrAccrual = ""; }
  if ($businessType eq $ledgerOrig->{businessType})
  {  $businessType = ""; }
  #must have 1 of three options
  my $num = 0;
  if ($args{color} ne "")
  {
    $num++;
    $sql .= " color = '$args{color}' ";
  }
  if ($num > 0)
  { $sql .= ", "; }
  if ($args{cashOrAccrual} ne "")
  {
    $num++;
    if ($args{cashOrAccrual} !~ /^(cash|accrual)$/i)
    {
      $self->prefixError("Invalid CashOrAccrual: '$args{cashOrAccrual}'.<br>\n");
      return 0;
    }
    my $cashOrAccrual = "";
    if ($args{cashOrAccrual} =~ /^cash$/i)
    {  $cashOrAccrual = "true";}
    elsif ($args{cashOrAccrual} =~ /^accrual$/i)
    { $cashOrAccrual = "false";}

    $sql .= " cash_or_accrual = '$cashOrAccrual'";
  }
  if ($num > 0)
  { $sql .= ", "; }
  if ($args{businessType} ne "")
  {
    $num++;
    $sql .= " business_type = '$args{businessType}' ";
  }
  if ($num < 1)
  {
    $self->prefixError("at least one field is required (color, cashOrAccrual, businessType).<br>\n");
    return 0;
  }
  $sql .= "WHERE code = '$args{ledgerCode}'";
  $self->{dbHandle}->write(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return 0;
  }

  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return 0;
  }

  return 1;
}

=item bool createLedger(Portal::Accounting::Objects::Ledger)

    Creates a database ledger entry when given a valid Ledger Object.

=cut
sub createLedger
{
  my $self = shift;
  my $l = shift;
  my $cashOrAccrual = "";
  my $ledgerName = "";

  if (not defined $l)
  {
    $self->prefixError("Error! Ledger not defined.  Cannot create!<br>\n");
    return 0;
  }
  if ($l->isValid == 0)
  {
    $self->prefixError("Error! Ledger is invalid.  Cannot create!<br>\n");
    return 0;
  }
  if ($l->basis eq "cash")
  {  $cashOrAccrual = "true"; }
  else
  {  $cashOrAccrual = "false"; }
  #technically this shoudl be done for each known year found in $l but there should only be one for creation
  my $fiscalYear = (sort keys %{$l->{knownYears}})[0];
  my ($fyYear, $fyMo, $fyDay) = split ("-",$l->end(month => 12, fiscalYear => $fiscalYear));
  my $insert = "insert into ledger_tb (name, code, color, cash_or_accrual, ";
  $insert .= "  fy_mo, fy_day, business_type," ;
  $insert .= "  start1, start2, start3, start4, start5, start6, start7, start8, start9, start10, start11, start12,";
  $insert .= "  offset1, offset2, offset3, offset4, offset5, offset6, offset7, offset8, offset9, offset10, offset11, offset12)";
  $insert .= " values ('".$l->name."', '".$l->code."', '".$l->color."', $cashOrAccrual, ";
  $insert .= "  '$fyMo', '$fyDay', ".$l->businessType.", ";
  $insert .= " '".$l->start(1)."',";
  $insert .= " '".$l->start(2)."',";
  $insert .= " '".$l->start(3)."',";
  $insert .= " '".$l->start(4)."',";
  $insert .= " '".$l->start(5)."',";
  $insert .= " '".$l->start(6)."',";
  $insert .= " '".$l->start(7)."',";
  $insert .= " '".$l->start(8)."',";
  $insert .= " '".$l->start(9)."',";
  $insert .= " '".$l->start(10)."',";
  $insert .= " '".$l->start(11)."',";
  $insert .= " '".$l->start(12)."',";
  $insert .= " '".$l->offset(1)."',";
  $insert .= " '".$l->offset(2)."',";
  $insert .= " '".$l->offset(3)."',";
  $insert .= " '".$l->offset(4)."',";
  $insert .= " '".$l->offset(5)."',";
  $insert .= " '".$l->offset(6)."',";
  $insert .= " '".$l->offset(7)."',";
  $insert .= " '".$l->offset(8)."',";
  $insert .= " '".$l->offset(9)."',";
  $insert .= " '".$l->offset(10)."',";
  $insert .= " '".$l->offset(11)."',";
  $insert .= " '".$l->offset(12)."')";
  $self->{dbHandle}->write(sql => $insert);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError("Could not insert '".$l->name."'<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  $insert = "insert into known_years_tb ";
  $insert .= "(ledger_code, closed, year) ";
  $insert .= "values ('".$l->code."', 'false', '$fiscalYear')";
  $self->{dbHandle}->write(sql => $insert);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return 0;
  }
  #############################
  my $entity = "insert into entity_tb (name, database_name) values ('".$l->name."', '".$l->name."')";
  $self->{dbHandle}->write(sql => $entity);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return 0;
  }
  ########################

  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return 0;
  }

  return 1;
}

=item int deleteLedger(ledgerCode)

 Attempts to delete the specified ledger.
 Returns 1 on success, 0 on errors and sets the errorMessage,
 -1 if the ledger doesn't exist anymore.

=cut

sub deleteLedger
{
  my $self = shift;
  my %args = ( ledgerCode => "", @_ );
  my $ledgerCode = $args{ledgerCode};

  # first verify the ledger still exists to delete.
  my @ledger = $self->getLedgers(ledgerCode => $ledgerCode);
  if (@ledger)
  {
    my $ledger = $ledger[0];
    # first we need to remove the ledger name from the entity_tb table.
    $self->{dbHandle}->write(sql => "DELETE FROM entity_tb WHERE name = ?", plug => [ $ledger->name ]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # now we need to remove all entries from the time_tracker_config_tb and tips_tb tables that reference this ledger.
    $self->{dbHandle}->write(sql => "DELETE FROM time_tracker_config_tb WHERE ledger_code = ?", plug => [ $ledgerCode ]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    $self->{dbHandle}->write(sql => "DELETE FROM tips_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # ledger_tb
    $self->{dbHandle}->write(sql => "DELETE FROM ledger_tb WHERE code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # known_years_tb
    $self->{dbHandle}->write(sql => "DELETE FROM known_years_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # sub_entity_tb
    $self->{dbHandle}->write(sql => "DELETE FROM sub_entity_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # journal_entry_tb
    $self->{dbHandle}->write(sql => "DELETE FROM journal_entry_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # debitcredit_tb
    $self->{dbHandle}->write(sql => "DELETE FROM debitcredit_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # invoice_tb
    $self->{dbHandle}->write(sql => "DELETE FROM invoice_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # invoice_items_tb
    $self->{dbHandle}->write(sql => "DELETE FROM invoice_items_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # billing_tb
    $self->{dbHandle}->write(sql => "DELETE FROM billing_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # group_tb
    $self->{dbHandle}->write(sql => "DELETE FROM group_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # client_group_tb
    $self->{dbHandle}->write(sql => "DELETE FROM client_group_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # asset_tb
    $self->{dbHandle}->write(sql => "DELETE FROM asset_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # reminder_tb
    $self->{dbHandle}->write(sql => "DELETE FROM reminder_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # custom_report_tb
    $self->{dbHandle}->write(sql => "DELETE FROM custom_report_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # acct_bal_tb
    $self->{dbHandle}->write(sql => "DELETE FROM acct_bal_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # beginning_balance_entry_tb
    $self->{dbHandle}->write(sql => "DELETE FROM beginning_balance_entry_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # local_accounts_tb
    $self->{dbHandle}->write(sql => "DELETE FROM local_accounts_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # sec_inv_tb
    $self->{dbHandle}->write(sql => "DELETE FROM sec_inv_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # sec_event_tb
    $self->{dbHandle}->write(sql => "DELETE FROM sec_event_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # dep_sched_tb
    $self->{dbHandle}->write(sql => "DELETE FROM dep_sched_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # report_index_tb
    $self->{dbHandle}->write(sql => "DELETE FROM report_index_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # report_tb
    $self->{dbHandle}->write(sql => "DELETE FROM report_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # memorized_report_tb
    $self->{dbHandle}->write(sql => "DELETE FROM memorized_report_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # tax_loss_tb
    $self->{dbHandle}->write(sql => "DELETE FROM tax_loss_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # scratch_tb
    $self->{dbHandle}->write(sql => "DELETE FROM scratch_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # lot_node_tb
    $self->{dbHandle}->write(sql => "DELETE FROM lot_node_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }

    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{dbHandle}->errorMessage);
      return 0;
    }
  }
  else
  {
    return -1;
  }

  return 1;
}

=item Accounting::Objects::Ledger closeoutLedger(ledger => , date => , accounts =>, commit =>  )

 Closes out ledger on the given date using the supplied equity account(s) to balance the transactions.
 Returns a ledgerObject with updated information on success,
 undef on errors and sets the errorMessage.
 commit is set to 1 by default.  This will commit all work done to
 the database at the end of the function by default.  To override, set
 commit to 0.  You will then need to commit the transaction(s) to the database.

=cut

sub closeoutLedger
{
  my $self = shift;
  my %args = ( ledger => undef, date => "", accounts => undef, commit => 1, @_ );
  my $ledger = $args{ledger};
  if (not defined $ledger->code || $ledger->code eq "")
  {
    $self->prefixError($self->{langObj}->map("error") . "ledger is not defined");
    return undef;
  }
  my $commit = $args{commit};
  my $closeoutDate = $args{date};
  my %eqAccounts;
  my @eqAccounts;
  if (ref $args{accounts} eq "ARRAY")
  {  @eqAccounts = @{$args{accounts}}; }
  else
  {  $eqAccounts[0] = $args{accounts};  }
  my $fromDate;
  my $sth;

  my $currentFiscalYear = $self->getCurrentFiscalYear($ledger->code);
  $fromDate = $ledger->start(month => 1, fiscalYear => $currentFiscalYear);
  my $yearEnd = $ledger->end(month => 12, fiscalYear => $currentFiscalYear);
  if ($ledger->didErrorOccur)
  {
    $self->prefixError($ledger->errorMessage);
    return undef;
  }

  # verify that the closeoutDate has not already been used or that the ledger has not been closed out after closeoutDate
  my $foo;
  my $bar;
  ($foo = $closeoutDate) =~ s/-//g;
  ($bar = $fromDate) =~ s/-//g;
  if ($foo < $bar)
  {
    $self->prefixError(sprintf $self->{langObj}->map("closeoutDateInvalid"), $closeoutDate, $fromDate);
    return undef;
  }

  ($foo = $closeoutDate) =~ s/-//g;
  ($bar = $self->getLastClose($ledger->code)) =~ s/-//g;
  if ($foo <= $bar)
  {
    $self->prefixError(sprintf $self->{langObj}->map("closeoutDateInvalid"), $closeoutDate, $bar);
    return undef;
  }

  # verify that debits == credits for all transactions in the system
  my $systemBalances = $self->doesSystemBalance(ledgerCode => $ledger->code, beginDate => undef, endDate => $closeoutDate);
  if ($self->didErrorOccur || $systemBalances == 0)
  {
    return undef;
  }
  if ($systemBalances == -1)
  {
    $self->prefixError(errorString => $self->{langObj}->map("debitsGreaterThanCredits"));
    return undef;
  }
  elsif ($systemBalances == -2)
  {
    $self->prefixError(errorString => $self->{langObj}->map("creditsGreaterThanDebits"));
    return undef;
  }
  my $retainedEarnings = $self->updateBalanceTable($ledger->code, $closeoutDate);
  if ($self->didErrorOccur)
  {  return undef;  }

  # adjust the user-entered equity account(s) by the Retained Earnings amount
  my $runningTotal = 0;
  my $divided = sprintf "%.2f", $retainedEarnings / @eqAccounts ;
  for my $x (1 .. @eqAccounts)
  {
    if ($x == @eqAccounts)
    {
      $eqAccounts{$eqAccounts[$x - 1]} = sprintf "%.2f", $retainedEarnings - $runningTotal;
    }
    else
    {
      $eqAccounts{$eqAccounts[$x]} = $divided;
      $runningTotal .= $divided;
    }
  }
  my $sql;
  for my $acct (keys %eqAccounts)
  {
    $sql = "UPDATE acct_bal_tb SET balance = balance + $eqAccounts{$acct}";
    $sql .=   " WHERE ledger_code = '".$ledger->code."' AND bal_date = '$closeoutDate' ";
    $sql .=   " AND acct_num = '$acct' AND (sub_acct_num = '-2' OR sub_acct_num = '-1'); ";
    $sth = $self->{dbHandle}->write($sql);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->postfixError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForPartnerEquityAccount"),) . "<br>\n" . $self->{dbHandle}->errorMessage."<br>$sql ACCT: $acct ".keys %eqAccounts);
      return undef;
    }
    if ($sth == 0)
    {
      $sql = "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, ledger_code, bal_date, balance)";
      $sql .= " VALUES ('$acct', '-2', '".$ledger->code."', '$closeoutDate', '$eqAccounts{$acct}'); ";
      $sql .= "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, ledger_code, bal_date, balance)";
      $sql .= " VALUES ('$acct', '-1', '".$ledger->code."', '$closeoutDate', '$eqAccounts{$acct}'); ";
      $sth = $self->{dbHandle}->write($sql);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->prefixError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForPartnerEquityAccount"),$eqAccounts{$acct},$closeoutDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return undef;
      }
    }
  }
  #UPDATE known years tb
  if ($closeoutDate eq $yearEnd)
  {
    my $sql = "UPDATE known_years_tb set closed = 'true' WHERE ledger_code = '".$ledger->code."' AND year = '".$currentFiscalYear++."'";
    $sth = $self->{dbHandle}->write($sql);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{langObj}->map("error") . "<br>$sql<br>\n" . $self->{dbHandle}->errorMessage);
      return undef;
    }
    $sql = "UPDATE known_years_tb set closed = 'false' WHERE ledger_code = '".$ledger->code."' AND year = '$currentFiscalYear'";
    $sth = $self->{dbHandle}->write($sql);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{langObj}->map("error") . "<br>$sql<br>\n" . $self->{dbHandle}->errorMessage);
      return undef;
    }
    if ($sth == 0)
    {
      $sql = "INSERT INTO known_years_tb (ledger_code, year, closed) VALUES ('".$ledger->code."', '".$currentFiscalYear."', false)";
      $sth = $self->{dbHandle}->write($sql);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->prefixError($self->{langObj}->map("error") . "<br>$sql<br>\n" . $self->{dbHandle}->errorMessage);
        return undef;
      }
    }
  }
  if ($commit)
  {
    $sth = $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError($self->{langObj}->map("error") . "<br>COMMIT<br>\n" . $self->{dbHandle}->errorMessage);
      return undef;
    }
  }
  return ($self->getLedgers($ledger->code))[0];
}

sub updateBalanceTable
{
  my $self = shift;
  my $ledgerCode = shift;
  my $closeoutDate = shift;
  # Use the balance stored procedures for the date passed in and the ledger code to gather the balances, accounts, and sub-accounts
  # that are needed to close out that period.
  my $subAccountCloseoutSQL = "INSERT INTO acct_bal_tb (ledger_code, acct_num, balance, sub_acct_num, bal_date) \
  (SELECT ledger_code, account_num, \
  sub_account_balance_plus_activity(account_num, sub_account_num, ledger_code, '$closeoutDate', 'custom'), sub_account_num, '$closeoutDate'::date \
  FROM debitcredit_tb \
  WHERE ledger_code = ? AND \
  trans_date <= ? AND \
  account_num = (SELECT number FROM account_tb WHERE number = account_num AND \
  type = (SELECT num FROM acct_type_tb WHERE num = type AND num <> 0 AND num <> 17 AND first_digit <> '4' AND first_digit <> '5')) \
  UNION \
  SELECT ledger_code, acct_num, \
  sub_account_balance_plus_activity(acct_num, sub_acct_num, ledger_code, '$closeoutDate', 'custom'), sub_acct_num, '$closeoutDate' \
  FROM acct_bal_tb \
  WHERE ledger_code = ? AND \
  bal_date <= ? AND \
  acct_num = (SELECT number FROM account_tb WHERE number = acct_num AND \
  type = (SELECT num FROM acct_type_tb WHERE num = type AND num <> 0 AND num <> 10 AND num <> 17 AND first_digit <> '4' AND first_digit <> '5')) \
  AND sub_acct_num <> -2 \
  GROUP BY ledger_code, acct_num, sub_acct_num)";
  my $sth = $self->{dbHandle}->write(sql => $subAccountCloseoutSQL, plug => [ $ledgerCode, $closeoutDate, $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError(errorString => $self->{langObj}->map("errorInsertingSubAccountBalances") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }

  # Use the balance stored procedures for the date passed in and the ledger code to gather the balances and accounts
  # that are needed to close out that period.
  my $accountCloseoutSQL = "INSERT INTO acct_bal_tb (ledger_code, acct_num, balance, sub_acct_num, bal_date) \
  (SELECT ledger_code, account_num, \
  account_balance_plus_activity(account_num, ledger_code, '$closeoutDate', 'custom'), -2, '$closeoutDate'::date \
  FROM debitcredit_tb \
  WHERE ledger_code = ? AND \
  trans_date <= ? AND \
  account_num = (SELECT number FROM account_tb WHERE number = account_num AND \
  type = (SELECT num FROM acct_type_tb WHERE num = type AND num <> 17 AND first_digit <> '4' AND first_digit <> '5')) \
  GROUP BY ledger_code, account_num \
  UNION \
  SELECT ledger_code, acct_num, \
  account_balance_plus_activity(acct_num, ledger_code, '$closeoutDate', 'custom'), -2, '$closeoutDate' \
  FROM acct_bal_tb \
  WHERE ledger_code = ? AND \
  bal_date <= ? AND \
  acct_num = (SELECT number FROM account_tb WHERE number = acct_num AND \
  type = (SELECT num FROM acct_type_tb WHERE num = type AND num <> 17 AND first_digit <> '4' AND first_digit <> '5')) \
  GROUP BY ledger_code, acct_num)";
  $sth = $self->{dbHandle}->write(sql => $accountCloseoutSQL, plug => [ $ledgerCode, $closeoutDate, $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError(errorString => $self->{langObj}->map("errorInsertingAccountBalances") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }

  # retrieve the last previous balance date
  $sth = $self->{dbHandle}->read(sql => "SELECT MAX(bal_date) FROM acct_bal_tb WHERE ledger_code = ? AND bal_date < ?", plug => [ $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError(errorString => $self->{langObj}->map("errorLookingUpPreviousCloseoutDate") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  my $lastPreviousCloseoutDate = "";
  my @info;
  if (defined (@info = $sth->fetchrow_array) && defined $info[0])
  {
    $lastPreviousCloseoutDate = $info[0];
  }
  # gather activity for Revenues and Expenses and Dividends Paid (account type == 17)
  my $dividendsPaidActivitySQL = "SELECT activity_sum(true, (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = 17) ##?## AND trans_date <= ? AND dorc = true), (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = 17) ##?## AND trans_date <= ? AND dorc = false))";
  if ($lastPreviousCloseoutDate eq "")
  {
    $dividendsPaidActivitySQL =~ s/\#\#\?\#\#//g;
  }
  else
  {
    $dividendsPaidActivitySQL =~ s/\#\#\?\#\#/AND trans_date > '$lastPreviousCloseoutDate'/g;
  }
  $sth = $self->{dbHandle}->read(sql => $dividendsPaidActivitySQL, plug => [ $ledgerCode, $closeoutDate, $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError(errorString => $self->{langObj}->map("errorRetrievingActivityForDividendsPaid") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  my $dividendsPaidTotal = "0.00";
  if (defined (@info = $sth->fetchrow_array) && defined $info[0])
  {
    $dividendsPaidTotal = $info[0];
  }

  my $revenuesActivitySQL = "SELECT activity_sum(false, (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = (SELECT num FROM acct_type_tb WHERE num = type AND first_digit = '4')) ##?## AND trans_date <= ? AND dorc = true), (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = (SELECT num FROM acct_type_tb WHERE num = type AND first_digit = '4')) ##?## AND trans_date <= ? AND dorc = false))";
  if ($lastPreviousCloseoutDate eq "")
  {
    $revenuesActivitySQL =~ s/\#\#\?\#\#//g;
  }
  else
  {
    $revenuesActivitySQL =~ s/\#\#\?\#\#/AND trans_date > '$lastPreviousCloseoutDate'/g;
  }
  $sth = $self->{dbHandle}->read(sql => $revenuesActivitySQL, plug => [ $ledgerCode, $closeoutDate, $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError(errorString => $self->{langObj}->map("errorRetrievingActivityForRevenues") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  my $revenuesTotal = "0.00";
  if (defined (@info = $sth->fetchrow_array) && defined $info[0])
  {
    $revenuesTotal = $info[0];
  }

  my $expensesActivitySQL = "SELECT activity_sum(true, (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = (SELECT num FROM acct_type_tb WHERE num = type AND first_digit = '5')) ##?## AND trans_date <= ? AND dorc = true), (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = (SELECT num FROM acct_type_tb WHERE num = type AND first_digit = '5')) ##?## AND trans_date <= ? AND dorc = false))";
  if ($lastPreviousCloseoutDate eq "")
  {
    $expensesActivitySQL =~ s/\#\#\?\#\#//g;
  }
  else
  {
    $expensesActivitySQL =~ s/\#\#\?\#\#/AND trans_date > '$lastPreviousCloseoutDate'/g;
  }
  $sth = $self->{dbHandle}->read(sql => $expensesActivitySQL, plug => [ $ledgerCode, $closeoutDate, $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError(errorString => $self->{langObj}->map("errorRetrievingActivityForExpenses") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  my $expensesTotal = "0.00";
  if (defined (@info = $sth->fetchrow_array) && defined $info[0])
  {
    $expensesTotal = $info[0];
  }

  # format activity totals
  $dividendsPaidTotal = sprintf("%.2f", $dividendsPaidTotal);
  $revenuesTotal = sprintf("%.2f", $revenuesTotal);
  $expensesTotal = sprintf("%.2f", $expensesTotal);

  # calculate the Retained Earnings balance
  my $retainedEarnings = $revenuesTotal - ($expensesTotal + $dividendsPaidTotal);
  $retainedEarnings = sprintf("%.2f", $retainedEarnings);

  return $retainedEarnings;
}

sub getCloseOutDates
{
  my $self = shift;
  my $ledgerCode = shift;
  my @dates;
  my $sql = "SELECT DISTINCT(bal_date) FROM acct_bal_tb WHERE ledger_code = '$ledgerCode' ORDER BY bal_date DESC";
  my $sth = $self->{dbHandle}->read($sql);
  my $index = 0;
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return undef;
  }
  while (my @date = $sth->fetchrow_array)
  {
    $dates[$index++] = $date[0];
  }

  return @dates;
}

=item int reopenLedger(ledgerCode, date, equityAccounts)

 Reopens a ledger on the given year or date period.
 Returns 1 on success, 0 on errors and sets the errorMessage.

=cut

sub reopenLedger
{
  my $self = shift;
  my %args = ( ledgerCode => "", date => "", commit => 1, @_ );
  my $ledgerCode = $args{ledgerCode};
  my $reopenDate = $args{date};
  my $lookupYear = (split "-", $reopenDate)[0];
  my $toDate;

  # validate input
  if (not defined $ledgerCode || $ledgerCode eq "")
  {
    $self->prefixError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "ledgerCode", $ledgerCode));
    return undef;
  }
  if (not defined $reopenDate || $reopenDate eq "")
  {
    $self->prefixError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "reopenDate", $reopenDate));
    return undef;
  }

  # remove any balance table entries that exist on or after the given reopenDate period
  my $sth = $self->{dbHandle}->write(sql => "DELETE FROM acct_bal_tb WHERE bal_date >= ? AND ledger_code = ?", plug => [ $reopenDate, $ledgerCode ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError(errorString => sprintf($self->{langObj}->map("errorRemovingAccountBalances"), $lookupYear) . "<br>\n" . $self->{dbHandle}->errorMessage);
    return undef;
  }

  my $sth = $self->{dbHandle}->write(sql => "UPDATE known_years_tb SET closed = false WHERE year >= ? AND ledger_code = ?", plug => [ $lookupYear, $ledgerCode ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError(errorString => sprintf($self->{langObj}->map("errorOpeningFiscalYears"), $lookupYear) . "<br>\n" . $self->{dbHandle}->errorMessage);
    return undef;
  }

  # commit the work
  if ($args{commit} == 1)
  {
    $sth = $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->prefixError(errorString => $self->{langObj}->map("errorCommittingTransaction") . "<br>\n" . $self->{dbHandle}->errorMessage);
      return undef;
    }
  }

  return ($self->getLedgers(ledgerCode => $args{ledgerCode}))[0];
}


=item Objects::Ledger[] getLedgers(ledgerCode =>)

    Returns an array of Ledger Objects matching the parameters passed.
    No params means select all ledgers.  Empty array is returned if
    no ledgers exist or match criteria.
    NOTE: Will not error on an invalid ledger.  You must do an
    ->isValid on each returned ledgerObj to decide what to do with it.
    Undef returned on errors.

=cut
sub getLedgers
{
  my $self = shift;
  my $ledgerCode;
  if (@_ == 1)
  {  $ledgerCode = @_[0];  }
  else
  {
    my %args = ( ledgerCode => "", @_ ) ;
    $ledgerCode = $args{ledgerCode};
  }
  my @Ledgers = ();
  my $where = "";

  if ($ledgerCode ne "")
  {
    $where .= " AND ";
    $where .= " ledger_tb.code = '$ledgerCode'";
  }

  my $sql = "SELECT ledger_tb.name, ledger_tb.code, ledger_tb.color";
  $sql .= ", ledger_tb.cash_or_accrual ";
  $sql .= ", business_type_tb.name, ledger_tb.business_type ";
  $sql .= ", offset1 ";
  $sql .= ", start1 ";
  $sql .= ", start2 ";
  $sql .= ", start3 ";
  $sql .= ", start4 ";
  $sql .= ", start5 ";
  $sql .= ", start6 ";
  $sql .= ", start7 ";
  $sql .= ", start8 ";
  $sql .= ", start9 ";
  $sql .= ", start10 ";
  $sql .= ", start11 ";
  $sql .= ", start12 ";
  $sql .= ", working_year(ledger_tb.code) ";
  $sql .= "FROM ledger_tb, business_type_tb ";
  $sql .= "WHERE ledger_tb.business_type = business_type_tb.num ".$where;
  $sql .= "ORDER BY ledger_tb.name, ledger_tb.code";
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($sql."<br>\n". $self->{dbHandle}->errorMessage);
    return undef;
  }
  while (my @info = $sth->fetchrow_array)
  {
    my %knownYears = $self->getKnownYears($info[1]);
    if ($self->didErrorOccur)
    {
      $self->prefixError(" Sent ledger '$ledgerCode' to function getKnownYears<br>\n");
      return undef;
    }

    my $cashOrAccrual = "";
    if ($info[3] =~ /^t|true|1$/)  #!!!!find DB independent way to do this!
    {  $cashOrAccrual = "cash";}
    elsif ($info[3] =~ /^f|false|0$/)
    { $cashOrAccrual = "accrual";}
    else
    { $cashOrAccrual = "error!";}

    my @start = ( $info[6], $info[7], $info[8], $info[9], $info[10], $info[11], $info[12], $info[13], $info[14], $info[15], $info[16], $info[17], $info[18]);

    my $ledgerObj = Portal::Accounting::Objects::Ledger->new( langObj => $self->{langObj});
    $ledgerObj->setup(\$self->getBusinessTypeNames);
    $ledgerObj->populate(name => $info[0], code => $info[1], color => $info[2],
                    basis => $cashOrAccrual,
                    businessType => $info[5],
                    knownYears => \%knownYears,
                    start => \@start,
                    fiscalYear => $info[19]
                    );

    $Ledgers[++$#Ledgers] = $ledgerObj;
  }

  return @Ledgers;
}

=item Objects::Ledger getValidLedger()

  Since ledgers need to show up whether they have all valid data getLedgers will return all of them.
  This routine will check the ledger returned and error if it is not valid.

=cut
sub getValidLedger
{
  my $self = shift;
  my @ledger = $self->getLedgers(@_);
  if (not defined @ledger)
  { return undef; }
  if ($ledger[0]->isValid)
  {  return $ledger[0]; }
  return undef;
}

=head2 Hash getBusinessTypeNames()

    This routine selects the business type names from the database and returns them as an hash
    of number = name values where the name is looked up from the Language Object so as to be
    Internationalized.

=cut
sub getBusinessTypeNames
{
  my $self = shift;
  my $errStr = "Portal::Accounting::DB::Ledger->getBusinessTypes() - Error!<br>\n";

  my $index = 0;
  my @info;
  my $sth;
  my %types;

  $sth = $self->{dbHandle}->read(sql => "SELECT name, num FROM business_type_tb ORDER BY num");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->postfixErrorString($self->{dbHandle}->errorMessage);
    return undef;
  }
  while (@info = $sth->fetchrow_array)
  {
    $types{$info[1]} = $self->{langObj}->map($info[0]);
  }

  return %types;
}

=head2 bool isValidData(field => businessType, data => "Corporation", ledger => $ledger)

=cut
sub isValidData
{
  my $self = shift;
  my $errStr = "Portal::Accounting::DB::Ledger->isValidData() - Error!<br>\n";

  return undef;
}

=head2 getArrayOfSubEntities

required args: ledgerCode
optional: orderBy (name, code)
  code:  array of specific sub entities to retrieve
returns: Array of SubEntityObjects
This routine selects all sub_entities for a given ledger.
The orderBy string must be a valid ORDER BY possible statement for this query.

=cut
sub getArrayOfSubEntities
{
  my $self = shift;
  my %args = ( orderBy => "name, code", code => undef, ledgerCode => "", @_ );
  my @SubEntitys;
  my $sth;
  my @info;

  # validate input
  if (length $args{orderBy} == 0)
  {
    $self->setError(errorString => "orderBy = '$args{orderBy}' is invalid!<br>\n");
    return @SubEntitys;
  }

  my $sql = "SELECT name, code, ledger_code FROM sub_entity_tb";
  if ($args{ledgerCode} ne "")
  {
    $sql .= " WHERE ledger_code = '$args{ledgerCode}'";
  }
  if (defined $args{code} && @{$args{code}} > 0)
  {
    $sql .= " AND code IN ('";
    $sql .= join( "', '", @{$args{code}}) ;
    $sql .= "') ";
  }
  $sql .= " ORDER BY $args{orderBy}";

  # now execute the sql statement
  $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    return @SubEntitys;
  }
  while (@info = $sth->fetchrow_array)
  {
    my $SubEntityObj = Portal::Accounting::Objects::SubEntityObject->new(langObj => $self->{langObj});
    $SubEntityObj->populate(name => $info[0], code => $info[1], ledgerCode =>$info[2]);
    if ($SubEntityObj->didErrorOccur)
    {
      $self->setError(errorString => $SubEntityObj->errorMessage);
      return @SubEntitys;
    }
    $SubEntitys[++$#SubEntitys] = $SubEntityObj;
  }

  return @SubEntitys;
}

=head2 <db sth> getSubEntitiesAsSth(ledgerCode =>, order =>)

=cut
sub getSubEntitiesAsSth
{
  my $self = shift;
  my $errStr = "Portal::Accounting::DB::Ledger->getSubEntitiesAsSth() - Error!<br>\n";
  my %args = (ledgerCode => "", order => "name, code", @_);

  my $sql = "SELECT name||' ('||code||')', code FROM sub_entity_tb WHERE ledger_code = '$args{ledgerCode}' ORDER BY $args{order}";
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    return undef;
  }

  return $sth;
}

=head2 bool updateSubEntity(ledgerCode =>, code =>, name =>)

  This routine will insert a new sub entity.
  If one already exists with code in given ledger it will
  update the name.

=cut
sub updateSubEntity
{
  my $self = shift;
  my %args = (name => "", code => "", ledgerCode => "", @_);
  my $sql = "UPDATE sub_entity_tb SET name = '$args{name}' WHERE ledger_code = '$args{ledgerCode}' AND code = '$args{code}'";
  my $sth = $self->{dbHandle}->write($sql);
  if ($self->didErrorOccur)
  {  return 0; }
  if ($sth < 1)
  {
    $sql = "INSERT INTO sub_entity_tb (name, code, ledger_code) VALUES ('$args{name}', '$args{code}', '$args{ledgerCode}')";
    $sth = $self->{dbHandle}->write($sql);
    if ($self->didErrorOccur)
    {  return 0; }
  }
  $self->{dbHandle}->commit();
  return 1;
}

=head2 string deleteSubEntity(ledgerCode =>, code =>)

  Returns a string of Success, Failed, or SubEntityInUse which is a failure.
  Cannot delete sub entities that are in use.

=cut
sub deleteSubEntity
{
  my $self = shift;
  my %args = (code => "", ledgerCode => "", @_);
  my $sql = "SELECT sub_e_code FROM debitcredit_tb WHERE ledger_code = '$args{ledgerCode}' AND sub_e_code = '$args{code}' LIMIT 1";
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->didErrorOccur)
  {  return "Failed"; }
  if (my @info = $sth->fetchrow_array)
  {
    return "SubEntityInUse";
  }
  $sql = "DELETE FROM sub_entity_tb WHERE ledger_code = '$args{ledgerCode}' AND code = '$args{code}'";
  $sth = $self->{dbHandle}->write($sql);
  if ($self->didErrorOccur)
  {  return "Failed"; }
  $self->{dbHandle}->commit();
  return "Success";
}

=item int doesSystemBalance()

     requires: ledgerCode, beginDate, endDate
     optional: if beginDate/endDate are not given, the balances will be calculated on the remaining date value.
               If neither date value is given, the balance will be calculated on the entire set of transactions
               for the given ledgerCode.
     returns: 1 if system balances, -1 if debits are greater than credits, -2 if credits are greater than debits,
              or 0 if an error occurred
    summary: This method sums all the debits and credits for the transactions in the debitcredit_tb table for the
             given ledgerCode and dates and compares them to see if they are the same.  It will return different
             error codes based on whether debits are greater than credits or vice versa.
=cut
sub doesSystemBalance
{
  my $self = shift;
  my %args = ( ledgerCode => "", beginDate => "", endDate => "", @_ );
  my $ledgerCode = $args{ledgerCode};
  my $beginDate = $args{beginDate};
  my $endDate = $args{endDate};
  my $sth;
  my @info;
  my $sqlString = "SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND ";
  my $sqlWhere = "";
  my @plug = ();
  my $sumDebits = "0.00";
  my $sumCredits = "0.00";
  my $result;

  if (not defined $ledgerCode || length $ledgerCode == 0)
  {
    $self->prefixError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "ledgerCode", $ledgerCode));
    return 0;
  }

  # put the valid ledgerCode onto the plug array
  push @plug, $ledgerCode;

  # debits
  if ((defined $beginDate) && ($beginDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/) && (defined $endDate) && ($endDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/))
  {
    $sqlWhere = "(trans_date BETWEEN ? AND ?) AND dorc = true";
    push @plug, $beginDate, $endDate;
  }
  elsif (((not defined $beginDate) || (length $beginDate == 0)) && ((defined $endDate) && ($endDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/)))
  {
    $sqlWhere = "trans_date <= ? AND dorc = true";
    push @plug, $endDate;
  }
  elsif (((not defined $endDate) || (length $endDate == 0)) && ((defined $beginDate) && ($beginDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/)))
  {
    $sqlWhere = "trans_date >= ? AND dorc = true";
    push @plug, $beginDate;
  }
  elsif (((not defined $beginDate) || (length $beginDate == 0)) && ((not defined $endDate) || (length $endDate == 0)))
  {
    $sqlWhere = "dorc = true";
  }
  else
  {
    $self->prefixError(errorString => sprintf($self->{langObj}->map("invalidDebitDate"), $beginDate, $endDate));
    return 0;
  }
  # retrieve the sum of all debit amounts
  $sth = $self->{dbHandle}->read(sql => $sqlString . $sqlWhere, plug => \@plug);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError(errorString => $self->{langObj}->map("errorSummingDebits") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  @info = $sth->fetchrow_array;
  if (defined @info && defined $info[0])
  {
    $sumDebits = $info[0];
  }
  else
  {
    $sumDebits = "0.00";
  }

  # refresh the plug array for the credits
  @plug = ();
  push @plug, $ledgerCode;

  # credits
  if ((defined $beginDate) && ($beginDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/) && (defined $endDate) && ($endDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/))
  {
    $sqlWhere = "(trans_date BETWEEN ? AND ?) AND dorc = false";
    push @plug, $beginDate, $endDate;
  }
  elsif (((not defined $beginDate) || (length $beginDate == 0)) && ((defined $endDate) && ($endDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/)))
  {
    $sqlWhere = "trans_date <= ? AND dorc = false";
    push @plug, $endDate;
  }
  elsif (((not defined $endDate) || (length $endDate == 0)) && ((defined $beginDate) && ($beginDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/)))
  {
    $sqlWhere = "trans_date >= ? AND dorc = false";
    push @plug, $beginDate;
  }
  elsif (((not defined $beginDate) || (length $beginDate == 0)) && ((not defined $endDate) || (length $endDate == 0)))
  {
    $sqlWhere = "dorc = false";
  }
  else
  {
    $self->prefixError(errorString => sprintf($self->{langObj}->map("invalidCreditDate"), $beginDate, $endDate));
    return 0;
  }
  # retrieve the sum of all credit amounts
  $sth = $self->{dbHandle}->read(sql => $sqlString . $sqlWhere, plug => \@plug);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError(errorString => $self->{langObj}->map("errorSummingCredits") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  @info = $sth->fetchrow_array;
  if (defined @info && defined $info[0])
  {
    $sumCredits = $info[0];
  }
  else
  {
    $sumCredits = "0.00";
  }

  # make sure both values are formatted correctly
  $sumDebits = sprintf("%.2f", $sumDebits);
  $sumCredits = sprintf("%.2f", $sumCredits);

  # compare to make sure that debits == credits
  if ($sumDebits eq $sumCredits)
  {
    # the system balances
    $result = 1;
  }
  elsif ($sumDebits > $sumCredits)
  {
    $result = -1;
  }
  elsif ($sumDebits < $sumCredits)
  {
    $result = -2;
  }

  return $result;
}

=item integer getCurrentFiscalYear(ledgerCode)

     requires: ledgerCode
     optional:
     returns: integer or undef if error
    summary: This method returns the current Fiscal Year for the specified ledger.
=cut
sub getCurrentFiscalYear
{
  my $self = shift;
  my $ledgerCode = shift;
  my $sth;
  my @info;

  if (length $ledgerCode == 0)
  {
    $self->prefixError( "ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return undef;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT working_year('$ledgerCode')");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return undef;
  }
  if (@info = $sth->fetchrow_array)
  { return $info[0]; }
  $self->prefixError("No data returned!<br>\n");
  return undef;
}

=item date getFiscalEnd(ledgerCode)

     requires: ledgerCode
     optional:
     returns: date or undef if error
    summary: This method returns the last day of the current Fiscal Year for the specified ledger.
=cut
sub getFiscalEnd
{
  my $self = shift;
  my $ledgerCode = shift;
  my $sth;
  my @info;
  my $date = undef;

  if (length $ledgerCode == 0)
  {
    $self->prefixError( "ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return undef;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT month_end('$ledgerCode',12)");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return undef;
  }
  @info = $sth->fetchrow_array;
  if (!defined $info[0])
  {
    $self->prefixError("No data returned!<br>\n");
    return undef;
  }
  return $date;
}

=head2 %years getKnownYears(ledgerCode)

    required: ledgerCode
    Returns hash of years as keys with YYYY/MM/DD as data for each key

=cut
sub getKnownYears
{
  my $self = shift;
  my $ledgerCode = shift;
  my %knownYears;
  my $sth;
  my @info;

  if (length $ledgerCode == 0)
  {
    $self->prefixError("ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return undef;
  }

  my $sql = "SELECT year, (SELECT month_end('$ledgerCode',12,year)) FROM known_years_tb WHERE ledger_code = '$ledgerCode' ORDER BY year";
  $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return undef;
  }
  while (@info = $sth->fetchrow_array)
  {
    $knownYears{$info[0]} = $info[1];
  }
  if (scalar %knownYears == 0)
  {
    $self->prefixError("No data returned!<br>\n");
    return undef;
  }

  return %knownYears;
}

=item date getLastClose(ledgerCode)

  This routine simply returns the last closeout date for this ledger.

=cut
sub getLastClose
{
  my $self = shift;
  my $ledgerCode = shift;
  if ($ledgerCode eq "")
  {
    $self->prefixError("Error! no ledgerCode sent to getLastClose!<br>\n");
    return undef;
  }
  my $sth = $self->{dbHandle}->write(sql => "SET datestyle TO 'ISO'");
  $sth = $self->{dbHandle}->read(sql => "SELECT MAX(bal_date) FROM acct_bal_tb WHERE ledger_code = '$ledgerCode'");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError("Error in getLastClose!<br>\n" . $self->{dbHandle}->errorMessage);
    return undef;
  }
  if (my @info = $sth->fetchrow_array)
  {
    return $info[0];
  }
  return "";
}

=item Objects::Register getRegister(ledgerCode => "", account => "", from => "", to => "", reconcile => 0, )

  This routine gathers the data that goes into a register screen.  Places it in an object and returns it.

=cut
sub getRegister
{
  my $self = shift;
  my %args = (ledgerCode => "", account => "", from => "", to => "", reconcile => 0, @_);
  my @account;
  my $sql;
  my $register = Portal::Accounting::Objects::Register->new(langObj => $self->{langObj});
  my $accountSql = "SELECT name, type, (SELECT SUM(d_amount) FROM debitcredit_tb WHERE account_num = '$args{account}' AND trans_date < '$args{from}') AS balance FROM account_tb WHERE number = $args{account}";
  my $sth = $self->{dbHandle}->read(sql => $accountSql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError("$sql".$self->{dbHandle}->errorMessage);
    return undef;
  }
  if (not @account = $sth->fetchrow_array)
  {
    $self->prefixError("No account found for '$args{account}'");
    return undef;
  }

  if ($account[1] == 0)  #if the type is a bank account do not select sub account names
  {
    $sql  = "SELECT parent_num, counter, '' AS outstanding, trans_date ";
    $sql .= ", (SELECT payee_payer FROM journal_entry_tb WHERE ref_num = parent_num) AS payee";
    $sql .= ", SUBSTRING(comment, 1, 15)";
    $sql .= ", checknum(dorc, sub_account_num), depositnum(dorc, sub_account_num), sub_account_num, cleared, d_amount";
    $sql .= " FROM debitcredit_tb";
    $sql .= " WHERE ledger_code = '$args{ledgerCode}' AND account_num = '$args{account}'";
    $sql .= " AND trans_date >= '$args{from}' AND trans_date <= '$args{to}'";
    if ($args{reconcile})
    {
      $sql .= " UNION ";
      $sql .= "SELECT parent_num, counter, '*' AS oustanding, trans_date ";
      $sql .= ", (SELECT payee_payer FROM journal_entry_tb WHERE ref_num = parent_num) AS payee";
      $sql .= ", SUBSTRING(comment, 1, 15)";
      $sql .= ", checknum(dorc, sub_account_num), depositnum(dorc, sub_account_num), sub_account_num, cleared, d_amount";
      $sql .= " FROM debitcredit_tb";
      $sql .= " WHERE ledger_code = '$args{ledgerCode}' AND account_num = '$args{account}'";
      $sql .= " AND trans_date < '$args{from}' AND cleared <> 'r'";
    }
    $sql .= " ORDER BY trans_date, payee";
  }
  elsif ($account[1] == 3)  #if the type is a stock account do not select sub accounts.  Only show cash transactions
  {
    $sql  = "SELECT parent_num, counter, '' AS outstanding, trans_date ";
    $sql .= ", (SELECT payee_payer FROM journal_entry_tb WHERE ref_num = parent_num) AS payee";
    $sql .= ", SUBSTRING(comment, 1, 15)";
    $sql .= ", '', '', sub_account_num, cleared, d_amount";
    $sql .= " WHERE ledger_code = '$args{ledgerCode}' AND account_num = '$args{account}'";
    $sql .= " FROM debitcredit_tb";
    $sql .= " AND trans_date >= '$args{from}' AND trans_date <= '$args{to}' AND sub_account_num = '-1'";
    if ($args{reconcile})
    {
      $sql .= " UNION ";
      $sql .= "SELECT parent_num, counter, '*' AS oustanding, trans_date ";
      $sql .= ", (SELECT payee_payer FROM journal_entry_tb WHERE ref_num = parent_num) AS payee";
      $sql .= ", SUBSTRING(comment, 1, 15)";
      $sql .= ", '', '', sub_account_num, cleared, d_amount";
      $sql .= " FROM debitcredit_tb";
      $sql .= " WHERE ledger_code = '$args{ledgerCode}' AND account_num = '$args{account}'";
      $sql .= " AND trans_date < '$args{from}' AND cleared <> 'r' AND sub_account_num = '-1'";
    }
    $sql .= " ORDER BY trans_date, payee";
  }
  else
  {
    $sql  = "SELECT parent_num, counter, '' AS outstanding, trans_date ";
    $sql .= ", (SELECT payee_payer FROM journal_entry_tb WHERE ref_num = parent_num)";
    $sql .= ", SUBSTRING(comment, 1, 15)";
    $sql .= ", '', (SELECT sub_acc_name from sub_acc_tb WHERE sub_acc_number = sub_account_num) AS sub_account, sub_account_num, cleared, d_amount";
    $sql .= " FROM debitcredit_tb";
    $sql .= " WHERE ledger_code = '$args{ledgerCode}' AND account_num = '$args{account}'";
    $sql .= " AND trans_date >= '$args{from}' AND trans_date <= '$args{to}'";
    if ($args{reconcile})
    {
      $sql .= " UNION ";
      $sql .= "SELECT parent_num, counter, '*' AS oustanding, trans_date ";
      $sql .= ", (SELECT payee_payer FROM journal_entry_tb WHERE ref_num = parent_num)";
      $sql .= ", SUBSTRING(comment, 1, 15)";
      $sql .= ", '', (SELECT sub_acc_name from sub_acc_tb WHERE sub_acc_number = sub_account_num) AS sub_account, sub_account_num, cleared, d_amount";
      $sql .= " FROM debitcredit_tb";
      $sql .= " WHERE ledger_code = '$args{ledgerCode}' AND account_num = '$args{account}'";
      $sql .= " AND trans_date < '$args{from}' AND cleared <> 'r'";
    }
    $sql .= " ORDER BY trans_date, sub_account_num";
  }
  my $sthData = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError("Error occured calling read<br>$sql".$self->{dbHandle}->errorMessage);
    return undef;
  }
  my $data;
  if ($data = $sthData->fetchall_arrayref)
  {
   ############!~~~~~~ uncomment to run over data set and output with error
#     $self->postfixError("Debug on <br>$sql<br>".scalar @$data."<hr>\n");
#     for (my $i = 0; $i < scalar @$data ; $i++)
#     {
#       $self->postfixError("start:");
#       for (my $j = 0; $j < scalar @{@$data[$i]} ; $j++)
#       {
#         $self->postfixError("$i,$j:".@$data[$i]->[$j]."\t");
#       }
#       $self->postfixError("<br>");
#     }
#     return undef;
    ##########################
    my $result = $register->populate(ledgerCode => $args{ledgerCode}, reconcile => $args{reconcile}
      , accountNumber => $args{account}
      , accountName => $account[0], type => $account[1], balance => $account[2]
      , from => $args{from}, to => $args{to}, data => $data
    );
    if (! $register)
    {
      $self->postfixError($register->errorMessage);
      return undef;
    }

  }
  return $register;
}

=item integer saveReconcile(ledgerCode, @Array)

  This routine takes a array of hash reconcile entries to mark in the database.
  array[row]{ref} =
  array[row]{counter} =
  array[row]{cleared} = 'x'
  returns: undef on error, number of rows updated on success;

=cut
sub saveReconcile
{
  my $self = shift;
  my $l = shift;
  my $data = shift;
  my @data = @$data;
  my $sql = "";
  my $result = 0;

  if ($l eq "")
  {
    $self->prefixError("LedgerCode is invalid in saveReconcile<br>\n");
    return undef;
  }
  if (! defined $data)
  {
    $self->prefixError("Data is invalid in saveReconcile<br>\n");
    return undef;
  }
  foreach my $row (@data)
  {
    if ($row->{ref} !~ /^\d+$/ || $row->{counter} !~ /^\d+$/) { next; }
    if ($row->{cleared} =~ /^[_xr]$/ )
    {
      $sql .= "UPDATE debitcredit_tb SET cleared = '$row->{cleared}' WHERE ledger_code = '$l' AND parent_num = '$row->{ref}' AND counter = '$row->{counter}';";
    }
  }
  if ($sql ne "")
  {
    $result = $self->{dbHandle}->write($sql);
    if ($self->checkDBError) { return undef;  }
    else {  $self->{dbHandle}->commit;  }
  }
  return $result;
}

=item integer finishReconcile(ledgerCode)

  converts all x to r in debitcredit_tb.cleared field.
  returns number of records updated.

=cut
sub finishReconcile
{
  my $self = shift;
  my %args = (ledgerCode => "", account => "", date => "", user => "", @_);
  my $lc = $args{ledgerCode};
  my $account = $args{account};
  my $id = $args{user};
  my $date = $args{date};
  my @t = Today_and_Now();

  if ($id eq "")
  {
    $self->prefixError("User is invalid in finishReconcile<br>\n");
    return undef;
  }
  $id .= "$args{id} $t[0]-$t[1]-$t[2] $t[3]:$t[4]:$t[5]";
  if ($lc eq "")
  {
    $self->prefixError("LedgerCode is invalid in finishReconcile<br>\n");
    return undef;
  }
  if ($date !~ /^\d{4}-\d{1,2}-\d{1,2}$/)
  {
    $self->prefixError("Date is invalid in finishReconcile: $date<br>\n");
    return undef;
  }
  if ($account !~ /^\d+$/)
  {
    $self->prefixError("Account is invalid in finishReconcile: $account<br>\n");
    return undef;
  }
  my $sql = "UPDATE debitcredit_tb SET cleared = 'r', reconcile = '$id' WHERE ledger_code = '$lc' AND cleared = 'x';";
  $sql .= " INSERT INTO reconcile_tb (id, ledger_code, account, date_of) VALUES ('$id', '$lc', '$account', '$date');";
  my $result = $self->{dbHandle}->write($sql);
  if ($self->checkDBError)
  {
    return undef;
  }
  else {  $self->{dbHandle}->commit;  }
  return $result;
}

sub checkDBError
{
  my $self = shift;
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->postfixError($self->{dbHandle}->errorMessage);
    return 1;
  }
  return 0;
}

sub getAccountsAsSth
{
  my $self = shift;
  my $l = shift;
  my $sql = "SELECT account_tb.name, local_accounts_tb.acct_num FROM account_tb, local_accounts_tb WHERE local_accounts_tb.acct_num = account_tb.number AND local_accounts_tb.ledger_code = '$l' AND account_tb.active = true ORDER BY name";
  my $sth = $self->{dbHandle}->read($sql);
  if ($self->didErrorOccur)
  {  return undef;  }
  return $sth;
}

=item @ getCommonAccounts(ledgerCode)

  this returns a 2D array with values for
  name, number, from, to

=cut
sub getCommonAccounts
{
  my $self = shift;
  my $l = shift;
  my $sql = <<"SQL";
SELECT
  DISTINCT(account)
  , (SELECT name FROM account_tb WHERE number = account) AS name
  , (SELECT MAX(date_of) FROM reconcile_tb WHERE account = reconcile_tb.account) +1 AS from
  , (SELECT MAX(date_of) FROM reconcile_tb WHERE account = reconcile_tb.account) +31 AS to
  FROM reconcile_tb
  WHERE ledger_code = '$l' AND  (SELECT active FROM account_tb WHERE number = account) = true
  GROUP BY account
SQL
  my $sth = $self->{dbHandle}->read($sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return undef;
  }
  if (my $data = $sth->fetchall_arrayref())
  {  return $data; }
  return undef;
}

1;
__END__

=pod

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;


=head1 AUTHOR

JT Moree, Xperience, Inc. <moreejt@pcxperience.org>

=head1 SEE ALSO

perl, Portal::Accounting

=cut

