# DBUpgrade.pm - The Object Class that provides a DBUpgrade Object
#$Id: DBUpgrade.pm,v 1.33 2004/09/14 14:29:47 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::DBUpgrade;
use strict;
use Portal::AppState;
use Portal::Accounting::DB::Main;
use Portal::Accounting::DB::Ledger;
use Portal::Accounting::DB::Transaction;
use Portal::Accounting::Objects::TransactionObject;
use Portal::Accounting::Objects::TransactionRowObject;
use Portal::Accounting::Objects::StockObject;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::AppState Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.05';

# global variables.
my $currDBVersion = "1.6";

# modify to represent the commands this state provides.
my %commands = ( display => "Initialize the Upgrade Process", phase1 => "Upgrade to next specified version", userinput => "Allow user to interact with Upgrade Process",
                 displayTop => "The Top Frame", displayPhase => "The Bottom Frame",
                 notAdmin => "Upgrade needed, but you are not an admin!",
               );
=head1 NAME

DBUpgrade - Object used to build a DBUpgrade Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::DBUpgrade;
  my $obj = Portal::Accounting::DBUpgrade->new;
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

DBUpgrade is a DBUpgrade class.
This class allows for migration from one version of the database to the
next and allows the user to interact with the migration in a nicer
environment than a command line provides.  This is mainly to allow the
user to specify the stock info or what accounts are to be taxable, etc.
If at all possible, the command line upgrade program should be used as
it is not affected by browser timeout, etc.

This is run from the Accounting module when it detects that a new
version of the database is available (code wise) and the database is not
at the version needed.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new()

 Creates a new instance of the Portal::Accounting::DBUpgrade module.
 See Portal::AppState(3) for a listing of required arguments.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item HTMLObject run(command)

    Runs the specified command and returns the resulting HTMLObject.

=cut

sub run
{
  my $self = shift;
  my %args = ( command => "", @_ );
  my $command = $args{command};
  my $doc;
  my $phrases = $self->langObj->phraseObj->phrases;

  if (!exists $commands{$command})
  {
    $self->setError(errorString => sprintf($phrases->{commandNotKnown}, $command));
    return undef;
  }
  $doc = HTMLObject::Normal->new;

  my $mainDBObj = Portal::Accounting::DB::Main->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return undef;
  }

  my $ledgerDBObj = Portal::Accounting::DB::Ledger->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($ledgerDBObj->didErrorOccur)
  {
    $self->setError(errorString => $ledgerDBObj->errorMessage);
    return undef;
  }

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  # check Database version
  my $dbVersion = $mainDBObj->getDBVersion;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return undef;
  }

  if ($command eq "notAdmin")
  {
    $doc->setTitle("DBUpgrade:  Needed!");
    $url .= "?" . (length $self->{sessionObj}->{store}->{originalDestination} > 0 ? $self->{sessionObj}->{store}->{originalDestination} : "app=Accounting&state=Main&command=display");
    $doc->setFocus("body");
    $doc->print("<center><h1>DB Upgrade Needed!</h1><br>But you are not an admin user!</center><br><br><br>\n");
    $doc->print("Go <a href=\"$url\">here</a> to continue using Xiwa after the database is upgraded.");

    return $doc;
  }

  if ($dbVersion >= $currDBVersion && $command ne "userinput")
  {
    # we don't need to upgrade!  Return to Main->display().
    $doc->setTitle("DBUpgrade:  Not needed!");
    $url .= "?" . (length $self->{sessionObj}->{store}->{originalDestination} > 0 ? $self->{sessionObj}->{store}->{originalDestination} : "app=Accounting&state=Main&command=display");
    $doc->setMetaTag('http-equiv' => "Refresh", content => "1; URL=$url");
    $doc->setFocus("body");
    $doc->print("Continuing...");

    # make sure the app is unlocked!
    if ($self->{applicationObj}->isAppLocked(app => $self->{appObj}->{id}))
    {
      # now unlock the app.
      $self->{applicationObj}->unlockApp(app => $self->{appObj}->{id}, type => "c", company => $self->{companyObj}->{id});
      if ($self->{applicationObj}->didErrorOccur)
      {
        $self->setError(errorString => $self->{applicationObj}->errorMessage);
        return undef;
      }
    }
    return $doc;
  }

  my @currentVersion = ("1.2", "1.3", "1.4", "1.5");
  my %updateFunctions = ( "1.2" => "updateToOneThree", "1.3" => "updateToOneFour", "1.4" => "updateToOneFive", "1.5" => "updateToOneSix" );
  my %userInputFunctions = ( "1.2" => "userInputToOneThree" );

  if ($command eq "display")
  {
    # generate frameset
    use HTMLObject::FrameSet;
    $doc = HTMLObject::FrameSet->new;

    $doc->setTitle("DBUpgrade: Database Upgrade");

    my $innerContent;
    $innerContent = $doc->createNoframe("You must have a frames capable browser!");
    $innerContent .= $doc->createFrame(src => "$url?app=Accounting&state=DBUpgrade&command=displayTop", frameborder => 'no', scrolling => 'auto', border => '1', name => 'upgradeTop');
    $innerContent .= $doc->createFrame(src => "$url?app=Accounting&state=DBUpgrade&command=displayPhase", frameborder => 'no', scrolling => 'auto', name => 'upgradeBottom');
    $doc->outerFrameset(rows => '*, 4*', content => $innerContent, border => '1', frameborder => 'no');
  }
  elsif ($command eq "displayTop")
  {
    $doc->setTitle("DBUpgrade: Top Frame");
    $doc->setFocus("body");
    if ($self->{sessionObj}->{store}->{dumpDB})
    {
      $doc->print("<center>Dumping Database at Version <b>$dbVersion</b>...</center>");
    }
    else
    {
      $doc->print("<center>Upgrading from Version <b>$dbVersion</b> to <b>$currDBVersion</b>...</center>");
    }
  }
  elsif ($command eq "displayPhase")
  {
    $doc = $self->displayPhaseScreen(doc => $doc, dbVersion => $dbVersion, mainDBObj => $mainDBObj, ledgerDBObj => $ledgerDBObj, currentVersion => \@currentVersion, updateFunctions => \%updateFunctions);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
  }
  elsif ($command eq "phase1")
  {
    $doc = $self->processPhase(doc => $doc, dbVersion => $dbVersion, mainDBObj => $mainDBObj, ledgerDBObj => $ledgerDBObj, currentVersion => \@currentVersion, updateFunctions => \%updateFunctions);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
  }
  elsif ($command eq "userinput")
  {
    $doc = $self->processUserInput(doc => $doc, dbVersion => $dbVersion, mainDBObj => $mainDBObj, ledgerDBObj => $ledgerDBObj, currentVersion => \@currentVersion, userInputFunctions => \%userInputFunctions);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
  }

  return $doc;
}

=item HTMLObject displayPhaseScreen(dbVersion, mainDBObj, ledgerDBObj, @currentVersion,
             %updateFunctions, doc, [errorStr, statusStr])

    This method decides what version we need to upgrade to and starts
    the process of calling the correct helper functions, etc.  If the
    helper function decides it needs user interaction, an HTMLObject
    document will be returned with the form, etc. that gets the user
    started on their part of the process.  If no user input is required
    and the upgrade is done for the current version, then undef is
    returned.  When all necessary upgrades are finished, a success
    document is displayed for the user and a link is provided to take
    them back to the main menu.

=cut
sub displayPhaseScreen
{
  my $self = shift;
  my %args = (dbVersion => "", mainDBObj => undef, ledgerDBObj => undef, currentVersion => undef, updateFunctions => undef, doc => undef, errorStr => "", statusStr => "", @_ );
  my $dbVersion = $args{dbVersion};
  my $mainDBObj = $args{mainDBObj};
  my $currentVersion = $args{currentVersion};
  my $updateFunctions = $args{updateFunctions};
  my $doc = $args{doc};
  my $errorStr = $args{errorStr};
  my $statusStr = $args{statusStr};

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  # figure out where the current DB Version is in relation to the latest version to upgrade.
  my $index;
  my $found = 0;
  my $count;
  my $nextVersion;
  for ($index=0; $index < scalar @{$currentVersion} && !$found; $index++)
  {
    if ($currentVersion->[$index] == $dbVersion)
    {
      $found = 1;
    }
  }
  if ($found)
  {
    $index--;
    $count = scalar(@{$currentVersion}) - $index;  # This tells me how many there are between it and the end of the array (number of upgrades to perform).
  }
  else
  {
    if ($nextVersion == $currDBVersion)
    {
      $count = 0;
    }
    else
    {
      # unknown version!
      $self->setError(errorString => "Current Database Version = '$dbVersion' is unknown!<br>\nSee the System Administrator!<br>\n");
      return undef;
    }
  }
  $nextVersion = (($index + 1) < scalar @{$currentVersion} ? $currentVersion->[$index+1] : $currDBVersion);

  # now generate the Request for confirmation of upgrade and tell the user how many steps to become current.
  $doc->setTitle("DBUpgrade: Bottom Frame");
  $doc->setFocus("body");
  if ($self->{sessionObj}->{store}->{dumpDB})
  {
    $doc->print(<<"END_OF_CODE");
<center>
<h1>Dumping Database...</h1>
</center>
END_OF_CODE
    if (length $errorStr > 0)
    {
      $doc->print("<font color=\"red\">Error:</font> $errorStr\n<br>\n");
    }
    if (length $statusStr > 0)
    {
      $doc->print("<font color=\"red\">Status:</font> $statusStr\n<br>\n");
    }
    $doc->print(<<"END_OF_CODE");
<center>
<form method="post" name="upgradeForm" action="$url">
  <input type="hidden" name="app" value="Accounting">
  <input type="hidden" name="state" value="DBUpgrade">
  <input type="hidden" name="command" value="displayPhase">
END_OF_CODE

    # dump the database
    my $dumpDir = $self->{configObj}->{webRoot} . $self->{configObj}->{siteDir} . "/db_dump/Accounting/$self->{companyObj}->{code}/";

    my $companyAppObj = $self->{sessionObj}->{store}->{companyAppObj};
    my $result = $self->{methods}->dbDump(type => $companyAppObj->{dbType}, host => $companyAppObj->{dbHost}, port => $companyAppObj->{dbPort},
                                          user => $self->{configObj}->{dbUser}, passwd => $self->{configObj}->{dbPasswd},
                                          dbName => $companyAppObj->{dbName}, dir => $dumpDir, fname => "xiwa_$dbVersion.dump", inserts => 1);
    if ($self->{methods}->didErrorOccur)
    {
      $self->setError(errorString => $self->{methods}->errorMessage);
      return undef;
    }

    $self->{sessionObj}->{store}->{dumpDB} = 0;  # indicate we no longer need to dump the database.

    $doc->print(<<"END_OF_CODE");
  <br>
  <br>
  $result
  <br>
  <br>
  <h1>Database Dumped...</h1>
  <br>
  <br>
  <h2>Do <font color="red">Not</font> reload your browser!</h2>
</form>
END_OF_CODE
  }
  else
  {
    $doc->print(<<"END_OF_CODE");
<center>
<h1>Database Upgrade In <font color="red">Progress</font></h1>
</center>
END_OF_CODE
  if (length $errorStr > 0)
  {
    $doc->print("<font color=\"red\">Error:</font> $errorStr\n<br>\n");
  }
  if (length $statusStr > 0)
  {
    $doc->print("<font color=\"red\">Status:</font> $statusStr\n<br>\n");
  }
  $doc->print(<<"END_OF_CODE");
<center>
<form method="post" name="upgradeForm" action="$url">
  <input type="hidden" name="app" value="Accounting">
  <input type="hidden" name="state" value="DBUpgrade">
  <input type="hidden" name="command" value="phase1">
  <input type="hidden" name="nextVersion" value="$dbVersion">
  <br>
  <input type="hidden" name="confirmUpgrade" value="YES">
  <br>
  Current step: Going from Version <b>$dbVersion</b> to <b>$nextVersion</b>.
  <br>
  There is a total of <b>$count</b> individual database upgrades to do. &nbsp;You may have to interact with the upgrade process and then you may not.
  <br>
  The upgrade process may take a while, so please be patient.
  <br>
  <br>
  <h2>Do <font color="red">Not</font> reload your browser!</h2>
</form>
END_OF_CODE
  }

  $doc->setOnload(code => "document.upgradeForm.submit();");

  return $doc;
}

=item HTMLObject processPhase(dbVersion, mainDBObj, ledgerDBObj, @currentVersion,
             %updateFunctions, doc)

    actually validates and then processes the database upgrade for the
    specified version.  Returns HTMLObject if user interaction required,
    else undef if upgrade is complete.

=cut
sub processPhase
{
  my $self = shift;
  my %args = (dbVersion => "", mainDBObj => undef, ledgerDBObj => undef, currentVersion => undef, updateFunctions => undef, doc => undef, @_ );
  my $dbVersion = $args{dbVersion};
  my $mainDBObj = $args{mainDBObj};
  my $ledgerDBObj = $args{ledgerDBObj};
  my $currentVersion = $args{currentVersion};
  my $updateFunctions = $args{updateFunctions};
  my $doc = $args{doc};

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  my $confirmUpgrade = $self->{input}->{confirmUpgrade};
  my $nextVersion = $self->{input}->{nextVersion};

  # set dbVersion into the input hash for children functions to get to.
  $self->{input}->{dbVersion} = $dbVersion;

  # make sure we have the input we need.
  if ($confirmUpgrade !~ /^YES|NO$/)
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "confirmUpgrade = '$confirmUpgrade' is invalid!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }
  if ($nextVersion !~ /^\d+\.\d+$/ || !exists $updateFunctions->{$nextVersion})
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "nextVersion = '$nextVersion' is invalid!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }

  if ($confirmUpgrade eq "NO")
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "You must specify <b>Yes</b> before I can Upgrade your database to the next version!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }

  # call the upgrade procedure for the specified version
  my $function = $updateFunctions->{$nextVersion};
  my @results;  # return results must be [0] = success/failure, [1] = doc, [2] = doc modified, [3] = success message
  eval "\@results = \$self->$function(doc => \$doc, mainDBObj => \$mainDBObj, ledgerDBObj => \$ledgerDBObj);";
  if ($@)
  {
    $self->setError(errorString => "Eval of Upgrade Function = '$function' failed!<br>\nError = '$@'.<br>\n");
    return undef;
  }

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return undef;
  }
  if ($results[2])
  {
    $doc = $results[1];  # return the updated version of $doc.  This is how user interaction is provided for.
  }
  else
  {
    # see if we have another version to upgrade or if we are done.
    # find nextVersion in the array.
    my $index;
    my $count;
    my $found = 0;
    for ($index=0; $index < scalar @{$currentVersion} && !$found; $index++)
    {
      if ($currentVersion->[$index] == $nextVersion)
      {
        $found = 1;
      }
    }
    if ($found)
    {
      $count = scalar(@{$currentVersion}) - $index;  # This tells me how many there are between it and the end of the array (number of upgrades to perform).
    }
    else
    {
      if ($nextVersion == $currDBVersion)
      {
        $count = 0;
      }
      else
      {
        # unknown version!
        $self->setError(errorString => "Current Database Version = '$nextVersion' is unknown!<br>\nSee the System Administrator!<br>\n");
        return undef;
      }
    }

    if ($count > 0)
    {
      $nextVersion = $currentVersion->[$index];  # get the next version to upgrade to.
      $args{dbVersion} = $nextVersion;
      $doc = $self->displayPhaseScreen(%args, statusStr => $results[3]);
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return undef;
      }
      return $doc;
    }
    else
    {
      # generate a page that indicates we are done and provides a link back to the main menu.
      $doc->setTitle("DBUpgrade:  Completed");
      $doc->setFocus("body");
      $doc->print(<<"END_OF_CODE");
<center>
<h1>Database Upgrade Completed!</h1>
<br>
<font color="red">Status:</font> $results[3]
<br>
<br>
</center>
<a href="$url?$self->{sessionObj}->{store}->{originalDestination}" onMouseOver="window.status='Original Destination'" target="_top">Continue</a>
END_OF_CODE
      $doc->setLinkDecorations; #/

      # now unlock the app.
      $self->{applicationObj}->unlockApp(app => $self->{appObj}->{id}, type => "c", company => $self->{companyObj}->{id});
      if ($self->{applicationObj}->didErrorOccur)
      {
        $self->setError(errorString => $self->{applicationObj}->errorMessage);
        return undef;
      }
    }
  }

  return $doc;
}

=item HTMLObject processUserInput(dbVersion, mainDBObj, ledgerDBObj, @currentVersion,
             %userInputFunctions, doc)

    actually validates and then processes any user input needed for the
    upgrade for the specified version.  Returns HTMLObject if more user
    input is required, or undef if done with the user input.

=cut
sub processUserInput
{
  my $self = shift;
  my %args = (dbVersion => "", mainDBObj => undef, ledgerDBObj => undef, currentVersion => undef, userInputFunctions => undef, doc => undef, @_ );
  my $dbVersion = $args{dbVersion};
  my $mainDBObj = $args{mainDBObj};
  my $ledgerDBObj = $args{ledgerDBObj};
  my $currentVersion = $args{currentVersion};
  my $userInputFunctions = $args{userInputFunctions};
  my $doc = $args{doc};

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  my $confirmUpgrade = $self->{input}->{confirmUpgrade};
  my $nextVersion = $self->{input}->{nextVersion};

  # set dbVersion into the input hash for children functions to get to.
  $self->{input}->{dbVersion} = $dbVersion;

  # make sure we have the input we need.
  if ($confirmUpgrade !~ /^YES|NO$/)
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "confirmUpgrade = '$confirmUpgrade' is invalid!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }
  if ($nextVersion !~ /^\d+\.\d+$/ || !exists $userInputFunctions->{$nextVersion})
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "nextVersion = '$nextVersion' is invalid!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }

  if ($confirmUpgrade eq "NO")
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "You must specify <b>Yes</b> before I can Upgrade your database to the next version!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }

  # call the user input procedure for the specified version
  my $function = $userInputFunctions->{$nextVersion};
  my @results;  # return results must be [0] = success/failure, [1] = doc, [2] = doc modified, [3] = success message
  eval "\@results = \$self->$function(doc => \$doc, mainDBObj => \$mainDBObj, ledgerDBObj => \$ledgerDBObj);";
  if ($@)
  {
    $self->setError(errorString => "Eval of User Input Function = '$function' failed!<br>\nError = '$@'.<br>\n");
    return undef;
  }

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return undef;
  }
  if ($results[2])
  {
    $doc = $results[1];  # return the updated version of $doc.  This is how user interaction is provided for.
  }
  elsif ($results[0])  # we successfully finished whatever we were doing.
  {
    # see if we have another version to upgrade or if we are done.
    # find nextVersion in the array.
    my $index;
    my $count;
    my $found = 0;
    for ($index=0; $index < scalar @{$currentVersion} && !$found; $index++)
    {
      if ($currentVersion->[$index] == $nextVersion)
      {
        $found = 1;
      }
    }
    if ($found)
    {
      $count = scalar(@{$currentVersion}) - $index;  # This tells me how many there are between it and the end of the array (number of upgrades to perform).
    }
    else
    {
      if ($nextVersion == $currDBVersion)
      {
        $count = 0;
      }
      else
      {
        # unknown version!
        $self->setError(errorString => "Current Database Version = '$nextVersion' is unknown!<br>\nSee the System Administrator!<br>\n");
        return undef;
      }
    }

    if ($count > 0)
    {
      $nextVersion = $currentVersion->[$index];  # get the next version to upgrade to.
      $args{dbVersion} = $nextVersion;
      $doc = $self->displayPhaseScreen(%args, statusStr => $results[3]);
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return undef;
      }
      return $doc;
    }
    else
    {
      # generate a page that indicates we are done and provides a link back to the main menu.
      $doc->setTitle("DBUpgrade:  Completed");
      $doc->setFocus("body");
      $doc->print(<<"END_OF_CODE");
<center>
<h1>Database Upgrade Completed!</h1>
<br>
<font color="red">Status:</font> $results[3]
<br>
<br>
</center>
<a href="$url?$self->{sessionObj}->{store}->{originalDestination}" onMouseOver="window.status='Original Destination'" target="_top">Continue</a>
END_OF_CODE
      $doc->setLinkDecorations;

      # now unlock the app.
      $self->{applicationObj}->unlockApp(app => $self->{appObj}->{id}, type => "c", company => $self->{companyObj}->{id});
      if ($self->{applicationObj}->didErrorOccur)
      {
        $self->setError(errorString => $self->{applicationObj}->errorMessage);
        return undef;
      }
    }
  }

  return $doc;
}

# version specific upgrade subroutines here....

=item @result userInputToOneThree(doc, mainDBObj)

    @result: [0]=success/failure (1/0), [1]=doc, [2]=doc modified (0/1),
             [3]=status message (if success)
    This migrates Stock SubAccount entries in the database from version
    1.2 to 1.3.  The user is prompted to fill out the info for each stock
    subaccount encountered.  This is used by the Investment and Portfolio
    code.

=cut
sub userInputToOneThree
{
  my $self = shift;
  my %args = ( doc => undef, mainDBObj => undef, @_ );
  my $doc = $args{doc};
  my $mainDBObj = $args{mainDBObj};
  my $success = 0;
  my $modified = 0;
  my $statusStr = "Successfully setup Stock Info for DB Version <b>1.2</b> to <b>1.3</b>.\n<br>\nCreated <b>COUNT</b> Stock Entries.";
  my @result = ($success, $doc, $modified, $statusStr);
  my $dbVersion = $self->{input}->{dbVersion};

  if (!defined $doc)
  {
    $self->setError(errorString => "doc is not defined!<br>\n");
    return @result;
  }

  # validate we are at version 1.3 before we do anything.
  my $currentVersion = $mainDBObj->getDBVersion;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return @result;
  }

  if ($currentVersion != 1.3)
  {
    $self->setError(errorString => "The database is not at version = <b>1.3</b>!<br>\nIt is currently at version <b>$currentVersion</b>.<br>\nI can not migrate the Stock Sub Accounts!<br>\n");
    return @result;
  }

  # determine if we are creating an entry or displaying the creation screen.
  my $mode = $self->{input}->{mode};
  my $stocks = $self->{input}->{stocks};
  my $currentStock = $self->{input}->{currentStock};

  if (length $stocks == 0)
  {
    $self->setError(errorString => "stocks must be defined to work with!<br>\n");
    return @result;
  }
  if ($currentStock !~ /^\d+$/)
  {
    $self->setError(errorString => "currentStock = '$currentStock' is invalid!<br>\n");
    return @result;
  }

  if ($mode eq "display")
  {
    $doc = $self->displayStockCreationScreen(doc => $doc, mainDBObj => $mainDBObj);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return @result;
    }
    $result[1] = $doc;
    $result[2] = 1;
  }
  elsif ($mode eq "createStock")
  {
    # we are creating the entry.
    # validate we have all the data we need.

    my $stockName = $self->{input}->{stockName};
    my $ticker = $self->{input}->{ticker};
    my $market = $self->{input}->{market};

    # get list of markets we are using for stocks.
    my @markets = $mainDBObj->getStockMarkets;
    if ($mainDBObj->didErrorOccur)
    {
      $self->setError(errorString => $mainDBObj->errorMessage);
      return @result;
    }

    if (length $stockName == 0)
    {
      $doc = $self->displayStockCreationScreen(doc => $doc, mainDBObj => $mainDBObj, errorString => "stockName = '$stockName' is invalid!<br>\n");
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return @result;
      }
      $result[1] = $doc;
      $result[2] = 1;
      return @result;
    }
    if (length $ticker > 10)
    {
      $doc = $self->displayStockCreationScreen(doc => $doc, mainDBObj => $mainDBObj, errorString => "ticker = '$ticker' is invalid!<br>\n");
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return @result;
      }
      $result[1] = $doc;
      $result[2] = 1;
      return @result;
    }
    my $found = 0;
    foreach my $tmpMarket (@markets)
    {
      if ($market =~ /^$tmpMarket$/)
      {
        $found = 1;
      }
    }
    if (!$found)
    {
      $doc = $self->displayStockCreationScreen(doc => $doc, mainDBObj => $mainDBObj, errorString => "market = '$market' is invalid!<br>\n");
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return @result;
      }
      $result[1] = $doc;
      $result[2] = 1;
      return @result;
    }

    my $stockObj = Portal::Accounting::Objects::StockObject->new(ticker => $ticker, name => $stockName,
                   number => $currentStock, market => $market, langObj => $self->{langObj});
    if ($stockObj->didErrorOccur)
    {
      $self->setError(errorString => $stockObj->errorMessage);
      return @result;
    }

    my @tmpResult = $mainDBObj->createStock(stockObj => $stockObj);
    my $result = $tmpResult[0];  # get the status value.
    if ($mainDBObj->didErrorOccur)
    {
      $self->setError(errorString => $mainDBObj->errorMessage);
      return @result;
    }
    if ($result == -1)
    {
      $doc = $self->displayStockCreationScreen(doc => $doc, mainDBObj => $mainDBObj, errorString => "Stock already exists with this <b>Ticker</b>, <b>Name</b>, <b>SubAccount Number</b> combination!");
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return @result;
      }
      $result[1] = $doc;
      $result[2] = 1;
      return @result;
    }
    elsif ($result == -2)
    {
      $doc = $self->displayStockCreationScreen(doc => $doc, mainDBObj => $mainDBObj, errorString => "Stock already exists with this <b>Name</b>!");
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return @result;
      }
      $result[1] = $doc;
      $result[2] = 1;
      return @result;
    }
    elsif ($result == -3)
    {
      $doc = $self->displayStockCreationScreen(doc => $doc, mainDBObj => $mainDBObj, errorString => "Stock already exists with this <b>SubAccount Number</b>!");
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return @result;
      }
      $result[1] = $doc;
      $result[2] = 1;
      return @result;
    }
    elsif ($result == -4)
    {
      $doc = $self->displayStockCreationScreen(doc => $doc, mainDBObj => $mainDBObj, errorString => "Stock already exists with this <b>Ticker</b>!");
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return @result;
      }
      $result[1] = $doc;
      $result[2] = 1;
      return @result;
    }
    else  # it succeeded.
    {
      # see if we have anymore stocks to create.
      my @stocks = split /\|/, $stocks;
      my $count = 0;
      my $location = -1;
      foreach my $stock (@stocks)
      {
        (my $number, my $name) = split /,/, $stock;
        if ($number == $currentStock)
        {
          $location = $count;
        }
        $count++;
      }
      if ($location == -1)
      {
        $self->setError(errorString => "currentStock = '$currentStock' was not found in the stocks string ($stocks)! &nbsp;This was after we just created this stock!<br>\n");
        return @result;
      }
      if ($location < $count - 1) # we have more stocks to create.
      {
        my $stock = $stocks[$location+1];
        ($stock, my $temp) = split /,/, $stock;
        $self->{input}->{currentStock} = $stock;  # point to the next stock to work with.
        $doc = $self->displayStockCreationScreen(doc => $doc, mainDBObj => $mainDBObj, statusString => "Stock successfully created!");
        if ($self->didErrorOccur)
        {
          $self->prefixError();
          return @result;
        }
        $result[1] = $doc;
        $result[2] = 1;
      }
      else  # we are done creating stocks.
      {
        $statusStr =~ s/(COUNT)/$count/m;
        $result[3] = $statusStr;
        $result[0] = 1;  # we are done
        $result[2] = 0;
      }
    }
  }
  else
  {
    $self->setError(errorString => "mode = '$mode' is invalid!<br>\n");
  }

  return @result;
}

=item HTMLObject displayStockCreationScreen(doc, mainDBObj, [errorString,
             statusString])

    This creates the data gathering screen for creating a Stock.  Used
    in the 1.2 to 1.3 migration (User Interaction process).

=cut
sub displayStockCreationScreen
{
  my $self = shift;
  my %args = ( doc => undef, mainDBObj => undef, errorString => "", statusString => "", @_ );
  my $doc = $args{doc};
  my $mainDBObj = $args{mainDBObj};
  my $errorString = $args{errorString};
  my $statusString = $args{statusString};
  my $currentStock = $self->{input}->{currentStock};
  my $stocks = $self->{input}->{stocks};
  my $dbVersion = $self->{input}->{dbVersion};
  
  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  if (!defined $doc)
  {
    $self->setError(errorString => "doc is not defined!<br>\n");
    return undef;
  }

  if (length $errorString > 0)
  {
    $errorString = "<font color=\"red\">Error:</font> &nbsp;$errorString\n<br>\n";
  }
  if (length $statusString > 0)
  {
    $statusString = "<font color=\"red\">Status:</font> &nbsp;$statusString\n<br>\n";
  }

  # we are displaying the creation screen.
  # now get the list of stock sub accounts and the one we are to currently work with and then display the create Stock Entry form.

  # get the info for this Sub Account (use DB::Main)
  my $subAccountObj = $mainDBObj->getSubAccountInfo(number => $currentStock);
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return undef;
  }
  # get list of markets we are using for stocks.
  my @markets = $mainDBObj->getStockMarkets;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return undef;
  }

  my $ticker = (exists $self->{input}->{ticker} ? $self->{input}->{ticker} : $subAccountObj->{name});
  my $stockName = (exists $self->{input}->{stockName} ? $self->{input}->{stockName} : $subAccountObj->{name});
  my $market = $self->{input}->{market};

  if (length $statusString > 0)
  {
    $ticker = $stockName = $subAccountObj->{name};
    $market = "";  #empty these values as we are starting another stock!
  }

  my $marketString = "";
  foreach my $tmpMarket (@markets)
  {
    $marketString .= "          <option value=\"$tmpMarket\"" . ($market eq "$tmpMarket" ? " selected" : "") . ">$tmpMarket\n";
  }

  $doc->setTitle("DBUpgrade: Bottom Frame - Create Stock");
  $doc->setFocus("body");
  $doc->print(<<"END_OF_CODE");
<center>
<h1>Create Stock</h1>
<br>
$statusString
$errorString
<br>
<form method="post" name="upgradeForm" action="$url">
  <input type="hidden" name="app" value="Accounting">
  <input type="hidden" name="state" value="DBUpgrade">
  <input type="hidden" name="command" value="userinput">
  <input type="hidden" name="nextVersion" value="1.2">
  <input type="hidden" name="confirmUpgrade" value="YES">
  <input type="hidden" name="stocks" value="$stocks">
  <input type="hidden" name="currentStock" value="$currentStock">
  <input type="hidden" name="mode" value="createStock">
  <table border="0" cellpadding="0" cellspacing="0" bgcolor="indianred">
    <tr>
      <td><font color="white">Sub Account Name</font></td>
      <td>$subAccountObj->{name}</td>
    </tr>
    <tr>
      <td><font color="white">Sub Account Number</font></td>
      <td>$subAccountObj->{number}</td>
    </tr>
    <tr>
      <td><font color="white">Sub Account Type</font></td>
      <td>Stock (70)</td>
    </tr>
    <tr>
      <td colspan="2"><hr noshade></td>
    </tr>
    <tr>
      <td><font color="white">Ticker</font></td>
      <td><input type="text" name="ticker" value="$ticker" size="10" maxlength="10"></td>
    </tr>
    <tr>
      <td><font color="white">Stock Name</font></td>
      <td><input type="text" name="stockName" value="$stockName">&nbsp;</td>
    </tr>
    <tr>
      <td><font color="white">Market</font></td>
      <td>
        <select name="market">
$marketString
        </select>
      </td>
    </tr>
  </table>
  <input type="submit" name="createStock" value="Create Stock">&nbsp;<input type="reset" value="Reset">
</form>
END_OF_CODE

  return $doc;
}

=item @result updateToOneThree(doc, mainDBObj)

    @result: [0]=success/failure (1/0), [1]=doc, [2]=doc modified (0/1),
             [3]=status message (if success)
    This updates the database from version 1.2 to 1.3.

=cut
sub updateToOneThree
{
  my $self = shift;
  my %args = ( doc => undef, mainDBObj => undef, @_ );
  my $doc = $args{doc};
  my $mainDBObj = $args{mainDBObj};
  my $success = 0;
  my $modified = 0;
  my $statusStr = "Successfully updated from <b>1.2</b> to <b>1.3</b>.\n<br>\nMigrated <b>COUNT</b> Journal Entries over. &nbsp;Please check the validity of your data.";
  my @result = ($success, $doc, $modified, $statusStr);
  my $dbVersion = $self->{input}->{dbVersion};

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  if (!defined $doc)
  {
    $self->setError(errorString => "doc is not defined!<br>\n");
    return @result;
  }

  # validate we are still at version 1.2 before we do anything.
  my $currentVersion = $mainDBObj->getDBVersion;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return @result;
  }

  if ($currentVersion != 1.2)
  {
    $self->setError(errorString => "The database is not at version = <b>1.2</b>!<br>\nI can not upgrade to <b>1.3</b>!<br>\n");
    return @result;
  }

  # modify the default value for the version table.
  $self->{dbHandle}->write(sql => "ALTER TABLE version_tb ALTER COLUMN version SET DEFAULT '1.3'");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }

  my @sequences = ( "CREATE SEQUENCE \"sec_event_tb_event_ref_seq\" start 1 increment 1 maxvalue 2147483647 minvalue 1  cache 1",
  		    "CREATE SEQUENCE \"lot_node_tb_node_id_seq\" start 1 increment 1 maxvalue 2147483647 minvalue 1  cache 1",
  		    "CREATE SEQUENCE \"new_lot_number_seq\" start 1 increment 1 maxvalue 2147483647 minvalue 1  cache 1",
  		    "CREATE SEQUENCE \"journal_entry_tb_ref_num_seq\" start 1 increment 1 maxvalue 2147483647 minvalue 1  cache 1"
  		  );
  # create the new sequences we need.
  foreach my $sql (@sequences)
  {
    $self->{dbHandle}->write(sql => $sql);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
  }

  my @newTables = ( "DROP TABLE sec_event_tb",
                    "CREATE TABLE sec_event_tb ( event_code character varying(4), master_ref int4, calc_date date, trans_date date, counter int4, sub_e_code text, a text, s text, l text, os numeric(30,6), ns numeric(30,6), cc numeric(30,6), ledger_code text, event_ref int4 NOT NULL, PRIMARY KEY (event_ref) )",
  		    "CREATE TABLE lot_node_tb ( ledger_code text, event_date date, parent_id int4, event_ref int4, journal_ref int4, sub_e_code text, account_num int4, sub_account_num int4, lot int4, shares numeric(30,6), cost numeric(15,2), node_id int4 DEFAULT nextval('lot_node_tb_node_id_seq'::text) NOT NULL )",
  		    "CREATE TABLE journal_entry_temp_tb ( ref_num int4 DEFAULT nextval('journal_entry_tb_ref_num_seq'::text) NOT NULL, journal_code text, trans_date date, payee_payer text, memo text, remember bool, ledger_code text )",
  		    "CREATE TABLE debitcredit_temp_tb ( parent_num int4, trans_date date, memo text, sub_e_code text, account_num int4, sub_account_num int4, amount numeric(15,2), dorc bool, cleared character, counter int4, ledger_code text )",
  		    "CREATE TABLE qif_class_tb ( name text, description text, map_sub_entity text )",
  		    "CREATE TABLE qif_category_tb ( name text, description text, tax_related_item bool, tax_ref_num text, income_expense_flag bool, map_account text, map_sub_account text )",
  		    "CREATE TABLE qif_account_tb ( name text, type text, sales_tax text, description text, credit_card_limit text, balance text, map_account text )",
  		    "CREATE TABLE qif_security_tb ( name text, symbol text, type text, goal text, map_sub_account text )",
  		    "CREATE TABLE qif_prices_tb ( price_symbol text, price_amount text, price_date date )",
  		    "CREATE TABLE qif_investment_list_tb ( trans_date date, action text, security text, price text, quantity text, cleared text, description text, memo text, commission text, category text, category_class text, transfer_account text, transfer_class text, amount text, transfer_amount text, account text )",
		    "CREATE TABLE qif_log_tb ( message text )",
		    "DROP TABLE stocks_tb",
		    "CREATE TABLE stocks_tb (ticker text, name text, sub_acct int4 PRIMARY KEY, type char, market text)",
		    "DROP TABLE stock_price_tb",
		    "CREATE TABLE stock_price_tb (ticker text, date_of date, price numeric(15,7), sub_acct int4)",
		    "CREATE TABLE stock_market_tb (name text PRIMARY KEY)",
		    "INSERT INTO stock_market_tb (name) VALUES ('NYSE')",
		    "INSERT INTO stock_market_tb (name) VALUES ('NASDAQ')",
		    "INSERT INTO stock_market_tb (name) VALUES ('AMEX')",
		    "INSERT INTO stock_market_tb (name) VALUES ('None')",
                  );
  # create all new tables needed.
  foreach my $sql (@newTables)
  {
    $self->{dbHandle}->write(sql => $sql);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
  }

  # massage the old journal_entry_tb and debitcredit_tb entries into using the new sequences.

  # first gather all journal_entry_tb entries sorted by ledger_code and ref_num
  my $sth = $self->{dbHandle}->read(sql => "SELECT ref_num, journal_code, trans_date, payee_payer, memo, remember, ledger_code FROM journal_entry_tb ORDER BY ledger_code, ref_num");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }
  my $count = 0;
  my %refNumEntries = ();
  while (my @info = $sth->fetchrow_array)
  {
    $count++;  # keep track of the total number of entries massaged.
    # get the next ref_num from the database sequence
    my $sth2 = $self->{dbHandle}->read(sql => "SELECT nextval('journal_entry_tb_ref_num_seq')");
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
    my @info2 = $sth2->fetchrow_array;
    if (!defined $info2[0])
    {
      $self->setError(errorString => "The next ref_num was not returned from the database!<br>\n");
      $self->{dbHandle}->rollback;
      return @result;
    }
    my $ref_num = $info2[0];
    my $index = "$info[6]" . "-" . "$info[0]";
    $refNumEntries{$index} = $ref_num;

    # insert this entry into the journal_entry_temp_tb.
    my $rc = $self->{dbHandle}->write(sql => "INSERT INTO journal_entry_temp_tb (ref_num, journal_code, trans_date, payee_payer, memo, remember, ledger_code) VALUES (?, ?, ?, ?, ?, ?, ?)",
                                      plug => [ $ref_num, $info[1], $info[2], $info[3], $info[4], $info[5], $info[6] ]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
    if ($rc == 0)
    {
      $self->setError(errorString => "The insert into journal_entry_temp_tb appeared to fail for ref_num = '$info[0]', ledger_code = '$info[6]'!<br>\n");
      $self->{dbHandle}->rollback;
      return @result;
    }
  }

  # now we migrate all debitcredit_tb entries using the refNumEntries hash as our guide.
  my $sth2 = $self->{dbHandle}->read(sql => "SELECT parent_num, trans_date, memo, sub_e_code, account_num, sub_account_num, amount, dorc, cleared, counter, ledger_code FROM debitcredit_tb ORDER BY ledger_code, parent_num, counter");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }
  while (my @info3 = $sth2->fetchrow_array)
  {
    my $index = "$info3[10]" . "-" . "$info3[0]";
    my $ref_num = $refNumEntries{$index};
    my $rc = $self->{dbHandle}->write(sql => "INSERT INTO debitcredit_temp_tb (parent_num, trans_date, memo, sub_e_code, account_num, sub_account_num, amount, dorc, cleared, counter, ledger_code) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                                      plug => [ $ref_num, $info3[1], $info3[2], $info3[3], $info3[4], $info3[5], $info3[6], $info3[7], $info3[8], $info3[9], $info3[10] ]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
    if ($rc == 0)
    {
      $self->setError(errorString => "The insert into debitcredit_temp_tb appeared to fail for old parent_num = '$info3[0]', new parent_num = '$ref_num', counter = '$info3[9]', ledger_code = '$info3[10]'!<br>\n");
      $self->{dbHandle}->rollback;
      return @result;
    }
  }

  # now update the result message to indicate how many journal_entry_tb entries were converted.
  $result[3] =~ s/COUNT/$count/;

  # drop journal_entry_tb and debitcredit_tb
  my @dropTables = ( "DROP TABLE journal_entry_tb",
                     "DROP TABLE debitcredit_tb"
                   );
  foreach my $sql (@dropTables)
  {
    $self->{dbHandle}->write(sql => $sql);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
  }

  # rename journal_entry_temp_tb to journal_entry_tb and debitcredit_temp_tb to debitcredit_tb
  my @renameTables = ( "ALTER TABLE journal_entry_temp_tb RENAME TO journal_entry_tb",
                       "ALTER TABLE debitcredit_temp_tb RENAME TO debitcredit_tb",
                     );
  foreach my $sql (@renameTables)
  {
    $self->{dbHandle}->write(sql => $sql);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
  }

  # create new indexes
  my @newIndexes = ( "CREATE UNIQUE INDEX lot_node_tb_node_id_key on lot_node_tb using btree ( node_id int4_ops )",
                     "CREATE UNIQUE INDEX journal_entry_tb_ref_num_key on journal_entry_tb using btree ( ref_num int4_ops )"
                   );

  foreach my $sql (@newIndexes)
  {
    $self->{dbHandle}->write(sql => $sql);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
  }

  my $sth3 = $self->{dbHandle}->read(sql => "SELECT name FROM sub_acct_type_tb WHERE num = 5");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }
  my @info4 = $sth3->fetchrow_array;
  if (!defined $info4[0])  # it doesn't exist, so create it.
  {
    $self->{dbHandle}->write(sql => "INSERT INTO sub_acct_type_tb (name, num) VALUES ('Asset', 5)");
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
  }

  # change the version in the version table
  $self->{dbHandle}->write(sql => "UPDATE version_tb SET version = '1.3'");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }

  # commit the work.
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }

  # now see if we have to create any Stock entries.
  my $sth = $self->{dbHandle}->read(sql => "SELECT sub_acc_number, sub_acc_name FROM sub_acc_tb WHERE type = 70 ORDER BY sub_acc_name");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }
  my @stocks = @{$sth->fetchall_arrayref};
  if (scalar @stocks > 0)
  {
    my $numStocks = scalar @stocks;
    my $firstStock = "";
    my $stocks = "";
    foreach my $stock (@stocks)
    {
      $stocks .= "|" if (length $stocks > 0);
      $stocks .= join (",", @{$stock});
    }
    qq/$stocks/;  # escape all escapable stuff in the string.

    $firstStock = $stocks[0][0];  # this is the first Stock to work with.

    $doc->setTitle("DBUpgrade: Bottom Frame - User Interaction Required");
    $doc->setFocus("body");
    $doc->print(<<"END_OF_CODE");
<center>
<h1>Database Upgrade In <font color="red">Progress</font></h1>
</center>
END_OF_CODE
    $doc->print("<font color=\"red\">Status:</font> $result[3]\n<br>\n");
    $doc->print(<<"END_OF_CODE");
<center>
<form method="post" name="upgradeForm" action="$url">
  <input type="hidden" name="app" value="Accounting">
  <input type="hidden" name="state" value="DBUpgrade">
  <input type="hidden" name="command" value="userinput">
  <input type="hidden" name="nextVersion" value="$dbVersion">
  <input type="hidden" name="confirmUpgrade" value="YES">
  <input type="hidden" name="stocks" value="$stocks">
  <input type="hidden" name="currentStock" value="$firstStock">
  <input type="hidden" name="mode" value="display">
  <br>
  <br>
  <h2>You must now create a Stock Entry for each Stock SubAccount!</h2>
  <br>
  There are a total of <b>$numStocks</b> Stocks to Create.
</form>
END_OF_CODE

    $doc->setOnload(code => "document.upgradeForm.submit();");

    $result[1] = $doc;  # make sure they know we want to have user input!
    $result[2] = 1;
  }
  else  # nothing more to do
  {
    # signal success.
    $result[0] = 1;
  }

  return @result;
}

# updateToOneFour
# requires: doc, mainDBObj
# optional:
# returns: array: [0]=success/failure (1/0), [1]=doc, [2]=doc modified (0/1), [3]=status message (if success)
# summary: This updates the database from version 1.3 to 1.4.
sub updateToOneFour
{
  my $self = shift;
  my %args = ( doc => undef, mainDBObj => undef, @_ );
  my $doc = $args{doc};
  my $mainDBObj = $args{mainDBObj};
  my $success = 0;
  my $modified = 0;
  my $statusStr = "Successfully updated from <b>1.3</b> to <b>1.4</b>.\n";
  my @result = ($success, $doc, $modified, $statusStr);
  my $dbVersion = $self->{input}->{dbVersion};

  if (!defined $doc)
  {
    $self->setError(errorString => "doc is not defined!<br>\n");
    return @result;
  }

  # validate we are still at version 1.3 before we do anything.
  my $currentVersion = $mainDBObj->getDBVersion;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return @result;
  }

  if ($currentVersion != 1.3)
  {
    $self->setError(errorString => "The database is not at version = <b>1.3</b>!<br>\nI can not upgrade to <b>1.4</b>!<br>\n");
    return @result;
  }

  # modify the default value for the version table.
  $self->{dbHandle}->write(sql => "ALTER TABLE version_tb ALTER COLUMN version SET DEFAULT '1.4'");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }

  my @sequences = ( "CREATE TABLE lot_parent_tb ( \"lot\" int4 NOT NULL UNIQUE, \"buy_lot\" int4 )",
  		    "ALTER TABLE lot_node_tb ADD COLUMN buy_lot int4",
  		    "ALTER TABLE acct_bal_tb ADD COLUMN sub_e_code text",
  		    "ALTER TABLE beginning_balance_entry_tb ADD COLUMN sub_e_code text",
  		    "ALTER TABLE qif_class_tb ADD COLUMN used bool",
  		    "ALTER TABLE qif_category_tb ADD COLUMN used bool",
  		    "ALTER TABLE qif_account_tb ADD COLUMN used bool",
  		    "ALTER TABLE qif_security_tb ADD COLUMN used bool",
  		    "DROP TABLE qif_investment_list_tb",
  		    "DROP TABLE qif_prices_tb",
  		    "CREATE FUNCTION capital_gain_loss(date, date) RETURNS text AS 'SELECT CASE WHEN ((\$1::date + ''1 year''::interval)::date > \$2::date) THEN ''0''::text ELSE ''1''::text END;' LANGUAGE 'sql'",
          "ALTER TABLE sec_event_tb DROP CONSTRAINT sec_event_tb_pkey",
          "DROP TABLE asset_tb",
          "CREATE TABLE \"asset_tb\" (\"purchase_date\" date,\"salvage_date\" date,\"purchase_price\" numeric(15,2),\"salvage_price\" numeric(15,2),\"sub_entity\" text,\"dep_acct\" int4,\"contra_acct\" int4,\"sub_acct\" int4,\"ledger_code\" text,\"depr_method\" text,\"depr_life\" int4,\"depr_increment\" text)",
          "CREATE FUNCTION \"get_asset_data\" (int4,int4,int4,bool,bool) RETURNS numeric AS 'SELECT CASE WHEN \$5 = true THEN (SELECT amount FROM debitcredit_tb WHERE parent_num = \$1 AND account_num <> \$2 AND sub_account_num = \$3 AND dorc = \$4)
          ELSE (SELECT amount FROM debitcredit_tb WHERE parent_num = \$1 AND account_num = \$2 AND sub_account_num = \$3 AND dorc = \$4) end;' LANGUAGE 'SQL'",
          "CREATE FUNCTION \"get_shares_data\" (text,int4) RETURNS numeric AS 'SELECT ns FROM sec_event_tb WHERE event_code = 20 AND l = \$1 AND event_ref = (SELECT event_ref FROM lot_node_tb WHERE journal_ref = \$2);' LANGUAGE 'SQL'",
          "CREATE FUNCTION \"dorcify\" (numeric,bool,bool) RETURNS numeric AS 'SELECT CASE WHEN \$2 = true AND \$3 = true THEN \$1
          WHEN \$2 = false AND \$3 = true THEN 0::numeric - \$1
          WHEN \$2 = true AND \$3 = false THEN 0::numeric - \$1
          WHEN \$2 = false AND \$3 = false THEN \$1 end;' LANGUAGE 'SQL'",
          "CREATE FUNCTION \"stock_purchased\" (int4) RETURNS date AS 'SELECT event_date FROM lot_node_tb WHERE lot = (SELECT buy_lot FROM lot_parent_tb WHERE lot = \$1) AND parent_id = -1;' LANGUAGE 'SQL'",
          "CREATE FUNCTION \"asset_purchased\" (int4,text) RETURNS date AS 'SELECT purchase_date FROM asset_tb WHERE ledger_code = \$2 AND sub_acct = \$1;' LANGUAGE 'SQL'",
          "INSERT INTO sub_acct_type_tb (name, num) VALUES ('Asset', '5')",
          "INSERT INTO sub_acct_type_tb (name, num) VALUES ('Bond', '7')",
          "INSERT INTO sub_acct_type_tb (name, num) VALUES ('CD', '9')",
          "INSERT INTO acct_type_tb (name, num, debit_increases, first_digit, sub_acct_type) VALUES ('Operating Revenue', 45, 'f', '4', '1,15,50')",
          "INSERT INTO acct_type_tb (name, num, debit_increases, first_digit, sub_acct_type) VALUES ('Operating Expenses', 51, 't', '5', '1,15,20,50')",
          "INSERT INTO acct_type_tb (name, num, debit_increases, first_digit, sub_acct_type) VALUES ('Admin Expenses', 52, 't', '5', '1,15,20,50')",
          "INSERT INTO acct_type_tb (name, num, debit_increases, first_digit, sub_acct_type) VALUES ('Other Payroll',58,'t','5','1,15,20,90')",
          "INSERT INTO acct_type_tb (name, num, debit_increases, first_digit, sub_acct_type) VALUES ('Interest Expense',59,'t','5','1,5,7,9,15,50,70')", 
          "UPDATE acct_type_tb SET sub_acct_type = '1,15,5,9,50' WHERE num = '2'",
          "UPDATE acct_type_tb SET sub_acct_type = '1,15,5' WHERE num = '4'",
          "UPDATE acct_type_tb SET sub_acct_type = '1,15,5,50' WHERE num = '16'",
          "UPDATE acct_type_tb SET sub_acct_type = '1,15,7,70' WHERE num = '3'",
          "UPDATE acct_type_tb SET sub_acct_type = '1,15,100,20,50' WHERE num = '5'",
          "UPDATE acct_type_tb SET sub_acct_type = '1,15,10,20' WHERE num = '111'",
          "UPDATE acct_type_tb SET sub_acct_type = '1,15,50,70' WHERE num = '9'",
          "UPDATE acct_type_tb SET sub_acct_type = '1,15,20,50,70' WHERE num = '13'",
          "DROP TABLE journal_tb",
          "CREATE TABLE \"journal_tb\" (\"name\" text, \"code\" text)",
          "INSERT INTO \"journal_tb\" VALUES ('General','G')",
          "INSERT INTO \"journal_tb\" VALUES ('Payroll','PAY')",
          "INSERT INTO \"journal_tb\" VALUES ('Investment','IS')",
          "INSERT INTO \"journal_tb\" VALUES ('Asset','A')",
          "INSERT INTO \"journal_tb\" VALUES ('Sales','S')",
          "INSERT INTO \"journal_tb\" VALUES ('Transfers','TX')",
          "INSERT INTO \"journal_tb\" VALUES ('Cash Purchases','CP')",
          "INSERT INTO \"journal_tb\" VALUES ('Cash Receipts','CSR')",
          "INSERT INTO \"journal_tb\" VALUES ('Credit Purchases','CDP')",
          "INSERT INTO \"journal_tb\" VALUES ('Credit Receipts','CDR')",
          "INSERT INTO \"journal_tb\" VALUES ('Inventory','IY')"
         		  );
  # create the new sequences we need.
  foreach my $sql (@sequences)
  {
    $self->{dbHandle}->write(sql => $sql);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
  }

  # index any existing lot nodes and determine their buy lots
  # retrieve all buy lot nodes from the lot_node_tb table
  my $sth = $self->{dbHandle}->read(sql => "SELECT lot, node_id FROM lot_node_tb WHERE parent_id = '-1'");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $result[0] = 0;
    $result[3] = $self->{dbHandle}->errorMessage;
    return @result;
  }
  # for each "buy" lot, retrieve the tree of nodes for that buy and update each node to have the lot for its buy lot
  while (my @info = $sth->fetchrow_array)
  {
    my @tree = $self->getNodeTree(nodeID => $info[1], mainDBObj => $self);
    if (not defined $tree[0])
    {
      $result[0] = 0;
      $result[3] = "Error retrieving lot node tree!<br>\n" . $self->{dbHandle}->errorMessage;
      return @result;
    }
    # make the first node be the "buy" node, so that it will be updated also
    unshift @tree, [ ($info[1], $info[0]) ];
    foreach my $nodeID (@tree)
    {
      if ($nodeID->[1] ne "")
      {
        # update the lot node with the lot number of the "buy" lot
        $self->{dbHandle}->write(sql => "UPDATE lot_node_tb SET buy_lot = ? WHERE node_id = ?", plug => [ $info[0], $nodeID->[0] ]);
        if ($self->{dbHandle}->didErrorOccur())
        {
          $result[0] = 0;
          $result[3] = $self->{dbHandle}->errorMessage;
          return @result;
        }
        # check to make sure the lot_parent_tb table does not have an entry with the current values
        my $sth2 = $self->{dbHandle}->read(sql => "SELECT count(lot) FROM lot_parent_tb WHERE lot = ? and buy_lot = ?", plug => [ $nodeID->[1], $info[0] ]);
        if ($self->{dbHandle}->didErrorOccur())
        {
          $result[0] = 0;
          $result[3] = $self->{dbHandle}->errorMessage;
          return @result;
        }
        my @numUnique = $sth2->fetchrow_array;
        if (defined $numUnique[0] && $numUnique[0] == 0)
        {
          # insert appropriate lot_parent_tb table entries
          $self->{dbHandle}->write(sql => "INSERT INTO lot_parent_tb (lot, buy_lot) VALUES (?, ?)", plug => [ $nodeID->[1], $info[0] ]);
          if ($mainDBObj->{dbHandle}->didErrorOccur())
          {
            $result[0] = 0;
            $result[3] = $mainDBObj->{dbHandle}->errorMessage;
            return @result;
          }
        }
      }
    }
  }

  # change the version in the version table
  $self->{dbHandle}->write(sql => "UPDATE version_tb SET version = '1.4'");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }

  # commit the work.
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }

  # signal success.
  $result[0] = 1;

  return @result;
}

# updateToOneFive
# requires: doc, mainDBObj
# optional:
# returns: array: [0]=success/failure (1/0), [1]=doc, [2]=doc modified (0/1), [3]=status message (if success)
# summary: This updates the database from version 1.4 to 1.5.
sub updateToOneFive
{
  my $self = shift;
  my %args = ( doc => undef, mainDBObj => undef, @_ );
  my $doc = $args{doc};
  my $mainDBObj = $args{mainDBObj};
  my $success = 0;
  my $modified = 0;
  my $statusStr = "Successfully updated from <b>1.4</b> to <b>1.5</b>.\n";
  my @result = ($success, $doc, $modified, $statusStr);
  my $dbVersion = $self->{input}->{dbVersion};

  if (!defined $doc)
  {
    $self->setError(errorString => "doc is not defined!<br>\n");
    return @result;
  }

  # validate we are still at version 1.4 before we do anything.
  my $currentVersion = $mainDBObj->getDBVersion;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return @result;
  }

  if ($currentVersion != 1.4)
  {
    $self->setError(errorString => "The database is not at version = <b>1.4</b>!<br>\nI can not upgrade to <b>1.5</b>!<br>\n");
    return @result;
  }

  # get the path that our upgrade.psql file is at.
  my $upgradePath = $self->{configObj}->{webRoot} . $self->{configObj}->{siteDir} . "/install/Accounting/db/1.5/";
  my $upgradeFile = $upgradePath . "upgrade.psql";

  # now validate we have an upgrade.psql file
  if (! -r $upgradeFile)
  {
    $self->setError(errorString => "Upgrade file '$upgradeFile' does not exist!  Error = '$!'.<br>\n");
    return @result;
  }

  # now use the Portal provided script to import/run the upgrade.psql file.
  my $companyAppObj = $self->{sessionObj}->{store}->{companyAppObj};
  my $result = $self->{methods}->dbImport(type => $companyAppObj->{dbType}, host => $companyAppObj->{dbHost}, port => $companyAppObj->{dbPort},
                                        user => $self->{configObj}->{dbUser}, passwd => $self->{configObj}->{dbPasswd},
                                        dbName => $companyAppObj->{dbName}, fname => "$upgradeFile");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  # signal success.
  $result[0] = 1;

  return @result;
}

# updateToOneSix
# requires: doc, mainDBObj, ledgerDBObj
# optional:
# returns: array: [0]=success/failure (1/0), [1]=doc, [2]=doc modified (0/1), [3]=status message (if success)
# summary: This updates the database from version 1.5 to 1.6.
sub updateToOneSix
{
  my $self = shift;
  my %args = ( doc => undef, mainDBObj => undef, ledgerDBObj => undef, @_ );
  my $doc = $args{doc};
  my $mainDBObj = $args{mainDBObj};
  my $ledgerDBObj = $args{ledgerDBObj};
  my $success = 0;
  my $modified = 0;
  my $lang = $self->{langObj};
  my $statusStr = "Successfully updated from <b>1.5</b> to <b>1.6</b>.\n";
  my @result = ($success, $doc, $modified, $statusStr);
  my $dbVersion = $self->{input}->{dbVersion};
  my @defaultEquityAccount = ( 3000 );

  if (!defined $doc)
  {
    $self->setError(errorString => "doc is not defined!<br>\n");
    return @result;
  }

  # validate we are still at version 1.5 before we do anything.
  my $currentVersion = $mainDBObj->getDBVersion;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return @result;
  }

  if ($currentVersion != 1.5)
  {
    $self->setError(errorString => "The database is not at version = <b>1.5</b>!<br>\nI can not upgrade to <b>1.6</b>!<br>\n");
    return @result;
  }

  # get the path that our upgrade.psql file is at.
  my $upgradePath = $self->{configObj}->{webRoot} . $self->{configObj}->{siteDir} . "/install/Accounting/db/1.6/";
  my $upgradeFile = $upgradePath . "upgrade.psql";

  # now validate we have an upgrade.psql file
  if (! -r $upgradeFile)
  {
    $self->setError(errorString => "Upgrade file '$upgradeFile' does not exist!  Error = '$!'.<br>\n");
    return @result;
  }

  # now use the Portal provided script to import/run the upgrade.psql file.
  my $companyAppObj = $self->{sessionObj}->{store}->{companyAppObj};
  my $result = $self->{methods}->dbImport(type => $companyAppObj->{dbType}, host => $companyAppObj->{dbHost}, port => $companyAppObj->{dbPort},
                                        user => $self->{configObj}->{dbUser}, passwd => $self->{configObj}->{dbPasswd},
                                        dbName => $companyAppObj->{dbName}, fname => "$upgradeFile");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return @result;
  }

  # Create a database Transaction object for use in inserting TransactionObject objects into the database
  $self->{transaction} = Portal::Accounting::DB::Transaction->new( dbHandle => $self->{dbHandle}, langObj => $self->{langObj} );

  if ($self->{transaction}->didErrorOccur())
  {
    $self->setError(errorString => "Error instantiating Transaction object!<br>\n" . $self->{transaction}->errorMessage);
    return @result;
  }

  # Convert beginning_balance_entry_tb table entries into Journal Entry transactions
  my $sql = "SELECT DISTINCT(beginning_balance_entry_tb.ledger_code), MIN(known_years_tb.year),  (SELECT start1 from ledger_tb WHERE code = beginning_balance_entry_tb.ledger_code) AS start FROM beginning_balance_entry_tb GROUP BY ledger_code;";
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->setError(errorString => "Error obtaining distinct list of used ledger codes from beginning_balance_tb table!<br>\n" . $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }
  while (my @info = $sth->fetchrow_array)
  {
    # Create journal entries using the books start date less one day
    my @date = ($info[1], split "-", $info[2]);
    @date = Date::Calc::Add_Delta_Days(@date, -1);
    my $fiscalPriorDate = "$date[0]-$date[1]-$date[2]";

    my $transObj = Portal::Accounting::Objects::TransactionObject->new(ledgerCode => $info[0], journalCode => "G", transDate => $fiscalPriorDate, langObj => $self->{langObj});

    my $sth2 = $self->{dbHandle}->read(sql => "SELECT account_num, sub_account_num, sub_e_code, amount, dorc FROM beginning_balance_entry_tb WHERE ledger_code = ?", plug => [ $info[0] ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->setError(errorString => "Error obtaining list of entries from beginning_balance_entry_tb for ledger_code = '" . $info[0] . "'<br>\n" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }

    my $debitSum = 0.00;
    my $creditSum = 0.00;
    my $problemString = "ledgerCode = '" . $info[0] . "'<br>\n";
    while (my @info2 = $sth2->fetchrow_array)
    {
      $problemString .= "fiscalPriorDate = '" . $fiscalPriorDate . "', subEntity = '" . $info2[2] . "', accountNum = '" . $info2[0] . "', subAccountNum = '" . $info2[1] . "', dorc = '" . $info2[4] . "', amount = '" . $info2[3] . "'";
      # build up the TransactionObject
      $transObj->addRow(rowDate => $fiscalPriorDate, subEntity => $info2[2], accountNum => $info2[0], subAccountNum => $info2[1], dorc => $info2[4], amount => $info2[3]);
      if ($transObj->didErrorOccur)
      {
        $self->setError(errorString => "Error adding row to TransactionObject:" . $transObj->errorMessage . "<br>\n" . $problemString);
        $self->{dbHandle}->rollback;
        return @result;
      }

      if ($info2[4] =~ /^(true|1)$/)
      {
        $debitSum += $info2[3];
      }
      else
      {
        $creditSum += $info2[3];
      }

      $problemString .= ", creditSum = '" . $creditSum . "', debitSum = '" . $debitSum . "'.<br>\n";
    }
    # add equity account amount to offset beginning_balance_entry_tb amounts
    my $sum = 0.00;
    my $dorc = 0;
    my $amount = 0.00;
    $sum = sprintf("%.2f", $debitSum - $creditSum);
    if ($sum < 0.00)
    {
      $dorc = 1;
    }
    else
    {
      $dorc = 0;
    }
    $amount = sprintf("%.2f", abs($sum));
    $problemString .= "equity sum = '" . $sum . "', equity dorc = '" . $dorc . "', equity amount = '" . $amount . "'.<br>\n";
    # hardcode Retained Earnings account as the equity account to offset all the values
    $transObj->addRow(rowDate => $fiscalPriorDate, subEntity => "", accountNum => "3000", subAccountNum => "-1", dorc => $dorc, amount => $amount);
    if ($transObj->didErrorOccur)
    {
      $self->setError(errorString => "Error adding row to TransactionObject:" . $transObj->errorMessage . "<br>\n" . $problemString);
      $self->{dbHandle}->rollback;
      return @result;
    }

    my $refNum = $self->{transaction}->insertTransaction(obj => $transObj, commit => 0);
    if (($self->{transaction}->didErrorOccur) || ($refNum == -1))
    {
      $self->setError(errorString => "Error inserting new transaction!\n" . $self->{transaction}->errorMessage . "<br>\n" . $problemString);
      $self->{dbHandle}->rollback;
      return @result;
    }

    # Delete all beginning_balance_entry_tb entries for the current ledger
    my $sth3 = $self->{dbHandle}->write(sql => "DELETE FROM beginning_balance_entry_tb WHERE ledger_code = ?", plug => [ $info[0] ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->setError(errorString => "Error deleting entries from beginning_balance_entry_tb table for ledger_code = '" . $info[0] . "'!\n" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }

    # Create new beginning_balance_entry_tb entry using reference number from journal entry and current ledger code
    $sth3 = $self->{dbHandle}->write(sql => "INSERT INTO beginning_balance_entry_tb (ref_num, ledger_code) VALUES (?, ?)", plug => [ $refNum, $info[0] ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->setError(errorString => "Error inserting updated entry into beginning_balance_entry_tb table for ref_num = '" . $refNum . "', ledger_code = '" . $info[0] . "'!\n" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }

    # make sure that there are no closeouts after the fiscalPriorDate
    if (not $ledgerDBObj->reopenLedger(ledgerCode => $info[0], reopenDate => $fiscalPriorDate, commit => 0))
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("errorReOpeningLedger"), $info[0], $fiscalPriorDate) . "<br>\n" . $ledgerDBObj->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }

    # "Closeout" the books for the day prior to the books start date for the current ledger
    if (not $ledgerDBObj->closeoutLedger(ledgerCode => $info[0], closeoutDate => $fiscalPriorDate, equityAccounts => \@defaultEquityAccount, checkDateRange => 0, commit => 0))
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("errorClosingOutLedger"), $info[0], $fiscalPriorDate) . "<br>\n" . $ledgerDBObj->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
  }
  # commit changes
  $sth = $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => "Error commiting upgrade of initial balance entries!\n" . $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }

  # Finish the upgrade by dropping unneeded columns in the beginning_balance_entry_tb table
  $upgradeFile = $upgradePath . "post-upgrade.psql";
  $result = $self->{methods}->dbImport(type => $companyAppObj->{dbType}, host => $companyAppObj->{dbHost}, port => $companyAppObj->{dbPort},
                                        user => $self->{configObj}->{dbUser}, passwd => $self->{configObj}->{dbPasswd},
                                        dbName => $companyAppObj->{dbName}, fname => "$upgradeFile");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return @result;
  }

  # signal success.
  $result[0] = 1;
  # use a JavaScript alert message box to indicate to the user that Retained Earnings account was used for all initial balance entries instead of
  # what might have originally been there and that any closeouts were reopened and will need to be closed out again
  $result[3] .= "<script language=\"Javascript1.2\"> alert('" . sprintf($lang->map("initialBalanceUpgrade"), $defaultEquityAccount[0]) . "'); </script>\n";

  return @result;
}

# getNodeTree
# Summary: retrieves a tree of nodes for the given node number from the lot_node_tb table
# Takes: nodeID, mainDBObj
# Returns: two-dimensional array of node ID's for the tree underneath the given node ID with the lot number associated with
#          the node ID
#          undef on failure of some kind
sub getNodeTree
{
  my $self = shift;
  my %args = ( mainDBObj => undef, nodeID => undef, @_ );
  my @treeNodes = [];
  my $mainDBObj = $args{'mainDBObj'};
  my $nodeID = $args{'nodeID'};

  # verify the arguments are valid
  if (not defined $mainDBObj)
  {
    return undef;
  }
  if (not defined $nodeID)
  {
    return undef;
  }

  # look for any children nodes that have the current node as their parent
  my $sth = $mainDBObj->{dbHandle}->read(sql => "SELECT lot, node_id FROM lot_node_tb WHERE parent_id = '$nodeID'");
  if ($mainDBObj->{dbHandle}->didErrorOccur())
  {
    return undef;
  }
  while (my @info = $sth->fetchrow_array)
  {
    push @treeNodes, [ ($info[1], $info[0]) ];
    my @tempNodes = $self->getNodeTree(nodeID => $info[1], mainDBObj => $mainDBObj);
    if (not defined $tempNodes[0])
    {
      return undef;
    }
    push @treeNodes, @tempNodes;
  }

  return @treeNodes;
}

# updateTemplate
# requires: doc, mainDBObj
# optional:
# returns: array: [0]=success/failure (1/0), [1]=doc, [2]=doc modified (0/1), [3]=status message (if success)
# summary: This updates the database from version 1.2 to 1.3.
sub updateTemplate
{
  my $self = shift;
  my %args = ( doc => undef, mainDBObj => undef, @_ );
  my $doc = $args{doc};
  my $mainDBObj = $args{mainDBObj};
  my $success = 0;
  my $modified = 0;
  my $statusStr = "Successfully updated from <b>1.2</b> to <b>1.3</b>.";
  my @result = ($success, $doc, $modified, $statusStr);

  if (!defined $doc)
  {
    $self->setError(errorString => "doc is not defined!<br>\n");
    return @result;
  }

  # validate we are still at version x.x before we do anything.
  my $currentVersion = $mainDBObj->getDBVersion;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return @result;
  }

  if ($currentVersion != 1.3)
  {
    $self->setError(errorString => "The database is not at version = <b>x.x</b>!<br>\nI can not upgrade to <b>x.y</b>!<br>\n");
    return @result;
  }

  # simulate success for now.
  $result[0] = 1;

  return @result;
}

=back

=cut
1;
__END__

=head1 NOTE

  See updateTemplate as a template for future DB version updates.

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::AppState(3)

=cut
