# WatchList.pm - The Object Class that provides a WatchList Object
# Copyright (c) 2000, 2002 PC & Web Xperience, Inc. http://www.pcxperience.com/
# All rights reserved.  This program is free software; you can redistribute it
# and/or modify it under the same terms of the GPL

=head1 NAME

WatchList - Object used to build a WatchList Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Investements::WatchList;
  my $obj = Portal::App::WatchList->new;
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

=head1

WatchList is a WatchList class.

=cut

package Portal::Accounting::Investments::WatchList;
use strict;
use Portal::Accounting::Ledger;
use Portal::Accounting::DB::Security;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Ledger Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.9 $';

# global variables.
# modify to represent the commands this state provides.
my %commands = ( display => "Main Menu", editWatchList => "Edit Watch List", saveWatchList => "Save Watch List",
  deleteWatchList => "Delete Watch List", newWatchList => "Create Watch List", statusScreen => "Status" );


=head1 Exported FUNCTIONS

=head1

=head2  scalar new(langObj)

    Creates a new instance of the Portal::Accounting::WatchList
    object.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # assign anything that you need
  $self->{ledgerCode} = $self->{input}->{ledgerCode};

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=head2 bool isValid(void)

    Returns 0 or 1 to indicate if the object is valid.  The error will be
    available via errorMessage().

=cut
sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # do validation code here.

  if ($self->{ledgerCode} eq "")
  {
    $self->missing("ledgerCode");
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=head2  HTMLObject run(command)

    Runs the specified command and returns the resulting HTMLObject.

=cut

sub run
{
  my $self = shift;
  my %args = ( command => "", @_ );
  my $command = $args{command};
  my $doc;
  my $phrases = $self->{langObj};
  my $message;
  $self->{DBSecurity} = Portal::Accounting::DB::Security->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($self->{DBSecurity}->didErrorOccur)
  {
    $self->prefixError("Error while initializing DB::Security<br>\n".$self->{DBSecurity}->errorMessage);
    return undef;
  }
  $self->{dbObj} = Portal::Accounting::DB::Main->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($self->{dbObj}->didErrorOccur)
  {
    $self->prefixError("Error while initializing DB::Main<br>\n".$self->{dbObj}->errorMessage);
    return undef;
  }

  if (!exists $commands{$command})
  {
    $self->prefixError(sprintf($phrases->map("commandNotKnown"), $command));
    return undef;
  }

  if ($command eq "editWatchList" || $command eq "newWatchList")
  {
    $doc = $self->editWatchList(command => $command);
  }
  elsif ($command eq "saveWatchList")
  {
    $doc = $self->saveWatchList();
  }
  elsif ($command eq "deleteWatchList")
  {
    $doc = $self->deleteWatchList();
  }
  elsif ($command eq "display")
  {
    $doc = $self->display();
  }
  elsif ($command eq "statusScreen")
  {
    $doc = $self->statusScreen();
  }
  else
  {
    $message .= "Command: '$command' is invalid.<br>\n";
    $doc = $self->display(message => $message);
  }
  if (not defined $doc)
  {
    $self->postfixError("Command: $command returned an undefined document<br>");
  }
  return $doc;
}

sub display
{
  my $self = shift;
  my %args = ( message => "", @_ );
  my $doc;
  my $message = $args{message};
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
  $message .= $self->{input}->{message};
  $doc = HTMLObject::Normal->new();
  $doc->setTitle("WatchList Editor " . $VERSION);

  $doc->setFocus("body");
  $doc->print("<h2>".$phrases->map("Watchlists")."</h2>\n");
  my @lists = $self->{DBSecurity}->getWatchLists(ledgerCode => $self->{ledgerCode});
  if ($self->{DBSecurity}->didErrorOccur)
  {
    $doc->print("There was an error while attempting to retrieve the watch lists\n".$self->{DBSecurity}->errorMessage);
    return $doc;
  }
  $doc->print("<hr>\n");
  if ($message ne "")
  {
    $doc->print("$message<br>\n");
    $doc->print("<hr>\n");
  }
  $doc->print("<form name=\"mainForm\" method=\"post\">\n");
  $doc->print("  <input type=\"hidden\" name=\"app\" value=\"Accounting\">\n");
  $doc->print("  <input type=\"hidden\" name=\"state\" value=\"Investments::WatchList\">\n");
  $doc->print("  <input type=\"hidden\" name=\"command\" value=\"\">\n");
  $doc->print("  <input type=\"hidden\" name=\"ledgerCode\" value=\"$ledgerCode\">\n");
  if (@lists > 0)
  {
    $doc->print("<select name=\"groupName\">\n");
    foreach my $list (@lists)
    {
      $doc->print("  <option value=\"$list\">$list</option>\n");
    }
    $doc->print("</select>\n");
    $doc->print("<a href=\"?$self->{getInfo}&command=editWatchList\" onClick=\"document.mainForm.command.value='editWatchList'; document.mainForm.submit(); return false;\" title=\"".$phrases->map("Edit")."\">".$phrases->map("Edit")."</a>");
    $doc->print("&nbsp;");
    $doc->print("<a href=\"?$self->{getInfo}&command=deleteWatchList\" onClick=\"if (confirm('" . $phrases->map("confirmDeleteWatchList") . "')) { document.mainForm.command.value='deleteWatchList'; document.mainForm.submit(); } return false;\" title=\"".$phrases->map("Delete")."\">".$phrases->map("Delete")."</a>");
    $doc->print("&nbsp;");
  }
  else
  {
    $doc->print($phrases->map("No Watchlists defined")."<br>\n");
  }
  $doc->print("<a href=\"?$self->{getInfo}&command=newWatchList\" onClick=\"document.mainForm.command.value='newWatchList'; document.mainForm.submit(); return false;\" title=\"".$phrases->map("New")."\">".$phrases->map("New")."</a>");
  $doc->print("</form>\n");
  return $doc;
}

sub editWatchList
{
  my $self = shift;
  my %args = ( message => "", command => "", groupName => "", oldGroupName => "", assignedTickers => "", @_ );
  my $command = $args{command};
  my $message = $args{message};
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $groupName = $args{groupName};
  my $oldGroupName = $args{oldGroupName};
  my $assignedTickers = $args{assignedTickers};
  my $title;
  my $oldCommand;

  # get list of all stocks in the system
  my @stocks = $self->{dbObj}->getListOfStocks(orderBy => "name,ticker");
  if ($self->{dbObj}->didErrorOccur)
  {
    return $self->display(message => $self->{dbObj}->errorMessage);
  }
  if (scalar @stocks == 0)
  {
    return $self->statusScreen(message => $phrases->map("noStocksCreated"));
  }

  if ($command eq "newWatchList")
  { #initialize params
    # we might be coming back from a failed save attempt, so leave the assignment of groupName
    # to what is in the arguments (which should be nothing if we just got here the first time)
    if ($self->{input}->{assignedTickers} ne "")
    {
      $assignedTickers = $self->{input}->{assignedTickers};
    }
    $title = $phrases->map("New");
    $oldCommand = "newWatchList";
    $oldGroupName = "";
  }
  else
  { #pull params from $self->input
    if ($self->{input}->{groupName} ne "")
    {
      $groupName = $self->{input}->{groupName};
    }
    if ($self->{input}->{oldGroupName} ne "")
    {
      $oldGroupName = $self->{input}->{oldGroupName};
    }
    else
    {
      $oldGroupName = $groupName;
    }
    if ($self->{input}->{assignedTickers} ne "")
    {
      $assignedTickers = $self->{input}->{assignedTickers};
    }
    else
    {
      # first time through
      # determine which stocks are assigned to the given groupName
      my @assignedStocks = $self->{DBSecurity}->getListOfWatchedStocks(ticker => "yes", orderBy => "name", byGroupName => "1", groupName => $groupName, ledgerCode => $self->{ledgerCode});
      if ($self->{DBSecurity}->didErrorOccur)
      {
        $message = $phrases->map("errorLookingUpStockWatchListInfo") . "<br>" . $self->{DBSecurity}->errorMessage . "<br>" . $message;
      }
      foreach my $assignedStock (@assignedStocks)
      {
        #(my $sTicker, my $sName, my $sSubAcct) = @{$assignedStock};
        if ($assignedTickers ne "")
        {
          $assignedTickers .= ", ";
        }
        #$assignedTickers .= $sTicker;
        $assignedTickers .= $assignedStock->[0];
      }
    }
    $title = $phrases->map("Edit");
    $oldCommand = "editWatchList";
  }
  my $doc = HTMLObject::Normal->new;
  $doc->setFocus("body");
  $doc->print("<h2>$title " . $phrases->map("WatchList") . "</h2>\n");
  $doc->print("<hr>\n");
  if ($message ne "")
  {
    $doc->print("$message<br>\n");
    $doc->print("<hr>\n");
  }
  # javascript code
  $doc->setFocus("javascript");
  $doc->print(<<"EOF");
function assignAllTickers()
{
  var allTickers = '';
  for (var count=0; count < document.editForm.availableTickersSelect.options.length; count++)
  {
    if (document.editForm.availableTickersSelect.options[count].value != '')
    {
      if (count > 0 && allTickers != '')
      {
        allTickers += ', ';
      }
      allTickers += document.editForm.availableTickersSelect.options[count].value;
    }
  }
  document.editForm.assignedTickers.value = allTickers;
}

function moveToAssignedTickers()
{
  if (document.editForm.availableTickersSelect.options[document.editForm.availableTickersSelect.selectedIndex].value != '')
  {
    if (document.editForm.assignedTickers.value != '')
    {
      document.editForm.assignedTickers.value += ', ';
    }
    document.editForm.assignedTickers.value += document.editForm.availableTickersSelect.options[document.editForm.availableTickersSelect.selectedIndex].value;
  }
}

function validateAssignedTickers()
{
  var regEXP = /^(\\s*([\\w\\d\\.]+),\\s*)*(\\s*[\\w\\d\\.]+\\s*)\$/;
  var result = document.editForm.assignedTickers.value.match(regEXP);
  if (result != null)
  {
    return true;
  }
  else
  {
    return false;
  }
}

function validateEditForm()
{
  if (!validateAssignedTickers())
  {
EOF
  $doc->print("    alert('" . $phrases->map("enterListOfCommaSeparatedTickers") . "');\n");
  $doc->print(<<"EOF");
    return false;
  }

  if (document.editForm.groupName.value.match(/^\\s*\$/))
  {
EOF
  $doc->print("    alert('" . sprintf($phrases->map("InvalidWatchListName"), '') . "');\n");
  $doc->print(<<"EOF");
    return false;
  }

  return true;
}
EOF
  #display text field with name of list above first multi-select
  $doc->setFocus("body");
  $doc->print(<<"EOF");
<form name="editForm" method="post">
  $self->{postInfo}
  <input type="hidden" name="command" value="saveWatchList">
  <input type="hidden" name="oldCommand" value="$oldCommand">
  <input type="hidden" name="oldGroupName" value="$oldGroupName">
<table><tr valign="bottom"><td>
  <input type="text" name="groupName" value="$groupName"><br>
EOF
  #display multi-select with all available tickers.
  $doc->print("  " . $phrases->map("Available") . "<br>\n");
  $doc->print("  <select name=\"availableTickersSelect\" size=\"10\" onChange=\"moveToAssignedTickers();\">\n");
  foreach my $stock (@stocks)
  {
    $doc->print("    <option value=\"" . $stock->ticker . "\">" . $stock->name . " (" . $stock->ticker . ")\n");
  }
  $doc->print("  </select>\n");
  $doc->print("</td><td><input type=\"button\" name=\"assignAllTickersButton\" value=\">>\" alt=\"" . $phrases->map("AssignAllTickers") . "\" title=\"" . $phrases->map("AssignAllTickers") . "\" onClick=\"assignAllTickers();\"><br>\n");
  $doc->print("<input type=\"button\" name=\"unAssignAllTickersButton\" value=\"<<\" alt=\"" . $phrases->map("unAssignAllTickers") . "\" title=\"" . $phrases->map("unAssignAllTickers") . "\" onClick=\"document.editForm.assignedTickers.value = '';\"></td><td>\n");
  #display textarea with tickers assigned in a comma separated format
  $doc->print("  " . $phrases->map("Assigned") . "<br>\n");
  $doc->print("  <textarea name=\"assignedTickers\" rows=\"13\" cols=\"40\" wrap=\"soft\">$assignedTickers</textarea>\n");
  $doc->print("</td></tr><tr><td>&nbsp;</td><td>&nbsp;</td><td align=\"right\">\n");
  $doc->print("<a href=\"\" onClick=\"if (validateEditForm()) { document.editForm.submit(); } return false;\">" . $phrases->map("Save") . "</a>\n");
  $doc->print("</td></tr></table>\n");
  $doc->print("</form>\n");
  return $doc;
}

sub saveWatchList
{
  my $self = shift;
  my %args = ( message => "", command => "", @_ );
  my $command = $args{command};
  my $message = $args{message};
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $groupName = $self->{input}->{groupName};
  my $oldGroupName = $self->{input}->{oldGroupName};
  my $assignedTickers = $self->{input}->{assignedTickers};
  my $oldCommand = $self->{input}->{oldCommand};

  #verify data
  if ($groupName eq "")
  {
    # if error, return editWatchList( message => "foo")
    return $self->editWatchList(message => sprintf($phrases->map("InvalidWatchListName"), $groupName), command => $oldCommand);
  }
  if ($oldGroupName eq "" and $oldCommand =~ /edit/) # oldCommand was an edit
  {
    # if error, return editWatchList( message => "foo")
    return $self->editWatchList(message => sprintf($phrases->map("InvalidWatchListName"), $oldGroupName), command => $oldCommand);
  }
  if ($assignedTickers eq "")
  {
    return $self->editWatchList(message => sprintf($phrases->map("ChooseTickersForWatchList"), $groupName), command => $oldCommand);
  }
  if (not ($assignedTickers =~ /^(\s*([\w\d\.]+),\s*)*(\s*[\w\d\.]+\s*)$/))
  {
    return $self->editWatchList(message => $phrases->map("enterListOfCommaSeparatedTickers"), command => $oldCommand);
  }
  # if oldCommand was new, assign the groupName to oldGroupName
  if ($oldCommand =~ /new/)
  {
    $oldGroupName = $groupName;
  }
  # remove leading and trailing whitespace from each ticker definition
  $assignedTickers =~ s/\s*//g;
  my @assignedTickers = split /,/, $assignedTickers;
  #attempt to put data in Database (delete old, insert new)
  my $result = $self->{DBSecurity}->unWatchAllStocks(ledgerCode => $ledgerCode, groupName => $oldGroupName);
  if ($self->{DBSecurity}->didErrorOccur || $result ne "1")
  {
    return $self->editWatchList(message => sprintf($phrases->map("UnWatchingAllStocksWhileSavingWatchListFailed"), $oldGroupName) . " " . $self->{DBSecurity}->errorMessage, command => $oldCommand, groupName => $groupName, oldGroupName => $oldGroupName);
  }
  foreach my $stock (@assignedTickers)
  {
    # lookup sub-account number for each ticker
    my $stockObj = $self->{dbObj}->getStockInfo(ticker => $stock);
    if ($self->{dbObj}->didErrorOccur)
    {
      return $self->editWatchList(message => $self->{dbObj}->errorMessage, command => $oldCommand, groupName => $groupName, oldGroupName => $oldGroupName);
    }
    # watch each stock
    my @result = $self->{DBSecurity}->watchStock(number => $stockObj->{number}, ledgerCode => $ledgerCode, groupName => $groupName);
    if (($result[0] ne "1" && $result[0] ne "-1") || $self->{DBSecurity}->didErrorOccur)
    {
      return $self->editWatchList(message => sprintf($phrases->map("SavingStockForWatchListFailed"), $stock, $groupName) . (($result[0] ne "1") ? " result ne 1 " : " ") . (($self->{DBSecurity}->didErrorOccur) ? $self->{DBSecurity}->errorMessage : "") , command => $oldCommand, groupName => $groupName, oldGroupName => $oldGroupName);
    }
  }
  return $self->statusScreen(message => sprintf($phrases->map("SavingWatchListSucceeded"), $groupName));
}

sub deleteWatchList
{
  my $self = shift;
  my %args = ( message => "", command => "", @_ );
  my $command = $args{command};
  my $message = $args{message};
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $groupName = $self->{input}->{groupName};

  if ($self->{input}->{groupName} eq "")
  {
    return $self->statusScreen(message => sprintf($phrases->map("InvalidWatchListName"), $groupName));
  }
  #attempt to delete specified list from db !!!!
  my $result = $self->{DBSecurity}->unWatchAllStocks(ledgerCode => $ledgerCode, groupName => $groupName);
  if ($self->{DBSecurity}->didErrorOccur || $result ne "1")
  {
    return $self->statusScreen(message => sprintf($phrases->map("DeletionOfWatchListFailed"), $groupName));
  }
  return $self->statusScreen(message => sprintf($phrases->map("DeletionOfWatchListSucceeded"), $groupName));
}

sub statusScreen
{
  my $self = shift;
  my %args = ( message => "", command => "", @_ );
  my $message = $args{message};
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};

  my $doc = HTMLObject::Normal->new;
  $doc->setTitle($commands{statusScreen});
  $doc->setFocus("body");
  $doc->print(<<"EOF");
<hr>
$message
<form name="statusForm" method="post">
  $self->{postInfo}
  <input type="hidden" name="command" value="display">
  <input type="hidden" name="message" value="$message">
EOF
  $doc->setOnload(code => "document.statusForm.submit();\n");

  return $doc;
}

1;
__END__

=head1

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

JT Moree
PC & Web Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::App(3)

=cut
