# CGLDisplayObject.pm - The Object Class that provides a CGLDisplayObject Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: CGLDisplayObject.pm,v 1.3 2003/11/19 23:19:28 pcxuser Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::CGLDisplayObject;
use strict;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.03';

# new
# instantiates an instance of a CGLDisplayObject
# requires: cost, market, gainLoss, cglStocks (ref to array), cglAssets (ref to array)
sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;
  my %args = ( cost => "0.00", market => "0.00", gainLoss => "0.00", cglStocks => undef,
               cglAssets => undef, cglManual => undef, @_ );
  my $errStr = "CGLDisplayObject: new() - Error!<br>\n";

  $self->{cost} = $args{cost};
  $self->{market} = $args{market};
  $self->{gainLoss} = $args{gainLoss};
  $self->{cglStocks} = $args{cglStocks};
  $self->{cglAssets} = $args{cglAssets};
  $self->{cglManual} = $args{cglManual};

  # do validation
  $self->{error} = !$self->isValid;  
  if ($self->{error})
  {
    $self->{errorString} = $errStr . $self->{errorString};
  }

  return $self;
}

# isValid - Returns 0 or 1 to indicate if the object is valid.
sub isValid
{
  my $self = shift;
  my $error = 0;
  my $errorString = "";
  my $errStr = "CGLDisplayObject: isValid() - Error!<br>\n";

  # do validation code here.
  if ($self->{cost} !~ /^(-?\d+\.\d{2})$/)
  {
    $errorString .= "cost = '$self->{cost}' is invalid!<br>\n";
  }
  if ($self->{market} !~ /^(-?\d+\.\d{2})$/)
  {
    $errorString .= "market = '$self->{market}' is invalid!<br>\n";
  }
  if ($self->{gainLoss} !~ /^(-?\d+\.\d{2})$/)
  {
    $errorString .= "gainLoss = '$self->{gainLoss}' is invalid!<br>\n";
  }
  if (! defined $self->{cglStocks})
  {
    $errorString .= "cglStocks is not defined!<br>\n";
  }
  elsif (scalar @{$self->{cglStocks}} == 0)
  {
    $errorString .= "You must define at least 1 CGL Stock Object to work with!<br>\n";
  }
  if (! defined $self->{cglAssets})
  {
    $errorString .= "cglAssets is not defined!<br>\n";
  }
  elsif (scalar @{$self->{cglAssets}} == 0)
  {
    $errorString .= "You must define at least 1 CGL Asset Object to work with!<br>\n";
  }

  $self->{error} = $error;
  $self->{errorString} = $errStr if $error;
  $self->{errorString} .= $errorString;

  return !$error;
}

sub DESTROY
{
  my $self = shift;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://;	# strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

# setError
# requires: errorString
# returns: nothing
sub setError
{
  my $self = shift;
  my %args = ( errorString => "", @_ );
  my $errorString = $args{errorString};
  
  $self->{errorString} = $errorString;
  $self->{error} = 1;
}

# didErrorOccur
# returns the value of error.
sub didErrorOccur
{
  my $self = shift;

  return $self->{error};
}

# errorMessage
# returns the value of errorString
sub errorMessage
{
  my $self = shift;

  return $self->{errorString};
}

1;
__END__

=head1 NAME

CGLDisplayObject - Object used to build a CGLDisplayObject Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::CGLDisplayObject;
  my $obj = Portal::Accounting::Objects::CGLDisplayObject->new;
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }

=head1 DESCRIPTION

CGLDisplayObject is a CGLDisplayObject class.

=head1 Exported FUNCTIONS

  scalar new(cost, market, gainLoss, \@cglStocks, \@cglAssets, \@cglManual)
    Creates a new instance of the
    Portal::Accounting::Objects::CGLDisplayObject object.

  bool isValid(void)
    Returns 0 or 1 to indicate if the object is valid.  The error will
    be available via errorMessage().

  void setError(errorString)
    Sets error = 1 and $self->{errorString} = errorString.

  scalar didErrorOccur(void)
    Returns the value of error.

  scalar errorMessage(void)
    Returns the value of errorString.


  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

PC & Web Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3)

=cut
