# Object that contains the referenceNum, ledgerCode, journalCode, date, payeePayer, memo, remember
# and an array of JournalEntryRows.
#$Id: JournalEntry.pm,v 1.6 2004/10/13 21:49:56 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::JournalEntry;

use strict;
use Portal::Accounting::Objects::JournalEntryRow;
use vars qw($VERSION @ISA @EXPORT $AUTOLOAD);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.6 $';
=head1 NAME

JournalEntry - Object used to build a complete JournalEntry for storage in the accounting database.

=head1 SYNOPSIS

  use JournalEntry;
  my $transobj = JournalEntry->new(referenceNum => '379',
  					ledgerCode => 'ts1',
  					journalCode => 'G',
  					date => '1999-12-14',
  					payeePayer => 'Bank',
  					memo => 'cashed check',
  					remember => 'false',
            origId => "");
  if ($transobj->didErrorOccur())
  {
    die $transobj->errorMessage();
  }

  # set object data by specifying the data field desired
  $transobj->journalCode("I");
  $transobj->memo("");

  # make sure the object is useful by calling the validation routine
  # before writing it to the database
  if ($transobj->isValid())
  {
    updateJournalEntry($transobj);
  }
  else
  {
    die $transobj->errorMessage();
  }

=head1 DESCRIPTION

JournalEntry is a class that stores data for an accounting
JournalEntry.  It provides a container for a JournalEntry that will be used
to write to the database or update an existing record in the database.

=head1 Exported FUNCTIONS

=item scalar new(langObj =>)

    Creates a new instance of the Portal::Accounting::Objects::JournalEntry
    object.  A new instance is created in an invalid state (i.e., missing
    fields, etc.).  isValid() will be called on the object before using the
    object in a database JournalEntry.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);

  $self->{referenceNum} = -1;
  $self->{ledgerCode}   = "G";
  $self->{journalCode}  = "";
  $self->{date}    = "";
  $self->{payeePayer}   = "";
  $self->{memo}         = "";
  $self->{remember}     = "false";
  $self->{rowObjects}   = [];
  $self->{trackingId}   = "";
  $self->{origId}       = "";

  return $self;
}

=item scalar populate(referenceNum, ledgerCode, journalCode, date, payeePayer, memo, remember, rowObjects, trackingId, origId)

  Fills in the data quickly for the object
    trackingId = the genSysId value from an XML JournalEntry Import. Can be "".
    origId = the original referenceNum value for this entry.  Can be "".

=cut
sub populate
{
  my $self = shift;
  my %args = ( referenceNum => -1,
      ledgerCode => "",
      journalCode => 'G',
      date => "",
      payeePayer => '',
      memo => '',
      remember => 'false',
      rowObjects => [],
      trackingId => "",
      origId => "",
      input => undef,
      @_ );

  $self->{referenceNum} = $args{referenceNum};
  $self->{ledgerCode}   = $args{'ledgerCode'};
  $self->{journalCode}  = $args{'journalCode'};
  $self->{date}         = $args{'date'};
  $self->{payeePayer}   = $args{'payeePayer'};
  $self->{memo}         = $args{'memo'};
  $self->{remember}     = $args{'remember'};
  foreach my $row (@{$self->{rowObjects}})
  {
    $self->addRow($row);
  }
  $self->{trackingId}   = $args{trackingId};
  $self->{origId}       = $args{origId};

  return 1;
}

# isValid
# validates the object in its current state
sub isValid
{
  my $self = shift;
  # Can we call the parent at the bottom to pick up on the error here without repeating code
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  my $errorHeader = "JournalEntry->isValid(): ";

  if ($self->{referenceNum} != -1 && $self->{referenceNum} !~ /^\d\d*/)
  {
    $self->invalid("referenceNum", $self->{referenceNum});
  }
  if ($self->{ledgerCode} eq "")
  {
    $self->missing( "ledgerCode");
  }
  if (not exists $self->{variables}->{journalCodes}->{$self->{journalCode}})
  {
    $self->invalid( "journalCode", $self->{journalCode});
  }
  if ($self->{d} !~ /^\d{8}$/ && $self->{date} !~ /^\d{4}[\-\\\/]\d{1,2}[\-\\\/]\d{1,2}$/)
  {
    $self->invalid( "date", $self->{date});
  }
  if (($self->{remember} ne "true" and $self->{remember} ne "false"))
  {
    $self->{remember} = "false";
  }
  if (scalar @{ $self->{rowObjects} } < 2)
  {
    $self->prefixError("The JournalEntry rows array contains too few items (" . (scalar @{ $self->{rowObjects} }) . ")!");
  }
  my $sum = 0;
  my $debit_sum = 0;
  my $credit_sum = 0;
  for (my $rowcount=0; $rowcount < scalar @{ $self->{rowObjects} }; $rowcount++)
  {
    if (ref($self->{rowObjects}[$rowcount]) ne "Portal::Accounting::Objects::JournalEntryRow")
    {
      $self->prefixError("Row $rowcount is not a valid JournalEntryRow (" . ref($self->{rowObjects}[$rowcount]) . ")!<br>");
    }
    elsif (!$self->{rowObjects}[$rowcount]->isValid())
    {
      $self->prefixError("JournalEntry->JournalEntryRow->isValid(): Row $rowcount has invalid data! " . $self->{rowObjects}[$rowcount]->errorMessage."<br>");
    }
    else
    {
      if ($self->{rowObjects}[$rowcount]->dorc eq "true")
      {
        $debit_sum += $self->{rowObjects}[$rowcount]->amount;
      }
      else
      {
        $credit_sum += $self->{rowObjects}[$rowcount]->amount;
      }
    }
  }
  $sum = sprintf("%.2f", $debit_sum - $credit_sum);
  if ($sum != 0.00)
  {
    if ($debit_sum == 0.00)
    {
      $self->prefixError( "No debits entered!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum);
    }
    elsif ($credit_sum == 0.00)
    {
      $self->prefixError("No credits entered!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum);
    }
    if ($sum < 0.00)
    {
      $self->prefixError("Credits are greater than debits!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum);
    }
    elsif ($sum > 0.00)
    {
      $self->prefixError("Debits are greater than credits!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum);
    }
  }
  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item scalar addRow(rowObject) OR (comment =>, subEntity =>, accountNum =>, subAccountNum =>, cleared =>, amount =>, [dorc =>])

    Adds a row to the rowObjects array in the JournalEntry.
    Returns the row number of the newly added row or undef on failure.
    takes a rowObject with the proper data or a list of data to create a row.
    dorc is optional.  based on the account number the rowObject knows how to set the dorc and other flags.

=cut
sub addRow
{
  my $self = shift;
  my $errorHeader = "JournalEntry->addRow(): ";
  if (@_ == 1)
  {
    #a row object was sent. check it and use it.
    my $row = @_[0];
    if (! ref($row) eq "Portal::Accounting::Objects::JournalEntryRow") { $self->prefixError("row is not a JournalEntryRow<br>"); return undef; }
    $row->date($self->{date});
    if ($row->isValid)
    {
      push (@{$self->{rowObjects}}, $row);
      return scalar @{$self->{rowObjects}} - 1;
    }
    else  { $self->prefixError("row is invalid<br>".$row->errorMessage); return undef; }
  }
  # create a new JournalEntryRow with the values passed in
  my $transrowobj = Portal::Accounting::Objects::JournalEntryRow->new(langObj => $self->{langObj});
  $transrowobj->populate(date => $self->{date}, @_);
  if ($transrowobj->didErrorOccur())
  {
    $self->prefixError($errorHeader . "Failed to add row!<br>\n" . $transrowobj->errorMessage);
    return undef;
  }
  # add the row object to the current array of row objects
  $self->{rowObjects}[scalar @{$self->{rowObjects}}] = $transrowobj;
  return (scalar @{$self->{rowObjects}} - 1);
}

=item scalar deleteRow(row)

    Removes the row specified from the rowObjects array.

=cut
sub deleteRow
{
  my $self = shift;
  my $errorHeader = "JournalEntry->deleteRow(): ";
  my %args = ( @_ );
  my $row = $args{'row'};

  if (not exists $args{'row'})
  {
    $self->prefixError( "Row number value argument missing!");
    return 0;
  }
  if ($row !~ /^[0-9]+$/)
  {
    $self->prefixError( "Invalid row number argument!  The value does not appear to be numeric.");
    return 0;
  }
  if ($row > 0 && $row < scalar @{$self->{rowObjects}})
  {
    # remove the row from the array using splice to "cut" the element out
    splice(@{$self->{rowObjects}}, $row, 1);
  }
  else
  {
    $self->prefixError( "Invalid row number!  That row does not exist.");
    return 0;
  }
  return 1;
}

=item boolean setDate('2004-9-9')

  Returns 0 for failure and sets error
  1 for success

  Will set all dates in rows and in object

=cut
sub setDate
{
  my $self = shift;
  my $date = shift;
  if ($date =~ /^(\d{4})[\-\\\/](\d{1,2})[\-\\\/](\d{1,2})$/)
  {
    $date = $1 ."-". sprintf("%02d", $2) ."-". sprintf("%02d", $3);
  }
  else
  {
    $self->postfixError("Invalid date: $date<br>\n");
    return 0;
  }
  $self->{date} = $date;
  foreach my $row (@{$self->{rowObjects}})
  {
    $row->setDate($date);
  }
  return 1;
}
1;
__END__

=pod

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.

=head1 AUTHOR

James A. Pattie (mailto:james@pcxperience.com), Jason A. Pattie (mailto:pattieja@pcxperience.com)

=head1 SEE ALSO

perl(1), JournalEntry(3), DBIWrapper(3)

=cut
