# Ledger.pm - The Object Class that provides a Ledger Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: Ledger.pm,v 1.27 2004/10/13 21:49:56 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::Ledger;
use strict;
use Portal::Accounting::Objects::Base;
use Date::Calc qw(Add_Delta_Days Today);
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.06';

=head1 NAME

Ledger - Object used to build a Ledger Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::Ledger;
  my $ledgerObj = Portal::Accounting::Objects::Ledger->new(
            name => "Testing", code => "test", color => "blue",
            basis => "true", businessType => 2, fyMo => 12,
            fyDay => 31, fyCurrent => "true");
  if ($ledgerObj->didErrorOccur())
  {
    die $ledgerObj->errorMessage();
  }

=head1 DESCRIPTION

Ledger is a Ledger class.  This represents an entry in the
ledger_tb table in the database.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(langObj => )

    Creates a new instance of Portal::Accounting::Objects::Ledger.
    See Portal::Base(3) for a listing of required arguments.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }
  my $l = $self->{langObj};

  # instantiate anything unique to this module
  my @date = Today;
  my $cash = $l->map("Cash");
  my $accrual = $l->map("Accrual");
  $self->{data} = {
    "new" => { -Type => "hidden", -Value => "1" },
    "color" => { -Label => $l->map("Color"), -Type => "color-picker", -Value => "#00FF00" },
    "name" => { -Label => $l->map("Ledger Name"), -Type => "text", size => 15 },
    "code" => { -Label => $l->map("Ledger Code"), -Type => "text", size => 15 },
    "fiscalYear" => { -Label => $l->map("Fiscal Year"), -Type => "text", size => 6 },
    "year" => { -Label => $l->map("Year"), -Type => "text", size => 4 },
    "month" => { -Label => $l->map("Month"), -Type => "text", size => 2 },
    "day" => { -Label => $l->map("Day"), -Type => "text", size => 2 },
    "command" => { -Type => "hidden",  -Value => "saveCreateLedger" },
    "app" => { -Type => "hidden", -Value => "Accounting" },
    "state" => { -Type => "hidden", -Value => "Main" },
    "windowName" => { -Type => "hidden", -Value => "" },
    "businessType" => { -Label => $l->map("Business Type"), -Type => "select", -Options => { names => [], values => [] }, },
    "basis" => { -Type => "radio", -Label => $l->map("Basis"), -Options => { names => [ qw($cash $accrual) ], values => [ qw(cash accrual) ] } },
    "start2" => { -Label => "2", -Type => "text", size => 6, -Value => "02-01" },
    "start3" => { -Label => "3", -Type => "text", size => 6, -Value => "03-01" },
    "start4" => { -Label => "4", -Type => "text", size => 6, -Value => "04-01" },
    "start5" => { -Label => "5", -Type => "text", size => 6, -Value => "05-01" },
    "start6" => { -Label => "6", -Type => "text", size => 6, -Value => "06-01" },
    "start7" => { -Label => "7", -Type => "text", size => 6, -Value => "07-01" },
    "start8" => { -Label => "8", -Type => "text", size => 6, -Value => "08-01" },
    "start9" => { -Label => "9", -Type => "text", size => 6, -Value => "09-01" },
    "start10" => { -Label => "10", -Type => "text", size => 6, -Value => "10-01" },
    "start11" => { -Label => "11", -Type => "text", size => 6, -Value => "11-01" },
    "start12" => { -Label => "12", -Type => "text", size => 6, -Value => "12-01" },
  };
  my $update = $l->map("Update Month Starts")."-->";
  my $Month = $l->map("Month");
  my $StartDate = $l->map("Start Date");
  $self->{template} = <<"ENDOF";
<form #FORMARGS#>
  #FORMERRORSTRING#
  #FIELD=app#
  #FIELD=state#
  #FIELD=command#
  #FIELD=new#
  <input type="hidden" name="ledgerOpen" value="0">
  #FIELD=windowName#
  <table width="100%" cellspacing="5" cellpadding="0" border="0">
    <tr>
      <td align="left" valign="top">
        <b>#LABEL=name#</b>&nbsp;#FI=name#
        &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
        <b>#LABEL=code#</b>&nbsp;#FI=code#
      </td>
    </tr>
    <tr>
      <td valign="top" align="left">
        <b>#LABEL=color#</b>&nbsp;#FI=color#
      </td>
    </tr>
    <tr>
      <td>
        <table width="100%"><tr>
        <td align="left" width="35%">
          <b>#LABEL=businessType#</b><br>
          #FI=businessType#
        </td>
        <td align="left" width="65%">
          <table width="50%" border="1"><tr><td>
          #LABEL=basis# #FI=basis#
          </td></tr></table>
        </td>
        </tr></table>
      </td>
    </tr>
  </table>

  <table border="1" width="100%">
    <tr>
      <td align="left">
        <b>#LABEL=fy#</b> #FI=fy#
        <b>Begins on</b><br>#FI=year##FIELD=year# #FI=month##FIELD=month# #FI=day##FIELD=day#
        <input type="button" value="$update" name="updateSI" onClick="updateStartInfo();">
      </td>
      <td>
        <table border="0" width="100%">
          <tr>
            <th align="left">$Month</th>
            <th align="center">$StartDate</th>
          </tr>
          <tr>
            <td align="left">#LABEL=start2#</td>
            <td align="center">#FIELD=start2#</td>
          </tr>
          <tr>
            <td align="left">#LABEL=start3#</td>
            <td align="center">#FIELD=start3#</td>
          </tr>
          <tr>
            <td align="left">#LABEL=start4#</td>
            <td align="center">#FIELD=start4#</td>
          </tr>
          <tr>
            <td align="left">#LABEL=start5#</td>
            <td align="center">#FIELD=start5#</td>
          </tr>
          <tr>
            <td align="left">#LABEL=start6#</td>
            <td align="center">#FIELD=start6#</td>
          </tr>
          <tr>
            <td align="left">#LABEL=start7#</td>
            <td align="center">#FIELD=start7#</td>
          </tr>
          <tr>
            <td align="left">#LABEL=start8#</td>
            <td align="center">#FIELD=start8#</td>
          </tr>
          <tr>
            <td align="left">#LABEL=start9#</td>
            <td align="center">#FIELD=start9#</td>
          </tr>
          <tr>
            <td align="left">#LABEL=start10#</td>
            <td align="center">#FIELD=start10#</td>
          </tr>
          <tr>
            <td align="left">#LABEL=start11#</td>
            <td align="center">#FIELD=start11#</td>
          </tr>
          <tr>
            <td align="left">#LABEL=start12#</td>
            <td align="center">#FIELD=start12#</td>
          </tr>
        </table>
      </td>
    </tr>
  </table>
  <input type="submit" name="Submit" value="Submit"></form>
ENDOF
  $self->{javascript} = <<"ENDOF";
function updateStartInfo()
{
  var d = document.mainForm;

  // get the first month and day values
  var month = d.month.value;
  var day = d.day.value;
  var n=new Number(month);

  // now update the start2-12 values to be month+1-day
  // take care to handle the month+1 > 12 case
  for (var i=2; i<13; i++)
  {
    // alert("i='"+i+"', month='"+n+"', day='"+day+"'");
    if (n + 1 > 12)
    {
      n = 1;
    }
    else
    {
      n++;
    }
    try
    {
      eval("d.start"+i+".value='"+(n.toString().length == 1 ? "0" : "") + n.toString() +"-"+day+"';");
    }
    catch(exception)
    {
      alert(exception);
    }
  }
}
ENDOF
  $self->{profile} = {
    required => [ qw( name code color basis businessType fiscalYear year month day start2 start3 start4 start5 start6 start7 start8 start9 start10 start11 start12 ) ],
    optional => [  ],
    constraints => {
      code => qr/^.$/,
      name => qr/^.+$/,
      color => qr/^((#([A-Fa-f0-9]){6})|transparent|inherit)( !important)?$/,
      businessType => qr/^\d$/,
      fiscalYear => qr/^\d{4}$/,
      month => qr/^\d{1,2}$/,
      day => qr/^\d{1,2}$/,
      year => qr/^\d{4}$/,
      start2 => qr/^\d{1,2}-\d{1,2}$/,
      start3 => qr/^\d{1,2}-\d{1,2}$/,
      start4 => qr/^\d{1,2}-\d{1,2}$/,
      start5 => qr/^\d{1,2}-\d{1,2}$/,
      start6 => qr/^\d{1,2}-\d{1,2}$/,
      start7 => qr/^\d{1,2}-\d{1,2}$/,
      start8 => qr/^\d{1,2}-\d{1,2}$/,
      start9 => qr/^\d{1,2}-\d{1,2}$/,
      start10 => qr/^\d{1,2}-\d{1,2}$/,
      start11 => qr/^\d{1,2}-\d{1,2}$/,
      start12 => qr/^\d{1,2}-\d{1,2}$/,
    },
  };

  # do anything else you might need to do.
  $self->{offsetCache} = ();
  return $self;
}

=item boolean setup( businessTypes)

  businessTypes is an hash of names of business types where the index is the number.

=cut
sub setup
{
  my $self = shift;
  my $b = shift;

  foreach my $i (keys %{$b})
  {
    push(@{$self->{data}->{businessType}->{-Options}->{names}}, $self->{langObj}->map($b->{$i}));
    push(@{$self->{data}->{businessType}->{-Options}->{values}}, $i);
  }
  return 1;
}

=item boolean populate(%input) or (name =>, code =>, color =>, basis =>, businessType =>, knownYears =>, start =>)

    name      - text
    code      - text
    color     - text
    basis - string (cash, accrual)
    businessType  - integer pertaining to a business type (see business_type_tb)
    knownYears- a hash containing fiscal years as keys
    fiscalYear -  the current fiscalYear
    start and offset   - array of month beginnings in format MM-DD or MMDD. Starting with index 1 for 1st month, etc.
        start[0]           - offset, if the fiscal year starts in the previous year this should be -1, otherwise 0. i.e. FY 2000 starts 1999-12-01
        start[0]  = "-1"
        start[1]  = "MMDD"

=cut
sub populate
{
  my $self = shift;
  if (scalar @_ == 1)
  {
    my $input = shift;
    $self->{new} = $input->{new};
    if ($self->{new})
    {
      $self->{data}->{command}->{-Value} = "saveCreateLedger";
    }
    else
    {
      $self->{data}->{command}->{-Value} = "saveEditLedger";
    }
    $self->{data}->{name}->{-Value} = $input->{name};
    $self->{data}->{code}->{-Value} = $input->{code};
    $self->{data}->{color}->{-Value} = $input->{color};
    $self->{data}->{basis}->{-Value} = $input->{basis};
    $self->{data}->{businessType}->{-Value} = $input->{businessType};
    $self->{data}->{fiscalYear}->{-Value} = $input->{fiscalYear};

    $self->{knownYears}->{$input->{fiscalYear}} = "$input->{year}-$input->{month}-$input->{day}";
    $self->{start}[0] = $input->{year} - $input->{fiscalYear};
    $self->{start}[1] = sprintf("%02d", $input->{month}).sprintf("%02d", $input->{day});
    for (my $m = 2; $m < 13 ; $m++)
    {
      my $start;
      if ($input->{start$m} =~ /^(\d{1,2})-(\d{1,2})$/) { $start = sprintf("%02d", $1).sprintf("%02d", $2); }
      else { $start = $input->{start$m} ; }
      $self->{start}[$m] = $input->{$start};
    }
  }
  else
  {
    my %args = (
      name => "",
      code => "",
      color => "",
      basis => "",
      businessType => "",
      fiscalYear => "",
      knownYears => undef,
      start => undef,
      @_ );

    $self->{data}->{name}->{-Value} = $args{name};
    $self->{data}->{code}->{-Value} = $args{code};
    $self->{data}->{color}->{-Value} = $args{color};
    $self->{data}->{basis}->{-Value} = $args{basis};
    $self->{data}->{businessType}->{-Value} = $args{businessType};
    $self->{data}->{fiscalYear}->{-Value} = $args{fiscalYear};
    $self->{start} = $args{start};
    $self->{knownYears} = $args{knownYears};
  }
  return $self->isValid;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self);
  if (!ref($self))
  {
    my $i=0;
    my $result = "";
    while (my @info=caller($i++))
    {
      $result .= "$info[2] - $info[3]: $info[4]\n";
    }
    $self->postfixError( "$self is not an object\nCalled from:\n$result");
  }
  my $name = $AUTOLOAD;

  # make sure we don't emulate the DESTROY() method.
  return if $name =~ /::DESTROY$/;

  $name =~ s/.*://;	# strip fully-qualified portion
  if (exists $self->{data}->{$name})
  {
    if (@_)
    {
      return $self->{data}->{$name}->{-Value} = shift;
    }
    else
    {
      return $self->{data}->{$name}->{-Value};
    }
  }
  return $self->SUPER->$name(@_);
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if (length $self->{data}->{name}->{-Value} == 0)
  {
    $self->missing("name");
  }
  if (length $self->{data}->{code}->{-Value} == 0)
  {
    $self->missing("code");
  }
  elsif ($self->{data}->{basis}->{-Value} !~ /^(cash|accrual)$/i)
  {
    $self->invalid("basis", $self->{data}->{basis}->{-Value});
  }
  if ($self->{data}->{businessType}->{-Value} !~ /^(\d)$/)
  {
    $self->invalid("businessType", $self->{data}->{businessType}->{-Value});
  }
  if ($self->{data}->{fiscalYear}->{-Value} !~ m/^\d{4}$/)
  {
    $self->invalid('fiscalYear', $self->{data}->{fiscalYear}->{-Value});
  }

  if (not defined $self->{knownYears})
  {
    $self->missing("knownYears");
  }
  else
  {
    foreach my $year (keys %{$self->{knownYears}} )
    {
      if ($year !~ /^(\d{4})$/) #check the key for a valid integer (year)
      {
        $self->invalid('knownYears_'.$year, $year);
      }
      elsif ($self->{knownYears}->{$year} !~ m/^(\d{8})$/ && $self->{knownYears}->{$year} !~ s/^(\d{4})([-\/\\])(\d{1,2})([-\/\\])(\d{1,2})$/$1.sprintf("%02d", $3 ).sprintf("%02d",$5)/e)
      {
        $self->invalid("knownYears->$year", $self->{knownYears}->{$year});
      }
    }
  }
  if ($self->{start}[0] !~ m/^(-?\d+)$/)
  {
    $self->invalid('start[0]/offset', $self->{start}[0]);
  }

  my $years = 0;
  for (my $index = 1; $index < 13; $index++)
  {
    $self->{start}[$index] =~ s/-//g;
    if ($self->{start}[$index] =~ m/^(\d{4})$/ || $self->{start}[$index] =~ s/^(\d{1,2})([\\\/-])(\d{1,2})$/sprintf("%02d",$1).sprintf("%02d",$3)/e)
    {
      #skip month1
      if ($index > 1)
      {
        #compare this date+offset with the last one+offset  #!!! does not compare correctly with '-'
        if ($self->{start}[$index] < $self->{start}[$index - 1])
        {
          $years++;
          if ( $years > 1 )
          {
            $self->invalid("start[".$index."]", $self->{start}[$index] . ": is out of order");
            $years = 1;
          }
        }
      }
    }
    else
    {
      $self->invalid("start[".$index."]", $self->{start}[$index]);
    }
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

# =item String[] businessType(void)
#
#  Returns the business type number
#
# =cut
#
# sub businessType
# {
#   my $self = shift;
#   return $self->{businessType};
# }

=item String[] businessTypeName(void)

 Returns the business type name from the proper lanquage
 if it is set in the object or an empty string if not set.
 There is no guarantee that it is set!

=cut

sub businessTypeName
{
  my $self = shift;
  for (my $i; $i < @{$self->{data}->{businessType}->{-Options}->{values}}; $i++)
  {
    if ($self->{data}->{businessType}->{-Options}->{values}[$i] == $self->{data}->{businessType}->{-Value})
    {
      return $self->{data}->{businessType}->{-Options}->{names}[$i];
    }
  }
  return "";
}

=item string end(month => , fiscalYear =>) or end(<month>)

 This routine returns the ending date of the month specified.
 If a year is given, it will be translated to actual (YYYY-MM-DD).
 Otherwise format will be MM-DD.

=cut
sub end
{
  my $self = shift;
  my $month;
  my $year = 0;
  my %args;
  my $return = "";
  if (scalar @_ == 1)
  {
    $month = @_[0];
  }
  else
  {
    %args = (month => "", fiscalYear => 0, @_);
    $month = $args{month};
    $year = $args{fiscalYear};
  }
  if ($year !~ m/^(\d)+$/)
  {
     $self->prefixError($self->{langObj}->map("Invalid") . " " . $self->{langObj}->map("year") . " '". $year . "'<br>\n");
     return undef;
  }
  if ($month !~ m/^(\d)+$/ || $month < 1 || $month > 12)
  {
     $self->prefixError($self->{langObj}->map("Invalid") . " " . $self->{langObj}->map("month") . " '". $month . "'<br>\n");
     return undef;
  }
  my @date = split '-', $self->start(month => ($month % 12) + 1, fiscalYear => $year  );
  if ($month == 12)
  {
    $date[0]++;
  }
  if (@date != 3)
  {
    $self->prefixError("Invalid split of date string: ".$self->start(month => $month % 12 + 1, fiscalYear => $year));
    return undef;
  }
  @date = Add_Delta_Days(@date, -1);
  if ($year == 0)
  {  return sprintf("%02d",$date[1])."-".sprintf("%02d",$date[2]);  }
  return sprintf("%02d",$date[0])."-".sprintf("%02d",$date[1])."-".sprintf("%02d",$date[2]);
}

=item date[] getQuarterEnds([fiscal year])

 Returns an array of strings representing the ending dates of Quarters
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut
sub getQuarterEnds
{
  my $self = shift;
  my @quarters;
  my $year = 0;
  my @month = qw [ 3 6 9 12 ];
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->fiscalYear;  }
    else
    {  $year = @_[0];  }
  }

  my $index = -1;
  for my $month (@month)
  {
    if ($year)
    {
      $quarters[++$index] = $self->end(month => $month, fiscalYear => $year);
    }
    else
    {
      $quarters[++$index] = $self->end($month);
    }
  }

  return @quarters;
}

=item date[] getQuarterStarts([fiscal year])

 Returns an array of strings representing the starting dates of Quarters
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut

sub getQuarterStarts
{
  my $self = shift;
  my @quarters = ();
  my $year = 0;
  my @month = qw [ 3 6 9 12 ];
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->fiscalYear;  }
    else
    {  $year = @_[0];  }
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  my $index = -1;
  for my $month (@month)
  {
    $quarters[++$index] = $self->start(month => $month, fiscalYear => $year);
    if ($year == 0)
    {
      my @date = split "-", $quarters[$index];
      $quarters[$index] = "$date[1]-$date[2]";
    }
  }

  return @quarters;
}

=item date[] getMonthEnds([fiscal year])

 Returns an array of strings representing the ending dates of months
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut

sub getMonthEnds
{
  my $self = shift;
  my @months = qw [ 1 2 3 4 5 6 7 8 9 10 11 12 ];
  my @return;
  my $year = 0;
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->fiscalYear;  }
    else
    {  $year = @_[0];  }
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  my $index = -1;
  for my $month (@months)
  {
    $return[++$index] = $self->end(month => $month, fiscalYear => $year);
    if ($year == 0)
    {
      my @date = split "-", $return[$index];
      $return[$index] = "$date[1]-$date[2]";
    }
  }

  return @return;
}

=item date[] getMonthStarts([fiscal year])

 Returns an array of strings representing the starting dates of months
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut

sub getMonthStarts
{
  my $self = shift;
  my @months = qw [ 1 2 3 4 5 6 7 8 9 10 11 12 ];
  my @return;
  my $year = 0;
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->fiscalYear;  }
    else
    {  $year = @_[0];  }
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  my $index = -1;
  for my $month (@months)
  {
    $return[++$index] = $self->start(month => $month, fiscalYear => $year);
    if ($year == 0)
    {
      my @date = split "-", $return[$index];
      $return[$index] = "$date[1]-$date[2]";
    }
  }

  return @return;
}

=item int offset(int)

 This routine calculates the offset for each month to change the
 fiscal year value into a calendar year value.

=cut

sub offset
{
  my $self = shift;
  my $month = shift;

  if ($month !~ m/^(\d)+$/ || $month < 1 || $month > 12)
  { return undef } ;
  if (not defined $self->{offsetCache}[$month])
  {
    $self->{offsetCache}[1] = $self->{start}[0];
    my $offset = $self->{start}[0];
    for (my $index = 2; $index < 13; $index++)
    {
      #compare this date+offset with the last one+offset
      if ($self->{start}[$index] < $self->{start}[$index - 1])
      {
        $offset++;
      }
      $self->{offsetCache}[$index] = $offset;
    }
  }
  return $self->{offsetCache}[$month];
}

=item string start(month => , fiscalYear =>) or start(<month>)

 This routine returns the starting date of the month specified.
 If a year is given, it will be translated to actual (YYYY-MM-DD).
 Otherwise format will be MM-DD.

=cut

sub start
{
  my $self = shift;
  my $month;
  my $year = "";
  my %args;
  my $return = "";
  if (@_ == 1)
  {
    $month = @_[0];
  }
  else
  {
    %args = (month => "", fiscalYear => "", @_);
    $month = $args{month};
    $year = $args{fiscalYear};
  }
  if (defined $self->{start}[$month])
  {
    if ($self->{start}[$month] !~ m/^(\d{2})(\d{2})$/)
    {
      $self->prefixError("Invalid internal month start: '".$self->{start}[$month]."'<br>\n");
      return undef;
    }
    $return = $1."-".$2;
    if ($year eq "")
    {  return $return; }
    elsif ($year =~ m/^\d+$/)
    {  return $year + $self->offset($month) ."-".$return;  }
    $self->prefixError("Invalid year: '".$year."'<br>\n");
  }
  $self->prefixError($self->{langObj}->map("Invalid") . " " . $self->{langObj}->map("month") . " '". $month . "'<br>\n");
  return undef;
}

=item ISO-Date firstDate()

 Returns the first date the books were opened for.

=cut

sub firstDate
{
  my $self = shift;

  # try not to calculate it every time.
  if (!defined $self->{firstDate})
  {
    my @years = keys %{$self->{knownYears}};
    (my $start1 = $self->{start}->[1]) =~ s/^(\d{2})(\d{2})$/$1-$2/;

    # we build this up based upon the oldest known year + offset of first month and
    # tack on the first month-day entry.
    my $firstDate = ((sort { $a <=> $b } @years)[0] + $self->offset(1)) . "-" . $start1;

    $self->{firstDate} = $firstDate;
  }

  return $self->{firstDate};
}

=item ISO-Date formatDate(string)

 This routine returns the input date formatted in the ISO standard
 if a year is specified.  In either case, we have -'s for seperators.

 If an invalid date is specified, we return "" and set an error.

=cut

sub formatDate
{
  my $self = shift;
  my $date = shift;

  if ($date =~ /^(\d{4})(\d{2})(\d{2})$/)
  {
    $date = "$1-$2-$3";
  }
  elsif ($date =~ /^(\d{2})(\d{2})$/)
  {
    $date = "$1-$2";
  }
  elsif ($date !~ /^((\d{4}-)?\d{2}-\d{2})$/)
  {
    $self->invalid("date", $date);
    $date = "";
    $self->postfixError($self->genErrorString("all"));
  }

  return $date;
}

=item string printXML(fullXML bool, indent int)

 This routine returns an XML snippet for this object for use in such
 things as exporting data

 fullXML - defaults to 0 (false).  If true, we output the xml header tag.
 indent - defaults to 2.  Specifies how many spaces in our top level tag
          needs to be.  If fullXML is true, then we force 0, otherwise we
          use what the user specifies.

=cut

sub printXML
{
  my $self = shift;
  my $result = "";
  my %args = ( fullXML => 0, indent => 2, @_ );
  my $fullXML = $args{fullXML};
  my $indent = ($fullXML ? 0 : $args{indent});  # start at 0 indentation if we are making a full XML document.
  my $spaceStr = " " x $indent;

  if ($self->isValid())
  {
    if ($fullXML)
    {
      $result .= "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>";
      $result .= $spaceStr . "<ledger name=\"$self->{ledger}->{name}\" code=\"$self->{ledger}->{code}\">\n";
      $indent += 2;
      $spaceStr = " " x $indent;
    }
    $result .= $spaceStr . "<ledgerInfo>\n";
    foreach my $name (qw{color basis businessType businessTypeName})
    {
      $result .= $spaceStr . "  <$name>" . ($name eq "firstDate" ? $self->formatDate($self->{$name}) : ($name eq "businessTypeName" ? $self->businessTypeName() : $self->{$name})) . "</$name>\n";
    }
    $result .= $spaceStr . "  <knownYears>\n";
    foreach my $year (keys %{$self->{knownYears}})
    {
      $result .= $spaceStr . "    <info year=\"$year\" endDate=\"" . $self->formatDate($self->{knownYears}->{$year}) . "\"/>\n";
    }
    $result .= $spaceStr . "  </knownYears>\n";
    $result .= $spaceStr . "  <dates>\n";
    for (my $i=1; $i <= 12; $i++)
    {
      $result .= $spaceStr . "    <info index=\"$i\" start=\"" . $self->start(month => $i) . "\" offset=\"" . $self->offset($i) . "\"/>\n";
    }
    $result .= $spaceStr . "  </dates>\n";
    $result .= $spaceStr . "</ledgerInfo>\n";
    if ($fullXML)
    {
      $indent -= 2;
      $spaceStr = " " x $indent;
      $result .= $spaceStr . "</ledger>\n";
    }
  }
  else
  {
    $result .= "Data not valid!\n\n";
    $result .= $self->errorMessage();
    die $result;
  }

  return $result;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal::Base(3)

=cut
