# Register.pm - The Object Class that provides a Register Object
#$Id: Register.pm,v 1.9 2004/10/13 21:49:56 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2003 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::Register;
use strict;
use Portal::Accounting::Objects::Base;
use Date::Calc qw (Today Today_and_Now);
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

my $Revision = '$Revision: 1.9 $';
($VERSION = $Revision) =~ s/^Revision:\s//;

=head1 NAME

Register - Object used to build a Register Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::Register;
  my $RegisterObj = Portal::Accounting::Objects::Register->new(   );
  $RegisterObj->populate(data => , ledgerCode =>, );
  if ($RegisterObj->didErrorOccur())
  {
    die $RegisterObj->errorMessage();
  }

=head1 DESCRIPTION

Register is a Register class.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new( )

    Creates a new instance of Portal::Accounting::Objects::Register.
    See Portal::Accounting::Objects::Base(3) for a listing of required arguments.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module

  # do anything else you might need to do.
  $self->{data} = [];
  return $self;
}

=item bool populate(ledgerCode =>, data => , from =>, to =>, balance =>, accountName =>, accountNumber =>, type =>, app =>, state =>, command => )

  this method inserts the data given into the object.
  balance defaults to 0, type defaults to cash/bank (0)

  data is a 2d array of register data  use something like
    SELECT parent_num, row, oustanding, trans_date, payee, memo, sub_account, cleared, debit, credit

=cut
sub populate
{
  my $self = shift;
  my %args = (data => undef,
    accountName => "", accountNumber => "", type => 0, balance => 0,
    from => "", to => "",
    ledgerCode => $self->{ledgerCode},
    app => $self->{app},
    state => $self->{state},
    command => $self->{command},
    reconcile => 0,
    @_);
  $self->{ledgerCode} = $args{ledgerCode};
  $self->{app} = $args{app};
  $self->{state} = $args{state};
  $self->{command} = $args{command};
  $self->{accountName} = $args{accountName};
  $self->{accountNumber} = $args{accountNumber};
  $self->{from} = $args{from};
  $self->{to} = $args{to};
  $self->{type} = $args{type};
  $self->{balance} = $args{balance};
  $self->{reconcile} = $args{reconcile};
  $self->{data} = $args{data};
  $self->{outstanding} = $self->outstanding();
  return $self->Portal::Accounting::Objects::Register::isValid;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.
  if ( $self->{ledgerCode} eq "")
  {
    $self->invalid("ledgerCode");
  }
  if ( $self->{accountName} eq "")
  {
    $self->invalid("accountName");
  }
  if ( $self->{accountNumber} =~ /^\d+$/)
  {
    $self->invalid("accountNumber");
  }
  if ( $self->{type} =~ /^\d+$/)
  {
    $self->invalid("type");
  }
  if ( $self->{balance} =~ /^\d(\.)?\d*$/)
  {
    $self->invalid("balance");
  }
  if ( $self->{from} =~ /^(\d{4})-(\d{1,2})-(\d{1,2})$/)
  {
    $self->{from} = $1 ."-". sprintf("%02d", $2) . "-" . sprintf("%02d", $3)
  }
  else
  {
    $self->invalid("from");
  }
  if ( $self->{to} =~ /^(\d{4})-(\d{1,2})-(\d{1,2})$/)
  {
    $self->{to} = $1 ."-". sprintf("%02d", $2) . "-" . sprintf("%02d", $3)
  }
  else
  {
    $self->invalid("to");
  }
  if (! defined $self->{data})
  {
    $self->invalid("data");
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

sub outstanding
{
  my $self = shift;
  my $total = 0;
  for (my $row = 0; $row < scalar @{$self->{data}}; $row++)
  {
    $self->{cashFlow} += $self->{data}->[$row]->[10];
    if ($self->{data}->[$row]->[2] ne "")  {  $total += $self->{data}->[$row]->[10];  }
  }
  return $total;
}

=item print

  This routine looks at the internal reconcile flag and calls the appropriate routine.

=cut
sub print
{
  my $self = shift;
  if ($self->{reconcile})
  {  return $self->printReconcile(@_);  }
  return $self->printPlain(@_);
}

=item string columnMarkup(text, row, column)

  This routine adds html that is appropriate to for each column in the register.
  Column is a integer starting from 0, text is the text to markup.
  The columns are based on the type of register
  Bank Acct: "ref_num", "counter", "$Outstanding", "$Date", "$Description", "$Comment", "$Chk", "$Dep", "$SubAccount", "$Cld", "$Amount"
  Invest Acct: "ref_num", "counter", "$Outstanding", "$Date", "$Description", "$Comment", "", "$SubAccountName", "$SubAccount", "$Cld", "$Amount"
  All others: "ref_num", "counter", "$Outstanding", "$Date", "$Description", "$Comment", "", "$SubAccountName", "$SubAccount", "$Cld", "$Amount"

=cut
sub columnMarkup
{
  my $self = shift;
  my $text = shift;
  my $row = shift;
  my $col = shift;
  my $return = "";
  if ($text eq "") { $text = "&nbsp;"; }
  if ($col == 9 && $self->{reconcile} == 1)
  {
    $return .= <<"ENDOF";
<input class="button" type="button" value="$text" name="button_$row" onclick="cyclebutton(this.form.cleared_$row, this); calculateBalance(this.form);">
<input type="hidden" name="cleared_$row" value="$text">
ENDOF
  }
  elsif ($col == 0 && $self->{reconcile} == 1)
  {
    $return .= <<"ENDOF";
<input type="hidden" value="$text" name="ref_$row">$text
ENDOF
  }
  elsif ($col == 1 && $self->{reconcile} == 1)
  {
    $return .= <<"ENDOF";
<input type="hidden" value="$text" name="counter_$row">$text
ENDOF
  }
  elsif ($col == 10)
  {
    $return .= "<input type=\"hidden\" name=\"amount_$row\" value=\"$text\">";
    if ($text < 0)
    {
      $return .= "<font class=\"negative\">$text</font>";
    }
    else
    {
      $return .= $text;
    }
  }
  elsif ($col == 11)
  {
    if ($text < 0)
    {
      $return .= "<font class=\"negative\">$text</font>";
    }
    else
    {
      $return .= $text;
    }
  }
  elsif ($col == 6) #check
  {
    my $value = $text;
    if ($text eq "&nbsp;") {  $value = "*"; }
    else
    {
      $return .= "<a style=\"text-decoration: none; color: #000000;\" href=\"?app=Accounting&state=JournalEntry&command=edit&ledgerCode=";
      $return .= $self->{ledgerCode}."&refNum=".$self->{data}->[$row]->[0];
      $return .= "\" target=\"_new\">$text</a>";
    }
    $return .= "<input type=\"hidden\" name=\"check_$row\" value=\"$value\">&nbsp;";
  }
  elsif ($col == 7) #dep
  {
    my $value = $text;
    if ($text eq "&nbsp;") {  $value = "*"; }
    else
    {
      $return .= "<a style=\"text-decoration: none; color: #000000;\" href=\"?app=Accounting&state=JournalEntry&command=edit&ledgerCode=";
      $return .= $self->{ledgerCode}."&refNum=".$self->{data}->[$row]->[0];
      $return .= "\" target=\"_new\">$text</a>";
    }
    $return .= "<input type=\"hidden\" name=\"deposit_$row\" value=\"$value\">&nbsp;";
  }
  elsif ($text ne "")
  {
    #/!!!! need to properly open a window.  use js code to do it?  does portal provide yet?
    $return .= "<a style=\"text-decoration: none; color: #000000;\" href=\"?app=Accounting&state=JournalEntry&command=edit&ledgerCode=";
    $return .= $self->{ledgerCode}."&refNum=".$self->{data}->[$row]->[0];
    $return .= "\" target=\"_new\">$text</a>";
  }

  return $return;
}

sub setupStyles
{
  my $self = shift;
  my $l = $self->{langObj};

  my $Date = $l->map("Date");
  my $Comment = $l->map("Comment");
  my $Outstanding = $l->map("Outstanding");
  my $SubAccount = $l->map("Sub-Accounts");
  my $Debit = $l->map("Debit");
  my $Amount = $l->map("Amount");
  my $Credit = $l->map("Credit");
  my $Cleared = $l->map("Cleared");
  my $Cld = $l->map("Cld");
  my $Balance = $l->map("Balance");
  my $Description = $l->map("Description");
  my $Chk = $l->map("Chk");
  my $Dep = $l->map("Dep");
  my %return;
  if ($self->{type} eq "")
  {
    $return{css} .= "<!-- Error setting up styles.  No type given to setupStyles. -->\n";
    $return{headers} .= "<!-- Error setting up headings.  No type given. -->\n";
  }
  elsif ($self->{type} == 0)
  {
    $return{headers} .= <<"ENDOF";
      <th class="active-column-0">ref#</th>
      <th class="active-column-1">#</th>
      <th class="active-column-2"><a title="$Outstanding">O</a></th>
      <th class="active-column-3">$Date</th>
      <th class="active-column-4">$Description</th>
      <th class="active-column-5">$Comment</th>
      <th class="active-column-6">$Chk </th>
      <th class="active-column-7">$Dep </th>
      <th class="active-column-8">$SubAccount</th>
      <th class="active-column-9">$Cld</th>
      <th class="active-column-10">$Amount</th>
      <th class="active-column-11">$Balance</th>
ENDOF
    $return{css} .= <<"ENDOF";
  .active-column-6 {text-align: right;} /* chk */
  .active-column-7 {text-align: right;} /* dep */
  .active-column-8 {display: none;}    /* sub account */
ENDOF
  }
  elsif ($self->{type} == 3) #investment accounts
  {
    $return{headers} .= <<"ENDOF";
      <th class="active-column-0">ref#</th>
      <th class="active-column-1">#</th>
      <th class="active-column-2"><a title="$Outstanding">O</a></th>
      <th class="active-column-3">$Date</th>
      <th class="active-column-4">$Description</th>
      <th class="active-column-5">$Comment</th>
      <th class="active-column-6">&nbsp;</th>
      <th class="active-column-7">&nbsp;</th>
      <th class="active-column-8">$SubAccount</th>
      <th class="active-column-9">$Cld</th>
      <th class="active-column-10">$Amount</th>
      <th class="active-column-11">$Balance</th>
ENDOF
    $return{css} .= <<"ENDOF";
  .active-column-6 {display: none;} /* sub account */
  .active-column-7 {display: none;} /* sub account */
  .active-column-8 {display: none;} /* sub account */
ENDOF
  }
  else #all others
  {
    $return{headers} .= <<"ENDOF";
      <th class="active-column-0">ref#</th>
      <th class="active-column-1">#</th>
      <th class="active-column-2"><a title="$Outstanding">O</a></th>
      <th class="active-column-3">$Date</th>
      <th class="active-column-4">$Description</th>
      <th class="active-column-5">$Comment</th>
      <th class="active-column-6">&nbsp;</th>
      <th class="active-column-7">$SubAccount</th>
      <th class="active-column-8">$SubAccount</th>
      <th class="active-column-9">$Cld</th>
      <th class="active-column-10">$Amount</th>
      <th class="active-column-11">$Balance</th>
ENDOF
    $return{css} = <<"ENDOF";
  .active-column-6 {display: none;}     /* sub account */
  .active-column-7 {text-align: left;}  /* sub account */
  .active-column-8 {text-align: right;} /* sub account */
ENDOF
  }
  my $outstanding = "display: none;";
  my $balance = "display: none;";
  if ($self->{reconcile})  { $outstanding = "width: 20px;"; }
  else { $balance = "text-align: right;"; }
  $return{css} = <<"ENDOF";
  body, html {margin:0px; padding: 0px; }
  input.button { font-size: 9; border-style:none;}

  .active-controls-grid {font: menu;}
  .active-grid-column {border-right: 1px solid threedlightshadow;}
  .active-grid-row {border-bottom: 1px solid threedlightshadow;}
  .negative { color: #ff0000; }  /* !!!! should be coming from system wide styles */

  .active-column-0 {display: none; text-align: right;}      /* ref */
  .active-column-1 {display: none; text-align: left;}      /* counter */
  .active-column-2 {$outstanding}      /* outstanding */
  .active-column-3 {text-align: left;}   /* date */
  .active-column-4 {text-align: left;}   /* payee */
  .active-column-5 {text-align: left;}   /* comment */
$return{css}
  .active-column-9 {width: 40px; text-align: center;}  /* cleared */
  .active-column-10 {text-align: right;} /* amount */
  .active-column-11 {$balance}  /* balance / / */
ENDOF
  return %return;
}

=item %hash printPlain()

  This is a register view with a running balance.  But no reconciliation forms.

=cut
sub printPlain
{
  my $self = shift;
  my %return;
  my %setup = $self->setupStyles();
  my $balance = $self->{balance};
  $return{css} = $setup{css};
  $return{body} .= <<"ENDOF";
  <center>$self->{accountName} - $self->{accountNumber} $self->{from} - $self->{to}</center><br>
  <table class="register" border="1" width="100%" cellpadding="0" cellspacing="0">
  $setup{headers}
  <tr>
    <td class="active-column-3">$self->{from}</td>
    <td class="active-column-4" colspan="5">Balance Forward</td>
    <td class="active-column-10">$balance</td>
    <td class="active-column-11">$balance</td></tr>
ENDOF
  for (my $row = 0; $row < scalar @{$self->{data}}; $row++)
  {
    $return{body} .= "  <tr>\n ";
    my $col;
    for ($col = 3; $col < scalar @{$self->{data}->[$row]}; $col++)
    {
      $return{body} .= "    <td class=\"active-column-$col\">".$self->columnMarkup($self->{data}->[$row]->[$col], $row, $col)."</td>\n";
    }
    $balance += $self->{data}->[$row]->[$col - 1];
    $return{body} .= "    <td class=\"active-column-$col\">".$self->columnMarkup(sprintf("%.02f",$balance))."</td>\n";
    $return{body} .= "  </tr>\n ";
  }
  $return{body} .= "</table>\n ";
  return %return;
}

=item %hash printReconcile()

  This register view contains forms for editing line items and marking entries as cleared.

=cut
sub printReconcile
{
  my $self = shift;
  my %args = (app => "Accounting", state => "Register", @_);
  $self->{app} = $args{app};
  $self->{state} = $args{state};
  my %return;
  my $l = $self->{langObj};
  my %setup = $self->setupStyles();
  my $headers;
  my $styles;
  my $rows = scalar @{$self->{data}};
  my $columns = scalar @{$self->{data}->[0]};
  my $Mark = $l->map("Mark");
  my $from = $l->map("From");
  my $to = $l->map("To");
  my $Dep = $l->map("Dep");
  my $Chk = $l->map("Chk");
  my $usethis = $l->map("usethis");

  $return{css} = $setup{css};
  $return{javascript} = <<"ENDOF";
function checkSubmit(myform)
{
  mark(myform);
  myform.submit();
}

function cyclebutton(button, cleared)
{
  if (button.value == "_")
  {
    button.value = "x";
    cleared.value = "x";
  }
  else if (button.value == "x")
  {
    button.value = "v";
    cleared.value = "v";
  }
  else if (button.value == "v")
  {
    button.value = "^";
    cleared.value = "^";
  }
  else if (button.value == "^")
  {
    button.value = "r";
    cleared.value = "r";
  }
  else
  {
    button.value = "_";
    cleared.value = "_";
  }
}

function mark(myform)
{
  markChecks(myform,'check');
  markChecks(myform,'deposit');
  markButtons(myform);
}

//takes: form, field
// form - form to work with
// field -  name of field to check for check or deposit number
function markChecks(myform, CorD)
{
  var count;
  var from;
  var to;
  if (CorD == "check")
  {
    from = myform.chkFrom.value;
    to = myform.chkTo.value;
    myform.chkFrom.value = '';
    myform.chkTo.value = '';
  }
  else if(CorD == "deposit")
  {
    from = myform.depFrom.value;
    to = myform.depTo.value;
    myform.depFrom.value = '';
    myform.depTo.value = '';
  }
  else
  {
    alert("Invalid field name for markChecks");
    return false;
  }
  if ((from != "" && isNaN(from)) || (to != "" && isNaN(to)) )
  { //either is set but not a number then blow error
    alert("FromToNotAnum");
    return false;
  }
  if (from != "" && to != "")
  {
    if (from > to)
    {
     alert("FromGTTo");
     return false;
    }
   //check each row for being a check by form.check_count is NAN
    for (count = 0; count < myform.rows.value; count++)
    {
      var check;
      eval('check = myform.' + CorD + '_' + count + '.value');
      if (! isNaN(check) )
      {
        var cleared;
        eval('cleared = myform.cleared_' + count);
        var button;
        eval('button = myform.button_' + count);
        if (cleared.value != "r" && check >= from && check <= to)
        {
          cleared.value = "x";
          button.value = "x";
        }
      }
    }
  }
  else if (from != "")
  {
    for (count = 0; count < myform.rows.value; count++)
    {
      var check;
      eval('check = myform.' + CorD + '_' + count + '.value');
      if (! isNaN(check) )
      {
        var cleared;
        eval('cleared = myform.cleared_' + count);
        var button;
        eval('button = myform.button_' + count);
        if (cleared.value != "r" && check == from)
        {
          cleared.value = "x";
          button.value = "x";
        }
      }
    }
  }
  else if (to != "")
  {
    for (count = 0; count < myform.rows.value; count++)
    {
      var check;
      eval('check = myform.' + CorD + '_' + count + '.value');
      if (! isNaN(check) )
      {
        var cleared;
        eval('cleared = myform.cleared_' + count);
        var button;
        eval('button = myform.button_' + count);
        if (cleared.value != "r" && check == to)
        {
          cleared.value = "x";
          button.value = "x";
        }
      }
    }
  }
  return true;
}

function markButtons(myform)
{
  var count;
  var markEntryX = 0;
  for (count = 0; count < myform.rows.value; count++)
  {
    eval('if (myform.cleared_' + count + '.value == "^") { markEntryX = 0; myform.cleared_' + count + '.value = "x"; myform.button_' + count + '.value = "x"; }' );
    eval('if (myform.cleared_' + count + '.value == "v") { markEntryX = 1; }');
    if (markEntryX == 1)
    {
      eval('myform.cleared_' + count + '.value = "x"; myform.button_' + count + '.value = "x";');
    }
  }
  calculateBalance(myform);
}

function calculateDiff(myform)
{
  if (myform.manualEndStatement.value == "")  { return true; }
  if (isNaN(myform.manualEndStatement.value))  { myform.manualEndStatement.value = ""; }
  else
  {
    var diff;
    diff = parseFloat(myform.manualEndStatement.value) - parseFloat(myform.endStatement.value);
    floatify(diff, 2);
    myform.difference.value = diff;
  }
  return true;
}

function calculateBalance(myform)
{
  //start with original balance
  //walk each row
  //check for x or r
  // if so, add to balance
  var balance = parseFloat(myform.beginStatement.value);
  var count;
  for (count = 0; count < myform.rows.value; count++)
  {
    var cleared;
    var amount = 0.0;
    eval('cleared = myform.cleared_' + count);
    if (cleared.value == "x" || cleared.value == "r")
    {
      eval('amount = myform.amount_' + count + '.value');
      balance += parseFloat(amount);
    }
  }
  myform.endStatement.value = balance;
  calculateDiff(myform);
}
ENDOF
  my $balance = $self->{balance} + $self->{outstanding};
  my $book = $self->{balance} + $self->{cashFlow};
  $return{body} .= <<"ENDOF";
<center>$self->{accountName} - $self->{accountNumber} $self->{from} - $self->{to}</center><br>
<form name="mainForm" method="post">
  <input type="hidden" name="app" value="$self->{app}">
  <input type="hidden" name="state" value="$self->{state}">
  <input type="hidden" name="command" value="reconcile">
  <table border="0" width="100%" cellpadding="0" cellspacing="0">
    <tr>
      <td style="text-align: left;">
      <input class="active-column-10" type="text" size="9" name="book" onfocus="this.form.chkFrom.focus();" style="border: medium none ;" tabIndex="-1" value="$book"> $self->{to} Book Balance<br>
      <input class="active-column-10" type="text" size="9" name="beginStatement" onfocus="this.form.chkFrom.focus();" style="border: medium none ;" value="$balance" tabIndex="-1"> <a title="If this does not match your statement correct it first with a non reconcile view.">Beginning Statement Balance</a><br>
      <input class="active-column-10" type="text" size="9" name="endStatement" onfocus="this.form.chkFrom.focus();" style="border: medium none ;"> <a title="When finished marking entries this should match your statement.">Ending Statement Balance</a><br>
      <input class="active-column-10" type="text" size="9" name="difference" onfocus="this.form.chkFrom.focus();" style="border: medium none ;" tabIndex="-1"> Difference between statement and calculations<br>
      <input class="active-column-10" type="text" size="9" name="manualEndStatement" onchange="calculateDiff(this.form); return true;"> Enter the ending balance from your statement
      </td>
      <td style="text-align: right;">
      <table border="0" align="right">
        <tr><th><font style="color: blue; font-size: large;"><input type="button" name="checks" value="$Mark" onclick="mark(this.form);" title="$usethis"></font></th><th>$from</th><th>$to</th></tr>
        <tr><th>$Chk</th><td><input type="text" size="5" name="chkFrom"></td> <td><input type="text" size="5" name="chkTo"></td></tr>
        <tr><th>$Dep</th><td><input type="text" size="5" name="depFrom"></td> <td><input type="text" size="5" name="depTo"></td></tr>
      </table>
      </td>
    </tr>
  </table>

  <input type="hidden" name="reconcile" value="1">
  <input type="hidden" name="rows" value="$rows">
  <input type="hidden" name="ledgerCode" value="$self->{ledgerCode}">
  <input type="hidden" name="accountName" value="$self->{accountName}">
  <input type="hidden" name="account" value="$self->{accountNumber}">
  <input type="hidden" name="accountType" value="$self->{type}">
  <input type="hidden" name="from" value="$self->{from}">
  <input type="hidden" name="to" value="$self->{to}">

  <table border="0" width="100%" cellspacing="0" cellpadding="0" bgcolor="white">
    <tbody>
    <tr>
    <td align="left"><input type="submit" name="Submit" value="Finished" onclick="checkSubmit(this.form);"></td>
    <td style="text-align: center;"><input type="submit" name="save" value="Save" onclick="this.form.command.value = 'saveReconcile'; checkSubmit(this.form);"><br>
    </td>
    <td align="right"><input type="reset" name="reset" value="Reset"></td>
    </tr>
    </tbody>
  </table>
ENDOF
  $return{body} .= <<"ENDOF";
  <table class="register" border="1" width="100%" cellpadding="0" cellspacing="0">
  $setup{headers}
ENDOF
  for (my $row = 0; $row < scalar @{$self->{data}}; $row++)
  {
    $return{body} .= "  <tr>\n ";
    my $col;
    for ($col = 0; $col < scalar @{$self->{data}->[$row]}; $col++)
    {
      $return{body} .= "    <td class=\"active-column-$col\">".$self->columnMarkup($self->{data}->[$row]->[$col], $row, $col)."</td>\n";
    }
    $return{body} .= "  </tr>\n ";
  }
  $return{body} .= "</table>\n ";

  return %return;
}

sub printOutstanding
{
  my $self = shift;
  my $l = $self->{langObj};
  my $debits;
  my $credits;
  my $sub = $l->map("Sub Account");
  my $Date = $l->map("Date");
  my $desc = $l->map("Description");
  my $amount = $l->map("Amount");
  my $Total = $l->map("SubTotal");
  my $Balance = $l->map("SubBalance");
  my $Outstanding = $l->map("Outstanding");
  my $cBalance = 0;
  my $dBalance = 0;

  if ($self->{type} == 0) { $sub = $l->map("Chk");  }
  $debits .= <<"ENDOF";
<table border="1" class="register">
  <tr>
    <th>$Date</th>
    <th>$desc</th>
    <th>$sub</th>
    <th>$amount</th>
  </tr>
ENDOF
  if ($self->{type} == 0) { $sub = $l->map("Dep");  }
  $credits .= <<"ENDOF";
<table border="1" class="register">
  <tr>
    <th>$Date</th>
    <th>$desc</th>
    <th>$sub</th>
    <th>$amount</th>
  </tr>
ENDOF
  for (my $i = 0; $i < scalar @{$self->{data}}; $i++)
  {
    if ($self->{data}->[$i]->[2] == "*")
    { #entry is outstanding
      my @data = @{$self->{data}->[$i]};
      if ($data[10] < 0)
      {
        $credits .= <<"ENDOF";
  <tr><td>$data[3]</td><td>$data[4]</td><td>$data[6]</td><td>$data[10]</td>
ENDOF
        $cBalance += $data[10];
      }
      else
      {
        $debits .= <<"ENDOF";
  <tr><td>$data[3]</td><td>$data[4]</td><td>$data[7]</td><td>$data[10]</td>
ENDOF
        $dBalance += $data[10];
      }
    }
  }
  my $o = sprintf ("%.02d", $dBalance + $cBalance);
  $dBalance = sprintf ("%.02d", $dBalance);
  $debits .= <<"ENDOF";
  <tr><td align="right">$Total</td><td align="right" colspan="3">$dBalance</td></tr>
</table>
ENDOF
  $cBalance = sprintf ("%.02d", $cBalance);
  $credits .= <<"ENDOF";
  <tr><td align="right">$Total</td><td align="right" colspan="3">$cBalance</td></tr>
</table>
<p>Summary<br>
$Balance:$self->{balance}<br>
$Outstanding:$o
</p>
ENDOF
  return $debits.$credits;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal::Accounting::Objects::Base(3)

=cut
