# SubEntityObject.pm - The Object Class that provides a SubEntity Object
#$Id: SubEntityObject.pm,v 1.5 2004/09/03 14:35:29 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::SubEntityObject;
use strict;
use Portal::Accounting::Objects::Base;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.03';

=head1 NAME

SubEntityObject - Object used to build a SubEntity Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::SubEntityObject;
  my $obj = Portal::Accounting::Objects::SubEntityObject->new;
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }

=head1 DESCRIPTION

SubEntityObject is a SubEntity class.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(langObj =>)

 Creates a new instance of the Portal::Accounting::Objects::SubEntityObject module.
 See Portal::Accounting::Objects::Base(3) for a listing of required arguments.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  my $phrases = $self->{langObj};
  $self->{data} = {
    "name" => { -Label => $phrases->map("Name"), -Value => "",
      -Type => "text"      },
    "code" => { -Label => $phrases->map("Code"), -Value => "",
      -Type => "text", -ReadOnly => 1      },
    "command" => { -Type => "hidden", -Value => "saveSubEntity" },      
    "ledgerCode" => { -Type => "hidden", -Value => "" },      
    "app" => { -Type => "hidden", -Value => "Accounting" },      
    "state" => { -Type => "hidden", -Value => "Ledger" },      
    "new" => { -Type => "hidden", -Value => 0 },      
  };
  
  my $message = $phrases->map("Sub Entity");
  my $submit = $phrases->map("Submit");
  $self->{template} = <<"ENDOF";
  #FORMERRORSTRING#
  <p><form #FORMARGS#>
  #FIELD=app#
  #FIELD=state#
  #FIELD=command#
  #FIELD=ledgerCode#
  #FIELD=new#
  #LABEL=name#:
  #FIR=name#<br>
  #LABEL=code#:
  #FIR=code#
  </p>
  <p><input type="submit" value="$submit">
  </p>
  #FORMREQUIREDSTRING#
  </form>
ENDOF

  $self->{profile} = {
    required => [ qw( ledgerCode code name ) ],
    constraints  =>
        {
        },
  };
  
  $self->{formObj} = HTMLObject::Form->new;

  return $self;
}

=item scalar populate(name, code, ledgerCode)
 requires:
  name
  code
  ledgerCode

=cut
sub populate
{
  my $self = shift;
  my %args = ( name => "", code => "", ledgerCode =>"", new => $self->{data}->{new}->{-Value}, @_ );
  
  $self->{data}->{new}->{-Value} = $args{new};
  if ($args{new} == 1)
  {  
    $self->{data}->{code}->{-ReadOnly} = 0;
  }
  $self->{data}->{name}->{-Value} = $args{name};
  $self->{data}->{code}->{-Value} = $args{code};
  $self->{data}->{ledgerCode}->{-Value} = $args{ledgerCode};
  return 1;
}
  
=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.
  if (length $self->{name} == 0)
  {
    $self->invalid("name", $self->{name});
  }
  if (length $self->{ledgerCode} == 0)
  {
    $self->invalid("ledgerCode", $self->{ledgerCode});
  }
  if (length $self->{code} == 0)
  {
    $self->invalid("code", $self->{code});
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://;	# strip fully-qualified portion
  unless (exists $self->{data}->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name}->{-Value} = shift;
  }
  else
  {
    return $self->{$name}->{-Value};
  }
}

=item string printXML(fullXML bool, indent int)

 This routine returns an XML snippet for this object for use in such
 things as exporting data

 fullXML - defaults to 0 (false).  If true, we output the xml header tag.
 indent - defaults to 2.  Specifies how many spaces in our top level tag
          needs to be.  If fullXML is true, then we force 0, otherwise we
          use what the user specifies.

=cut

sub printXML
{
  my $self = shift;
  my $result = "";
  my %args = ( fullXML => 0, indent => 2, @_ );
  my $fullXML = $args{fullXML};
  my $indent = ($fullXML ? 0 : $args{indent});  # start at 0 indentation if we are making a full XML document.
  my $spaceStr = " " x $indent;

  if ($self->isValid())
  {
    if ($fullXML)
    {
      $result .= "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>";
    }
    $result .= $spaceStr . "<subEntity>\n";
    my @entries = qw {name code};
    if ($fullXML)
    {
      push @entries, "ledgerCode";
    }
    foreach my $name (@entries)
    {
      $result .= $spaceStr . "  <$name>$self->$name</$name>\n";
    }
    $result .= $spaceStr . "</subEntity>\n";
  }
  else
  {
    $result .= "Data not valid!\n\n";
    $result .= $self->errorMessage();
    die $result;
  }

  return $result;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Accounting::Objects::Base(3)

=cut
