# TransactionImport.pm - Represents an Transaction Import XML document.
# Created by James A. Pattie.  Copyright (c) 2003, Xperience, Inc.

package Portal::Accounting::Objects::TransactionImport;

use strict;
use XML::LibXML;
use Portal::Base;
use vars qw ($AUTOLOAD @ISA @EXPORT $VERSION);

require Exporter;
@ISA = qw(Portal::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = "1.0";

=head1 NAME

TransactionImport - The XML Transaction Import Data Module.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::TransactionImport;
  my $obj = Portal::Accounting::Objects::TransactionImport->new();

=head1 DESCRIPTION

Portal::Accounting::Objects::TransactionImport will contain the parsed XML file.
It provides a method to validate that it is complete and also a method
to generate a valid XML file from the data stored in the data hash.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new()

 Creates a new instance of the Portal::Accounting::Objects::TransactionImport module.
 See Portal::Base(3) for a listing of required arguments.

 If an error occurs, it will die, so eval it.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );
  my $errStr = "Portal::Accounting::Objects::TransactionImport->new()  - Error:";

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{version} = "1.0";
  $self->{type} = "transactionImport";
  $self->{dataFile} = "";  # only used for debugging purposes.
  $self->{timestamp} = "";
  $self->{genSysId} = "";
  $self->{accounts} = [];
  $self->{subAccounts} = [];
  $self->{stocks} = [];
  $self->{ledgers} = [];
  $self->{errorCodes} = {
    0  => "version = '%s' is invalid",
    1  => "%s must be specified",
    2  => "%s = '%s' is invalid",
    3  => "no %s defined",
    4  => "%s: %s = '%s' already defined",
    5  => "%s invalid.  Error = '%s'",
    6  => "%s: %s with code = '%s' not defined",
    7  => "JournalEntry Row #%s: %s with code = '%s' not defined",
  };

  # do anything else you might need to do.

  return $self;
}

=item array isValid(void)

    Determines if the data structure is complete and usable for
    generating an XML file from.
    Returns an array.  The first index is boolean (1 or 0 to indicate
    if the object is valid or not).  The second index is an array of
    error messages that were generated based upon the errors found.
    (This is stored as an array ref.)

=cut

sub isValid
{
  my $self = shift;
  my @errors = ();
  my %accountCodes = ();
  my %subAccountCodes = ();

  if ($self->{version} !~ /^(1.0)$/)
  {
    push @errors, sprintf($self->{errorCodes}->{0}, $self->{version});
  }
  if (length $self->{timestamp} == 0)
  {
    push @errors, sprintf($self->{errorCodes}->{1}, "timestamp");
  }
  elsif ($self->{timestamp} !~ /^(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})$/)
  {
    push @errors, sprintf($self->{errorCodes}->{2}, "timestamp", $self->{timestamp});
  }
  if (length $self->{genSysId} == 0)
  {
    push @errors, sprintf($self->{errorCodes}->{2}, "genSysId", $self->{genSysId});
  }
  if (scalar @{$self->{accounts}} > 0)
  {
    # validate accounts
    my %encounteredAccounts = ();
    foreach my $account (@{$self->{accounts}})
    {
      if (!$account->isValid())
      {
        push @errors, sprintf($self->{errorCodes}->{5}, "account", $account->errorMessage);
      }
      # finish up processing the account.
      if (exists $encounteredAccounts{$account->{name}})
      {
        push @errors, sprintf($self->{errorCodes}->{4}, "account", "name", $account->{name});
      }
      elsif ($encounteredAccounts{$account->{name}} eq $account->{number})
      {
        push @errors, sprintf($self->{errorCodes}->{4}, "account", "number", $account->{number});
      }
      else
      {
        # first make sure the number hasn't been duplicated under a different name.
        my $found = 0;
        foreach my $number (keys %encounteredAccounts)
        {
          if ($account->{number} eq $number)
          {
            push @errors, sprintf($self->{errorCodes}->{4}, "account", "number", $account->{number});
            $found = 1;
            last;
          }
        }
        $encounteredAccounts{$account->{name}} = $account->{number} if (!$found);
        $accountCodes{$account->{number}} = $account->{name} if (!$found);
      }
    }
  }
  if (scalar @{$self->{subAccounts}} > 0)
  {
    # validate subAccounts
    my %encounteredSubAccounts = ();
    foreach my $subAccount (@{$self->{subAccounts}})
    {
      if (!$subAccount->isValid())
      {
        push @errors, sprintf($self->{errorCodes}->{5}, "subAccount", $subAccount->errorMessage);
      }
      # finish up processing the subAccount.
      if (exists $encounteredSubAccounts{$subAccount->{name}})
      {
        push @errors, sprintf($self->{errorCodes}->{4}, "subAccount", "name", $subAccount->{name});
      }
      elsif ($encounteredSubAccounts{$subAccount->{name}} eq $subAccount->{number})
      {
        push @errors, sprintf($self->{errorCodes}->{4}, "subAccount", "number", $subAccount->{number});
      }
      else
      {
        # first make sure the number hasn't been duplicated under a different name.
        my $found = 0;
        foreach my $number (keys %encounteredSubAccounts)
        {
          if ($subAccount->{number} eq $number)
          {
            push @errors, sprintf($self->{errorCodes}->{4}, "subAccount", "number", $subAccount->{number});
            $found = 1;
            last;
          }
        }
        $encounteredSubAccounts{$subAccount->{name}} = $subAccount->{number} if (!$found);
        $subAccountCodes{$subAccount->{number}} = $subAccount->{name} if (!$found);
      }
    }
  }
  if (scalar @{$self->{stocks}} > 0)
  {
    # validate stocks
    my %encounteredStocks = ();
    foreach my $stock (@{$self->{stocks}})
    {
      if (!$stock->isValid())
      {
        push @errors, sprintf($self->{errorCodes}->{5}, "stock", $stock->errorMessage);
      }
      # finish up processing the stock.
      if (exists $encounteredStocks{$stock->{name}})
      {
        push @errors, sprintf($self->{errorCodes}->{4}, "stock", "name", $stock->{name});
      }
      elsif ($encounteredStocks{$stock->{name}}->{number} eq $stock->{number})
      {
        push @errors, sprintf($self->{errorCodes}->{4}, "stock", "number", $stock->{number});
      }
      elsif ($encounteredStocks{$stock->{name}}->{ticker} eq $stock->{ticker})
      {
        push @errors, sprintf($self->{errorCodes}->{4}, "stock", "ticker", $stock->{ticker});
      }
      else
      {
        # first make sure the number hasn't been duplicated under a different name.
        my $found = 0;
        foreach my $tmpStock (keys %encounteredStocks)
        {
          if ($stock->{number} eq $encounteredStocks{$tmpStock}->{number})
          {
            push @errors, sprintf($self->{errorCodes}->{4}, "stock", "number", $stock->{number});
            $found = 1;
            last;
          }
          if ($stock->{ticker} eq $encounteredStocks{$tmpStock}->{ticker})
          {
            push @errors, sprintf($self->{errorCodes}->{4}, "stock", "ticker", $stock->{ticker});
            $found = 1;
            last;
          }
        }
        $encounteredStocks{$stock->{name}} = $stock if (!$found);
        $subAccountCodes{$stock->{number}} = $stock->{name} if (!$found);  # not 100% this is needed.
      }
    }
  }
  if (scalar @{$self->{ledgers}} > 0)
  {
    # validate ledgers
    my %encounteredLedgers = ();
    foreach my $ledger (@{$self->{ledgers}})
    {
      if (!$ledger->isValid())
      {
        push @errors, sprintf($self->{errorCodes}->{5}, "ledger", $ledger->errorMessage);
      }
      # finish up processing the ledger.
      if (exists $encounteredLedgers{$ledger->{ledger}->{name}})
      {
        push @errors, sprintf($self->{errorCodes}->{4}, "ledger", "name", $ledger->{ledger}->{name});
      }
      elsif ($encounteredLedgers{$ledger->{ledger}->{name}}->{ledger}->{code} eq $ledger->{ledger}->{code})
      {
        push @errors, sprintf($self->{errorCodes}->{4}, "ledger", "code", $ledger->{ledger}->{code});
      }
      else
      {
        # first make sure the code hasn't been duplicated under a different name.
        my $found = 0;
        foreach my $tmpLedger (keys %encounteredLedgers)
        {
          if ($ledger->{ledger}->{code} eq $encounteredLedgers{$tmpLedger}->{ledger}->{code})
          {
            push @errors, sprintf($self->{errorCodes}->{4}, "ledger", "code", $ledger->{ledger}->{code});
            $found = 1;
            last;
          }
        }
        if (!$found)
        {
          $encounteredLedgers{$ledger->{ledger}->{name}} = $ledger;
          my %subEntityCodes = ();
          my %journalCodes = ();

          foreach my $subEntity (@{$ledger->{subEntities}})
          {
            $subEntityCodes{$subEntity->{code}} = $subEntity->{name};
          }
          foreach my $journal (@{$ledger->{journals}})
          {
            $journalCodes{$journal->{code}} = $journal->{name};
          }

          # now, if there are any journalEntries defined, make sure that all the accounts, subAccounts, subEntities,
          # journals are defined and available for reference.
          foreach my $journalEntry (@{$ledger->{journalEntries}})
          {
            if (!exists $journalCodes{$journalEntry->{journalCode}})
            {
              push @errors, sprintf($self->{errorCodes}->{6}, "JournalEntry", "journal", $journalEntry->{journalCode});
            }
            # now process all the rows for this journal entry.
            for (my $i=0; $i < scalar (@{$journalEntry->{rowObjects}}); $i++)
            {
              my $row = $journalEntry->{rowObjects}[$i];
              if (!exists $accountCodes{$row->{accountNum}})
              {
                push @errors, sprintf($self->{errorCodes}->{7}, $i, "account", $row->{accountNum});
              }
              if ($row->{subAccountNum} ne "-1" && !exists $subAccountCodes{$row->{subAccountNum}})
              {
                push @errors, sprintf($self->{errorCodes}->{7}, $i, "subAccount", $row->{subAccountNum});
              }
              if ($row->{subEntity} && !exists $subEntityCodes{$row->{subEntity}})
              {
                push @errors, sprintf($self->{errorCodes}->{7}, $i, "subEntity", $row->{subEntity});
              }
            }
          }
        }
      }
    }
  }

  return ((scalar @errors > 0 ? 0 : 1), \@errors);
}

=item string generateXML(void)
    Creates an XML file based upon the info stored in the
    Portal::Accounting::Objects::TransactionImport object.
    It first calls isValid() to make sure this is possible.
    If not then we die with an informative error message.

=cut

sub generateXML
{
  my $self = shift;
  my $result = "";
  my $errStr = "Portal::Accounting::Objects::TransactionImport->generateXML()  - Error:";

  my @valid = $self->isValid();
  if ($valid[0])
  {
    $result .= <<"END_OF_XML";
<?xml version="1.0" encoding="ISO-8859-1"?>
<transactionImport version="$self->{version}" timestamp="$self->{timestamp}" genSysId="$self->{genSysId}">
END_OF_XML
    $result .= "  <accounts>\n";
    foreach my $account (@{$self->{accounts}})
    {
      $result .= $account->printXML(fullXML => 0, indent => 4);
    }
    $result .= "  </accounts>\n";
    $result .= "  <subAccounts>\n";
    foreach my $subAccount (@{$self->{subAccounts}})
    {
      $result .= $subAccount->printXML(fullXML => 0, indent => 4);
    }
    $result .= "  </subAccounts>\n";
    $result .= "  <stocks>\n";
    foreach my $stock (@{$self->{stocks}})
    {
      $result .= $stock->printXML(fullXML => 0, indent => 4);
    }
    $result .= "  </stocks>\n";
    $result .= "  <ledgers>\n";
    foreach my $ledger (@{$self->{ledgers}})
    {
      $result .= $ledger->printXML(fullXML => 0, indent => 4);
    }
    $result .= "  </ledgers>\n";
    $result .= <<"END_OF_XML";
</transactionImport>
END_OF_XML
  }
  else
  {
    $result .= "$errStr Data not valid!\n\n";
    $result .= join("\n", @{$valid[1]}) . "\n";
    die $result;
  }

  return $result;
}

=item string encodeEntities(string)

 requires: string - string to encode
 optional:
 returns: string that has been encoded.
 summary: replaces all special characters with their XML entity
          equivalent. " => &quot;

=cut

sub encodeEntities
{
  my $self = shift;
  my %args = ( string => "", @_ );
  my $string = $args{string};

  my @entities = ('&', '"', '<', '>', '\n');
  my %entities = ('&' => '&amp;', '"' => '&quot;', '<' => '&lt;', '>' => '&gt;', '\n' => '\\n');

  return $string if (length $string == 0);

  foreach my $entity (@entities)
  {
    $string =~ s/$entity/$entities{$entity}/g;
  }

  return $string;
}

=back

=cut

1;
__END__

=head1 VARIABLES

  version - version of the XML file parsed

  dataFile - the name of the file parsed or the string of XML

  timestamp - the timestamp when this file was created

  genSysId - the id used by the generating system to identify this Import

  accounts - array of Account Objects

  subAccounts - array of SubAccount Objects

  ledgers - array of LedgerContainer Objects


  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin at pcxperience.com)

=head1 SEE ALSO

perl(1), Portal::Accounting::XML::TransactionParser(3), Portal::Accounting::Objects::Account(3), Portal::Accounting::Objects::SubAccount(3), Portal::Accounting::Objects::LedgerContainer(3)

=cut
