# TransactionRowObject.pm - Contains the data for a debitcredit_tb table row.
# Created by Jason A. Pattie, 11/07/2000.
#$Id: TransactionRowObject.pm,v 1.6 2004/05/28 22:15:34 pattieja Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::TransactionRowObject;
use strict;
use vars qw($VERSION @ISA @EXPORT $AUTOLOAD);

require Exporter;

@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.02';

# new
# instantiates an instance of the TransactionRowObject for the Accounting Package
sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;
  my $errorHeader = "TransactionRowObject->new(): ";
  my %args = ( rowDate => undef, comment => '', subEntity => '', accountNum => undef, subAccountNum => '-1',
               dorc => undef, amount => undef, d_amount => undef, c_amount => undef, cleared => ' ', @_ );

  $self->{rowDate}       = $args{'rowDate'};
  $self->{comment}       = $args{'comment'};
  $self->{subEntity}     = $args{'subEntity'};
  $self->{accountNum}    = $args{'accountNum'};
  $self->{subAccountNum} = $args{'subAccountNum'};
  $self->{dorc}          = $args{'dorc'};
  $self->{amount}        = $args{'amount'};
  $self->{d_amount}      = $args{'d_amount'};
  $self->{c_amount}      = $args{'c_amount'};
  $self->{cleared}       = $args{'cleared'};

  $self->{error} = !$self->isValid; # validation
  if ($self->{error})
  {
    $self->{errorString} = $errorHeader . "Error!<br>\n" . $self->{errorString};
  }

  return $self;
}

sub DESTROY
{
  my $self = shift;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://;	# strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

# isValid
# validates the object in its current state
sub isValid
{
  my $self = shift;
  my $errorHeader = "TransactionRowObject->isValid(): ";

  if ((not defined $self->{rowDate}) or $self->{rowDate} eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "The row date does not have a value!";
    return 0;
  }
  if (not defined $self->{comment})
  {
    $self->{comment} = "";
  }
  if (not defined $self->{subEntity})
  {
    $self->{subEntity} = "";
  }
  if ((not defined $self->{accountNum}) or $self->{accountNum} !~ /^-?[0-9][0-9]*$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid account number!  The value is either missing or does not appear to be numeric.";
    return 0;
  }
  if (not defined $self->{subAccountNum})
  {
    $self->{subAccountNum} = -1;
  }
  if (defined $self->{subAccountNum} and $self->{subAccountNum} !~ /^-?[0-9][0-9]*$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid sub-account number!  The value does not appear to be numeric.";
    return 0;
  }
  if ((not defined $self->{dorc}) or $self->{dorc} eq "" or ($self->{dorc} ne "true" and $self->{dorc} ne "false" and $self->{dorc} ne "1" and $self->{dorc} ne "0"))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid debit/credit indicator! ('$self->{dorc}')";
    return 0;
  }
  if (not defined $self->{amount})
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Amount value missing!";
    return 0;
  }
  if ($self->{amount} !~ /[0-9]*.[0-9]{2}/)
  {
    $self->{error} = 1;
    if ($self->{amount} =~ /\-/)
    {
      $self->{errorString} = $errorHeader . "The amount cannot be a negative value!";
    }
    else
    {
      $self->{errorString} = $errorHeader . "Invalid amount!  The value does not appear to be numeric.";
    }
    return 0;
  }
  if ((not defined $self->{cleared}) or $self->{cleared} eq "")
  {
    $self->{cleared} = ' ';
  }

  return 1;
}

# setError
# requires: errorString
# returns: nothing
sub setError
{
  my $self = shift;
  my %args = ( errorString => "", @_ );
  my $errorString = $args{errorString};
  
  $self->{errorString} = $errorString;
  $self->{error} = 1;
}

# didErrorOccur
# returns the value of error.
sub didErrorOccur
{
  my $self = shift;

  return $self->{error};
}

# errorMessage
# returns the value of errorString
sub errorMessage
{
  my $self = shift;

  return $self->{errorString};
}

=item string printXML(fullXML bool, indent int)

 This routine returns an XML snippet for this object for use in such
 things as exporting data

 fullXML - defaults to 0 (false).  If true, we output the xml header tag.
 indent - defaults to 2.  Specifies how many spaces in our top level tag
          needs to be.  If fullXML is true, then we force 0, otherwise we
          use what the user specifies.

=cut

sub printXML
{
  my $self = shift;
  my $result = "";
  my %args = ( fullXML => 0, indent => 2, @_ );
  my $fullXML = $args{fullXML};
  my $indent = ($fullXML ? 0 : $args{indent});  # start at 0 indentation if we are making a full XML document.
  my $spaceStr = " " x $indent;

  my %mapping = ("comment" => "comment", "accountCode" => "accountNum", "date" => "rowDate",
                 "subAccountCode" => "subAccountNum", "dorc" => "dorc", "amount" => "amount",
                 "subEntityCode" => "subEntity", "cleared" => "cleared");

  if ($self->isValid())
  {
    if ($fullXML)
    {
      $result .= "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>";
    }
    $result .= $spaceStr . "<line>\n";
    my @tags = qw{dorc accountCode amount subAccountCode comment subEntityCode cleared};
    push @tags, "date" if ($fullXML);
    foreach my $name (@tags)
    {
      $result .= $spaceStr . "  <$name>" . $self->{$mapping{$name}} . "</$name>\n";
    }
    $result .= $spaceStr . "</line>\n";
  }
  else
  {
    $result .= "Data not valid!\n\n";
    $result .= $self->errorMessage();
    die $result;
  }

  return $result;
}

1;
__END__

=head1 NAME

TransactionRowObject - Object that contains a row of data used to build a transaction in the accounting database.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::TransactionRowObject;
  my $transrowobj = TransactionRowObject->new(rowDate => '12-14-1999',
                             comment => '',
                             subEntity => '',
                             accountNum => 1001,
                             subAccountNum => -1,
                             dorc => 'true',
                             amount => '2579.63',
                             cleared => 'r');
  if ($transrowobj->didErrorOccur())
  {
    die $transrowobj->errorMessage();
  }

  # this object is generally used to build an array passed to
  # the constructor of a TransactionObject
  my @rowObjectArray;
  $rowObjectArray[0] = $transrowobj;
  $rowObjectArray[1] = $transrowobj;
  $rowObjectArray[1]->dorc('false');

  # check for validness of row objects values
  # before adding to TransactionObject
  if ($rowObjectArray[1]->isValid())
  {
    ...
  }
  else
  {
    die $rowObjectArray[1]->errorMessage();
  }

=head1 DESCRIPTION

TransactionRowObject is used as a container class that is incorporated
into the TransactionObject class.  This class is used to build an array
that is given to the TransactionObject class.

=head1 Exported FUNCTIONS

  scalar new(rowDate, comment, subEntity, accountNum, subAccountNum, dorc,
             amount, cleared)
    Creates a new instance of the Portal::Accounting::Objects::TransactionRowObject
    and validates the values contained in it.

  void setError(errorString)
    Sets error = 1 and $self->{errorString} = errorString.

  scalar didErrorOccur(void)
    Returns the value of error.

  scalar errorMessage(void)
    Returns the value of errorString.

  scalar isValid(void)
    Runs through the data fields contained in the class to make sure
    they contain valid data.  This function will be called
    automatically when embedded in a TransactionObject.

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.

=head1 AUTHOR

Jason A. Pattie (mailto:pattieja@pcxperience.com)

=head1 SEE ALSO

perl(1), TransactionObject(3)

=cut
