# Security.pm - The Object Class that provides methods to work with SecurityEventObject objects
# and TransactionObject objects.  Added routines to work with LotNodeObjects.
# Created by Jason A. Pattie, 11/13/2000.

#$Id: Security.pm,v 1.55 2004/06/21 19:48:55 pattieja Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::DB::Security;
use strict;
use Portal::Accounting::Objects::SecurityEventObject;
use Portal::Accounting::DB::Transaction;
use Portal::Accounting::DB::Main;
use Portal::Accounting::Objects::TransactionObject;
use Portal::Accounting::Objects::TransactionRowObject;
use Portal::Accounting::Objects::LotNodeObject;
use Portal::Accounting::Objects::PortfolioObject;
use Portal::Accounting::Objects::PortfolioAccountObject;
use Portal::Accounting::Objects::PortfolioLotObject;
use Portal::Accounting::Objects::PortfolioStockObject;
use Portal::Accounting::Objects::InvestmentJournalViewObject;
use Portal::Accounting::Objects::CGLObject;
use Portal::Accounting::Objects::CGLDisplayObject;
use Portal::Accounting::Objects::CGLAssetObject;
use Portal::Accounting::Objects::CGLAssets;
use Portal::Accounting::Objects::CGLStockLotObject;
use Portal::Accounting::Objects::CGLStockObject;
use Portal::Accounting::Objects::CGLStocks;
use Portal::Accounting::Objects::CGLManual;
use Portal::Accounting::Objects::CGLManualEntry;
use DBIWrapper;
use vars qw($VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::DB::Main Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.02';

my $shortTermCG = 4200;
my $longTermCG = 4201;
my $shortTermCL = 5200;
my $longTermCL = 5201;

=head1 NAME

Security - Object that contains functions that work with the sec_event_tb table and other associated database tables.

=head1 SYNOPSIS

  use Portal::Accounting::DB::Security;
  my $secobj = Portal::Accounting::DB::Security->new( dbHandle => $dbHandle );
  if ($secobj->didErrorOccur())
  {
    die $secobj->errorMessage();
  }

=head1 DESCRIPTION

Security is a class that provides data access functions to do work
on security transactions.  These methods can be used to write
SecurityEventObject objects to the database or update an existing
record in the database.  These methods will also do the necessary
work in the journal_entry_tb and debitcredit_tb tables (building
TransactionObject objects and calling Transaction class methods
when necessary).

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(dbHandle)

 Creates a new instance of the Portal::Accounting::DB::Security object.
 See Portal::Base(3) for a listing of required arguments.

 requires:
   dbHandle

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{debug} = 0;

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.
  $self->{transaction} = Portal::Accounting::DB::Transaction->new( dbHandle => $self->{dbHandle}, langObj => $self->{langObj} );

  if ($self->{transaction}->didErrorOccur())
  {
    $self->setError("Error instantiating Transaction object!<br>\n" . $self->{transaction}->errorMessage);
    return $self;
  }

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

# getNewLotNumber
# looks up the next number from the new_lot_number_seq sequence in the database
# returns undef on failure or the new lot number on success
sub getNewLotNumber
{
  my $self = shift;
  my $errorHeader = "Security->getNewLotNumber(): ";
  if ($self->{debug})
  {
     $self->{errorString} .= "Method: getNewLotNumber<br>\n";
  }
  my %args = ( lot => undef, @_ );

  my $sth = $self->{dbHandle}->read(sql => "SELECT nextval('new_lot_number_seq')");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving new lot number!\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }
  my @info = $sth->fetchrow_array;
  my $newLot = $info[0];

  return $newLot;
}

# propagateLot
# this function recursively calls itself until there are no more nodes/events to work on that the node it is called with
# has any interactions with
sub propagateLot
{
  my $self = shift;
  my $errorHeader = "Security->propagateLot(): ";
  if ($self->{debug})
  {
     $self->{errorString} .= "Method: propogateLot<br>\n";
  }
  my %args = ( lotNodeObj => undef, date => undef, @_ );
  my $lotNodeObj = $args{'lotNodeObj'};

  # validate node and date sent in
  if ((not defined $args{'lotNodeObj'}) || $lotNodeObj eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "LotNodeObject not specified or empty!";
    return 0;
  }
  if (ref($args{'lotNodeObj'}) ne "Portal::Accounting::Objects::LotNodeObject")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid LotNodeObject! (" . ref($args{'lotNodeObj'}) . ")!";
    return 0;
  }

  # determine if there is the need to continue based on whether the number of shares for the lot is zero or not
  if ($lotNodeObj->shares == 0.0)
  {
    # do not propagate this node; it will only retain the values it currently has for balances and would only clutter
    # the database
    return 1;
  }
  my $secObj = $self->getSecurityEvent(eventRefNum => $lotNodeObj->eventRef);
  if (not defined $secObj)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving security event object for event = " . $lotNodeObj->eventRef . "!";
    return 0;
  }
  if ($secObj->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving security event object!<br>\n" . $secObj->errorMessage;
    return 0;
  }
  # swap the values in the secObj object if they are backwards compared to the lot node object
  if ($lotNodeObj->subAccountNum ne $secObj->subAccountNum)
  {
    # swap counters
    my $tempCounter = $secObj->counter2;
    $secObj->counter2($secObj->counter);
    $secObj->counter($tempCounter);
    # swap sub-account numbers
    my $tempSubAccountNum = $secObj->subAccountNum2;
    $secObj->subAccountNum2($secObj->subAccountNum);
    $secObj->subAccountNum($tempSubAccountNum);
    # swap new shares with old shares
    my $tempOldShares = $secObj->oldShares;
    $secObj->oldShares($secObj->newShares);
    $secObj->newShares($tempOldShares);
    # swap cost percentage
    $secObj->costPercentage(1.0 - $secObj->costPercentage);
  }

  # get the starting date of the ledger for the current lot node


  # get the next event to process based on the node and the security event object
  # also make sure that the dates for the events gathered are greater than or equal to the beginning date of the ledger for the lots
  my $sth = $self->{dbHandle}->read(sql => "SELECT event_ref FROM sec_event_tb WHERE event_code <> 'memo' AND ((L = '" . $lotNodeObj->lot . "' AND calc_date = '" . $lotNodeObj->eventDate . "' AND counter > '" . $secObj->counter . "') OR (L = '" . $lotNodeObj->lot . "' AND calc_date = (SELECT min(calc_date) FROM sec_event_tb WHERE L = '" . $lotNodeObj->lot . "' AND calc_date > '" . $lotNodeObj->eventDate . "')) OR (L IS NULL AND S = '" . $lotNodeObj->subAccountNum . "' AND calc_date = '" . $lotNodeObj->eventDate . "' AND counter > '" . $secObj->counter . "') OR (L IS NULL AND S = '" . $lotNodeObj->subAccountNum . "' AND calc_date = (SELECT min(calc_date) FROM sec_event_tb WHERE L IS NULL AND S = '" . $lotNodeObj->subAccountNum . "' AND calc_date > '" . $lotNodeObj->eventDate . "'))) ORDER BY calc_date ASC, counter ASC LIMIT 1");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error getting next security event to process!<br>\n" . $self->{dbHandle}->errorMessage;
    return 0;
  }
  my @info = $sth->fetchrow_array;
  if ((not defined @info) || (not defined $info[0]) || $info[0] eq "")
  {
    # no more events to process for this node
    return 1;
  }

  my $nextSecObj = $self->getSecurityEvent(eventRefNum => $info[0]);
  if (not defined $nextSecObj)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving next security event from database!  Event reference = '$info[0]'";
    return 0;
  }
  if ($nextSecObj->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving next security event from database!<br>\n" . $nextSecObj->errorMessage;
    return 0;
  }

  # calculate the new balance for the lot based on the next event
  if ($nextSecObj->eventCode eq "20") # sell
  {
# ***--- Sell code ---***
    # use the node's values as the current balance of cost and shares for the lot
    # calculate the sell value and the sell cost to determine if the sell results in a loss or a gain
    my $sellValue;
    my $sellCost;
    my $gainLoss;
    my @transrowObject = [];
    my $sharesLeft;
    my $costBalance;

    my $capitalGainAccount;
    my $capitalLossAccount;

    # determine if the sell is occurring after one year that the security has been held in order to assign long or short
    # term gains or losses
    my $sth = $self->{dbHandle}->read(sql => "SELECT capital_gain_loss((SELECT event_date FROM lot_node_tb WHERE lot = '" . $lotNodeObj->buyLot . "' AND parent_id = '-1'), '" . $nextSecObj->transDate . "')");
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error evaluating date of sell for capital gains/losses!<br>\n" . $self->{dbHandle}->errorMessage;
      return 0;
    }
    my @info = $sth->fetchrow_array;
    if (defined @info)
    {
      if ($info[0] eq "0")
      {
        $capitalGainAccount = $shortTermCG;
        $capitalLossAccount = $shortTermCL;
      }
      else
      {
        $capitalGainAccount = $longTermCG;
        $capitalLossAccount = $longTermCL;
      }
    }
    else
    {
      $capitalGainAccount = $shortTermCG;
      $capitalLossAccount = $shortTermCL;
    }

    if ($nextSecObj->shares == $lotNodeObj->shares)
    {
      $sellCost = sprintf("%.2f", $lotNodeObj->cost);
      $sharesLeft = 0;
      $costBalance = 0.00;
    }
    else
    {
      my $costPerShare = ($lotNodeObj->cost / $lotNodeObj->shares);  # /
      $sellCost = sprintf("%.2f", $costPerShare * $nextSecObj->shares);
      $sharesLeft = $lotNodeObj->shares - $nextSecObj->shares;
      $costBalance = $lotNodeObj->cost - $sellCost;
    }
    $sellValue = sprintf("%.2f", ($nextSecObj->price * $nextSecObj->shares) - $nextSecObj->commission);
    $gainLoss = sprintf("%.2f", abs($sellValue - $sellCost));

    $transrowObject[0] = Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $nextSecObj->accountNum, subAccountNum => $nextSecObj->subAccountNum, dorc => "false", amount => "$sellCost", langObj => $self->{langObj});
    $transrowObject[1] = Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $nextSecObj->accountNum, subAccountNum => "-1", dorc => "true", amount => "$sellValue", langObj => $self->{langObj});
    if ($sellValue > $sellCost)
    {
      # sold for more than what it cost; gain (use capital gain account)
      $transrowObject[2] = Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => "$capitalGainAccount", subAccountNum => $nextSecObj->subAccountNum, dorc => "false", amount => "$gainLoss", langObj => $self->{langObj});
    }
    elsif ($sellValue < $sellCost)
    {
      # sold for less than what it cost; loss (use capital loss account)
      $transrowObject[2] = Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => "$capitalLossAccount", subAccountNum => $nextSecObj->subAccountNum, dorc => "true", amount => "$gainLoss", langObj => $self->{langObj});
    }
    else
    {
      # sold for what it cost; no gain/loss (use capital gain account)
      $transrowObject[2] = Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => "$capitalGainAccount", subAccountNum => $nextSecObj->subAccountNum, dorc => "false", amount => "0.00", langObj => $self->{langObj});
    }
    my $transobj = Portal::Accounting::Objects::TransactionObject->new(ledgerCode => $nextSecObj->ledgerCode,
                                                                       journalCode => "IS", # hardcoded investments journal code
                                                                       transDate => $nextSecObj->transDate,
                                                                       rowObjects => \@transrowObject, langObj => $self->{langObj});
    if ($transobj->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Failed to create new transaction object!<br>\n" . $transobj->errorMessage;
      return 0;
    }

    # insert the transaction into the database
    my $refNum = $self->{transaction}->insertTransaction(obj => $transobj);
    if ($refNum == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting journal transaction for sell!<br>\n" . $self->{transaction}->errorMessage;
      return 0;
    }
    # insert a new lot node into the database
    my $newLotNodeObj = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $nextSecObj->ledgerCode,
                                                                        eventDate => $nextSecObj->transDate,
                                                                        parentId => $lotNodeObj->nodeId,
                                                                        eventRef => $nextSecObj->eventRefNum,
                                                                        journalRef => $refNum,
                                                                        subEntity => $lotNodeObj->subEntity,
                                                                        accountNum => $nextSecObj->accountNum,
                                                                        subAccountNum => $lotNodeObj->subAccountNum,
                                                                        lot => $lotNodeObj->lot,
                                                                        buyLot => $lotNodeObj->buyLot,
                                                                        cost => sprintf("%.2f", $costBalance),
                                                                        shares => sprintf("%.6f", $sharesLeft), 
                                                                        langObj => $self->{langObj});
    my $lotID = $self->insertLotNodeObject(obj => $newLotNodeObj);
    if ($lotID == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting new lot for sell!<br>\n" . $self->{errorString};
      return 0;
    }

    if (not $self->propagateLot(lotNodeObj => $newLotNodeObj))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating sell event!<br>\n" . $self->{errorString};
      return 0;
    }
  }
  elsif ($nextSecObj->eventCode eq "30") # stock transfer
  {
# ***--- Stock Transfer code ---***
    if ($nextSecObj->shares > $lotNodeObj->shares)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Attempting to transfer more shares than exist in the lot node table! (transfer: " . $nextSecObj->shares . ", exist: " . $lotNodeObj->shares . ")";
      return 0;
    }
    # calculate the transfer value and the transfer cost
    my $transferValue;
    my $costPerShare;
    my @transrowObject = [];
    my $costBalance;
    my $sharesLeft;

    if ($nextSecObj->shares == $lotNodeObj->shares)
    {
      $transferValue = sprintf("%.2f", $lotNodeObj->cost);
      $costBalance = 0;
      $sharesLeft = 0;
    }
    else
    {
      $costPerShare = sprintf("%.2f", ($lotNodeObj->cost / $lotNodeObj->shares));  # / just a comment that changes the colors back to normal after the editor gets confused on a division instead of a pattern match
      $transferValue = sprintf("%.2f", $costPerShare * $nextSecObj->shares);
      $costBalance = $lotNodeObj->cost - $transferValue;
      $sharesLeft = $lotNodeObj->shares - $nextSecObj->shares;
    }

    $transrowObject[0] = Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $nextSecObj->oldAccountNum, subAccountNum => $nextSecObj->subAccountNum, dorc => "false", amount => sprintf("%.2f", $transferValue), langObj => $self->{langObj});
    $transrowObject[1] = Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $nextSecObj->newAccountNum, subAccountNum => $nextSecObj->subAccountNum, dorc => "true", amount => sprintf("%.2f", $transferValue), langObj => $self->{langObj});
    my $transobj = Portal::Accounting::Objects::TransactionObject->new(ledgerCode => $nextSecObj->ledgerCode,
                                                                       journalCode => "IS", # hardcoded investments journal code
                                                                       transDate => $nextSecObj->transDate,
                                                                       rowObjects => \@transrowObject, langObj => $self->{langObj});
    if ($transobj->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Failed to create new transaction object!<br>\n" . $transobj->errorMessage;
      return 0;
    }

    # insert the transaction into the database
    my $refNum = $self->{transaction}->insertTransaction(obj => $transobj);
    if ($refNum == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting journal transaction for stock transfer!<br>\n" . $self->{transaction}->errorMessage;
      return 0;
    }

    # create two new lot nodes, one for the new lot that is generated from the event and one for the old lot to decrease
    # its balances (shares, cost)
    my $newLotNodeObj1 = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $nextSecObj->ledgerCode,
                                                                         eventDate => $nextSecObj->transDate,
                                                                         parentId => $lotNodeObj->nodeId,
                                                                         eventRef => $nextSecObj->eventRefNum,
                                                                         journalRef => $refNum,
                                                                         subEntity => $nextSecObj->subEntity,
                                                                         accountNum => $nextSecObj->newAccountNum,
                                                                         subAccountNum => $lotNodeObj->subAccountNum,
                                                                         lot => $nextSecObj->newLot,
                                                                         buyLot => $lotNodeObj->buyLot,
                                                                         cost => sprintf("%.2f", $transferValue),
                                                                         shares => sprintf("%.6f", $nextSecObj->shares), langObj => $self->{langObj});
    my $lotID1 = $self->insertLotNodeObject(obj => $newLotNodeObj1);
    if ($lotID1 == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting new lot node (1) for stock transfer!<br>\n" . $self->{errorString};
      return 0;
    }

    # insert an entry in the lot_parent_tb table for the new lot
    $self->{dbHandle}->write(sql => "INSERT INTO lot_parent_tb (lot, buy_lot) VALUES (?, ?)", plug => [ $newLotNodeObj1->lot, $newLotNodeObj1->buyLot ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting lot entry into lot_parent_tb table for new lot!<br>\n" . $self->{dbHandle}->errorMessage;
      $self->{dbHandle}->rollback;
      return 0;
    }

    my $newLotNodeObj2 = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $nextSecObj->ledgerCode,
                                                                         eventDate => $nextSecObj->transDate,
                                                                         parentId => $lotNodeObj->nodeId,
                                                                         eventRef => $nextSecObj->eventRefNum,
                                                                         journalRef => $refNum,
                                                                         subEntity => $lotNodeObj->subEntity,
                                                                         accountNum => $nextSecObj->oldAccountNum,
                                                                         subAccountNum => $lotNodeObj->subAccountNum,
                                                                         lot => $lotNodeObj->lot,
                                                                         buyLot => $lotNodeObj->buyLot,
                                                                         cost => sprintf("%.2f", $costBalance),
                                                                         shares => sprintf("%.6f", $sharesLeft), langObj => $self->{langObj});
    my $lotID2 = $self->insertLotNodeObject(obj => $newLotNodeObj2);
    if ($lotID2 == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting new lot node (2) for stock transfer!<br>\n" . $self->{errorString};
      return 0;
    }

    # propagate both nodes
    if (not $self->propagateLot(lotNodeObj => $newLotNodeObj1))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating stock transfer event (node 1)!<br>\n" . $self->{errorString};
      return 0;
    }
    if (not $self->propagateLot(lotNodeObj => $newLotNodeObj2))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating stock transfer event (node 2)!<br>\n" . $self->{errorString};
      return 0;
    }
  }
  elsif ($nextSecObj->eventCode eq "40") # fee
  {
# ***--- Fee code ---***
    my $amount;
    my $dorc;
    my $cord;
    my $costBalance;

    $amount = sprintf("%.2f", abs($nextSecObj->amount));
    if ($nextSecObj->amount < 0)
    {
      $dorc = "false";
      $cord = "true";
      $costBalance = $lotNodeObj->cost - $amount;
    }
    else
    {
      $dorc = "true";
      $cord = "false";
      $costBalance = $lotNodeObj->cost + $amount;
    }
    # build a TransactionObject for the data
    my $transobj = Portal::Accounting::Objects::TransactionObject->new(ledgerCode => $nextSecObj->ledgerCode,
                                                                       journalCode => "IS", # investments journal code hardcoded
                                                                       transDate => $nextSecObj->transDate,
                                                                       rowObjects => [ Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $nextSecObj->accountNum, subAccountNum => $nextSecObj->subAccountNum, dorc => "$dorc", amount => sprintf("%.2f", $amount), langObj => $self->{langObj}),
                                                                       Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $nextSecObj->accountNum, subAccountNum => "-1", dorc => "$cord", amount => sprintf("%.2f", $amount), langObj => $self->{langObj}) ], langObj => $self->{langObj} );
    # insert the transaction into the database
    my $refNum = $self->{transaction}->insertTransaction(obj => $transobj);
    if ($refNum == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting journal transaction for fee!<br>\n" . $self->{transaction}->errorMessage;
      return undef;
    }

    # insert a new lot node into the database
    my $newLotNodeObj = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $nextSecObj->ledgerCode,
                                                                        eventDate => $nextSecObj->transDate,
                                                                        parentId => $lotNodeObj->nodeId,
                                                                        eventRef => $nextSecObj->eventRefNum,
                                                                        journalRef => $refNum,
                                                                        subEntity => $lotNodeObj->subEntity,
                                                                        accountNum => $nextSecObj->accountNum,
                                                                        subAccountNum => $lotNodeObj->subAccountNum,
                                                                        lot => $lotNodeObj->lot,
                                                                        buyLot => $lotNodeObj->buyLot,
                                                                        cost => sprintf("%.2f", $costBalance),
                                                                        shares => sprintf("%.6f", $lotNodeObj->shares), langObj => $self->{langObj});
    my $lotID = $self->insertLotNodeObject(obj => $newLotNodeObj);
    if ($lotID == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting new lot for fee!<br>\n" . $self->{errorString};
      return 0;
    }

    if (not $self->propagateLot(lotNodeObj => $newLotNodeObj))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating fee event!<br>\n" . $self->{errorString};
      return 0;
    }
  }
  elsif ($nextSecObj->eventCode eq "50") # return of capital
  {
# ***--- Return of Capital code ---***
    my $newCost;
    my $capitalGain = "";
    my @transrowObject = [];
    my $returnAmount;
    my $returnCost;

    my $capitalGainAccount;
    my $capitalLossAccount;

    # calculate the amount for the new cost (on a per share basis) and potentially capital gains
    $returnAmount = $lotNodeObj->shares * $nextSecObj->amount;
    if ($returnAmount > $lotNodeObj->cost)
    {
      $newCost = 0.00;
      $returnCost = sprintf("%.2f", abs($lotNodeObj->cost));
      $capitalGain = sprintf("%.2f", abs($returnAmount - $lotNodeObj->cost));
    }
    else
    {
      $newCost = sprintf("%.2f", abs($lotNodeObj->cost - $returnAmount));
      $returnCost = sprintf("%.2f", abs($returnAmount));
    }
    $returnAmount = sprintf("%.2f", abs($returnAmount));

    $transrowObject[0] = Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $lotNodeObj->accountNum, subAccountNum => $nextSecObj->subAccountNum, dorc => "false", amount => "$returnCost", langObj => $self->{langObj});
    $transrowObject[1] = Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $lotNodeObj->accountNum, subAccountNum => "-1", dorc => "true", amount => "$returnAmount", langObj => $self->{langObj});
    if ($capitalGain ne "")
    {
      # determine if the return of capital is occurring after one year that the security has been held in order to assign
      # long or short term gains or losses
      my $sth = $self->{dbHandle}->read(sql => "SELECT capital_gain_loss((SELECT event_date FROM lot_node_tb WHERE lot = '" . $lotNodeObj->buyLot . "' AND parent_id = '-1'), '" . $nextSecObj->transDate . "')");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error evaluating date of sell for capital gains/losses!<br>\n" . $self->{dbHandle}->errorMessage;
        return 0;
      }
      my @info = $sth->fetchrow_array;
      if (defined @info)
      {
        if ($info[0] eq "0")
        {
          $capitalGainAccount = $shortTermCG;
          $capitalLossAccount = $shortTermCL;
        }
        else
        {
          $capitalGainAccount = $longTermCG;
          $capitalLossAccount = $longTermCL;
        }
      }
      else
      {
        $capitalGainAccount = $shortTermCG;
        $capitalLossAccount = $shortTermCL;
      }

      $transrowObject[2] = Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => "$capitalGainAccount", subAccountNum => $nextSecObj->subAccountNum, dorc => "false", amount => "$capitalGain", langObj => $self->{langObj});
    }
    # build a TransactionObject for the data
    my $transobj = Portal::Accounting::Objects::TransactionObject->new(ledgerCode => $lotNodeObj->ledgerCode,
                                                                       journalCode => "IS", # investements journal code hardcoded
                                                                       transDate => $nextSecObj->transDate,
                                                                       rowObjects => \@transrowObject, langObj => $self->{langObj});
    # insert the transaction into the database
    my $refNum = $self->{transaction}->insertTransaction(obj => $transobj);
    if ($refNum == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting journal transaction for return of capital!<br>\n" . $self->{transaction}->errorMessage;
      return undef;
    }

    # insert a new lot node into the database
    my $newLotNodeObj = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $lotNodeObj->ledgerCode,
                                                                        eventDate => $nextSecObj->transDate,
                                                                        parentId => $lotNodeObj->nodeId,
                                                                        eventRef => $nextSecObj->eventRefNum,
                                                                        journalRef => $refNum,
                                                                        subEntity => $lotNodeObj->subEntity,
                                                                        accountNum => $lotNodeObj->accountNum,
                                                                        subAccountNum => $lotNodeObj->subAccountNum,
                                                                        lot => $lotNodeObj->lot,
                                                                        buyLot => $lotNodeObj->buyLot,
                                                                        cost => sprintf("%.2f", $newCost),
                                                                        shares => sprintf("%.6f", $lotNodeObj->shares), langObj => $self->{langObj});
    my $lotID = $self->insertLotNodeObject(obj => $newLotNodeObj);
    if ($lotID == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting new lot for return of capital!<br>\n" . $self->{errorString};
      return 0;
    }

    if (not $self->propagateLot(lotNodeObj => $newLotNodeObj))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating return of capital event!<br>\n" . $self->{errorString};
      return 0;
    }
  }
  elsif ($nextSecObj->eventCode eq "60" || $nextSecObj->eventCode eq "70") # stock swap/spin off
  {
# ***--- Spin Off/Stock Swap code ---***
    # swap the values in the nextSecObj object if they are backwards compared to the lot node object
    if ($lotNodeObj->subAccountNum ne $nextSecObj->subAccountNum)
    {
      # swap counters
      my $tempCounter = $nextSecObj->counter2;
      $nextSecObj->counter2($nextSecObj->counter);
      $nextSecObj->counter($tempCounter);
      # swap sub-account numbers
      my $tempSubAccountNum = $nextSecObj->subAccountNum2;
      $nextSecObj->subAccountNum2($nextSecObj->subAccountNum);
      $nextSecObj->subAccountNum($tempSubAccountNum);
      # swap new shares with old shares
      my $tempOldShares = $nextSecObj->oldShares;
      $nextSecObj->oldShares($nextSecObj->newShares);
      $nextSecObj->newShares($tempOldShares);
      # swap cost percentage
      $nextSecObj->costPercentage(1.0 - $nextSecObj->costPercentage);
    }
    # calculate new lot cost
    my $newLotCost = $nextSecObj->costPercentage * $lotNodeObj->cost;
    my $oldLotCost = $lotNodeObj->cost - $newLotCost;
    # calculate new lot's number of shares
    my $newShares = $nextSecObj->newShares / $nextSecObj->oldShares * $lotNodeObj->shares;  # /
    # get a new lot number
    my $newLotNumber = $self->getNewLotNumber();
    if (not defined $newLotNumber)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error getting new lot number for stock swap/spin off!<br>\n" . $self->{errorString};
      return 0;
    }
    # create a transaction to generate a new lot with this cost
    my $transobj = Portal::Accounting::Objects::TransactionObject->new(ledgerCode => $lotNodeObj->ledgerCode,
                                                                       journalCode => "IS", # hardcoded investment journal code
                                                                       transDate => $nextSecObj->transDate,
                                                                       memo => "$newShares shares for lot $newLotNumber",
                                                                       rowObjects => [ Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $lotNodeObj->accountNum, subAccountNum => $nextSecObj->subAccountNum, dorc => "false", amount => sprintf("%.2f", $newLotCost), langObj => $self->{langObj}),
                                                                       Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $lotNodeObj->accountNum, subAccountNum => $nextSecObj->subAccountNum2, dorc => "true", amount => sprintf("%.2f", $newLotCost), langObj => $self->{langObj}) ], langObj => $self->{langObj} );
    # insert transaction
    my $refNum = $self->{transaction}->insertTransaction(obj => $transobj);
    if ($refNum == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting journal transaction for spin off/stock swap!<br>\n" . $self->{transaction}->errorMessage;
      return undef;
    }

    # create two new lot nodes, one for the new lot that is generated from the event and one for the old lot to decrease
    # its balances (shares, cost)
    my $newLotNodeObj1 = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $lotNodeObj->ledgerCode,
                                                                         eventDate => $nextSecObj->transDate,
                                                                         parentId => $lotNodeObj->nodeId,
                                                                         eventRef => $nextSecObj->eventRefNum,
                                                                         journalRef => $refNum,
                                                                         subEntity => $lotNodeObj->subEntity,
                                                                         accountNum => $lotNodeObj->accountNum,
                                                                         subAccountNum => $nextSecObj->subAccountNum2,
                                                                         lot => $newLotNumber,
                                                                         buyLot => $lotNodeObj->buyLot,
                                                                         cost => sprintf("%.2f", $newLotCost),
                                                                         shares => sprintf("%.6f", $newShares), langObj => $self->{langObj});
    my $lotID1 = $self->insertLotNodeObject(obj => $newLotNodeObj1);
    if ($lotID1 == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting new lot node (1) for spin off/stock swap!<br>\n" . $self->{errorString};
      return 0;
    }

    # insert an entry in the lot_parent_tb table for the new lot
    $self->{dbHandle}->write(sql => "INSERT INTO lot_parent_tb (lot, buy_lot) VALUES (?, ?)", plug => [ $newLotNodeObj1->lot, $newLotNodeObj1->buyLot ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting lot entry into lot_parent_tb table for new lot!<br>\n" . $self->{dbHandle}->errorMessage;
      $self->{dbHandle}->rollback;
      return 0;
    }

    my $newLotNodeObj2 = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $lotNodeObj->ledgerCode,
                                                                         eventDate => $nextSecObj->transDate,
                                                                         parentId => $lotNodeObj->nodeId,
                                                                         eventRef => $nextSecObj->eventRefNum,
                                                                         journalRef => $refNum,
                                                                         subEntity => $lotNodeObj->subEntity,
                                                                         accountNum => $lotNodeObj->accountNum,
                                                                         subAccountNum => $lotNodeObj->subAccountNum,
                                                                         lot => $lotNodeObj->lot,
                                                                         buyLot => $lotNodeObj->buyLot,
                                                                         cost => sprintf("%.2f", $oldLotCost),
                                                                         shares => sprintf("%.6f", $lotNodeObj->shares), langObj => $self->{langObj});
    my $lotID2 = $self->insertLotNodeObject(obj => $newLotNodeObj2);
    if ($lotID2 == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting new lot node (2) for spin off/stock swap!<br>\n" . $self->{errorString};
      return 0;
    }

    # propagate both nodes
    if (not $self->propagateLot(lotNodeObj => $newLotNodeObj1))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating spin off/stock swap event (node 1)!<br>\n" . $self->{errorString};
      return 0;
    }
    if (not $self->propagateLot(lotNodeObj => $newLotNodeObj2))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating spin off/stock swap event (node 2)!<br>\n" . $self->{errorString};
      return 0;
    }
  }
  elsif ($nextSecObj->eventCode eq "80") # acquisition
  {
# ***--- Acquisition code ---***
    # calculate new lot's number of shares
    my $newShares = $nextSecObj->newShares / $nextSecObj->oldShares * $lotNodeObj->shares;  # /
    # get a new lot number
    my $newLotNumber = $self->getNewLotNumber();
    if (not defined $newLotNumber)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error getting new lot number for acquisition!<br>\n" . $self->{errorString};
      return 0;
    }
    # create a transaction to generate a new lot moving the cost of the old lot to the new lot
    my $transobj = Portal::Accounting::Objects::TransactionObject->new(ledgerCode => $lotNodeObj->ledgerCode,
                                                                       journalCode => "IS", # hardcoded investment journal code
                                                                       transDate => $nextSecObj->transDate,
                                                                       memo => "$newShares shares for lot $newLotNumber",
                                                                       rowObjects => [ Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $lotNodeObj->accountNum, subAccountNum => $nextSecObj->subAccountNum, dorc => "false", amount => sprintf("%.2f", $lotNodeObj->cost), langObj => $self->{langObj}),
                                                                       Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $nextSecObj->transDate, subEntity => $lotNodeObj->subEntity, accountNum => $lotNodeObj->accountNum, subAccountNum => $nextSecObj->subAccountNum2, dorc => "true", amount => sprintf("%.2f", $lotNodeObj->cost), langObj => $self->{langObj}) ], langObj => $self->{langObj} );
    # insert transaction
    my $refNum = $self->{transaction}->insertTransaction(obj => $transobj);
    if ($refNum == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting journal transaction for acquisition!<br>\n" . $self->{transaction}->errorMessage;
      return undef;
    }

    # create two new lot nodes, one for the new lot that is generated from the event and one for the old lot to decrease
    # its balances (shares, cost)
    my $newLotNodeObj1 = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $lotNodeObj->ledgerCode,
                                                                         eventDate => $nextSecObj->transDate,
                                                                         parentId => $lotNodeObj->nodeId,
                                                                         eventRef => $nextSecObj->eventRefNum,
                                                                         journalRef => $refNum,
                                                                         subEntity => $lotNodeObj->subEntity,
                                                                         accountNum => $lotNodeObj->accountNum,
                                                                         subAccountNum => $nextSecObj->subAccountNum2,
                                                                         lot => $newLotNumber,
                                                                         buyLot => $lotNodeObj->buyLot,
                                                                         cost => sprintf("%.2f", $lotNodeObj->cost),
                                                                         shares => sprintf("%.6f", $newShares), langObj => $self->{langObj});
    my $lotID1 = $self->insertLotNodeObject(obj => $newLotNodeObj1);
    if ($lotID1 == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting new lot (1) for acquisition!<br>\n" . $self->{errorString};
      return 0;
    }

    # insert an entry in the lot_parent_tb table for the new lot
    $self->{dbHandle}->write(sql => "INSERT INTO lot_parent_tb (lot, buy_lot) VALUES (?, ?)", plug => [ $newLotNodeObj1->lot, $newLotNodeObj1->buyLot ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting lot entry into lot_parent_tb table for new lot!<br>\n" . $self->{dbHandle}->errorMessage;
      $self->{dbHandle}->rollback;
      return 0;
    }

    my $newLotNodeObj2 = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $lotNodeObj->ledgerCode,
                                                                         eventDate => $nextSecObj->transDate,
                                                                         parentId => $lotNodeObj->nodeId,
                                                                         eventRef => $nextSecObj->eventRefNum,
                                                                         journalRef => $refNum,
                                                                         subEntity => $lotNodeObj->subEntity,
                                                                         accountNum => $lotNodeObj->accountNum,
                                                                         subAccountNum => $lotNodeObj->subAccountNum,
                                                                         lot => $lotNodeObj->lot,
                                                                         buyLot => $lotNodeObj->buyLot,
                                                                         cost => "0.00",
                                                                         shares => 0, langObj => $self->{langObj});
    my $lotID2 = $self->insertLotNodeObject(obj => $newLotNodeObj2);
    if ($lotID2 == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting new lot (2) for acquisition!<br>\n" . $self->{errorString};
      return 0;
    }

    # propagate both nodes
    if (not $self->propagateLot(lotNodeObj => $newLotNodeObj1))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating acquisition event (node 1)!<br>\n" . $self->{errorString};
      return 0;
    }
    if (not $self->propagateLot(lotNodeObj => $newLotNodeObj2))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating acquisition event (node 2)!<br>\n" . $self->{errorString};
      return 0;
    }
  }
  elsif ($nextSecObj->eventCode eq "90" || $nextSecObj->eventCode eq "100") # stock split/dividend
  {
# ***--- Stock Split/Dividend ---***
    my $totalShares = $nextSecObj->newShares / $nextSecObj->oldShares * $lotNodeObj->shares; # /

    # insert a new lot node into the database
    my $newLotNodeObj = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $lotNodeObj->ledgerCode,
                                                                        eventDate => $nextSecObj->transDate,
                                                                        parentId => $lotNodeObj->nodeId,
                                                                        eventRef => $nextSecObj->eventRefNum,
                                                                        subEntity => $lotNodeObj->subEntity,
                                                                        accountNum => $lotNodeObj->accountNum,
                                                                        subAccountNum => $lotNodeObj->subAccountNum,
                                                                        lot => $lotNodeObj->lot,
                                                                        buyLot => $lotNodeObj->buyLot,
                                                                        cost => sprintf("%.2f", $lotNodeObj->cost),
                                                                        shares => sprintf("%.6f", $totalShares), langObj => $self->{langObj});
    my $lotID = $self->insertLotNodeObject(obj => $newLotNodeObj);
    if ($lotID == -1)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting new lot for stock split/dividend!<br>\n" . $self->{errorString};
      return 0;
    }

    if (not $self->propagateLot(lotNodeObj => $newLotNodeObj))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating stock split/dividend event!<br>\n" . $self->{errorString};
      return 0;
    }
  }

  return 1;
}

# propagateDeleteLot
sub propagateDeleteLot
{
  my $self = shift;
  my $errorHeader = "Security->propagateDeleteLot(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "propagateDeleteLot<br>\n";
  }
  my %args = ( nodeId => undef, includeNodeId => 0, commit => 0, @_ );
  my $nodeId = $args{'nodeId'};
  my $includeNodeId = $args{'includeNodeId'};

  # validate node sent in
  if ((not defined $args{'nodeId'}) || $nodeId eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Node ID not specified or empty!";
    return 0;
  }
  if ($nodeId !~ /^[0-9]+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Node ID does not appear to be numeric!";
    return 0;
  }
  if ((not defined $args{'includeNodeId'}) || $includeNodeId eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "includeNodeId not specified or empty!";
    return 0;
  }
  if ($includeNodeId ne "0" && $includeNodeId ne "1")
  {
    $includeNodeId = "0";
  }

  # get the sub-tree for this lot
  my %lotNodes = $self->getLotTree(nodeId => $nodeId);
  # delete each node in the tree removing journal entries for those nodes that have an associated journal entry
  foreach my $lotkey (keys %lotNodes)
  {
    if (($includeNodeId eq "1") || ($includeNodeId eq "0" && ($lotkey ne $nodeId)))
    {
      if ($lotNodes{$lotkey}->journalRef > -1)
      {
        # delete the journal transaction associated with this node
        if (not $self->{transaction}->deleteTransaction(refNum => $lotNodes{$lotkey}->journalRef))
        {
          $self->{error} = 1;
          $self->{errorString} = $errorHeader . "Error deleting journal entry (ref = " . $lotNodes{$lotkey}->journalRef . ")!<br>\n" . $self->{transaction}->errorMessage;
          return 0;
        }
      }
      # remove the node from the lot_node_tb table
      my $rowCount = $self->{dbHandle}->write(sql => "DELETE FROM lot_node_tb WHERE node_id = " . $lotNodes{$lotkey}->nodeId);
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error deleting lot node from database!<br>\n" . $self->{dbHandle}->errorMessage;
        $self->{dbHandle}->rollback;
        return 0;
      }
      if ($rowCount == 0)
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "No rows were deleted when attempting to delete lot node (nodeID = " . $lotNodes{$lotkey}->nodeId . ")!";
        $self->{dbHandle}->rollback;
        return 0;
      }
    }
  }

  # delete all entries in the lot_parent_tb table that no longer exist in the lot_node_tb table
  $self->{dbHandle}->write(sql => "DELETE FROM lot_parent_tb WHERE lot NOT IN (SELECT DISTINCT lot FROM lot_node_tb)");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error deleting lot entries in lot_parent_tb table that no longer exist in the lot_node_tb table!<br>\n" . $self->{dbHandle}->errorMessage;
    $self->{dbHandle}->rollback;
    return 0;
  }

  if ($args{'commit'} eq "1")
  {
    # finish the transaction
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
      return 0;
    }
  }

  return 1;
}

# deleteSecurityEvent
# takes a SecurityEventObject reference number and uses the local Transaction object to delete any necessary journal
# entries from the database that are affected by deletion of the security event;  determines which event code is being
# used and validates the SecurityEventObject before working with the data
sub deleteSecurityEvent
{
  my $self = shift;
  my $errorHeader = "Security->deleteSecurityEvent(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "deleteSecurityEvent<br>\n";
  }
  my %args = ( eventRefNum => undef, @_ );
  my $eventRefNum = $args{'eventRefNum'};

  if (not defined $args{'eventRefNum'})
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Event reference number not specified!";
    return 0;
  }
  if ($eventRefNum !~ /^(\d+)$/)
  {
    $self->setError(errorString => $errorHeader . "Event reference number = '$eventRefNum' is invalid!\n");
    return 0;
  }
  # retrieve the security event object for this reference number
  my $secobj = $self->getSecurityEvent(eventRefNum => $eventRefNum);
  if (not defined $secobj)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving security event object!";
    return 0;
  }
  if ($secobj->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving security event object!<br>\n" . $secobj->errorMessage;
    return 0;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid security event object!<br>\n" . $secobj->errorMessage;
    return 0;
  }

  # get nodes to delete based on the security event object's event reference number
  my %nodeArray = ();
  my $sth = $self->{dbHandle}->read(sql => "SELECT node_id FROM lot_node_tb WHERE event_ref = '$eventRefNum'");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving lot nodes with event_ref = " . $eventRefNum . "!<br>\n" . $self->{dbHandle}->errorMessage;
    return 0;
  }
  # build a list with all returned node IDs
  while ((my @info = $sth->fetchrow_array))
  {
    my $lotNode = $self->getLotNode(nodeId => $info[0]);
    if (not defined $lotNode || $self->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error retrieving lot node object for node ID = '" . $info[0] . "'!<br>\n" . $self->errorMessage;
      return 0;
    }
    $nodeArray{$lotNode->{nodeId}} = $lotNode;
  }

  # get a list of all previous nodes (unless security event is a buy[10])
  my %previousNodeArray = ();
  my $listOfNodes2 = join(',', (keys %nodeArray));
  if (length $listOfNodes2 == 0)
  {
    $self->setError(errorString => $errorHeader . "No lot node entries returned for eventRef = '$eventRefNum'!  Probable data corruption!<br>\n");
    return 0;
  }
  if ($secobj->eventCode ne "10")
  {
    $sth = $self->{dbHandle}->read(sql => "SELECT DISTINCT parent_id FROM lot_node_tb WHERE node_id IN ($listOfNodes2)");
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error retrieving parent nodes for node ID list = ($listOfNodes2)!<br>\n" . $self->{dbHandle}->errorMessage;
      return 0;
    }
    while ((my @info = $sth->fetchrow_array))
    {
      my $lotNode = $self->getLotNode(nodeId => $info[0]);
      if (not defined $lotNode || $self->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error retrieving lot node object for node ID = '" . $info[0] . "'!<br>\n" . $self->errorMessage;
        return 0;
      }
      $previousNodeArray{$lotNode->nodeId} = $lotNode;
    }
  }
  else
  {
    %previousNodeArray = %nodeArray;
  }

  # determine whether this node should be included in the list of nodes to retrieve local associated events to delete
  # for buys, all local associated events are deleted for the entire lot node tree
  # for stock transfers and acquisitions, all local associated events are deleted for the lot node tree of the node that
  # does not have the same lot as the parent node
  # for stock swaps and spin offs, all local events associated with nodes that were generated by the stock swap or spin off
  # for lots that do not match the parent node's lot are deleted
  # for all other events, no local events are affected
  my $listOfNodes = "";
  if ($secobj->eventCode eq "30" || $secobj->eventCode eq "60" || $secobj->eventCode eq "70" || $secobj->eventCode eq "80")
  {
    foreach my $lotNode (values %nodeArray)
    {
      if ($lotNode->lot ne $previousNodeArray{$lotNode->parentId}->lot)
      {
        my $nodeId = $lotNode->nodeId;
        my %lotNodes = $self->getLotTree(nodeId => $nodeId);
        $listOfNodes .= (length $listOfNodes > 0 ? ',' : "") . join(',', (keys %lotNodes));
      }
    }
  }
  elsif ($secobj->eventCode eq "10")
  {
    foreach my $lotNode (values %nodeArray)
    {
      my $nodeId = $lotNode->nodeId;
      my %lotNodes = $self->getLotTree(nodeId => $nodeId);
      $listOfNodes .= (length $listOfNodes > 0 ? ',' : "") . join(',', (keys %lotNodes));
    }
  }

  # get a list of local (event code = 20, 30, 40; 10[buy] is not included) event reference numbers
  # that are dependent on the nodes based on the type of event being deleted
  my @eventArray = ();
  my $listOfEvents = "";
  if ($listOfNodes ne "")
  {
    $sth = $self->{dbHandle}->read(sql => "SELECT DISTINCT event_ref FROM sec_event_tb WHERE (event_code = '20' OR event_code = '30' OR event_code = '40') AND event_ref IN (SELECT event_ref FROM lot_node_tb WHERE node_id IN ($listOfNodes))");
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error retrieving dependent local events for node IDs!<br>\n" . $self->{dbHandle}->errorMessage;
      return 0;
    }
    while ((my @info = $sth->fetchrow_array))
    {
      $eventArray[scalar @eventArray] = $info[0];
    }
    $listOfEvents = join(',', @eventArray);
  }

  # delete the nodes
  # since the previousNodeArray will be the nodeArray in the event of a buy, include the buy node for deletion
  my $includeNode = 0;
  if ($secobj->eventCode eq "10")
  {
    $includeNode = 1;
  }
  foreach my $lotNode (values %previousNodeArray)
  {
    my $nodeId = $lotNode->nodeId;
    if (not $self->propagateDeleteLot(nodeId => $nodeId, includeNodeId => $includeNode))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating the deletion of node ID = " . $nodeId . "!<br>\n" . $self->{errorString};
      return 0;
    }
  }

  if ($listOfEvents ne "")
  {
    # delete the list of dependent security events
    $self->{dbHandle}->write(sql => "DELETE FROM sec_event_tb WHERE event_ref IN ($listOfEvents)");
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error deleting dependent list of security events!<br>\n" . $self->{dbHandle}->errorMessage;
      $self->{dbHandle}->rollback;
      return 0;
    }
  }

  # delete the security event
  $self->{dbHandle}->write(sql => "DELETE FROM sec_event_tb WHERE event_ref = $eventRefNum");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error deleting security event (" . $eventRefNum . ") from database!";
    $self->{dbHandle}->rollback;
    return 0;
  }

  # repropagate the previous nodes (unless event was a buy)
  if ($secobj->eventCode ne "10")
  {
    foreach my $previousLot (values %previousNodeArray)
    {
      if (not $self->propagateLot(lotNodeObj => $previousLot))
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error propagating previous Node = '" . $previousLot->{nodeId} . "'!<br>\n" . $self->{errorString};
        return 0;
      }
    }
  }

  # everything successful
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
    return 0;
  }

  return 1;
}

# getSecurityEvent
# retrieves the desired security event from the database and returns a SecurityEventObject object
# modified to take a list of comma seperated event Refs and return a reference to an array of SecurityEventObjects.  If only
# a single event ref passed in, then only a single SecurityEventObject is returned and not an array ref.
# returns undef if error occurred and it hasn't gotten to the point where it knows if we are working with multiple event refs.
# returns undef if only a single event ref passed in and an error occurred, else returns ref to empty array if event occurred and
# multiple event refs passed in.  It is up to the user to check and handle this.
sub getSecurityEvent
{
  my $self = shift;
  my $errorHeader = "Security->getSecurityEvent(): ";
  my $mode = "scalar";
  my @securityEvents = ();
  
  if ($self->{debug})
	  {
    $self->{errorString} .= "getSecurityEvent<br>\n";
  }
  my %args = ( eventRefNum => undef, @_ );
  my $eventRefNum = $args{'eventRefNum'};

  if ((not defined $args{'eventRefNum'}) || $eventRefNum eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Event reference number not specified!";
    return undef;
  }
  if ($eventRefNum !~ /^([0-9]+(,[0-9]+)*)$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid event reference number (" . $eventRefNum . ")!";
    return undef;
  }
  if ($eventRefNum =~ /,/) # a comma exists
  {
    $mode = "array";
  }

  my $sth = $self->{dbHandle}->read(sql => "SELECT event_code, master_ref, calc_date, trans_date, counter, sub_e_code, A, S, L, OS, NS, CC, ledger_code, event_ref FROM sec_event_tb WHERE event_ref IN ($eventRefNum) GROUP BY calc_date, trans_date, master_ref, event_ref, s, counter, l, ns, os, cc, sub_e_code, a, event_code, ledger_code ORDER BY calc_date, event_ref");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving event!<br>\n" . $self->{dbHandle}->errorMessage;
    if ($mode eq "scalar")
    {
      return undef;
    }
    else
    {
      return \@securityEvents;
    }
  }
  while (my @info = $sth->fetchrow_array)
  {
    my $eventCode = $info[0];
    my $masterEventRef = $info[1];
    my $calcDate = $info[2];
    my $transDate = $info[3];
    my $counter = $info[4];
    my $subEntity = $info[5];
    my $accountNum = $info[6];
    my $subAccountNum = $info[7];
    my $lot = $info[8];
    my $oldShares = $info[9];
    my $newShares = $info[10];
    my $commission = $info[11];
    my $ledgerCode = $info[12];
    my $eventRefNum = $info[13];
  
    my %securityArgs = {};
    if ($eventCode eq "10") # buy
    {
      %securityArgs = (subEntity => "$subEntity", accountNum => "$accountNum", subAccountNum => "$subAccountNum", lot => "$lot", price => "$oldShares", shares => "$newShares", commission => "$commission");
    }
    elsif ($eventCode eq "20") # sell
    {
      %securityArgs = (masterEventRef => "$masterEventRef", accountNum => "$accountNum", subAccountNum => "$subAccountNum", lot => "$lot", price => "$oldShares", shares => "$newShares", commission => "$commission");
    }
    elsif ($eventCode eq "90") # stock split
    {
      %securityArgs = (subAccountNum => "$subAccountNum", list => "$lot", oldShares => "$oldShares", newShares => "$newShares", costPercentage => "$commission");
    }
    elsif ($eventCode eq "100") # stock dividend
    {
      %securityArgs = (subAccountNum => "$subAccountNum", list => "$lot", oldShares => "$oldShares", newShares => "$newShares", costPercentage => "$commission");
    }
    elsif ($eventCode eq "60") # stock swap
    {
      my @info = $sth->fetchrow_array;
      if (not defined @info)
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Empty result set from database for event_ref = " . $eventRefNum;
        if ($mode eq "scalar")
        {
          return undef;
        }
        else
        {
          return ();
        }
      }
      elsif ($info[13] != $eventRefNum)
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Next row from database for event_ref = '" . $eventRefNum . "' has event_ref = '" . $info[13] . "'!";
        if ($mode eq "scalar")
        {
          return undef;
        }
        else
        {
          return ();
        }
      }
      my $counter2 = $info[4];
      %securityArgs = (counter2 => "$counter2", subAccountNum => "$subAccountNum", subAccountNum2 => "$accountNum", oldShares => "$oldShares", newShares => "$newShares", costPercentage => "$commission");
    }
    elsif ($eventCode eq "80") # acquisition
    {
      %securityArgs = (subAccountNum => "$subAccountNum", subAccountNum2 => "$accountNum", oldShares => "$oldShares", newShares => "$newShares", costPercentage => "$commission");
    }
    elsif ($eventCode eq "70") # spin off
    {
      %securityArgs = (subAccountNum => "$subAccountNum", subAccountNum2 => "$accountNum", oldShares => "$oldShares", newShares => "$newShares", costPercentage => "$commission");
    }
    elsif ($eventCode eq "30") # stock transfer
    {
      my $newAccount;
      my $oldAccount;
      my $newLot;
      my $oldLot;
  
      my @info = $sth->fetchrow_array;
      if (not defined @info)
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Empty result set from database for event_ref = " . $eventRefNum;
        if ($mode eq "scalar")
        {
          return undef;
        }
        else
        {
          return ();
        }
      }
      elsif ($info[13] != $eventRefNum)
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Next row from database for event_ref = '" . $eventRefNum . "' has event_ref = '" . $info[13] . "'!";
        if ($mode eq "scalar")
        {
          return undef;
        }
        else
        {
          return ();
        }
      }
      my $subEntity2 = $info[5];
      my $accountNum2 = $info[6];
      my $lot2 = $info[8];
      my $newShares2 = $info[10];
  
      if ($newShares < 0)
      {
        $oldAccount = $accountNum;
        $newAccount = $accountNum2;
        $oldLot = $lot;
        $newLot = $lot2;
        $newShares = $newShares2;
        $subEntity = $subEntity2;
      }
      else
      {
        $oldAccount = $accountNum2;
        $newAccount = $accountNum;
        $oldLot = $lot2;
        $newLot = $lot;
      }
      %securityArgs = (masterEventRef => "$masterEventRef", subEntity => "$subEntity", newLot => "$newLot", oldLot => "$oldLot", newAccountNum => "$newAccount", oldAccountNum => "$oldAccount", subAccountNum => "$subAccountNum", shares => "$newShares");
    }
    elsif ($eventCode eq "memo") # memo
    {
      %securityArgs = (memoText => "$accountNum", subAccountNum => "$subAccountNum");
    }
    elsif ($eventCode eq "40") # fee
    {
      %securityArgs = (accountNum => "$accountNum", subAccountNum => "$subAccountNum", lot => "$lot", amount => "$commission");
    }
    elsif ($eventCode eq "50") # return of capital
    {
      %securityArgs = (accountNum => "$accountNum", subAccountNum => "$subAccountNum", amount => "$commission");
    }
  
    my $secobj = Portal::Accounting::Objects::SecurityEventObject->new(eventCode => "$eventCode", calcDate => "$calcDate", transDate => "$transDate", counter => "$counter", %securityArgs, ledgerCode => "$ledgerCode", eventRefNum => $eventRefNum, langObj => $self->{langObj});
    if ((not defined $secobj) || (not ref($secobj)))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Failed to create new security event object!";
      if ($mode eq "scalar")
      {
        return undef;
      }
      else
      {
        return ();
      }
    }
    if ($secobj->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Failed to create new security event object!<br>\n" . $secobj->errorMessage;
      if ($mode eq "scalar")
      {
        return undef;
      }
      else
      {
        return ();
      }
    }
    $securityEvents[++$#securityEvents] = $secobj;  # add to array of Security Events to return.
  }
  if (scalar @securityEvents == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Empty result set from database for event_ref = " . $eventRefNum;
    if ($mode eq "scalar")
    {
      return undef;
    }
    else
    {
      return \@securityEvents;
    }
  }

  if ($mode eq "scalar")
  {
    return $securityEvents[0];
  }
  else
  {
    return \@securityEvents;
  }
}

# getLotBalance
# opt: ledgerCode
# req: eventDate, lot  --want to make evenDate optional as well
# returns: LotNodeObject or undef if nothing found.
sub getLotBalance
{
  my $self = shift;
  my $errorHeader = "Security->getLotBalance(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "getLotBalance<br>\n";
  }
  my %args = ( ledgerCode => undef, eventDate => undef, lot => undef, @_ );
  my $ledgerCode = $args{'ledgerCode'};
  my $lot = $args{'lot'};
  my $eventDate = $args{'eventDate'};
  my $lotBalObj = undef;

  if (not defined $args{'ledgerCode'})
  {
    $ledgerCode = "";
  }
  if (not defined $args{'eventDate'})
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "The event date must be specified!";
    return undef;
  }
  if (not defined $args{'lot'})
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "The lot must be specified!";
    return undef;
  }

  my $sqlstr ="SELECT sub_e_code, account_num, sub_account_num, shares, cost, event_date, event_ref, journal_ref, parent_id, node_id, ledger_code, buy_lot FROM lot_node_tb AS a WHERE ";
  if ($ledgerCode ne "")
  {
    $sqlstr .="ledger_code = '$ledgerCode' AND ";
  }
  $sqlstr .= "lot = '$lot' AND event_date = (SELECT max(event_date) FROM lot_node_tb WHERE lot = a.lot AND event_date <= '$eventDate') AND node_id = (SELECT max(node_id) FROM lot_node_tb WHERE lot = a.lot AND event_date = a.event_date)";
  my $sth = $self->{dbHandle}->read(sql => $sqlstr);
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Database error retrieving lot balance!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }
  my @info = $sth->fetchrow_array;
  if (defined @info)
  {
    $lotBalObj = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $info[10], subEntity => $info[0], accountNum => $info[1], subAccountNum => $info[2], lot => $lot, shares => $info[3], cost => $info[4], eventDate => $info[5], eventRef => $info[6],
                    journalRef => $info[7], parentId => $info[8], nodeId => $info[9], buyLot => $info[11], langObj => $self->{langObj});
    if ($lotBalObj->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error creating lot balance object!<br>\n" . $lotBalObj->errorMessage;
      return $lotBalObj;
    }
  }

  return $lotBalObj;
}

# getLotBalanceEntriesFor
#sum: retrieves the latest lot nodes for each lot as of the given date
#can be narrowed to given sub-account and/or account and/or ledger and/or lot
#then orders by an order by clause if given
#pre: required: date
#Optional: accountNum, subAccountNum, ledgerCode, lot, order
#order can have contain the arguments comma separated
#NOTE: the argments do not have to be passed in to use them in order
#post: returns an array of lotNodes or undefined if there is a problem
sub getLotBalanceEntriesFor
{
  my $self = shift;
  my $errorHeader = "Security->getLotBalanceEntriesFor(): ";
  if ($self->{debug} = 1)
  {
    $self->{errorString} .= "getLotBalanceEntriesFor<br>\n";
  }
  my %args = ( eventDate => undef, subAccountNum => undef, @_ );
  my @lotNodes = ();
  my $eventDate = $args{'eventDate'};
  my $subAccountNum = $args{'subAccountNum'};
  my $accountNum = $args{'accountNum'};
  my $ledgerCode = $args{'ledgerCode'};
  my $order = $args{'order'};
  my $lot = $args{'lot'};

  # validate input
  if ((not defined $args{'eventDate'}) || $eventDate eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Event date not specified or empty!";
    return undef;
  }
  if (not defined $args{'subAccountNum'})
  {
    $subAccountNum = "";
  }
  if (not defined $args{'accountNum'})
  {
    $accountNum = "";
  }
  if (not defined $args{'ledgerCode'})
  {
    $ledgerCode = "";
  }
  if (not defined $args{'order'})
  {
    $order = "";
  }
  if (not defined $args{'lot'})
  {
    $lot = "";
  }

  if ($order ne "")
  {
    #replace the argumentformat sent in with the database format for the fields
    $order =~ s/accountNum/account_num/g;
    $order =~ s/subAccountNum/sub_account_num/g;
    $order =~ s/ledgerCode/ledger_code/g;
    $order =~ s/lot/a.lot/g;
  }

  my $sqlstr = "SELECT distinct a.node_id";
  if ($order ne "")
  {
    $sqlstr .= ", $order";
  }
  $sqlstr .= " FROM lot_node_tb AS a WHERE a.lot IN (SELECT DISTINCT lot FROM lot_node_tb WHERE event_date <= '$eventDate'";
  if ($subAccountNum ne "")
  {
    $sqlstr .= " AND sub_account_num = $subAccountNum";
  }
  if ($accountNum ne "")
  {
    $sqlstr .= " AND account_num = $accountNum";
  }
  if ($lot ne "")
  {
    $sqlstr .= " AND lot = $lot";
  }
  if ($ledgerCode ne "")
  {
    $sqlstr .= " AND ledger_code = '$ledgerCode'";
  }
  $sqlstr .= ") AND a.event_date = (SELECT max(event_date) FROM lot_node_tb WHERE lot = a.lot AND event_date <= '$eventDate') AND a.node_id = (SELECT max(node_id) FROM lot_node_tb WHERE lot = a.lot AND event_date = a.event_date)";
  if ($order ne "")
  {
    $sqlstr .= " ORDER BY $order";
  }
  my $sth = $self->{dbHandle}->read(sql => $sqlstr);
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving lot node ID's from lot node table!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }
  while (my @info = $sth->fetchrow_array)
  {
    my $nodeID = $info[0];
    if ($nodeID eq "")
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error! Node ID is empty.";
      return undef;
    }
    my $lotNode = $self->getLotNode(nodeId => $nodeID);
    if ($self->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error retrieving lot node!<br>\n" . $self->{errorString};
      return undef;
    }
    if ($self->{debug})
    {
      $self->{errorString} .= "Got LotNode in getLotBalanceEntriesFor<br>\n";
    }

    if (not defined $lotNode)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error retrieving lot node!  No lot node object returned.";
      return undef;
    }
    if (not $lotNode->isValid())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Node returned is invalid!<br>\n" . $lotNode->errorMessage;
      return undef;
    }
    # add the lot node to the end of the lotNodes array
    $lotNodes[scalar @lotNodes] = $lotNode;
  }
  if ($self->{debug})
  {
    $self->{errorString} .= "Made it to end of getLotBalanceEntriesFor<br>\n";
  }
  return @lotNodes;
}

# insertSecurityEventObject
# takes obj and an optional commit argument
# returns the newly inserted reference number of the security event object on success
# returns -1 on failure
sub insertSecurityEventObject
{
  my $self = shift;
  my $errorHeader = "Security->insertSecurityEventObject(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "insertSecurityEventObject<br>\n";
  }
  my %args = ( obj => undef, commit => 0, counter => undef, counter2 => undef, @_ );
  my $eventRefNum = -1;
  my $secObj = $args{'obj'};
  my $counter = $args{'counter'};
  my $counter2 = $args{'counter2'};

  if (not defined $args{'obj'})
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Object reference missing!";
    return -1;
  }
  # make sure the object passed in is really a SecurityEventObject
  if (ref($args{'obj'}) ne "Portal::Accounting::Objects::SecurityEventObject")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid SecurityEventObject reference (" . ref($args{'obj'}) . ")!";
    return -1;
  }
  # check the validity of the security event object
  if ($args{'obj'}->isValid())
  {
    # get the nextval(sec_event_tb_event_ref_seq) of the serial sequence for the table
    my $sth = $self->{dbHandle}->read(sql => "SELECT nextval('sec_event_tb_event_ref_seq')");
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Failed to get the nextval() of the security event reference number sequence!<br>\n" . $self->{dbHandle}->errorMessage;
      return -1;
    }
    my @info = $sth->fetchrow_array;
    $eventRefNum = $info[0];
    # set the eventRefNum of the security object to the reference number returned from the database
    $secObj->eventRefNum("$eventRefNum");

    # determine if the event object is to be inserted in a specific order
    if (defined $args{'counter'})
    {
      # move all events on the same day for the same sub-account up one counter in order to insert the current event at the
      # counter location
      $self->{dbHandle}->write(sql => "UPDATE sec_event_tb SET counter = counter + 1 WHERE event_ref = (SELECT DISTINCT event_ref FROM sec_event_tb WHERE calc_date = '" . $secObj->calcDate . "' AND S = '" . $secObj->subAccountNum . "' AND counter >= '$counter')");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error updating counter locations to insert new event in order!<br>\n" . $self->{dbHandle}->errorMessage;
        $self->{dbHandle}->rollback;
        return -1;
      }
    }
    else
    {
      # determine if there are any events on the same day for the same sub-account and obtain the last counter location
      # and increase the counter by one to use for inserting the event
      my $sth = $self->{dbHandle}->read(sql => "SELECT max(counter) FROM sec_event_tb WHERE calc_date = '" . $secObj->calcDate . "' AND S = '" . $secObj->subAccountNum . "'");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error retrieving maximum counter location for the date of the event!<br>\n" . $self->{dbHandle}->errorMessage;
        return -1;
      }
      my @info2 = $sth->fetchrow_array;
      if ((not defined @info2) || $info2[0] eq "")
      {
        $counter = 0;
      }
      else
      {
        $counter = $info2[0] + 1;
      }
    }
    if ($secObj->eventCode eq "60") # stock swap; need to get the counter to be used for subAccountNum2
    {
      # determine if the event object is to be inserted in a specific order
      if (defined $args{'counter2'})
      {
        # move all events on the same day for the same sub-account up one counter in order to insert the current event at the
        # counter location
        $self->{dbHandle}->write(sql => "UPDATE sec_event_tb SET counter = counter + 1 WHERE event_ref = (SELECT DISTINCT event_ref FROM sec_event_tb WHERE calc_date = '" . $secObj->calcDate . "' AND S = '" . $secObj->subAccountNum2 . "' AND counter >= '$counter2')");
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errorHeader . "Error updating counter locations to insert new stock swap event in order!<br>\n" . $self->{dbHandle}->errorMessage;
          $self->{dbHandle}->rollback;
          return -1;
        }
      }
      else
      {
        # determine if there are any events on the same day for the same sub-account and obtain the last counter location
        # and increase the counter by one to use for inserting the event
        my $sth = $self->{dbHandle}->read(sql => "SELECT max(counter) FROM sec_event_tb WHERE calc_date = '" . $secObj->calcDate . "' AND S = '" . $secObj->subAccountNum2 . "'");
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errorHeader . "Error retrieving maximum counter location for the date of the stock swap event!<br>\n" . $self->{dbHandle}->errorMessage;
          return -1;
        }
        my @info2 = $sth->fetchrow_array;
        if ((not defined @info2) || $info2[0] eq "")
        {
          $counter2 = 0;
        }
        else
        {
          $counter2 = $info2[0] + 1;
        }
      }
      $secObj->counter2("$counter2");
    }

    $secObj->counter("$counter");

    # insert the security event into the database
    if ($secObj->eventCode eq "10") # buy
    {
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, A, S, L, OS, NS, CC, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->transDate . "', '" . $secObj->transDate . "', '$counter', '" . $secObj->accountNum . "', '" . $secObj->subAccountNum  . "', '" . $secObj->lot . "', '" . $secObj->price . "', '" . $secObj->shares . "', '" . $secObj->commission . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    elsif ($secObj->eventCode eq "20") # sell
    {
      if ($secObj->masterEventRef == -1)
      {
        $secObj->masterEventRef($secObj->eventRefNum);
      }
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, master_ref, calc_date, trans_date, counter, A, S, L, OS, NS, CC, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->masterEventRef  . "', '" . $secObj->transDate . "', '" . $secObj->transDate . "', '$counter', '" . $secObj->accountNum . "', '" . $secObj->subAccountNum . "', '" . $secObj->lot . "', '" . $secObj->price . "', '" . $secObj->shares . "', '" . $secObj->commission . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    elsif ($secObj->eventCode eq "30") # stock transfer
    {
      # insert two records for a stock transfer
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, A, S, L, NS, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->transDate . "', '" . $secObj->transDate . "', '$counter', '" . $secObj->oldAccountNum . "', '" . $secObj->subAccountNum . "', '" . $secObj->oldLot . "', '-" . $secObj->shares . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, sub_e_code, A, S, L, NS, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->transDate . "', '" . $secObj->transDate . "', '$counter', '" . $secObj->subEntity . "', '" . $secObj->newAccountNum . "', '" . $secObj->subAccountNum . "', '" . $secObj->newLot . "', '" . $secObj->shares . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    elsif ($secObj->eventCode eq "40") # fee
    {
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, A, S, L, CC, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->transDate . "', '" . $secObj->transDate . "', '" . $counter . "', '" . $secObj->accountNum . "', '" . $secObj->subAccountNum . "', '" . $secObj->lot . "', '" . $secObj->amount . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    elsif ($secObj->eventCode eq "50") # return of capital
    {
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, S, CC, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->transDate . "', '" . $secObj->transDate . "', '" . $counter . "', '" . $secObj->subAccountNum . "', '" . $secObj->amount . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    elsif ($secObj->eventCode eq "60") # stock swap
    {
      # insert two records for a stock swap
      my $tempCostPercentage = 1.0 - $secObj->costPercentage;
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, A, S, OS, NS, CC, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->calcDate . "', '" . $secObj->transDate . "', '" . $counter . "', '" . $secObj->subAccountNum2 . "', '" . $secObj->subAccountNum . "', '" . $secObj->oldShares . "', '" . $secObj->newShares . "', '" . $secObj->costPercentage . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, A, S, OS, NS, CC, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->calcDate . "', '" . $secObj->transDate . "', '" . $counter2 . "', '" . $secObj->subAccountNum . "', '" . $secObj->subAccountNum2 . "', '" . $secObj->newShares . "', '" . $secObj->oldShares . "', '" . $tempCostPercentage . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    elsif ($secObj->eventCode eq "70") # spin off
    {
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, A, S, OS, NS, CC, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->calcDate . "', '" . $secObj->transDate . "', '" . $counter . "', '" . $secObj->subAccountNum2 . "', '" . $secObj->subAccountNum . "', '" . $secObj->oldShares . "', '" . $secObj->newShares . "', '" . $secObj->costPercentage . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    elsif ($secObj->eventCode eq "80") # acquisition
    {
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, A, S, OS, NS, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->calcDate . "', '" . $secObj->transDate . "', '" . $counter . "', '" . $secObj->subAccountNum2 . "', '" . $secObj->subAccountNum . "', '" . $secObj->oldShares . "', '" . $secObj->newShares . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    elsif ($secObj->eventCode eq "90") # stock split
    {
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, S, OS, NS, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->calcDate . "', '" . $secObj->transDate . "', '" . $counter . "', '" . $secObj->subAccountNum . "', '" . $secObj->oldShares . "', '" . $secObj->newShares . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    elsif ($secObj->eventCode eq "100") # stock dividend
    {
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, S, OS, NS, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->calcDate . "', '" . $secObj->transDate . "', '" . $counter . "', '" . $secObj->subAccountNum . "', '" . $secObj->oldShares . "', '" . $secObj->newShares . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    elsif ($secObj->eventCode eq "memo")
    {
      $self->{dbHandle}->write(sql => "INSERT INTO sec_event_tb (event_code, calc_date, trans_date, counter, A, S, ledger_code, event_ref) VALUES ('" . $secObj->eventCode . "', '" . $secObj->calcDate . "', '" . $secObj->transDate . "', '" . $counter . "', '" . $secObj->memoText . "', '" . $secObj->subAccountNum . "', '" . $secObj->ledgerCode . "', " . $secObj->eventRefNum . ")");
    }
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting security event object!<br>\n" . $self->{dbHandle}->errorMessage;
      $self->{dbHandle}->rollback;
      return -1;
    }
    if ($args{'commit'} eq "1")
    {
      # finish the transaction
      $self->{dbHandle}->commit;
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
        return -1;
      }
    }
  }
  else
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid security event object!<br>\n" . $secObj->errorMessage;
    return -1;
  }

  # return the created reference number from the database
  return $eventRefNum;
}

# insertLotNodeObject
# takes obj and an optional commit argument that determines whether the transaction is committed immediately or not
# returns the newly inserted node number of the lot node object on success
# returns -1 on failure
sub insertLotNodeObject
{
  my $self = shift;
  my $errorHeader = "Security->insertLotNodeObject(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "insertLotNodeObject<br>\n";
  }
  my %args = ( obj => undef, commit => 0, @_ );
  my $lotNodeID = -1;
  my $nodeObj = $args{'obj'};

  if (not defined $args{'obj'})
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Object reference missing!";
    return -1;
  }
  # make sure the object passed in is really a LotNodeObject
  if (ref($args{'obj'}) ne "Portal::Accounting::Objects::LotNodeObject")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid LotNodeObject reference (" . ref($args{'obj'}) . ")!";
    return -1;
  }
  # get the nextval(lot_node_tb_node_id_seq) of the serial sequence for the table
  my $sth = $self->{dbHandle}->read(sql => "SELECT nextval('lot_node_tb_node_id_seq')");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to get the nextval() of the lot node ID number sequence!<br>\n" . $self->{dbHandle}->errorMessage;
    return -1;
  }
  my @info = $sth->fetchrow_array;
  $lotNodeID = $info[0];
  if ($lotNodeID !~ /^[0-9]+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid lot node ID number (" . $lotNodeID . ")!  The value does not appear to be numeric.";
    return -1;
  }
  # set the node ID of the lot node object to the number returned from the database
  $nodeObj->nodeId("$lotNodeID");
  # check the validity of the lot node object
  if ($nodeObj->isValid())
  {
    # insert the lot node object into the database
    $self->{dbHandle}->write(sql => "INSERT INTO lot_node_tb (node_id, ledger_code, event_date, parent_id, event_ref, journal_ref, account_num, sub_account_num, lot, buy_lot, shares, cost) VALUES ('" . $nodeObj->nodeId . "', '" . $nodeObj->ledgerCode . "', '" . $nodeObj->eventDate . "', '" . $nodeObj->parentId . "', '" . $nodeObj->eventRef . "', '" . $nodeObj->journalRef . "', '" . $nodeObj->accountNum . "', '" . $nodeObj->subAccountNum . "', '" . $nodeObj->lot . "', '" . $nodeObj->buyLot . "', '" . $nodeObj->shares . "', '" . $nodeObj->cost . "')");
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error inserting lot node object!<br>\n" . $self->{dbHandle}->errorMessage;
      $self->{dbHandle}->rollback;
      return -1;
    }
    if ($args{'commit'} eq "1")
    {
      # finish the transaction
      $self->{dbHandle}->commit;
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
        return -1;
      }
    }
  }
  else
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid lot node object!<br>\n" . $nodeObj->errorMessage;
    return -1;
  }

  # return the created node ID from the database
  return $lotNodeID;
}

# insertBuy - local event
# adds a buy to the sec_event_tb table and to the journal_entry_tb and debitcredit_tb tables
sub insertBuy
{
  my $self = shift;
  my $errorHeader = "Security->insertBuy(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "insertBuy<br>\n";
  }
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  if ($secobj->eventCode ne "10") # buy
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a buy using a non-buy object (" . $secobj->eventCode . ")!";
    return undef;
  }

  # set the lot number of the security event object to the next new lot number
  my $lotNum = $self->getNewLotNumber();
  if (not defined $lotNum)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error getting new lot!<br>\n" . $self->{errorString};
    return undef;
  }
  $secobj->lot("$lotNum");

  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }

  # insert the security event into the database
  my $eventRef = $self->insertSecurityEventObject(obj => $secobj, counter => $counter);
  if ($eventRef == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting security event object!<br>\n" . $self->{errorString};
    return undef;
  }

  # calculate total price
  my $totalCost = sprintf("%.2f", ($secobj->price * $secobj->shares) + $secobj->commission);
  # build a TransactionObject for the data
  my $transobj = Portal::Accounting::Objects::TransactionObject->new(ledgerCode => $secobj->ledgerCode,
                                                                     journalCode => "IS", # investments journal code hardcoded
                                                                     transDate => $secobj->transDate,
                                                                     rowObjects => [ Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $secobj->transDate, subEntity => $args{'subEntity'}, accountNum => $secobj->accountNum, subAccountNum => $secobj->subAccountNum, dorc => "true", amount => "$totalCost", langObj => $self->{langObj}),
                                                                     Portal::Accounting::Objects::TransactionRowObject->new(rowDate => $secobj->transDate, subEntity => "$args{'subEntity'}", accountNum => $secobj->accountNum, subAccountNum => "-1", dorc => "false", amount => "$totalCost", langObj => $self->{langObj}) ], langObj => $self->{langObj} );
  if ($transobj->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to create new transaction object!<br>\n" . $transobj->errorMessage;
    return undef;
  }

  # insert the transaction into the database
  my $refNum = $self->{transaction}->insertTransaction(obj => $transobj);
  if ($refNum == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting TransactionObject!<br>\n" . $self->{transaction}->errorMessage;
    return undef;
  }
  # insert a new lot node into the database (node_id will fill itself in from the unique sequence on the table definition)
  my $lotNodeObj = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $secobj->ledgerCode,
                                                                   eventDate => $secobj->calcDate,
                                                                   parentId => -1,
                                                                   eventRef => $eventRef,
                                                                   journalRef => $refNum,
                                                                   accountNum => $secobj->accountNum,
                                                                   subAccountNum => $secobj->subAccountNum,
                                                                   lot => $secobj->lot,
                                                                   buyLot => $secobj->lot,
                                                                   cost => sprintf("%.2f", $totalCost),
                                                                   shares => sprintf("%.6f", $secobj->shares), langObj => $self->{langObj});
  my $lotNodeID = $self->insertLotNodeObject(obj => $lotNodeObj);
  if ($lotNodeID == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting lot node!<br>\n" . $self->{errorString};
    return undef;
  }

  # insert an entry in the lot_parent_tb table for the new lot
  $self->{dbHandle}->write(sql => "INSERT INTO lot_parent_tb (lot, buy_lot) VALUES (?, ?)", plug => [ $lotNodeObj->lot, $lotNodeObj->buyLot ]);
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting lot entry into lot_parent_tb table for new lot!<br>\n" . $self->{dbHandle}->errorMessage;
    $self->{dbHandle}->rollback;
    return 0;
  }

  # propagate the newly created lot into the future
  if (not $self->propagateLot(lotNodeObj => $lotNodeObj))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error propagating lot node (" . $lotNodeID . ") into the future!<br>\n" . $self->{errorString};
    $self->{dbHandle}->rollback;
    return undef;
  }

  # everything about the transaction succeeded
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }

  return $eventRef;
}

# insertSell - local event
# adds a sell to the sec_event_tb table
sub insertSell
{
  my $self = shift;
  my $errorHeader = "Security->insertSell(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "insertSell<br>\n";
  }
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  if ($secobj->eventCode ne "20") # sell
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a sell using a non-sell object (" . $secobj->eventCode . ")!";
    return undef;
  }

  # insert the security event into the database
  my $eventRef = $self->insertSecurityEventObject(obj => $secobj, counter => $counter);
  if ($eventRef == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting security event object!<br>\n" . $self->{errorString};
    return undef;
  }
  # retrieve the lot balance entry from the lot_node_tb table
  my $lotbalobj = $self->getLotBalance(ledgerCode => $secobj->ledgerCode, eventDate => $secobj->transDate, lot => $secobj->lot);
  if (not defined $lotbalobj)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to retrieve lot balance!<br>\n" . $self->errorMessage;
    return undef;
  }
  if ($lotbalobj->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to retrieve lot balance!<br>\n" . $lotbalobj->errorMessage;
    return undef;
  }
  # delete the tree for this node
  if (not $self->propagateDeleteLot(nodeId => $lotbalobj->nodeId))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error deleting lot node (" . $lotbalobj->nodeId . ")!<br>\n" . $self->errorMessage;
    return undef;
  }
  # propagate the node into the future
  if (not $self->propagateLot(lotNodeObj => $lotbalobj))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error propagating node (" . $lotbalobj->nodeId . ") into the future!<br>\n" . $self->{errorString};
    $self->{dbHandle}->rollback;
    return undef;
  }

  # everything about the transaction succeeded
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }

  return $eventRef;
}

# insertStockSplit - global event
# adds a split to the sec_event_tb table
sub insertStockSplit
{
  my $self = shift;
  my $errorHeader = "Security->insertStockSplit(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "insertStockSplit<br>\n";
  }
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  if ($secobj->eventCode ne "90") # stock split
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a split using a non-split object (" . $secobj->eventCode . ")!";
    return undef;
  }

  # make the calcDate reflect the transDate value
  $secobj->calcDate($secobj->transDate);

  my $eventRef = $self->insertSplitDividend(secobj => $secobj, counter => $counter);
  if (not defined $eventRef)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to insert split event!<br>\n" . $self->{errorString};
    return undef;
  }

  return $eventRef;
}

# insertStockDividend - global event
# adds a stock dividend to the sec_event_tb table
sub insertStockDividend
{
  my $self = shift;
  my $errorHeader = "Security->insertStockDividend(): ";
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  if ($secobj->eventCode ne "100") # stock dividend
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a stock dividend using a non-stock dividend object (" . $secobj->eventCode . ")!";
    return undef;
  }

  # make the calcDate reflect the transDate value
  $secobj->calcDate($secobj->transDate);

  my $eventRef = $self->insertSplitDividend(secobj => $secobj, counter => $counter);
  if (not defined $eventRef)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to insert stock dividend event!<br>\n" . $self->{errorString};
    return undef;
  }

  return $eventRef;
}

# insertSplitDividend
# this function performs the exact same actions on the two different object types that can be passed to it; it will insert
# a record into the sec_event_tb table and update all affected ledgers and lots in the lot_node_tb table
sub insertSplitDividend
{
  my $self = shift;
  my $errorHeader = "Security->insertSplitDividend(): ";
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  if ($secobj->eventCode ne "90" && $secobj->eventCode ne "100")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a stock split/dividend with a non-stock split/dividend object (" . $secobj->eventCode . ")!";
    return undef;
  }

  # insert the security event object into the database
  my $eventRefNum = $self->insertSecurityEventObject(obj => $secobj, counter => $counter);
  if ($eventRefNum == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting security event object!<br>\n" . $self->{errorString};
    return undef;
  }

  # lookup all ledgers that have current closeout/minimum open fiscal dates less than or equal to $secobj->calcDate
  #my @ledgers = $self->openLedgers(date => $secobj->{calcDate});

  # retrieve the affected lots for given ledgers
  my @lotBalanceEntries = $self->getLotBalanceEntriesFor(eventDate => $secobj->{calcDate}, subAccountNum => $secobj->{subAccountNum}); #ledgers => \@ledgers,
  if ($self->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error getting lot balance entries!<br>\n" . $self->errorMessage;
    return undef;
  }
  # for each affected lot, calculate the new lot balance
  for (my $lotcount=0; $lotcount < scalar @lotBalanceEntries; $lotcount++)
  {
    # get the sub-tree of nodes attached to this node
    if ((not defined $lotBalanceEntries[$lotcount]) || (not $lotBalanceEntries[$lotcount]->isValid()))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error getting lot balance entries!  Element $lotcount not defined.";
      return undef;
    }
    # delete the tree for this node
    if (not $self->propagateDeleteLot(nodeId => $lotBalanceEntries[$lotcount]->nodeId))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error deleting lot node (" . $lotBalanceEntries[$lotcount]->nodeId . ")!<br>\n" . $self->errorMessage;
      return undef;
    }
    # propagate the node into the future
    if (not $self->propagateLot(lotNodeObj => $lotBalanceEntries[$lotcount]))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating node (" . $lotBalanceEntries[$lotcount]->nodeId . ") into the future!<br>\n" . $self->{errorString};
      $self->{dbHandle}->rollback;
      return undef;
    }
  }

  # everything about the transaction succeeded
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }

  return $eventRefNum;
}

# insertAcquisition - global event
# adds an acquisition event to the sec_event_tb table
sub insertAcquisition
{
  my $self = shift;
  my $errorHeader = "Security->insertAcquisition(): ";
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  if ($secobj->eventCode ne "80") # acquisition
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert an acquisition using a non-acquisition object (" . $secobj->eventCode . ")!";
    return undef;
  }

  # insert the security event object into the database
  my $eventRefNum = $self->insertSecurityEventObject(obj => $secobj, counter => $counter);
  if ($eventRefNum == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to insert security event!<br>\n" . $self->{errorString};
    return undef;
  }
  # retrieve the affected lots for all ledgers
  my @lotBalanceEntries = $self->getLotBalanceEntriesFor(eventDate => $secobj->calcDate, subAccountNum => $secobj->subAccountNum);
  if ($self->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error getting lot balance entries!<br>\n" . $self->errorMessage;
    return undef;
  }
  # for each lot entry, create a new journal entry with the new cost placing the number of shares and the lot number
  # into the memo field and building up a list of journal reference numbers (new lot numbers)
  for (my $lotcount=0; $lotcount < scalar @lotBalanceEntries; $lotcount++)
  {
    # get the sub-tree of nodes attached to this node
    if ((not defined $lotBalanceEntries[$lotcount]) || (not $lotBalanceEntries[$lotcount]->isValid()))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error getting lot balance entries!  Element $lotcount not defined.";
      return undef;
    }
    # delete the tree for this node
    if (not $self->propagateDeleteLot(nodeId => $lotBalanceEntries[$lotcount]->nodeId))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error deleting lot node (" . $lotBalanceEntries[$lotcount]->nodeId . ")!<br>\n" . $self->errorMessage;
      return undef;
    }
    # propagate the previous node into the future
    if (not $self->propagateLot(lotNodeObj => $lotBalanceEntries[$lotcount]))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating node (" . $lotBalanceEntries[$lotcount]->nodeId . ") into the future!<br>\n" . $self->{errorString};
      $self->{dbHandle}->rollback;
      return undef;
    }
  }

  # everything about the transaction succeeded
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }

  return $eventRefNum;
}

# insertStockSwap - global event
# adds a stock swap event to the sec_event_tb table
sub insertStockSwap
{
  my $self = shift;
  my $errorHeader = "Security->insertStockSwap(): ";
  my %args = ( secobj => undef, counter => undef, counter2 => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};
  my $counter2 = $args{'counter2'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  if ($secobj->eventCode ne "60") # stock swap
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a stock swap using a non-stock swap object (" . $secobj->eventCode . ")!";
    return undef;
  }

  my $refNum = $self->insertSpinSwap(secobj => $secobj, counter => $counter, counter2 => $counter2);
  if (not defined $refNum)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to insert stock swap event!<br>\n" . $self->{errorString};
    return undef;
  }

  return $refNum;
}

# insertSpinOff - global event
# adds a spin off event to the sec_event_tb table
sub insertSpinOff
{
  my $self = shift;
  my $errorHeader = "Security->insertSpinOff(): ";
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  if ($secobj->eventCode ne "70") # spin off
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a spinoff using a non-spinoff object (" . $secobj->eventCode . ")!";
    return undef;
  }

  my $refNum = $self->insertSpinSwap(secobj => $secobj, counter => $counter);
  if (not defined $refNum)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to insert spinoff event!<br>\n" . $self->{errorString};
    return undef;
  }

  return $refNum;
}

# insertSpinSwap
# takes the spinoff or stock swap event and inserts the proper records into the security event table and the journal tables
sub insertSpinSwap
{
  my $self = shift;
  my $errorHeader = "Security->insertSpinSwap(): ";
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};
  my $counter2 = $args{'counter2'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  if ($secobj->eventCode ne "60" && $secobj->eventCode ne "70") # stock swap/spin off
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a stock swap/spin off with a non-stock swap/spin off object (" . $secobj->eventCode . ")!";
    return undef;
  }

  # insert the security event object into the database
  my $eventRefNum = $self->insertSecurityEventObject(obj => $secobj, counter => $counter, counter2 => $counter2);
  if ($eventRefNum == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to insert security event!<br>\n" . $self->{errorString};
    return undef;
  }
  # retrieve the affected lots for all ledgers for the first sub-account
  my @lotBalanceEntries1;
  my @lotBalanceEntries2;
  @lotBalanceEntries1 = $self->getLotBalanceEntriesFor(eventDate => $secobj->calcDate, subAccountNum => $secobj->subAccountNum);
  if ($self->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error getting lot balance entries!<br>\n" . $self->errorMessage;
    return undef;
  }
  if ($secobj->eventCode eq "60") # stock swap
  {
    # retrieve the affected lots for all ledgers for the second sub-account
    @lotBalanceEntries2 = $self->getLotBalanceEntriesFor(eventDate => $secobj->calcDate, subAccountNum => $secobj->subAccountNum2);
    if ($self->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error getting lot balance entries!<br>\n" . $self->errorMessage;
      return undef;
    }
  }
  for (my $lotcount=0; $lotcount < scalar @lotBalanceEntries1; $lotcount++)
  {
    # get the sub-tree of nodes attached to this node
    if ((not defined $lotBalanceEntries1[$lotcount]) || (not $lotBalanceEntries1[$lotcount]->isValid()))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error getting lot balance entries!  Element $lotcount not defined.";
      return undef;
    }
    # delete the tree for this node
    if (not $self->propagateDeleteLot(nodeId => $lotBalanceEntries1[$lotcount]->nodeId))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error deleting lot node (" . $lotBalanceEntries1[$lotcount]->nodeId . ")!<br>\n" . $self->errorMessage;
      return undef;
    }
    # propagate the previous node into the future
    if (not $self->propagateLot(lotNodeObj => $lotBalanceEntries1[$lotcount]))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating node (" . $lotBalanceEntries1[$lotcount]->nodeId . ") into the future!<br>\n" . $self->{errorString};
      $self->{dbHandle}->rollback;
      return undef;
    }
  }
  if ($secobj->eventCode eq "60") # stock swap
  {
    for (my $lotcount=0; $lotcount < scalar @lotBalanceEntries2; $lotcount++)
    {
      # get the sub-tree of nodes attached to this node
      if ((not defined $lotBalanceEntries2[$lotcount]) || (not $lotBalanceEntries2[$lotcount]->isValid()))
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error getting lot balance entries!  Element $lotcount not defined.";
        return undef;
      }
      # delete the tree for this node
      if (not $self->propagateDeleteLot(nodeId => $lotBalanceEntries2[$lotcount]->nodeId))
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error deleting lot node (" . $lotBalanceEntries2[$lotcount]->nodeId . ")!<br>\n" . $self->errorMessage;
        return undef;
      }
      # propagate the previous node into the future
      if (not $self->propagateLot(lotNodeObj => $lotBalanceEntries2[$lotcount]))
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error propagating node (" . $lotBalanceEntries2[$lotcount]->nodeId . ") into the future!<br>\n" . $self->{errorString};
        $self->{dbHandle}->rollback;
        return undef;
      }
    }
  }

  # everything about the transaction succeeded
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }

  return $eventRefNum;
}

# insertStockTransfer - local event
# adds a stock transfer event to the sec_event_tb table
sub insertStockTransfer
{
  my $self = shift;
  my $errorHeader = "Security->insertStockTransfer(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "insertStockTransfer<br>\n";
  }
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  if ($secobj->eventCode ne "30") # stock transfer
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a stock transfer using a non-stock transfer object (" . $secobj->eventCode . ")!";
    return undef;
  }

  # set the lot number of the security event object to the next new lot number
  my $lotNum = $self->getNewLotNumber();
  if (not defined $lotNum)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error getting new lot!<br>\n" . $self->{errorString};
    return undef;
  }
  $secobj->newLot("$lotNum");

  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }

  # insert the security event into the database
  my $eventRefNum = $self->insertSecurityEventObject(obj => $secobj, counter => $counter);
  if ($eventRefNum == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting security event object!<br>\n" . $self->{errorString};
    return undef;
  }

  # retrieve the lot balance entry from the lot_node_tb table for the "old" lot
  my $lotbalobj = $self->getLotBalance(ledgerCode => $secobj->ledgerCode, eventDate => $secobj->transDate, lot => $secobj->oldLot);
  if (not defined $lotbalobj)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to retrieve lot balance!<br>\n" . $self->errorMessage;
    return undef;
  }
  if ($lotbalobj->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to retrieve lot balance!<br>\n" . $lotbalobj->errorMessage;
    return undef;
  }
  if (not $lotbalobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $lotbalobj->errorMessage;
    return undef;
  }
  # delete the tree for this node
  if (not $self->propagateDeleteLot(nodeId => $lotbalobj->nodeId))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error deleting lot node (" . $lotbalobj->nodeId . ")!<br>\n" . $self->errorMessage;
    return undef;
  }
  # propagate the node into the future
  if (not $self->propagateLot(lotNodeObj => $lotbalobj))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error propagating node (" . $lotbalobj->nodeId . ") into the future!<br>\n" . $self->{errorString};
    $self->{dbHandle}->rollback;
    return undef;
  }

  # everything about the transaction succeeded
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }

  return $eventRefNum;
}

# insertMemo
# adds a memo to the sec_event_tb table
sub insertMemo
{
  my $self = shift;
  my $errorHeader = "Security->insertMemo(): ";
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  if ($secobj->eventCode ne "memo")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a memo using a non-memo object (" . $secobj->eventCode . ")!";
    return undef;
  }

  my $eventRefNum = $self->insertSecurityEventObject(obj => $secobj, counter => $counter);
  if ($eventRefNum == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting memo event!<br>\n" . $self->{errorString};
    return undef;
  }

  # everything about the transaction succeeded
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }

  return $eventRefNum;
}

# insertFee - local event
# adds a fee to the sec_event_tb table
sub insertFee
{
  my $self = shift;
  my $errorHeader = "Security->insertFee(): ";
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  if ($secobj->eventCode ne "40") # fee
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a fee using a non-fee object (" . $secobj->eventCode . ")!";
    return undef;
  }

  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  # insert the security event into the database
  my $eventRefNum = $self->insertSecurityEventObject(obj => $secobj, counter => $counter);
  if ($eventRefNum == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting fee event!<br>\n" . $self->{errorString};
    return undef;
  }

  # retrieve the lot balance entry from the lot_node_tb table for the lot
  my $lotbalobj = $self->getLotBalance(ledgerCode => $secobj->ledgerCode, eventDate => $secobj->transDate, lot => $secobj->lot);
  if (not defined $lotbalobj)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to retrieve lot balance!<br>\n" . $self->errorMessage;
    return undef;
  }
  if ($lotbalobj->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to retrieve lot balance!<br>\n" . $lotbalobj->errorMessage;
    return undef;
  }
  if (not $lotbalobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $lotbalobj->errorMessage;
    return undef;
  }
  # delete the tree for this node
  if (not $self->propagateDeleteLot(nodeId => $lotbalobj->nodeId))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error deleting lot node (" . $lotbalobj->nodeId . ")!<br>\n" . $self->errorMessage;
    return undef;
  }
  # propagate the node into the future
  if (not $self->propagateLot(lotNodeObj => $lotbalobj))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error propagating node (" . $lotbalobj->nodeId . ") into the future!<br>\n" . $self->{errorString};
    $self->{dbHandle}->rollback;
    return undef;
  }

  # everything about the transaction succeeded
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }

  return $eventRefNum;
}

# insertReturnOfCapital - global event
# adds a return of capital entry to the sec_event_tb table
sub insertReturnOfCapital
{
  my $self = shift;
  my $errorHeader = "Security->insertReturnOfCapital(): ";
  my %args = ( secobj => undef, counter => undef, @_ );
  my $secobj = $args{'secobj'};
  my $counter = $args{'counter'};

  if ((not defined $args{'secobj'}) || (not ref($secobj)))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "SecurityEventObject not specified!";
    return undef;
  }
  # validate the security event object
  if (not $secobj->isValid())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid object!<br>\n" . $secobj->errorMessage;
    return undef;
  }
  if ($secobj->eventCode ne "50") # return of capital
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Attempting to insert a return-of-capital using a non-return-of-capital object (" . $secobj->eventCode . ")!";
    return undef;
  }

  # insert the security event into the database
  my $eventRefNum = $self->insertSecurityEventObject(obj => $secobj, counter => $counter);
  if ($eventRefNum == -1)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting security event object!<br>\n" . $self->{errorString};
    return undef;
  }
  # retrieve the affected lots for all ledgers
  my @lotBalanceEntries = $self->getLotBalanceEntriesFor(eventDate => $secobj->calcDate, subAccountNum => $secobj->subAccountNum);
  if ($self->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error getting lot balance entries!<br>\n" . $self->errorMessage;
    return undef;
  }
  for (my $lotcount=0; $lotcount < scalar @lotBalanceEntries; $lotcount++)
  {
    # get the sub-tree of nodes attached to this node
    if ((not defined $lotBalanceEntries[$lotcount]) || (not $lotBalanceEntries[$lotcount]->isValid()))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error getting lot balance entries!  Element $lotcount not defined.";
      return undef;
    }
    # delete the tree for this node
    if (not $self->propagateDeleteLot(nodeId => $lotBalanceEntries[$lotcount]->nodeId))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error deleting lot node (" . $lotBalanceEntries[$lotcount]->nodeId . ")!<br>\n" . $self->errorMessage;
      return undef;
    }
    # propagate the previous node into the future
    if (not $self->propagateLot(lotNodeObj => $lotBalanceEntries[$lotcount]))
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error propagating node (" . $lotBalanceEntries[$lotcount]->{nodeId} . ") into the future!<br>\n" . $self->{errorString};
      $self->{dbHandle}->rollback;
      return undef;
    }
  }

  # everything about the transaction succeeded
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error committing transaction to database!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }

  return $eventRefNum;
}

# minimumOpenFiscalStartDate
# this function returns the minimum open fiscal start date for the specified ledger code
sub minimumOpenFiscalStartDate
{
  my $self = shift;
  my $errorHeader = "Security->minimumOpenStartDate():";
  my %args = (ledgerCode => undef, @_ );
  my $ledgerCode = $args{'ledgerCode'};

  if ((not defined $args{'ledgerCode'}) || $ledgerCode eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "ledgerCode not specified!";
    return undef;
  }

  my $sth = $self->{dbHandle}->read(sql => "SELECT min(start_date) FROM known_years_tb WHERE ledger_code = '$ledgerCode' AND closed = false");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving minimum open start date for ledger code = '$ledgerCode'!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }
  my @info = $sth->fetchrow_array;
  if ((not defined @info) || ($info[0] eq ""))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "No minimum open start date returned for ledger code = '$ledgerCode'!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }

  return $info[0];
}

# openLedgers
sub openLedgers
{
  my $self = shift;
  my $errorHeader = "Security->openLedgers():";
  my %args = (date => undef, @_ );
  my $date = $args{'date'};
  my @ledgers;

  if ((not defined $args{'date'}) || $date eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "date not specified!";
    return undef;
  }

  my $sth = $self->{dbHandle}->read(sql => "SELECT ledger_code FROM known_years_tb AS a WHERE a.start_date <= '$date' AND a.start_date = (SELECT min(start_date) FROM known_years_tb WHERE closed = false AND ledger_code = a.ledger_code)");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving open ledger codes for date  = '$date'!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }
  while (my @info = $sth->fetchrow_array)
  {
    $ledgers[scalar @ledgers] = $info[0];
  }

  return @ledgers;
}

# calculateNumShares
sub calculateNumShares
{
  my $self = shift;
  my $errorHeader = "Security->calculateNumShares(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "calculateNumShares<br>\n";
  }
  my %args = ( lot => undef, date => undef, @_ );
  my $lot = $args{'lot'};
  my $date = $args{'date'};

  if ((not defined $args{'lot'}) || $lot eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Lot not specified!";
    return undef;
  }
  if ($lot !~ /^[0-9]+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid lot value!  Lot does not appear to be numeric.";
    return undef;
  }

  my $sqlString = "SELECT event_code, OS, NS, CC FROM sec_event_tb WHERE L = '$lot'";
  if (defined $args{'date'} && $date ne "")
  {
    $sqlString .= " AND calc_date <= '$date'";
  }
  # need smallest date order first, since the earliest value should be the first instance of the lot
  # all lots must be processed in the order the transactions occurred in time
  $sqlString .= " ORDER BY calc_date asc, event_code asc";
  my $sth = $self->{dbHandle}->read(sql => "$sqlString");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving list of events for calculating number of shares!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }
  my $totalShares = 0.0;
  my @info;
  while (defined (@info = $sth->fetchrow_array))
  {
    my $eventCode = $info[0];
    my $oldShares = $info[1];
    my $newShares = $info[2];
    my $costPercentage = $info[3];

    if ($eventCode eq "10" || $eventCode eq "20" || $eventCode eq "30") # buy/sell/stock transfer
    {
      $totalShares += $newShares;
    }
    elsif ($eventCode eq "90" || $eventCode eq "100" || $eventCode eq "60" || $eventCode eq "80" || $eventCode eq "70") # stock split/stock dividend/stock swap/acquisition/spin off
    {
      $totalShares = $newShares / $oldShares * $totalShares;  # /
    }
  }

  return $totalShares;
}

# calculateCostShares
# this function takes the lot and optional date to search for the total cost and number of shares for the lot
# returns a two element array with the total cost as the first element and the total number of shares as the second element
sub calculateCostShares
{
  my $self = shift;
  my $errorHeader = "Security->calculateCostShares(): ";
  if ($self->{debug})
  {
    $self->{errorString} .= "calculateCostShares<br>\n";
  }
  my %args = ( lot => undef, date => undef, @_ );
  my $lot = $args{'lot'};
  my $date = $args{'date'};

  if ((not defined $args{'lot'}) || $lot eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Lot not specified!";
    return undef;
  }
  if ($lot !~ /^[0-9]+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid lot value!  Lot does not appear to be numeric.";
    return undef;
  }

  (my $accountNum, my $subAccountNum, my $ledgerCode) = getAccountSubAccountLedgerForLot(lot => "$lot");
  if (not defined $subAccountNum)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to retrieve sub-account number for lot (" . $lot . ")!<br>\n" . $self->{errorString};
    return undef;
  }

  my $sqlString = "SELECT event_code, OS, NS, CC, event_ref FROM sec_event_tb WHERE L = '$lot'";
  if (defined $args{'date'} && $date ne "")
  {
    $sqlString .= " AND calc_date <= '$date'";
  }
  # need smallest date order first, since the earliest value should be the first instance of the lot
  # all lots must be processed in the order the transactions occurred in time
  $sqlString .= " ORDER BY calc_date asc, event_code asc";
  my $sth = $self->{dbHandle}->read(sql => "$sqlString");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving list of events for calculating number of shares!<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }
  my $totalShares = 0.0;
  my $totalCost = 0.0;
  my @info;
  while (defined (@info = $sth->fetchrow_array))
  {
    my $eventCode = $info[0];
    my $oldShares = $info[1];
    my $newShares = $info[2];
    my $costPercentage = $info[3];
    my $eventRefNum = $info[4];

    if ($eventCode eq "10" || $eventCode eq "20" || $eventCode eq "30") # buy/sell
    {
      $totalShares += $newShares;
      my $sth2 = $self->{dbHandle}->read(sql => "SELECT dorc, amount FROM debitcredit_tb WHERE parent_num = $eventRefNum AND sub_account_num = $subAccountNum");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error retrieving value for buy/sell from debitcredit_tb table!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
      my @info2 = $sth2->fetchrow_array;
      my $dorc = $info2[0];
      my $amount = $info2[1];
      if ($dorc)
      {
        $totalCost += $amount;
      }
      else
      {
        $totalCost -= $amount;
      }
    }
    elsif ($eventCode eq "30") # stock transfer
    {
      $totalShares += $newShares;
      my $tempDORC;
      if ($eventRefNum == $lot)
      {
        $tempDORC = "true";
      }
      else
      {
        $tempDORC = "false";
      }
      my $sth2 = $self->{dbHandle}->read(sql => "SELECT amount FROM debitcredit_tb WHERE parent_num = $eventRefNum AND sub_account_num = $subAccountNum AND dorc = $tempDORC");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error retrieving value for stock transfer from debitcredit_tb table!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
      my @info2 = $sth2->fetchrow_array;
      my $amount = $info2[0];
      if ($tempDORC eq "true")
      {
        $totalCost += $amount;
      }
      else
      {
        $totalCost -= $amount;
      }
    }
    elsif ($eventCode eq "90" || $eventCode eq "100") # stock split/stock dividend
    {
      $totalShares = $newShares / $oldShares * $totalShares;  # /
    }
    elsif ($eventCode eq "60" || $eventCode eq "80" || $eventCode eq "70") # stock swap/acquisition/spin off
    {
      $totalShares = $newShares / $oldShares * $totalShares;  # /
      my $tempDORC;
      if ($eventRefNum == $lot)
      {
        $tempDORC = "true";
      }
      else
      {
        $tempDORC = "false";
      }
      my $sth2 = $self->{dbHandle}->read(sql => "SELECT amount FROM debitcredit_tb WHERE parent_num = $eventRefNum AND sub_account_num = $subAccountNum AND dorc = $tempDORC");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error retrieving value for stock swap/acquisition/spin off from debitcredit_tb table!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
      my @info2 = $sth2->fetchrow_array;
      my $amount = $info2[0];
      if ($tempDORC eq "true")
      {
        $totalCost += $amount;
      }
      else
      {
        $totalCost -= $amount;
      }
    }
  }

  return ($totalCost, $totalShares);
}

sub debug
{
  my $self = shift;
  $self->{debug} = @_;
}

# getLotNode
# takes: nodeId
# returns: LotNodeObject or undef if not found
# This routines looks up the specified nodeId in the database and returns the LotNodeObject that represents it.
sub getLotNode
{
  my $self = shift;
  my %args = ( nodeId => -1, @_ );
  if ($self->{debug})
  {
    $self->{errorString} .= "getLotNode<br>\n";
  }
  my $nodeObj = undef;

  my $nodeId = $args{'nodeId'};

  if ($nodeId !~ /^\d+$/ || $nodeId < 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "nodeId = '$nodeId' is invalid!<br>\n";
    return undef;
  }

  # lookup the specified node in the lot_node_tb.
  my $sth = $self->{dbHandle}->read(sql => "SELECT ledger_code, event_date, parent_id, event_ref, journal_ref, account_num, sub_account_num, lot, buy_lot, shares, cost, node_id FROM lot_node_tb WHERE node_id = ?",
                                plug => [ $nodeId ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = "DBIW Error in getLotNode<br>\n" . $self->{dbHandle}->errorMessage;
    return undef;
  }
  my @info = $sth->fetchrow_array;
  if (defined $info[0])
  {
    $nodeObj = Portal::Accounting::Objects::LotNodeObject->new(ledgerCode => $info[0], eventDate => $info[1], parentId => $info[2], eventRef => $info[3],
              journalRef => $info[4], accountNum => $info[5], subAccountNum => $info[6], lot => $info[7], buyLot => $info[8], shares => $info[9], cost => $info[10], nodeId => $info[11], langObj => $self->{langObj});
    if ($nodeObj->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = "Node Error in getLotNode<br>\n" . $nodeObj->errorMessage;
      return undef;
    }
  }

  return $nodeObj;
}

# getLotTree
# takes: nodeId
# returns: hash of LotNodeObjects
# This routine, starts with the Lot Node = nodeId and gathers all nodes that are children of it.  It creates their objects and stores them in the hash
# which is returned.  The nodeId for each node is the key in the hash.
sub getLotTree
{
  my $self = shift;
  my %args = ( nodeId => -1, @_ );
  if ($self->{debug})
  {
    $self->{errorString} .= "getLotTree<br>\n";
  }
  my $nodeId = $args{nodeId};
  my %nodes = ();

  if ($nodeId !~ /^\d+$/ || $nodeId < 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "nodeId = '$nodeId' is invalid!<br>\n";
    return %nodes;
  }

  # get the top node.
  my $nodeObj = $self->getLotNode(nodeId => $nodeId);
  if ($self->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = "Node Error in getLotTree<br>\n" . $self->errorMessage;
    return %nodes;
  }
  if (!defined $nodeObj)
  {
    $self->{error} = 1;
    $self->{errorString} = "The Parent Node = '$nodeId' does not exist!<br>\n";
    return %nodes;
  }
  $nodes{$nodeObj->nodeId} = $nodeObj;  # insert the parent node into the hash.

  # now get all children nodes of the parent.
  my %children = $self->getChildLotNodes(parentId => $nodeId);
  foreach my $nodeId (keys %children)
  {
    $nodes{$nodeId} = $children{$nodeId};
  }

  return %nodes;
}

# getChildLotNodes
# takes: parentId
# returns: hash of LotNodeObjects that represent all child nodes of the specified parent
# This is a potentially recursive function.
sub getChildLotNodes
{
  my $self = shift;
  my %args = ( parentId => -1, @_ );
  if ($self->{debug})
  {
    $self->{errorString} .= "getChildLotNodes<br>\n";
  }
  my $parentId = $args{parentId};
  my %nodes = ();

  if ($parentId !~ /^\d+$/ || $parentId < 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "parentId = '$parentId' is invalid!<br>\n";
    return %nodes;
  }

  # query the database for all nodes that have parent_id = $parentId
  my $sth = $self->{dbHandle}->read(sql => "SELECT node_id FROM lot_node_tb WHERE parent_id = ? ORDER BY node_id", plug => [ $parentId ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = "DBIW Error in getChildLotNodes<br>\n" . $self->{dbHandle}->errorMessage;
    return %nodes;
  }

  while (my @info = $sth->fetchrow_array)
  {
    # lookup the node info for this node.
    my $nodeObj = $self->getLotNode(nodeId => $info[0]);
    if ($self->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = "Node Error in getChildLotNodes<br>\n" . $self->errorMessage;
      return %nodes;
    }
    if (!defined $nodeObj)
    {
      $self->{error} = 1;
      $self->{errorString} = "Lot Node with ID = '$info[0]' no longer exists!  Getting Children of Node = '$parentId'.<br>\n";
      return %nodes;
    }

    # insert the node into the hash
    $nodes{$nodeObj->nodeId} = $nodeObj;

    # now get all children of this node.
    my %children = $self->getChildLotNodes(parentId => $nodeObj->nodeId);
    if ($self->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = "Children Error in getChildLotNodes<br>\n" . $self->errorMessage;
      return %nodes;
    }
    foreach my $nodeId (keys %children)
    {
      $nodes{$nodeId} = $children{$nodeId};  # insert the children into the parents node hash to build the tree.
    }
  }

  return %nodes;
}

# getCurrentStockPrice
# required: stock, date
# optional:
# returns: array [ticker, price, date, stock]
# This routine gathers the current price for the specified stock (subaccount) along with the current ticker it is under.
sub getCurrentStockPrice
{
  my $self = shift;
  my %args = ( stock => "", date => "", year => "", @_ );
  my $date = $args{date};
  my $year = $args{year};
  my $stock = $args{stock};
  my @result = ();

  if ($stock !~ /^\d+$/)
  {
    $self->setError(errorString => "stock = '$stock' is invalid!<br>\n");
    return @result;
  }
  if ($date !~ /^(now\(\))$/)
  {
    my @dateResult = $self->{methods}->fixupDate(date => $date, year => $year);
    if (!$dateResult[0])
    {
      $self->setError(errorString => "date = '$date' is invalid!<br>$dateResult[2]\n");
      return @result;
    }
    else
    {
      # make sure the date is "fixed".
      $date = $dateResult[1];
    }
  }

  # get the price for the stock (sub-account) for the date
  my $sql = "SELECT price, date_of, sub_acct FROM stock_price_tb WHERE sub_acct = ? AND date_of <= ? ORDER BY date_of DESC LIMIT 1";
  my $sth = $self->{dbHandle}->read(sql => $sql, plug => [ $stock, $date ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => "DBIW Error in getCurrentStockPrice<br>\n" . $self->{dbHandle}->errorMessage);
    return @result;
  }
  my @info = $sth->fetchrow_array;
  if (!defined @info)
  {
    # no data returned.  Give them bogus values, but a cost of "0.00"!
    $result[0] = "0.00";
    $result[1] = $date;  # echo back the date they gave us.  Usually this won't be used.
    $result[2] = $stock;
  }
  else
  {
    # make a copy of the data to return to the user.
    $result[0] = $info[0];
    $result[1] = $info[1];
    $result[2] = $info[2];
  }

  return @result;
}

# getPortfolioData
# required: accounts, date, ledgerCode
# optional: stocks, displayZeroShares
# returns: PortfolioObject or undef if error.
# This routine takes the array of accounts, stocks and gathers all info for a Portfolio view upto the specified date.
sub getPortfolioData
{
  my $self = shift;
  my %args = ( accounts => undef, stocks => undef, date => "", ledgerCode => "", displayZeroShares => 0, year => "", @_ );
  if ($self->{debug})
  {
    $self->{errorString} .= "getPortfolioData<br>\n";
  }
  my @accounts = @{$args{accounts}};
  my @stocks = @{$args{stocks}};
  my $date = $args{date};
  my $ledgerCode = $args{ledgerCode};
  my $portfolioObj = undef;
  my $displayZeroShares = $args{displayZeroShares};
  my $year = $args{year};
  my %times = ();

  if (!defined @accounts)
  {
    $self->setError(errorString => "accounts must be defined!<br>\n");
    return $portfolioObj;
  }
  my @dateResult = $self->{methods}->fixupDate(date => $date, year => $year);
  if (!$dateResult[0])
  {
    $self->setError(errorString => "date = '$date' is invalid!<br>$dateResult[2]\n");
    return $portfolioObj;
  }
  else
  {
    # make sure the date is "fixed".
    $date = $dateResult[1];
  }
  if (length $ledgerCode == 0)
  {
    $self->setError(errorString => "ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return $portfolioObj;
  }

  $times{start} = time();

  # build up the sql string to execute.
  my $sql = "SELECT DISTINCT(a.lot), a.event_date, a.node_id, a.account_num, a.sub_account_num, a.shares, a.cost, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = a.sub_account_num) AS sub_account_name, (SELECT MIN(event_date) FROM lot_node_tb WHERE lot = (SELECT buy_lot FROM lot_parent_tb WHERE lot = a.lot)) as buy_date, (SELECT name FROM account_tb WHERE number = a.account_num) AS account_name FROM lot_node_tb AS a WHERE a.ledger_code = '$ledgerCode'";
  $sql .= " AND a.event_date = (SELECT max(event_date) FROM lot_node_tb WHERE lot = a.lot AND event_date <= '$date')";
  $sql .= " AND a.node_id = (SELECT max(node_id) FROM lot_node_tb WHERE lot = a.lot AND event_date = a.event_date)";
  $sql .= " AND a.shares > '0.0'" if (!$displayZeroShares);
  $sql .= " AND a.account_num IN (";
  for (my $i=0; $i < scalar @accounts; $i++)
  {
    $sql .= "," if ($i > 0);
    $sql .= $accounts[$i]->number;
  }
  $sql .= ")";
  if (scalar @stocks > 0)
  {
    $sql .= " AND a.sub_account_num IN (";
    for (my $i=0; $i < scalar @stocks; $i++)
    {
      $sql .= "," if ($i > 0);
      $sql .= $stocks[$i]->number;
    }
    $sql .= ")";
  }
  $sql .= " ORDER BY account_name, sub_account_name, a.lot";

  # gather the info for the specified data.
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => "DBIW Error in getPortfolioData<br>\n" . $self->{dbHandle}->errorMessage);
    return $portfolioObj;
  }

  $times{sql} = time();

  my $result = $sth->fetchall_arrayref;  # get all the items from the database.
  my @result = @{$result};  # cast it to an array.

  $times{cast} = time();

  my @accountObjs = ();
  my $startAccount = 1;
  my $prevAccount = "";
  my $startAccountIndex = 0;
  my $endAccountIndex = 0;

  my %accountNames;
  my $stockName = "";

  foreach my $accountObj (@accounts)
  {
    $accountNames{$accountObj->number} = $accountObj->name;
  }

  if (scalar @result == 0)
  {
    $self->{errorCode} = -1;  # signal we didn't get any rows back.
    return undef;
  }

  my $lotIndex = 0;
  my $accountIndex = 3;
  my $stockIndex = 4;
  my $dateIndex = 1;
  my $nodeIndex = 2;
  my $sharesIndex = 5;
  my $costIndex = 6;
  my $accountNameIndex = 9;
  my $subAccountNameIndex = 7;
  my $buyDateIndex = 8;

  my $processed_last_account = 0;  # flag to catch the case where the last account only had 1 lot row returned and wasn't being included in result.

  my %currentStockPrices = ();

  # now walk over the result set and for all rows that have the same account, process all stock and lot pairs and produce a single PortfolioAccountObject.
  for (my $i=0; $i < scalar @result; $i++)
  {
    if ($startAccount)
    {
      $startAccount = 0;  # signal we are done just blindly taking the first account we meet.
      $prevAccount = $result[$i][$accountIndex];
      $startAccountIndex = $i;
      $endAccountIndex = $i;
    }
    elsif ($result[$i][$accountIndex] == $prevAccount)
    {
      $endAccountIndex = $i;
    }
    if (($result[$i][$accountIndex] != $prevAccount && $startAccount == 0) || ($i == scalar(@result) - 1))  # now we need to process all entries for the previous account.
    {
      my @stockObjs = ();
      my @lotObjs = ();
      my %accountTotals = ( totalCost => "0.00", totalMarketValue => "0.00", totalGainLoss => "0.00" );
      my %stockTotals = ( totalShares => "0.00", totalCost => "0.00", totalMarketValue => "0.00", totalGainLoss => "0.00" );
      my $startStocks = 1;
      my $prevStock = "";
      for (my $j=$startAccountIndex; $j <= $endAccountIndex; $j++)
      {
        # create the PortfolioLotObject
        my $shares = $result[$j][$sharesIndex];
        my $cost   = $result[$j][$costIndex];
        my $pricePerShare = "0.00";

        if (not exists $currentStockPrices{$result[$j][$stockIndex]})
        {
          # lookup the current price from the database.
          my @stockPriceInfo = $self->getCurrentStockPrice(stock => $result[$j][$stockIndex], date => $date, year => $year);
          if ($self->didErrorOccur)
          {
            $self->prefixError();
            return $portfolioObj;
          }
          $currentStockPrices{$result[$j][$stockIndex]} = $stockPriceInfo[0];
        }
        $pricePerShare = sprintf("%.2f", $currentStockPrices{$result[$j][$stockIndex]});

        my $marketValue = $shares * $pricePerShare;
        $marketValue = sprintf("%.2f", $marketValue);
        my $gainLoss = $marketValue - $cost;
        $gainLoss = sprintf("%.2f", $gainLoss);
        my $costPerShare = ($shares !~ /^0+(\.0+)?$/ && $shares ne "" ? $cost / $shares : "0.00");#/
        $costPerShare = sprintf("%.2f", $costPerShare);
        $shares =~ s/^(\d+\.\d)(0+)$/$1/;
        $cost = sprintf("%.2f", $cost);

        my $portfolioLotObj = Portal::Accounting::Objects::PortfolioLotObject->new(accountNum => $result[$j][$accountIndex], subAccountNum => $result[$j][$stockIndex],
                              lotRefNum => $result[$j][$lotIndex], date => $result[$j][$dateIndex], shares => $shares, cost => $cost,
                              newPricePerShare => $pricePerShare, marketValue => $marketValue, gainLoss => $gainLoss, costPerShare => $costPerShare,
                              buyDate => $result[$j][$buyDateIndex], langObj => $self->{langObj});
        if ($portfolioLotObj->didErrorOccur)
        {
          $self->setError(errorString => "PortfolioLotObject Error in getPortfolioData<br>\n" . $portfolioLotObj->errorMessage);
          return $portfolioObj;
        }

        # now see what StockObject this goes with.
        if ($result[$j][$stockIndex] == $prevStock || $startStocks)
        {
          $startStocks = 0;
          $prevStock = $result[$j][$stockIndex];
          $lotObjs[++$#lotObjs] = $portfolioLotObj;

          $stockName = $result[$j][$subAccountNameIndex];

          $stockTotals{totalShares} += $shares;
          $stockTotals{totalCost} += $cost;
          $stockTotals{totalMarketValue} += $marketValue;
          $stockTotals{totalGainLoss} += $gainLoss;

          $accountTotals{totalCost} += $cost;
          $accountTotals{totalMarketValue} += $marketValue;
          $accountTotals{totalGainLoss} += $gainLoss;
        }
        if (($result[$j][$stockIndex] != $prevStock && $startStocks == 0) || ($j == $endAccountIndex))  # create this stock and start next one.
        {
          # make sure all figures are 2 decimal points.
          foreach my $total (keys %stockTotals)
          {
            if ($total eq "totalShares")
            {
              $stockTotals{$total} = sprintf("%.6f", $stockTotals{$total});
              $stockTotals{$total} =~ s/^(\d+\.\d)(0+)$/$1/;
            }
            else
            {
              $stockTotals{$total} = sprintf("%.2f", $stockTotals{$total});
            }
          }

          my $averageCostPerShare = ($stockTotals{totalShares} !~ /^0+\.0+$/ && $stockTotals{totalShares} ne "" ? $stockTotals{totalCost} / $stockTotals{totalShares} : "0.00");#/
          $averageCostPerShare = sprintf("%.2f", $averageCostPerShare);

          my @tmpLotObjs = @lotObjs;
          my $portfolioStockObj = Portal::Accounting::Objects::PortfolioStockObject->new(subAccountName => $stockName, subAccountNum => $prevStock,
                                  totalShares => $stockTotals{totalShares}, totalCost => $stockTotals{totalCost}, totalMarketValue => $stockTotals{totalMarketValue},
                                  accountNum => $prevAccount, totalGainLoss => $stockTotals{totalGainLoss}, status => "c", lots => \@tmpLotObjs,
                                  averageCostPerShare => $averageCostPerShare, newPricePerShare => $tmpLotObjs[0]->{newPricePerShare}, langObj => $self->{langObj});
          if ($portfolioStockObj->didErrorOccur)
          {
            $self->setError(errorString =>  "PortfolioStockObject Error in getPortfolioData<br>\n" .$portfolioStockObj->errorMessage);
            return $portfolioObj;
          }
          $stockObjs[++$#stockObjs] = $portfolioStockObj;

          if (($j + 1) <= $endAccountIndex)
          {
            # reset Lot* variables and Stock* totals
            @lotObjs = ();
            $prevStock = $result[$j][$stockIndex];
            $startStocks = 0;

            $stockName = $result[$j][$subAccountNameIndex];

            $lotObjs[++$#lotObjs] = $portfolioLotObj;

            %stockTotals = ( totalShares => "0.00", totalCost => "0.00", totalMarketValue => "0.00", totalGainLoss => "0.00" );

            $stockTotals{totalShares} += $shares;
            $stockTotals{totalCost} += $cost;
            $stockTotals{totalMarketValue} += $marketValue;
            $stockTotals{totalGainLoss} += $gainLoss;

            $accountTotals{totalCost} += $cost;
            $accountTotals{totalMarketValue} += $marketValue;
            $accountTotals{totalGainLoss} += $gainLoss;
          }
          elsif (($result[$j][$stockIndex] != $prevStock && $startStocks == 0) && ($j == $endAccountIndex))
          {  # did we just switch stocks and we are the last one in the returned list?
            # we must make the PortfolioStockObj entry!
            @lotObjs = ();
            $prevStock = $result[$j][$stockIndex];
            $startStocks = 0;

            $stockName = $result[$j][$subAccountNameIndex];

            $lotObjs[++$#lotObjs] = $portfolioLotObj;

            %stockTotals = ( totalShares => "0.00", totalCost => "0.00", totalMarketValue => "0.00", totalGainLoss => "0.00" );

            $stockTotals{totalShares} += $shares;
            $stockTotals{totalCost} += $cost;
            $stockTotals{totalMarketValue} += $marketValue;
            $stockTotals{totalGainLoss} += $gainLoss;

            $accountTotals{totalCost} += $cost;
            $accountTotals{totalMarketValue} += $marketValue;
            $accountTotals{totalGainLoss} += $gainLoss;

            # make sure all figures are 2 decimal points.
            foreach my $total (keys %stockTotals)
            {
              if ($total eq "totalShares")
              {
                $stockTotals{$total} = sprintf("%.6f", $stockTotals{$total});
                $stockTotals{$total} =~ s/^(\d+\.\d)(0+)$/$1/;
              }
              else
              {
                $stockTotals{$total} = sprintf("%.2f", $stockTotals{$total});
              }
            }

            my $averageCostPerShare = ($stockTotals{totalShares} !~ /^0+\.0+$/ && $stockTotals{totalShares} ne "" ? $stockTotals{totalCost} / $stockTotals{totalShares} : "0.00");#/
            $averageCostPerShare = sprintf("%.2f", $averageCostPerShare);

            my @tmpLotObjs = @lotObjs;
            my $portfolioStockObj = Portal::Accounting::Objects::PortfolioStockObject->new(subAccountName => $stockName, subAccountNum => $prevStock,
                                    totalShares => $stockTotals{totalShares}, totalCost => $stockTotals{totalCost}, totalMarketValue => $stockTotals{totalMarketValue},
                                    accountNum => $prevAccount, totalGainLoss => $stockTotals{totalGainLoss}, status => "o", lots => \@tmpLotObjs,
                                    averageCostPerShare => $averageCostPerShare, newPricePerShare => $tmpLotObjs[0]->{newPricePerShare}, langObj => $self->{langObj});
            if ($portfolioStockObj->didErrorOccur)
            {
              $self->setError(errorString =>  "PortfolioStockObject Error in getPortfolioData<br>\n" .$portfolioStockObj->errorMessage);
              return $portfolioObj;
            }
            $stockObjs[++$#stockObjs] = $portfolioStockObj;
          }
        }
      }

      # make sure all figures are 2 decimal points.
      foreach my $total (keys %accountTotals)
      {
        $accountTotals{$total} = sprintf("%.2f", $accountTotals{$total});
      }

      # create the PortfolioAccountObject and insert into accountObjs
      my @tmpStockObjs = @stockObjs;
      my $portfolioAccountObj = Portal::Accounting::Objects::PortfolioAccountObject->new(accountName => $accountNames{$prevAccount}, accountNum => $prevAccount,
                                totalCost => $accountTotals{totalCost}, totalMarketValue => $accountTotals{totalMarketValue}, totalGainLoss => $accountTotals{totalGainLoss},
                                status => "o", stocks => \@tmpStockObjs, langObj => $self->{langObj});
      if ($portfolioAccountObj->didErrorOccur)
      {
        $self->setError(errorString => "PortfolioAccountObject Error in getPortfolioData<br>\n" . $portfolioAccountObj->errorMessage);
        return $portfolioObj;
      }
      $accountObjs[++$#accountObjs] = $portfolioAccountObj;

      # reset Account* variables.
      $startAccount = 0;
      $startAccountIndex = $i;
      $endAccountIndex = $i;
      $prevAccount = $result[$i][$accountIndex];
      if ($i > 0 && $result[$i-1][$accountIndex] != $prevAccount && ($i == scalar(@result) - 1) && !$processed_last_account)
      {
        # signal we have already done this.
        $processed_last_account = 1;
        redo;  # process last entry.
      }
    }
  }

  # now build the portfolioObj.
  $portfolioObj = Portal::Accounting::Objects::PortfolioObject->new(accounts => \@accountObjs, langObj => $self->{langObj});
  if ($portfolioObj->didErrorOccur)
  {
    $self->setError(errorString => "PortfolioObject Error in getPortfolioData<br>\n" . $portfolioObj->errorMessage);
    return $portfolioObj;
  }

  $times{end} = time();

  $portfolioObj->{times} = \%times;

  return $portfolioObj;
}

# getCGLData
# required: startDate, endDate, orderBy, ledgerCode
# optional: invAccount, stocks, assets, getManualEntries
# returns: CGLDisplayObject or undef if error.
# This routine takes the invAccount, array of assets, stocks and gathers all info for a CGL view in the specified date range.
sub getCGLData
{
  my $self = shift;
  my %args = ( invAccount => undef, assets => undef, stocks => undef, startDate => "", endDate => "", orderBy => 0, ledgerCode => "", getManualEntries => 0, @_ );
  if ($self->{debug})
  {
    $self->{errorString} .= "getCGLData<br>\n";
  }
  my $invAccount = $args{invAccount};
  my @assets = @{$args{assets}};
  my @stocks = @{$args{stocks}};
  my $startDate = $args{startDate};
  my $endDate = $args{endDate};
  my $ledgerCode = $args{ledgerCode};
  my $orderBy = $args{orderBy};
  my $getManualEntries = $args{getManualEntries};
  my $cglDisplayObj = undef;

  my @dateResult = $self->{methods}->fixupDate(date => $startDate);
  if (!$dateResult[0])
  {
    $self->setError(errorString => "startDate = '$startDate' is invalid!<br>$dateResult[2]\n");
    return $cglDisplayObj;
  }
  else
  {
    # make sure the date is "fixed".
    $startDate = $dateResult[1];
  }
  @dateResult = $self->{methods}->fixupDate(date => $endDate);
  if (!$dateResult[0])
  {
    $self->setError(errorString => "endDate = '$endDate' is invalid!<br>$dateResult[2]\n");
    return $cglDisplayObj;
  }
  else
  {
    # make sure the date is "fixed".
    $endDate = $dateResult[1];
  }
  if (length $ledgerCode == 0)
  {
    $self->setError("ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return $cglDisplayObj;
  }
  if ($orderBy !~ /^(0|1)$/)
  {
    $self->setError(errorString => "orderBy = '$orderBy' is invalid!<br>\n");
    return $cglDisplayObj;
  }

  my $assets = join(", ", @assets);
  my $stocks = join(", ", @stocks);

  if ($orderBy)  # SubAccount - everything lumped together sorted by subaccount
  {
    # generate the Sql statements needed to gather the data
    my $slcapglAccounts = "$shortTermCG, $longTermCG, $shortTermCL, $longTermCL";
    my $stockSQLShortLong = "SELECT d.trans_date AS sell_date, c_amount AS amount, d.account_num, d.sub_account_num, get_asset_data(d.parent_num, d.account_num, d.sub_account_num, false, true) AS cost, get_asset_data(d.parent_num, d.account_num, -1, true, true) AS market, stock_purchased(l.lot) AS purchase_date, l.lot, get_shares_data(l.lot, d.parent_num) AS shares, d.parent_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = d.sub_account_num) AS sub_account_name, (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') AS invaccount FROM debitcredit_tb AS d, lot_node_tb AS l WHERE d.ledger_code = '$ledgerCode' AND d.ledger_code = l.ledger_code AND d.account_num IN ($slcapglAccounts)INVACCOUNTSTOCKS AND d.trans_date BETWEEN '$startDate' AND '$endDate' AND d.parent_num = l.journal_ref ORDER BY sub_account_name, account_num, sell_date";
    my $assetSQLShortLong = "SELECT trans_date AS sell_date, c_amount AS amount, account_num, sub_account_num, get_asset_data(parent_num, account_num, sub_account_num, false, true) AS cost, get_asset_data(parent_num, account_num, -1, true, true) AS market, asset_purchased(sub_account_num, ledger_code) AS purchase_date, parent_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = sub_account_num) AS sub_account_name FROM debitcredit_tb WHERE ledger_code = '$ledgerCode' AND account_num IN ($slcapglAccounts)STOCKS AND trans_date between '$startDate' AND '$endDate' ORDER BY sub_account_name, account_num, sell_date";
    my $manualSQLShortLong = "SELECT d.account_num, trans_date AS sell_date, c_amount AS amount, d.sub_account_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = d.sub_account_num) AS sub_account_name, d.parent_num FROM debitcredit_tb AS d, lot_node_tb AS l WHERE d.ledger_code = '$ledgerCode' AND d.ledger_code = l.ledger_code AND d.account_num IN ($slcapglAccounts) AND d.trans_date BETWEEN '$startDate' AND '$endDate' EXCEPT SELECT d.account_num, trans_date AS sell_date, c_amount AS amount, d.sub_account_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = d.sub_account_num) AS sub_account_name, d.parent_num FROM debitcredit_tb AS d, lot_node_tb AS l WHERE d.ledger_code = '$ledgerCode' AND d.ledger_code = l.ledger_code AND d.account_num IN ($slcapglAccounts) AND d.trans_date BETWEEN '$startDate' AND '$endDate' AND d.parent_num = l.journal_ref ORDER BY sub_account_name, account_num, sell_date";

    if (length $stocks > 0)
    {
      $stockSQLShortLong =~ s/(STOCKS)/ AND d.sub_account_num IN ($stocks)/;
    }
    else
    {
      $stockSQLShortLong =~ s/(STOCKS)//;
    }
    if (length $assets > 0)
    {
      $assetSQLShortLong =~ s/(STOCKS)/ AND sub_account_num IN ($assets)/;
    }
    else
    {
      $assetSQLShortLong =~ s/(STOCKS)//;
    }
    if (length $invAccount > 0)
    {
      $stockSQLShortLong =~ s/(INVACCOUNT)/ AND (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') = $invAccount/;
      $assetSQLShortLong =~ s/(INVACCOUNT)/ AND (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') = $invAccount/;
    }
    else
    {
      $stockSQLShortLong =~ s/(INVACCOUNT)//;
      $assetSQLShortLong =~ s/(INVACCOUNT)//;
    }

    # gather the info for the specified data.
    my $sth = $self->{dbHandle}->read(sql => $stockSQLShortLong);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => "DBIW Error in getCGLData<br>\n" . $self->{dbHandle}->errorMessage);
      return $cglDisplayObj;
    }

    my $cglStocks = $self->buildCGLStocks(sth => $sth);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return $cglDisplayObj;
    }

    #$sth = $self->{dbHandle}->read(sql => $assetSQLShortLong);
    #if ($self->{dbHandle}->didErrorOccur)
    #{
    #  $self->setError(errorString => "DBIW Error in getCGLData<br>\n" . $self->{dbHandle}->errorMessage);
    #  return $cglDisplayObj;
    #}

    my $cglAssets = $self->buildCGLAssets(sth => $sth);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return $cglDisplayObj;
    }
    # remove when the Assets are actually being gathered for real.
    $cglAssets->{cost} = "0.00";
    $cglAssets->{market} = "0.00";
    $cglAssets->{gainLoss} = "0.00";

    $sth = $self->{dbHandle}->read(sql => $manualSQLShortLong);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => "DBIW Error in getCGLData<br>\n" . $self->{dbHandle}->errorMessage);
      return $cglDisplayObj;
    }

    my $cglManual = undef;
    if ($getManualEntries)
    {
      $cglManual = $self->buildCGLManual(sth => $sth);
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return $cglDisplayObj;
      }
    }

    my $grandTotalCost = $cglStocks->{cost} + $cglAssets->{cost};
    my $grandTotalMarket = $cglStocks->{market} + $cglAssets->{market};
    my $grandTotalGainLoss = $cglStocks->{gainLoss} + $cglAssets->{gainLoss} + ($getManualEntries ? $cglManual->{gainLoss} : "00.0");

    $grandTotalCost = sprintf("%.2f", $grandTotalCost);
    $grandTotalMarket = sprintf("%.2f", $grandTotalMarket);
    $grandTotalGainLoss = sprintf("%.2f", $grandTotalGainLoss);

    my $cglStockObject = Portal::Accounting::Objects::CGLObject->new(type => "both", state => "o", cglObjects => $cglStocks, cost => $cglStocks->{cost}, market => $cglStocks->{market}, gainLoss => $cglStocks->{gainLoss}, langObj => $self->{langObj}, name => "stocks");
    if ($cglStockObject->didErrorOccur)
    {
      $self->setError(errorString => $cglStockObject->errorMessage);
      return $cglDisplayObj;
    }

    if ((scalar @{$cglStockObject->{cglObjects}->{cglStockObjects}} == 0))
    {
      $cglStockObject->{state} = "c";
    }

    my $cglAssetObject = Portal::Accounting::Objects::CGLObject->new(type => "both", state => "o", cglObjects => $cglAssets, cost => $cglAssets->{cost}, market => $cglAssets->{market}, gainLoss => $cglAssets->{gainLoss}, langObj => $self->{langObj}, name => "assets");
    if ($cglAssetObject->didErrorOccur)
    {
      $self->setError(errorString => $cglAssetObject->errorMessage);
      return $cglDisplayObj;
    }

    if ((scalar @{$cglAssetObject->{cglObjects}->{cglAssetObjects}} == 0))
    {
      $cglAssetObject->{state} = "c";
    }

    my $cglManualObject = undef;
    if ($getManualEntries)
    {
      $cglManualObject = Portal::Accounting::Objects::CGLObject->new(type => "both", state => "o", cglObjects => $cglManual, cost => "0.00", market => "0.00", gainLoss => $cglManual->{gainLoss}, langObj => $self->{langObj}, name => "manual");
      if ($cglManualObject->didErrorOccur)
      {
        $self->setError(errorString => $cglManualObject->errorMessage);
        return $cglDisplayObj;
      }

      if ((scalar @{$cglManualObject->{cglObjects}->{cglManualObjects}} == 0))
      {
        $cglManualObject->{state} = "c";
      }
    }

    $cglDisplayObj = Portal::Accounting::Objects::CGLDisplayObject->new(cost => $grandTotalCost, market => $grandTotalMarket, gainLoss => $grandTotalGainLoss, cglStocks => [ $cglStockObject ], cglAssets => [ $cglAssetObject ], cglManual => [ $cglManualObject ], langObj => $self->{langObj});
    if ($cglDisplayObj->didErrorOccur)
    {
      $self->setError(errorString => $cglDisplayObj->errorMessage);
      return $cglDisplayObj;
    }
  }
  else # ST/LT - they are broken apart into 2 distinct objects.
  {
    # generate the Sql statements needed to gather the data
    my $stcapglAccounts = "$shortTermCG, $shortTermCL";
    my $ltcapglAccounts = "$longTermCG, $longTermCL";
    my $stockSQLShort = "SELECT d.trans_date AS sell_date, c_amount AS amount, d.account_num, d.sub_account_num, get_asset_data(d.parent_num, d.account_num, d.sub_account_num, false, true) AS cost, get_asset_data(d.parent_num, d.account_num, -1, true, true) AS market, stock_purchased(l.lot) AS purchase_date, l.lot, get_shares_data(l.lot, d.parent_num) AS shares, d.parent_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = d.sub_account_num) AS sub_account_name, (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') AS invaccount FROM debitcredit_tb AS d, lot_node_tb AS l WHERE d.ledger_code = '$ledgerCode' AND d.ledger_code = l.ledger_code AND d.account_num IN ($stcapglAccounts)INVACCOUNTSTOCKS AND d.trans_date BETWEEN '$startDate' AND '$endDate' AND d.parent_num = l.journal_ref ORDER BY sub_account_name, account_num, sell_date, lot";
    my $assetSQLShort = "SELECT trans_date AS sell_date, c_amount AS amount, account_num, sub_account_num, get_asset_data(parent_num, account_num, sub_account_num, false, true) AS cost, get_asset_data(parent_num, account_num, -1, true, true) AS market, asset_purchased(sub_account_num, ledger_code) AS purchase_date, parent_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = sub_account_num) AS sub_account_name, (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') AS invaccount FROM debitcredit_tb WHERE ledger_code = '$ledgerCode' AND account_num IN ($stcapglAccounts)INVACCOUNTSTOCKS AND trans_date between '$startDate' AND '$endDate' ORDER BY sub_account_name, account_num, sell_date";
    my $stockSQLLong = "SELECT d.trans_date AS sell_date, c_amount AS amount, d.account_num, d.sub_account_num, get_asset_data(d.parent_num, d.account_num, d.sub_account_num, false, true) AS cost, get_asset_data(d.parent_num, d.account_num, -1, true, true) AS market, stock_purchased(l.lot) AS purchase_date, l.lot, get_shares_data(l.lot, d.parent_num) AS shares, d.parent_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = d.sub_account_num) AS sub_account_name, (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') AS invaccount FROM debitcredit_tb AS d, lot_node_tb AS l WHERE d.ledger_code = '$ledgerCode' AND d.ledger_code = l.ledger_code AND d.account_num IN ($ltcapglAccounts)INVACCOUNTSTOCKS AND d.trans_date BETWEEN '$startDate' AND '$endDate' AND d.parent_num = l.journal_ref ORDER BY sub_account_name, account_num, sell_date, lot";
    my $assetSQLLong = "SELECT trans_date AS sell_date, c_amount AS amount, account_num, sub_account_num, get_asset_data(parent_num, account_num, sub_account_num, false, true) AS cost, get_asset_data(parent_num, account_num, -1, true, true) AS market, asset_purchased(sub_account_num, ledger_code) AS purchase_date, parent_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = sub_account_num) AS sub_account_name, (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') AS invaccount FROM debitcredit_tb WHERE ledger_code = '$ledgerCode' AND account_num IN ($ltcapglAccounts)INVACCOUNTSTOCKS AND trans_date between '$startDate' AND '$endDate' ORDER BY sub_account_name, account_num, sell_date";
    my $manualSQLShort = "SELECT d.account_num, trans_date AS sell_date, c_amount AS amount, d.sub_account_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = d.sub_account_num) AS sub_account_name, d.parent_num FROM debitcredit_tb AS d, lot_node_tb AS l WHERE d.ledger_code = '$ledgerCode' AND d.ledger_code = l.ledger_code AND d.account_num IN ($stcapglAccounts) AND d.trans_date BETWEEN '$startDate' AND '$endDate' EXCEPT SELECT d.account_num, trans_date AS sell_date, c_amount AS amount, d.sub_account_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = d.sub_account_num) AS sub_account_name, d.parent_num FROM debitcredit_tb AS d, lot_node_tb AS l WHERE d.ledger_code = '$ledgerCode' AND d.ledger_code = l.ledger_code AND d.account_num IN ($stcapglAccounts) AND d.trans_date BETWEEN '$startDate' AND '$endDate' AND d.parent_num = l.journal_ref ORDER BY sub_account_name, account_num, sell_date";
    my $manualSQLLong = "SELECT d.account_num, trans_date AS sell_date, c_amount AS amount, d.sub_account_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = d.sub_account_num) AS sub_account_name, d.parent_num FROM debitcredit_tb AS d, lot_node_tb AS l WHERE d.ledger_code = '$ledgerCode' AND d.ledger_code = l.ledger_code AND d.account_num IN ($ltcapglAccounts) AND d.trans_date BETWEEN '$startDate' AND '$endDate' EXCEPT SELECT d.account_num, trans_date AS sell_date, c_amount AS amount, d.sub_account_num, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = d.sub_account_num) AS sub_account_name, d.parent_num FROM debitcredit_tb AS d, lot_node_tb AS l WHERE d.ledger_code = '$ledgerCode' AND d.ledger_code = l.ledger_code AND d.account_num IN ($ltcapglAccounts) AND d.trans_date BETWEEN '$startDate' AND '$endDate' AND d.parent_num = l.journal_ref ORDER BY sub_account_name, account_num, sell_date";

    if (length $stocks > 0)
    {
      $stockSQLShort =~ s/(STOCKS)/ AND d.sub_account_num IN ($stocks)/;
      $stockSQLLong =~ s/(STOCKS)/ AND d.sub_account_num IN ($stocks)/;
    }
    else
    {
      $stockSQLShort =~ s/(STOCKS)//;
      $stockSQLLong =~ s/(STOCKS)//;
    }
    if (length $assets > 0)
    {
      $assetSQLShort =~ s/(STOCKS)/ AND sub_account_num IN ($assets)/;
      $assetSQLLong =~ s/(STOCKS)/ AND sub_account_num IN ($assets)/;
    }
    else
    {
      $assetSQLShort =~ s/(STOCKS)//;
      $assetSQLLong =~ s/(STOCKS)//;
    }
    if (length $invAccount > 0)
    {
      $stockSQLShort =~ s/(INVACCOUNT)/ AND (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') = $invAccount/;
      $stockSQLLong =~ s/(INVACCOUNT)/ AND (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') = $invAccount/;
      $assetSQLShort =~ s/(INVACCOUNT)/ AND (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') = $invAccount/;
      $assetSQLLong =~ s/(INVACCOUNT)/ AND (SELECT account_num FROM debitcredit_tb WHERE parent_num = d.parent_num AND sub_account_num = '-1') = $invAccount/;
    }
    else
    {
      $stockSQLShort =~ s/(INVACCOUNT)//;
      $stockSQLLong =~ s/(INVACCOUNT)//;
      $assetSQLShort =~ s/(INVACCOUNT)//;
      $assetSQLLong =~ s/(INVACCOUNT)//;
    }

    # gather the info for the specified data. - Short Term
    my $sth = $self->{dbHandle}->read(sql => $stockSQLShort);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = "DBIW Error in getCGLData - Short Term Capital GL - Stocks<br>\n" . $self->{dbHandle}->errorMessage;
      return $cglDisplayObj;
    }

    my $cglStocks = $self->buildCGLStocks(sth => $sth);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return $cglDisplayObj;
    }

    #$sth = $self->{dbHandle}->read(sql => $assetSQLShort);
    #if ($self->{dbHandle}->didErrorOccur)
    #{
    #  $self->{error} = 1;
    #  $self->{errorString} = "DBIW Error in getCGLData - Short Term Capital GL - Stocks<br>\n" . $self->{dbHandle}->errorMessage;
    #  return $cglDisplayObj;
    #}

    my $cglAssets = $self->buildCGLAssets(sth => $sth);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return $cglDisplayObj;
    }
    # remove when the Assets are actually being gathered for real.
    $cglAssets->{cost} = "0.00";
    $cglAssets->{market} = "0.00";
    $cglAssets->{gainLoss} = "0.00";

    $sth = $self->{dbHandle}->read(sql => $manualSQLShort);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = "DBIW Error in getCGLData - Short Term Capital GL - Manual<br>\n" . $self->{dbHandle}->errorMessage;
      return $cglDisplayObj;
    }

    my $cglManual = undef;
    if ($getManualEntries)
    {
      $cglManual = $self->buildCGLManual(sth => $sth);
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return $cglDisplayObj;
      }
    }

    my $grandTotalCost = $cglStocks->{cost} + $cglAssets->{cost};
    my $grandTotalMarket = $cglStocks->{market} + $cglAssets->{market};
    my $grandTotalGainLoss = $cglStocks->{gainLoss} + $cglAssets->{gainLoss} + ($getManualEntries ? $cglManual->{gainLoss} : "00.0");

    $grandTotalCost = sprintf("%.2f", $grandTotalCost);
    $grandTotalMarket = sprintf("%.2f", $grandTotalMarket);
    $grandTotalGainLoss = sprintf("%.2f", $grandTotalGainLoss);

    my $cglStockObject = Portal::Accounting::Objects::CGLObject->new(type => "ST", state => "o", cglObjects => $cglStocks, cost => $cglStocks->{cost}, market => $cglStocks->{market}, gainLoss => $cglStocks->{gainLoss}, langObj => $self->{langObj}, name => "stocks");
    if ($cglStockObject->didErrorOccur)
    {
      $self->setError(errorString => $cglStockObject->errorMessage);
      return $cglDisplayObj;
    }

    if ((scalar @{$cglStockObject->{cglObjects}->{cglStockObjects}} == 0))
    {
      $cglStockObject->{state} = "c";
    }

    my $cglAssetObject = Portal::Accounting::Objects::CGLObject->new(type => "ST", state => "o", cglObjects => $cglAssets, cost => $cglAssets->{cost}, market => $cglAssets->{market}, gainLoss => $cglAssets->{gainLoss}, langObj => $self->{langObj}, name => "assets");
    if ($cglAssetObject->didErrorOccur)
    {
      $self->setError(errorString => $cglAssetObject->errorMessage);
      return $cglDisplayObj;
    }

    if ((scalar @{$cglAssetObject->{cglObjects}->{cglAssetObjects}} == 0))
    {
      $cglAssetObject->{state} = "c";
    }

    my $cglManualObject = undef;
    if ($getManualEntries)
    {
      $cglManualObject = Portal::Accounting::Objects::CGLObject->new(type => "ST", state => "o", cglObjects => $cglManual, cost => "0.00", market => "0.00", gainLoss => $cglManual->{gainLoss}, langObj => $self->{langObj}, name => "manual");
      if ($cglManualObject->didErrorOccur)
      {
        $self->setError(errorString => $cglManualObject->errorMessage);
        return $cglDisplayObj;
      }

      if ((scalar @{$cglManualObject->{cglObjects}->{cglManualObjects}} == 0))
      {
        $cglManualObject->{state} = "c";
      }
    }

    # Deal with the Long Term data
    $sth = $self->{dbHandle}->read(sql => $stockSQLLong);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = "DBIW Error in getCGLData - Long Term Capital GL - Stocks<br>\n" . $self->{dbHandle}->errorMessage;
      return $cglDisplayObj;
    }

    my $cglStocks = $self->buildCGLStocks(sth => $sth);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return $cglDisplayObj;
    }

    my $cglAssets = $self->buildCGLAssets(sth => $sth);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return $cglDisplayObj;
    }
    # remove when the Assets are actually being gathered for real.
    $cglAssets->{cost} = "0.00";
    $cglAssets->{market} = "0.00";
    $cglAssets->{gainLoss} = "0.00";

    $sth = $self->{dbHandle}->read(sql => $manualSQLLong);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = "DBIW Error in getCGLData - Long Term Capital GL - Manual<br>\n" . $self->{dbHandle}->errorMessage;
      return $cglDisplayObj;
    }

    my $cglManual = undef;
    if ($getManualEntries)
    {
      $cglManual = $self->buildCGLManual(sth => $sth);
      if ($self->didErrorOccur)
      {
        return $cglDisplayObj;
      }
    }

    my $grandTotalCost = $cglStocks->{cost} + $cglAssets->{cost};
    my $grandTotalMarket = $cglStocks->{market} + $cglAssets->{market};
    my $grandTotalGainLoss = $cglStocks->{gainLoss} + $cglAssets->{gainLoss} + ($getManualEntries ? $cglManual->{gainLoss} : "00.0");

    $grandTotalCost = sprintf("%.2f", $grandTotalCost);
    $grandTotalMarket = sprintf("%.2f", $grandTotalMarket);
    $grandTotalGainLoss = sprintf("%.2f", $grandTotalGainLoss);

    my $cglStockObject2 = Portal::Accounting::Objects::CGLObject->new(type => "LT", state => "o", cglObjects => $cglStocks, cost => $cglStocks->{cost}, market => $cglStocks->{market}, gainLoss => $cglStocks->{gainLoss}, langObj => $self->{langObj}, name => "stocks");
    if ($cglStockObject2->didErrorOccur)
    {
      $self->setError(errorString => $cglStockObject2->errorMessage);
      return $cglDisplayObj;
    }

    if ((scalar @{$cglStockObject2->{cglObjects}->{cglStockObjects}} == 0))
    {
      $cglStockObject2->{state} = "c";
    }

    my $cglAssetObject2 = Portal::Accounting::Objects::CGLObject->new(type => "LT", state => "o", cglObjects => $cglAssets, cost => $cglAssets->{cost}, market => $cglAssets->{market}, gainLoss => $cglAssets->{gainLoss}, langObj => $self->{langObj}, name => "assets");
    if ($cglAssetObject2->didErrorOccur)
    {
      $self->setError(errorString => $cglAssetObject2->errorMessage);
      return $cglDisplayObj;
    }

    if ((scalar @{$cglAssetObject2->{cglObjects}->{cglAssetObjects}} == 0))
    {
      $cglAssetObject2->{state} = "c";
    }

    my $cglManualObject2 = undef;
    if ($getManualEntries)
    {
      $cglManualObject2 = Portal::Accounting::Objects::CGLObject->new(type => "LT", state => "o", cglObjects => $cglManual, cost => "0.00", market => "0.00", gainLoss => $cglManual->{gainLoss}, langObj => $self->{langObj}, name => "manual");
      if ($cglManualObject2->didErrorOccur)
      {
        $self->setError(errorString => $cglManualObject2->errorMessage);
        return $cglDisplayObj;
      }

      if ((scalar @{$cglManualObject2->{cglObjects}->{cglManualObjects}} == 0))
      {
        $cglManualObject2->{state} = "c";
      }
    }

    my $grandTotalCost = $cglStockObject->{cost} + $cglStockObject2->{cost} + $cglAssetObject->{cost} + $cglAssetObject2->{cost};
    my $grandTotalMarket = $cglStockObject->{market} + $cglStockObject2->{market} + $cglAssetObject->{market} + $cglAssetObject2->{market};
    my $grandTotalGainLoss = $cglStockObject->{gainLoss} + $cglStockObject2->{gainLoss} + $cglAssetObject->{gainLoss} + $cglAssetObject2->{gainLoss} + ($getManualEntries ? ($cglManualObject->{gainLoss} + $cglManualObject2->{gainLoss}) : "00.0");

    $grandTotalCost = sprintf("%.2f", $grandTotalCost);
    $grandTotalMarket = sprintf("%.2f", $grandTotalMarket);
    $grandTotalGainLoss = sprintf("%.2f", $grandTotalGainLoss);


    $cglDisplayObj = Portal::Accounting::Objects::CGLDisplayObject->new(cost => $grandTotalCost, market => $grandTotalMarket, gainLoss => $grandTotalGainLoss, cglStocks => [ $cglStockObject, $cglStockObject2 ], cglAssets => [ $cglAssetObject, $cglAssetObject2 ], cglManual => [ $cglManualObject, $cglManualObject2 ], langObj => $self->{langObj});
    if ($cglDisplayObj->didErrorOccur)
    {
      $self->setError(errorString => $cglDisplayObj->errorMessage);
      return $cglDisplayObj;
    }
  }

  return $cglDisplayObj;
}

# buildCGLStocks
# requires: sth - handle of sql statement just executed.
# returns: cglStocks Object
sub buildCGLStocks
{
  my $self = shift;
  my %args = ( sth => undef, @_ );
  my $sth = $args{sth};
  my $cglStocks = undef;

  if (!defined $sth)
  {
    $self->setError(errorString => "sth must be defined!<br>\n");
    return $cglStocks;
  }

  my $result = $sth->fetchall_arrayref;  # get all items from the database.
  my @result = @{$result};  # cast it to an array for easier working with.

  if (scalar @result == 0)
  {
    my @emptyArray = ();
    $cglStocks = Portal::Accounting::Objects::CGLStocks->new(cglStockObjects => \@emptyArray, langObj => $self->{langObj});
    if ($cglStocks->didErrorOccur)
    {
      $self->setError(errorString => $cglStocks->errorMessage);
      return $cglStocks;
    }
  }
  else  # we have data to work with!
  {
    # define the index into the array for each item that we are working with.
    my $sellDateIndex = 0;
    my $amountIndex = 1;
    my $accountIndex = 2;
    my $subAccountIndex = 3;
    my $costIndex = 4;
    my $marketIndex = 5;
    my $purchaseDateIndex = 6;
    my $lotIndex = 7;
    my $sharesIndex = 8;
    my $parentNumIndex = 9;
    my $subAccountNameIndex = 10;

    my $startAccount = 1;
    my $prevAccount = "";
    my @AccountInfo = ();
    my @cglStockObjs = ();

    my $stockName = "";
    my %globalTotals = ( cost => "0.00", market => "0.00", gainLoss => "0.00" );
    my $accountType = "";  # short term or long term flag (S|L)

    my $processed_last_account = 0;  # flag to catch the case where the last account only had 1 lot row returned and wasn't being included in result.

    # now walk over the result set and for all rows that have the same account, process all stock and lot pairs and produce a single CGLStockObject.
    for (my $i=0; $i < scalar @result; $i++)
    {
      if (($startAccount || $result[$i][$accountIndex] == $prevAccount) && ($i < scalar(@result) - 1 || scalar(@result) == 1))
      {
        $startAccount = 0;  # signal we are done just blindly taking the first account we meet.
        $prevAccount = $result[$i][$accountIndex];
        $AccountInfo[++$#AccountInfo] = $result[$i];  # make a copy of this row.
        if ($prevAccount == $shortTermCG || $prevAccount == $shortTermCL)
        {
          $accountType = "S";
        }
        else
        {
          $accountType = "L";
        }
        $processed_last_account = 1 if (scalar(@result) == 1);
      }
      if (($result[$i][$accountIndex] != $prevAccount && $startAccount == 0) || ($i == scalar(@result) - 1))  # now we need to process all entries for the previous account.
      {
        # make sure the last row is being inserted in the case that the account was last and had multiple entries. :(
        if ($i == scalar(@result) - 1 && scalar(@result) > 1 && $result[$i][$accountIndex] == $result[$i-1][$accountIndex])
        {
          $AccountInfo[++$#AccountInfo] = $result[$i];  # make a copy of this row.
        }
        my @lotObjs = ();
        my %stockTotals = ( shares => "0.00", cost => "0.00", market => "0.00", gainLoss => "0.00" );
        my $startStocks = 1;
        my $prevStock = "";
        for (my $j=0; $j < scalar @AccountInfo; $j++)
        {
          # create the CGLStockLotObject
          my $shares = $AccountInfo[$j][$sharesIndex];
          my $cost   = $AccountInfo[$j][$costIndex];

          my $market = $AccountInfo[$j][$marketIndex];
          my $gainLoss = $market - $cost;
          $gainLoss = sprintf("%.2f", $gainLoss);
          $shares =~ s/^(\d+\.\d)(0+)$/$1/;
          $cost = sprintf("%.2f", $cost);

          my $cglStockLotObj = Portal::Accounting::Objects::CGLStockLotObject->new(refNum => $AccountInfo[$j][$parentNumIndex], purchased => $AccountInfo[$j][$purchaseDateIndex], shares => $shares, cost => $cost,
                                market => $market, gainLoss => $gainLoss, sold => $AccountInfo[$j][$sellDateIndex], type => $accountType, langObj => $self->{langObj});
          if ($cglStockLotObj->didErrorOccur)
          {
            $self->setError(errorString => "CGLStockLotObject Error in buildCGLStocks<br>\n" . $cglStockLotObj->errorMessage);
            return $cglStocks;
          }

          # now see what StockObject this goes with.
          if ($AccountInfo[$j][$subAccountIndex] == $prevStock || $startStocks)
          {
            $startStocks = 0;
            $prevStock = $AccountInfo[$j][$subAccountIndex];
            $lotObjs[++$#lotObjs] = $cglStockLotObj;

            $stockName = $AccountInfo[$j][$subAccountNameIndex];

            $stockTotals{shares} += $shares;
            $stockTotals{cost} += $cost;
            $stockTotals{market} += $market;
            $stockTotals{gainLoss} += $gainLoss;

            $globalTotals{cost} += $cost;
            $globalTotals{market} += $market;
            $globalTotals{gainLoss} += $gainLoss;
          }
          if (($AccountInfo[$j][$subAccountIndex] != $prevStock && $startStocks == 0) || ($j == scalar(@AccountInfo) - 1))  # create this stock and start next one.
          {
            # make sure all figures are 2 decimal points.
            foreach my $total (keys %stockTotals)
            {
              if ($total eq "shares")
              {
                $stockTotals{$total} = sprintf("%.6f", $stockTotals{$total});
                $stockTotals{$total} =~ s/^(\d+\.\d)(0+)$/$1/;
              }
              else
              {
                $stockTotals{$total} = sprintf("%.2f", $stockTotals{$total});
              }
            }

            my @tmpLotObjs = @lotObjs;
            my $cglStockObj = Portal::Accounting::Objects::CGLStockObject->new(name => $stockName, number => $prevStock,
                                    shares => $stockTotals{shares}, cost => $stockTotals{cost}, market => $stockTotals{market},
                                    gainLoss => $stockTotals{gainLoss}, state => "o", cglStockLotObjects => \@tmpLotObjs,
                                    type => $accountType . ($stockTotals{gainLoss} =~ /^-/ ? "L" : "G"), langObj => $self->{langObj});
            if ($cglStockObj->didErrorOccur)
            {
              $self->{error} = 1;
              $self->{errorString} =  "CGLStockObject Error in buildCGLStocks<br>\n" .$cglStockObj->errorMessage;
              return $cglStocks;
            }
            $cglStockObjs[++$#cglStockObjs] = $cglStockObj;

            if (($j + 1) < scalar @AccountInfo)
            {
              # reset Lot* variables and Stock* totals
              @lotObjs = ();
              $prevStock = $AccountInfo[$j][$subAccountIndex];
              $startStocks = 0;

              $stockName = $AccountInfo[$j][$subAccountNameIndex];

              $lotObjs[++$#lotObjs] = $cglStockLotObj;

              %stockTotals = ( shares => "0.00", cost => "0.00", market => "0.00", gainLoss => "0.00" );

              $stockTotals{shares} += $shares;
              $stockTotals{cost} += $cost;
              $stockTotals{market} += $market;
              $stockTotals{gainLoss} += $gainLoss;

              $globalTotals{cost} += $cost;
              $globalTotals{market} += $market;
              $globalTotals{gainLoss} += $gainLoss;
            }
            elsif (($AccountInfo[$j][$subAccountIndex] != $prevStock && $startStocks == 0) && ($j == scalar(@AccountInfo) - 1))
            {  # did we just switch stocks and we are the last one in the returned list?
              # we must make the CGLStockObj entry!
              @lotObjs = ();
              $prevStock = $AccountInfo[$j][$subAccountIndex];
              $startStocks = 0;

              $stockName = $AccountInfo[$j][$subAccountNameIndex];

              $lotObjs[++$#lotObjs] = $cglStockLotObj;

              %stockTotals = ( shares => "0.00", cost => "0.00", market => "0.00", gainLoss => "0.00" );

              $stockTotals{shares} += $shares;
              $stockTotals{cost} += $cost;
              $stockTotals{market} += $market;
              $stockTotals{gainLoss} += $gainLoss;

              $globalTotals{cost} += $cost;
              $globalTotals{market} += $market;
              $globalTotals{gainLoss} += $gainLoss;

              # make sure all figures are 2 decimal points.
              foreach my $total (keys %stockTotals)
              {
                if ($total eq "shares")
                {
                  $stockTotals{$total} = sprintf("%.6f", $stockTotals{$total});
                  $stockTotals{$total} =~ s/^(\d+\.\d)(0+)$/$1/;
                }
                else
                {
                  $stockTotals{$total} = sprintf("%.2f", $stockTotals{$total});
                }
              }

              my @tmpLotObjs = @lotObjs;
              my $cglStockObj = Portal::Accounting::Objects::CGLStockObject->new(name => $stockName, number => $prevStock,
                                      shares => $stockTotals{shares}, cost => $stockTotals{cost}, market => $stockTotals{market},
                                      gainLoss => $stockTotals{gainLoss}, state => "o", cglStockLotObjects => \@tmpLotObjs,
                                      type => $accountType . ($stockTotals{gainLoss} =~ /^-/ ? "L" : "G"), langObj => $self->{langObj});
              if ($cglStockObj->didErrorOccur)
              {
                $self->{error} = 1;
                $self->{errorString} =  "CGLStockObject Error in buildCGLStocks<br>\n" .$cglStockObj->errorMessage;
                return $cglStocks;
              }
              $cglStockObjs[++$#cglStockObjs] = $cglStockObj;
            }
          }
        }

        # reset Account* variables.
        @AccountInfo = ();
        $startAccount = 0;
        $AccountInfo[0] = $result[$i];  # make a copy of the current row.
        $prevAccount = $result[$i][$accountIndex];
        if ($prevAccount == $shortTermCG || $prevAccount == $shortTermCL)
        {
          $accountType = "S";
        }
        else
        {
          $accountType = "L";
        }
        if ((($i > 0 && $result[$i-1][$accountIndex] != $prevAccount && $i == scalar(@result) - 1) || ($i == 0 && $i == scalar(@result) - 1)) && !$processed_last_account)
        {
          # signal we have already done this.
          $processed_last_account = 1;
          redo;  # process the last entry.
        }
      }
    }

    # make sure all figures are 2 decimal points.
    foreach my $total (keys %globalTotals)
    {
      $globalTotals{$total} = sprintf("%.2f", $globalTotals{$total});
    }

    # now build the cglStocks Object.
    $cglStocks = Portal::Accounting::Objects::CGLStocks->new(cglStockObjects => \@cglStockObjs, %globalTotals, langObj => $self->{langObj});
    if ($cglStocks->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = "CGLStocks Error in buildCGLStocks<br>\n" . $cglStocks->errorMessage;
      return $cglStocks;
    }
  }

  return $cglStocks;
}

# buildCGLManual
# requires: sth - handle of sql statement just executed.
# returns: cglManual Object
sub buildCGLManual
{
  my $self = shift;
  my %args = ( sth => undef, @_ );
  my $sth = $args{sth};
  my $cglManual = undef;

  if (!defined $sth)
  {
    $self->setError(errorString => "sth must be defined!<br>\n");
    return $cglManual;
  }

  my $result = $sth->fetchall_arrayref;  # get all items from the database.
  my @result = @{$result};  # cast it to an array for easier working with.

  if (scalar @result == 0)
  {
    my @emptyArray = ();
    $cglManual = Portal::Accounting::Objects::CGLManual->new(cglManualObjects => \@emptyArray, langObj => $self->{langObj});
    if ($cglManual->didErrorOccur)
    {
      $self->setError(errorString => $cglManual->errorMessage);
      return $cglManual;
    }
  }
  else  # we have data to work with!
  {
    # define the index into the array for each item that we are working with.
    my $sellDateIndex = 1;
    my $amountIndex = 2;
    my $accountIndex = 0;
    my $subAccountIndex = 3;
    my $parentNumIndex = 5;
    my $subAccountNameIndex = 4;

    my $startAccount = 1;
    my $prevAccount = "";
    my @AccountInfo = ();
    my @cglManualObjs = ();
    my @cglManualEntryObjs = ();

    my $subAccountName = "";
    my %globalTotals = ( gainLoss => "0.00" );
    my $accountType = "";  # short term or long term flag (S|L)

    my $processed_last_account = 0;  # flag to catch the case where the last account only had 1 row returned and wasn't being included in result.

    # now walk over the result set and for all rows that have the same account, process all stock and lot pairs and produce a single CGLManualObject.
    for (my $i=0; $i < scalar @result; $i++)
    {
      if (($startAccount || $result[$i][$accountIndex] == $prevAccount) && $i < scalar(@result) - 1)
      {
        $startAccount = 0;  # signal we are done just blindly taking the first account we meet.
        $prevAccount = $result[$i][$accountIndex];
        $AccountInfo[++$#AccountInfo] = $result[$i];  # make a copy of this row.
        if ($prevAccount == $shortTermCG || $prevAccount == $shortTermCL)
        {
          $accountType = "S";
        }
        else
        {
          $accountType = "L";
        }
      }
      if (($result[$i][$accountIndex] != $prevAccount && $startAccount == 0) || ($i == scalar(@result) - 1))  # now we need to process all entries for the previous account.
      {
        my $startSubAccount = 1;
        my $prevSubAccount = "";
        for (my $j=0; $j < scalar @AccountInfo; $j++)
        {
          # create the CGLManualEntry
          my $gainLoss = $AccountInfo[$j][$amountIndex];
          my $sellDate = $AccountInfo[$j][$sellDateIndex];
          $prevSubAccount = $AccountInfo[$j][$subAccountIndex];
          $subAccountName = $AccountInfo[$j][$subAccountNameIndex];

          my $cglManualEntryObj = Portal::Accounting::Objects::CGLManualEntry->new(refNum => $AccountInfo[$j][$parentNumIndex],
                                name => $subAccountName, number => $prevSubAccount, state => "o",
                                gainLoss => $gainLoss, type => $accountType . ($gainLoss =~ /^-/ ? "L" : "G"),
                                langObj => $self->{langObj}, sellDate => $sellDate);
          if ($cglManualEntryObj->didErrorOccur)
          {
            $self->setError(errorString => "CGLManualEntry Error in buildCGLManual<br>\n" . $cglManualEntryObj->errorMessage);
            return $cglManual;
          }
          $cglManualEntryObjs[++$#cglManualEntryObjs] = $cglManualEntryObj;

          $globalTotals{gainLoss} += $gainLoss;
        }

        # reset Account* variables.
        @AccountInfo = ();
        $startAccount = 0;
        $AccountInfo[0] = $result[$i];  # make a copy of the current row.
        $prevAccount = $result[$i][$accountIndex];
        if ($prevAccount == $shortTermCG || $prevAccount == $shortTermCL)
        {
          $accountType = "S";
        }
        else
        {
          $accountType = "L";
        }
        if ((($i > 0 && $result[$i-1][$accountIndex] != $prevAccount && $i == scalar(@result) - 1) || ($i == 0 && $i == scalar(@result) - 1)) && !$processed_last_account)
        {
          # signal we have already done this.
          $processed_last_account = 1;
          redo;  # process the last entry.
        }
      }
    }

    # make sure all figures are 2 decimal points.
    foreach my $total (keys %globalTotals)
    {
      $globalTotals{$total} = sprintf("%.2f", $globalTotals{$total});
    }

    # now build the cglManual Object.
    $cglManual = Portal::Accounting::Objects::CGLManual->new(cglManualObjects => \@cglManualEntryObjs, %globalTotals, langObj => $self->{langObj});
    if ($cglManual->didErrorOccur)
    {
      $self->setError(errorString => "CGLManual Error in buildCGLManual<br>\n" . $cglManual->errorMessage);
      return $cglManual;
    }
  }

  return $cglManual;
}

#
# requires: sth - handle of sql statement just executed.
# returns: cglAssets Object
sub buildCGLAssets
{
  my $self = shift;
  my %args = ( sth => undef, @_ );
  my $sth = $args{sth};
  my $cglAssets = undef;

  if (!defined $sth)
  {
    $self->setError(errorString => "sth must be defined!<br>\n");
    return $cglAssets;
  }

  #my $result = $sth->fetchall_arrayref;  # get all items from the database.
  #my @result = @{$result};  # cast it to an array for easier working with.
  my @result = ();  # force it to always be empty for now.

  if (scalar @result == 0)
  {
    my @emptyArray = ();
    $cglAssets = Portal::Accounting::Objects::CGLAssets->new(cglAssetObjects => \@emptyArray, langObj => $self->{langObj});
    if ($cglAssets->didErrorOccur)
    {
      $self->setError(errorString => $cglAssets->errorMessage);
      return $cglAssets;
    }
  }
  else  # we have data to work with!
  {
    # This still needs to be fleshed out, but we are waiting for the Asset System to be developed.
    my @emptyArray = ();
    $cglAssets = Portal::Accounting::Objects::CGLAssets->new(cglAssetObjects => \@emptyArray, langObj => $self->{langObj});
    if ($cglAssets->didErrorOccur)
    {
      $self->setError(errorString => $cglAssets->errorMessage);
      return $cglAssets;
    }

    return $cglAssets;

    # define the index into the array for each item that we are working with.
    my $sellDateIndex = 0;
    my $amountIndex = 1;
    my $accountIndex = 2;
    my $subAccountIndex = 3;
    my $costIndex = 4;
    my $marketIndex = 5;
    my $purchaseDateIndex = 6;
    my $parentNumIndex = 7;
    my $subAccountNameIndex = 8;

    my $startAccount = 1;
    my $prevAccount = "";
    my @AccountInfo = ();
    my @cglAssetObjs = ();

    my $assetName = "";
    my %globalTotals = ( cost => "0.00", market => "0.00", gainLoss => "0.00" );
    my $accountType = "";  # short term or long term flag (S|L)

    my $processed_last_account = 0;  # flag to catch the case where the last account only had 1 asset row returned and wasn't being included in result.

    # now walk over the result set and for all rows that have the same account, process all assets and produce a single CGLAssetObject.
    for (my $i=0; $i < scalar @result; $i++)
    {
      if (($startAccount || $result[$i][$accountIndex] == $prevAccount) && $i < scalar(@result) - 1)
      {
        $startAccount = 0;  # signal we are done just blindly taking the first account we meet.
        $prevAccount = $result[$i][$accountIndex];
        $AccountInfo[++$#AccountInfo] = $result[$i];  # make a copy of this row.
        if ($prevAccount == $shortTermCG || $prevAccount == $shortTermCL)
        {
          $accountType = "S";
        }
        else
        {
          $accountType = "L";
        }
      }
      if (($result[$i][$accountIndex] != $prevAccount && $startAccount == 0) || ($i == scalar(@result) - 1))  # now we need to process all entries for the previous account.
      {
        # make sure the last row is being inserted in the case that the account was last and had multiple entries. :(
        if ($i == scalar(@result) - 1 && $result[$i][$accountIndex] == $result[$i-1][$accountIndex])
        {
          $AccountInfo[++$#AccountInfo] = $result[$i];  # make a copy of this row.
        }
        my %assetTotals = ( shares => "0.00", cost => "0.00", market => "0.00", gainLoss => "0.00" );
        my $startAssets = 1;
        my $prevAsset = "";
        for (my $j=0; $j < scalar @AccountInfo; $j++)
        {
          # create the CGLAssetObject
          my $cost   = $AccountInfo[$j][$costIndex];

          my $market = $AccountInfo[$j][$marketIndex];
          my $gainLoss = $market - $cost;
          $gainLoss = sprintf("%.2f", $gainLoss);
          $cost = sprintf("%.2f", $cost);

          $assetName = $AccountInfo[$j][$subAccountNameIndex];

          my $cglAssetObj = Portal::Accounting::Objects::CGLAssetObject->new(refNum => $AccountInfo[$j][$parentNumIndex], purchased => $AccountInfo[$j][$purchaseDateIndex], cost => $cost,
                                market => $market, gainLoss => $gainLoss, sold => $AccountInfo[$j][$sellDateIndex], type => $accountType,
                                name => $assetName, number => $AccountInfo[$j][$subAccountIndex], langObj => $self->{langObj});
          if ($cglAssetObj->didErrorOccur)
          {
            $self->setError(errorString => "CGLAssetObject Error in buildCGLAssets<br>\n" . $cglAssetObj->errorMessage);
            return $cglAssets;
          }
          $cglAssetObjs[++$#cglAssetObjs] = $cglAssetObj;

          $globalTotals{cost} += $cost;
          $globalTotals{market} += $market;
          $globalTotals{gainLoss} += $gainLoss;
        }

        # reset Account* variables.
        @AccountInfo = ();
        $startAccount = 0;
        $AccountInfo[0] = $result[$i];  # make a copy of the current row.
        $prevAccount = $result[$i][$accountIndex];
        if ($prevAccount == $shortTermCG || $prevAccount == $shortTermCL)
        {
          $accountType = "S";
        }
        else
        {
          $accountType = "L";
        }
        if ((($i > 0 && $result[$i-1][$accountIndex] != $prevAccount && $i == scalar(@result) - 1) || ($i == 0 && $i == scalar(@result) - 1)) && !$processed_last_account)
        {
          # signal we have already done this.
          $processed_last_account = 1;
          redo;  # process the last entry.
        }
      }
    }

    # make sure all figures are 2 decimal points.
    foreach my $total (keys %globalTotals)
    {
      $globalTotals{$total} = sprintf("%.2f", $globalTotals{$total});
    }

    # now build the cglAssets Object.
    $cglAssets = Portal::Accounting::Objects::CGLAssets->new(cglAssetObjects => \@cglAssetObjs, %globalTotals, langObj => $self->{langObj});
    if ($cglAssets->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = "CGLAssets Error in buildCGLAssets<br>\n" . $cglAssets->errorMessage;
      return $cglAssets;
    }
  }

  return $cglAssets;
}

# getSecurityEventsForDisplay
# requires: startIndex, numEntriesToDisplay, ledgerCode, showEvents
# optional: accounts, stocks, actions, startDate, endDate
# returns: InvestmentJournalViewObject or undef if error
# summary: Builds up the SQL statement to gather SecurityEvents from the database based upon the possible accounts, stocks, actions, and dates to
# search across.  accounts, stocks and actions are comma (,) separated strings of the numbers that represent the various items.  If nothing is specified other
# than the ledgerCode, startIndex and numEntriesToDisplay, then all security event entries will be searched in the specified ledger.  The number of rows
# that would have been returned before limiting will be calculated and then the actual data will be gathered and stored in the InvestmentJournalViewObject.
sub getSecurityEventsForDisplay
{
  my $self = shift;
  my %args = ( startIndex => 1, numEntriesToDisplay => 20, ledgerCode => "", accounts => "", stocks => "", actions => "", startDate => "", endDate => "",
               showEvents => "global", @_ );
  my $investmentJournalViewObj = undef;

  my $startIndex = $args{startIndex};
  my $numEntriesToDisplay = $args{numEntriesToDisplay};
  my $ledgerCode = $args{ledgerCode};
  my $accounts = $args{accounts};
  my $stocks = $args{stocks};
  my $actions = $args{actions};
  my $startDate = $args{startDate};
  my $endDate = $args{endDate};
  my $showEvents = $args{showEvents};

  # validate input.
  if ($startIndex !~ /^\d+$/ || $startIndex < 1)
  {
    $self->{error} = 1;
    $self->{errorString} = "startIndex = '$startIndex' is invalid!<br>\n";
    return $investmentJournalViewObj;
  }
  if ($numEntriesToDisplay !~ /^\d+$/ || $numEntriesToDisplay < 5)
  {
    $self->{error} = 1;
    $self->{errorString} = "numEntriesToDisplay = '$numEntriesToDisplay' is invalid!<br>\n";
    return $investmentJournalViewObj;
  }
  if (length $ledgerCode == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "ledgerCode = '$ledgerCode' is invalid!<br>\n";
    return $investmentJournalViewObj;
  }
  if (length $accounts > 0 && $accounts !~ /^\d+(,\d+)*$/)
  {
    $self->{error} = 1;
    $self->{errorString} = "accounts = '$accounts' is invalid!<br>\n";
    return $investmentJournalViewObj;
  }
  if (length $stocks > 0 && $stocks !~ /^\d+(,\d+)*$/)
  {
    $self->{error} = 1;
    $self->{errorString} = "stocks = '$stocks' is invalid!<br>\n";
    return $investmentJournalViewObj;
  }
  if (length $actions > 0 && $actions !~ /^(\d+|memo)(,(\d+|memo))*$/)
  {
    $self->{error} = 1;
    $self->{errorString} = "actions = '$actions' is invalid!<br>\n";
    return $investmentJournalViewObj;
  }
  if (length $startDate > 0)
  {
    my @dateResult = $self->{methods}->fixupDate(date => $startDate);
    if (!$dateResult[0])
    {
      $self->setError(errorString => "startDate = '$startDate' is invalid!<br>$dateResult[2]\n");
      return $investmentJournalViewObj;
    }
    else
    {
      # make sure the date is "fixed".
      $startDate = $dateResult[1];
    }
  }
  if (length $endDate > 0)
  {
    my @dateResult = $self->{methods}->fixupDate(date => $endDate);
    if (!$dateResult[0])
    {
      $self->setError(errorString => "endDate = '$endDate' is invalid!<br>$dateResult[2]\n");
      return $investmentJournalViewObj;
    }
    else
    {
      # make sure the date is "fixed".
      $endDate = $dateResult[1];
    }
  }
  if ($showEvents !~ /^(global|global\+local|local)$/)
  {
    $self->{error} = 1;
    $self->{errorString} = "showEvents = '$showEvents' is invalid!<br>\n";
    return $investmentJournalViewObj;
  }

  # fixup the accounts, actions and stocks to be strings not numbers for SQL.
  $accounts =~ s/(\d+)/'$1'/g;
  $stocks =~ s/(\d+)/'$1'/g;
  $actions =~ s/(\d+|memo)/'$1'/g;

  # now build up the sql statement to gather the count of all security_event_tb entries that would possibly be returned.
  my $sql = "SELECT COUNT(DISTINCT(event_ref)) FROM sec_event_tb WHERE (";
  $sql .= "ledger_code = '$ledgerCode'" if ($showEvents eq "local" || $showEvents eq "global+local");
  $sql .= "ledger_code = '*'" if ($showEvents eq "global");
  $sql .= " AND (calc_date BETWEEN '$startDate' AND '$endDate')" if (length $startDate > 0 && length $endDate > 0);
  $sql .= " AND calc_date = '$startDate'" if (length $startDate > 0 && length $endDate == 0);
  $sql .= " AND a IN ($accounts)" if (length $accounts > 0);
  $sql .= " AND s IN ($stocks)" if (length $stocks > 0);
  $sql .= " AND event_code IN ($actions)" if (length $actions > 0);
  $sql .= ")";
  if ($showEvents eq "global+local")
  {
    $sql .= " OR (ledger_code = '*'";
    $sql .= " AND (calc_date BETWEEN '$startDate' AND '$endDate')" if (length $startDate > 0 && length $endDate > 0);
    $sql .= " AND calc_date = '$startDate'" if (length $startDate > 0 && length $endDate == 0);
    $sql .= " AND s IN ($stocks)" if (length $stocks > 0);
    $sql .= " AND event_code IN ($actions)" if (length $actions > 0);
    $sql .= ")";
  }

  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $self->{dbHandle}->errorMessage;
    return $investmentJournalViewObj;
  }
  my @info = $sth->fetchrow_array;
  if (! defined $info[0])
  {
    $self->{error} = 1;
    $self->{errorString} = "No COUNT returned!<br>\n";
    return $investmentJournalViewObj;
  }
  my $totalEntries = $info[0];
  my @securityEvents = ();
  if ($totalEntries == 0)
  {
    # exit early as there is nothing to gather.
    $investmentJournalViewObj = Portal::Accounting::Objects::InvestmentJournalViewObject->new(startIndex => $startIndex, numEntriesToDisplay => $numEntriesToDisplay,
                                totalEntries => $totalEntries, securityEvents => \@securityEvents, langObj => $self->{langObj});
    if ($investmentJournalViewObj->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = $investmentJournalViewObj->errorMessage;
      return undef;
    }
    return $investmentJournalViewObj;
  }

  # now gather all event_refs of the rows that are supposed to be returned.
  my $sql = "SELECT DISTINCT(event_ref) FROM sec_event_tb WHERE (";
  $sql .= "ledger_code = '$ledgerCode'" if ($showEvents eq "local" || $showEvents eq "global+local");
  $sql .= "ledger_code = '*'" if ($showEvents eq "global");
  $sql .= " AND (calc_date BETWEEN '$startDate' AND '$endDate')" if (length $startDate > 0 && length $endDate > 0);
  $sql .= " AND calc_date = '$startDate'" if (length $startDate > 0 && length $endDate == 0);
  $sql .= " AND a IN ($accounts)" if (length $accounts > 0);
  $sql .= " AND s IN ($stocks)" if (length $stocks > 0);
  $sql .= " AND event_code IN ($actions)" if (length $actions > 0);
  $sql .= ")";
  if ($showEvents eq "global+local")
  {
    $sql .= " OR (ledger_code = '*'";
    $sql .= " AND (calc_date BETWEEN '$startDate' AND '$endDate')" if (length $startDate > 0 && length $endDate > 0);
    $sql .= " AND calc_date = '$startDate'" if (length $startDate > 0 && length $endDate == 0);
    $sql .= " AND s IN ($stocks)" if (length $stocks > 0);
    $sql .= " AND event_code IN ($actions)" if (length $actions > 0);
    $sql .= ")";
  }
  $sql .= " ORDER BY event_ref LIMIT $numEntriesToDisplay OFFSET " . ($startIndex - 1);
  
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $self->{dbHandle}->errorMessage;
    return $investmentJournalViewObj;
  }
  my $eventRefs = "";
  while (my @info = $sth->fetchrow_array)
  {
    $eventRefs .= "," if (length $eventRefs > 0);
    $eventRefs .= $info[0];
  }
  if (length $eventRefs == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "No Event Refs returned but the database claimed '$totalEntries' entries existed!<br>\n";
    return $investmentJournalViewObj;
  }
  
  # get the Security Event Objects to display (possibly only 1 so we will have to build the array in that case)
  my $securityEvents = $self->getSecurityEvent(eventRefNum => $eventRefs);
  if ($self->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $self->errorMessage;
    return $investmentJournalViewObj;
  }
  if ($eventRefs =~ /,/)
  {
    @securityEvents = @{$securityEvents};  # cast it as an array
  }
  else
  {
    $securityEvents[0] = $securityEvents;  # put it in the array.
  }
  
  # now build the InvestmentJournalViewObject
  $investmentJournalViewObj = Portal::Accounting::Objects::InvestmentJournalViewObject->new(startIndex => $startIndex, numEntriesToDisplay => $numEntriesToDisplay,
                              totalEntries => $totalEntries, securityEvents => \@securityEvents, langObj => $self->{langObj});
  if ($investmentJournalViewObj->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $investmentJournalViewObj->errorMessage;
    return undef;
  }
  
  return $investmentJournalViewObj;
}

=item array[] getWatchLists(ledgerCode =>)

  This routine selects all of the watchlists from the stock_watch_tb.

=cut
sub getWatchLists
{
  my $self = shift;
  my %args = (ledgerCode => "", @_);
  my $ledger = $args{ledgerCode};
  if ($ledger eq "")
  {
     $self->setError("Ledger Code was not defined!<br>\n");
     return undef;
  }
  my @lists = ();

  my $sth = $self->{dbHandle}->read(sql => "SELECT DISTINCT(group_name) FROM stock_watch_tb ORDER BY group_name");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->setError("Error retrieving watch lists!\n" . $self->{dbHandle}->errorMessage);
    return undef;
  }
  my $index = 0;
  while ((my @info = $sth->fetchrow_array))
  {
    $lists[$index++] = $info[0];
  }
  return @lists;
}

=back

=cut

# getListOfWatchedStocks
# required args: ledgerCode
# optional: orderBy, groupName
# returns: Array of Stock SubAccounts, or tickers, and/or portfolio information depending
# on the arguments given
# This routine selects all stocks from the stock_watch_tb with a group = 'Default' or NULL unless byGroupName == 1
# and groupName is specified as something different.  If byGroupName == 2, all stocks for all groups are selected
# and returned ordered and grouped by groupName
# The orderBy string must be a valid ORDER BY possible statement for this query.
sub getListOfWatchedStocks
{
  my $self = shift;
  my %args = ( orderBy => "sub_acct", ledgerCode => "", ticker => "no", byGroupName => "0", groupName => "Default", @_ );
  my $orderBy = $args{orderBy};
  my $ledgerCode = $args{ledgerCode};
  my $ticker = $args{ticker};
  my $byGroupName = $args{byGroupName};
  my $groupName = $args{groupName};
  my @stocks = ();
  my %stocks = ();
  my $sth;
  my @info;

  # validate input
  if (length $orderBy == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "orderBy = '$orderBy' is invalid!<br>\n";
    return @stocks;
  }
  if (length $ledgerCode == 0)
  {
    $self->setError(errorString => "ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return @stocks;
  }

  my $sql = "";
  if ($ticker eq "yes")
  {
    if ($byGroupName eq "0" || ($byGroupName ne "1" && $byGroupName ne "2"))
    {
      $sql = "SELECT distinct ticker, name FROM stocks_tb, stock_watch_tb WHERE ledger_code = '$ledgerCode' AND stocks_tb.sub_acct = stock_watch_tb.sub_acct AND (group_name IS NULL OR group_name = 'Default') ORDER BY $orderBy ASC";
    }
    elsif ($byGroupName eq "1")
    {
      $sql = "SELECT ticker, name, stock_watch_tb.sub_acct FROM stocks_tb, stock_watch_tb WHERE ledger_code = '$ledgerCode' AND stocks_tb.sub_acct = stock_watch_tb.sub_acct AND group_name = '$groupName' ORDER BY $orderBy ASC";
    }
    elsif ($byGroupName eq "2")
    {
      $sql = "SELECT stocks_tb.ticker, stocks_tb.name, group_name, SUM(shares), SUM(cost), SUM(cost)/SUM(shares) FROM lot_node_tb AS ltb, stocks_tb, stock_watch_tb WHERE ltb.ledger_code = '$ledgerCode' AND stock_watch_tb.ledger_code = ltb.ledger_code AND ltb.sub_account_num = stock_watch_tb.sub_acct AND stocks_tb.sub_acct = ltb.sub_account_num AND ltb.node_id = (SELECT MAX(node_id) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND lot = ltb.lot AND event_date <= DATE('now') AND sub_account_num = stock_watch_tb.sub_acct) AND (SELECT shares FROM lot_node_tb WHERE node_id = ltb.node_id) <> 0 GROUP BY group_name, name, ticker UNION SELECT stocks_tb.ticker, stocks_tb.name, group_name, 0, 0, 0 FROM stocks_tb, stock_watch_tb WHERE stock_watch_tb.ledger_code = '$ledgerCode' AND stock_watch_tb.sub_acct = stocks_tb.sub_acct AND ((SELECT COUNT(lot) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND sub_account_num = stock_watch_tb.sub_acct) = 0 OR (SELECT shares FROM lot_node_tb WHERE node_id = (SELECT MAX(node_id) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND sub_account_num = stock_watch_tb.sub_acct)) = 0) GROUP BY group_name, name, ticker ORDER BY group_name ASC, name ASC, ticker ASC";
    }
  }
  else
  {
    if ($byGroupName eq "0" || ($byGroupName ne "1" && $byGroupName ne "2"))
    {
      $sql = "SELECT sub_acct FROM stock_watch_tb WHERE ledger_code = '$ledgerCode' AND (group_name IS NULL OR group_name = 'Default') ORDER BY $orderBy";
    }
    elsif ($byGroupName eq "1")
    {
      $sql = "SELECT sub_acct, ticker, name FROM stocks_tb, stock_watch_tb WHERE ledger_code = '$ledgerCode' AND stocks_tb.sub_acct = stock_watch_tb.sub_acct AND group_name = '$groupName' ORDER BY $orderBy ASC";
    }
    elsif ($byGroupName eq "2")
    {
      $sql = "SELECT sub_account_num, stocks_tb.name, group_name, SUM(shares), SUM(cost), SUM(cost)/SUM(shares) FROM lot_node_tb AS ltb, stock_watch_tb, stocks_tb WHERE ltb.ledger_code = '$ledgerCode' AND stock_watch_tb.ledger_code = ltb.ledger_code AND ltb.sub_account_num = stock_watch_tb.sub_acct AND stocks_tb.sub_acct = ltb.sub_account_num AND ltb.node_id = (SELECT MAX(node_id) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND lot = ltb.lot AND event_date <= DATE('now') AND sub_account_num = stock_watch_tb.sub_acct) AND (SELECT shares FROM lot_node_tb WHERE node_id = ltb.node_id) <> 0 GROUP BY group_name, name, sub_account_num UNION SELECT stocks_tb.sub_acct, stocks_tb.name, group_name, 0, 0, 0 FROM stocks_tb, stock_watch_tb WHERE stock_watch_tb.ledger_code = '$ledgerCode' AND stock_watch_tb.sub_acct = stocks_tb.sub_acct AND ((SELECT COUNT(lot) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND sub_account_num = stock_watch_tb.sub_acct) = 0  OR (SELECT shares FROM lot_node_tb WHERE node_id = (SELECT MAX(node_id) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND sub_account_num = stock_watch_tb.sub_acct)) = 0) GROUP BY group_name, name, sub_account_num ORDER BY group_name ASC, name ASC, sub_account_num ASC";
    }
  }

  # now execute the sql statement
  $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $self->{dbHandle}->errorMessage;
    return @stocks;
  }
  if ($ticker ne "yes" || ($ticker eq "yes" && ($byGroupName eq "0" || ($byGroupName ne "1" && $byGroupName ne "2"))))
  {
    while (@info = $sth->fetchrow_array)
    {
      $stocks[++$#stocks] = $info[0];
    }

    return @stocks;
  }
  elsif ($byGroupName eq "1")
  {
    while (@info = $sth->fetchrow_array)
    {
      $stocks[++$#stocks] = [$info[0], $info[1], $info[2]];
    }

    return @stocks;
  }
  else
  {
    $stocks{dataRows} = [];
    $stocks{stocks} = {};
    while (@info = $sth->fetchrow_array)
    {
      $stocks{dataRows}[scalar @{$stocks{dataRows}}] = [ $info[2], $info[0] ];  # store the ordered row information as a two-element array with the groupname and ticker symbol
      if (not defined $stocks{stocks}{$info[0]})
      {
        $stocks{stocks}{$info[0]} = {};
      }
      $stocks{stocks}{$info[0]}{name} = $info[1];
      $stocks{stocks}{$info[0]}{totalShares} = $info[3];
      $stocks{stocks}{$info[0]}{totalCost} = $info[4];
      $stocks{stocks}{$info[0]}{avgCostPerShare} = $info[5];
    }
    return %stocks;
  }
}

# watchStock
# requires: number, ledgerCode
# optional: groupName (if not specified, defaults to 'Default')
# returns: 1=OK, 0=Error, -1=Stock Sub Account Already exists
# summary: creates the watch this stock entry in the database.
sub watchStock
{
  my $self = shift;
  my %args = ( number => "", ledgerCode => "", groupName => "Default", @_ );
  my $number = $args{number};
  my $ledgerCode = $args{ledgerCode};
  my $groupName = $args{groupName};
  my $sth;
  my @info;

  if ($number !~ /^\d+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = "Stock Sub Account = '$number' is invalid!<br>\n";
    return (0);
  }

  if (length $ledgerCode == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "ledgerCode = '$ledgerCode' is invalid!<br>\n";
    return (0);
  }

  if (length $groupName == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "groupName = '$groupName' is invalid!<br>\n";
    return (0);
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT ledger_code FROM stock_watch_tb WHERE sub_acct = ? AND ledger_code = ? AND group_name = ?",
                                    plug => [  $number, $ledgerCode, $groupName ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $self->{dbHandle}->errorMessage;
    return (0);
  }
  @info = $sth->fetchrow_array;
  if (defined $info[0])
  {
    return (-1, $info[0]);  # The stock sub account is already being watched!
  }

  my $rc = $self->{dbHandle}->write(sql => "INSERT INTO stock_watch_tb (sub_acct, ledger_code, group_name) VALUES (?, ?, ?)",
                                    plug => [ $number, $ledgerCode, $groupName ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $self->{dbHandle}->errorMessage;
    return (0);
  }
  $self->{dbHandle}->commit;

  return (1);
}

# unWatchStock
# requires: number, ledgerCode
# optional: groupName (if not specified, defaults to 'Default')
# returns: 1=OK, 0=Error, -1=Stock Sub Account No Longer exists
# summary: deletes the watch this stock entry in the database.
sub unWatchStock
{
  my $self = shift;
  my %args = ( number => "", ledgerCode => "", groupName => "Default", @_ );
  my $number = $args{number};
  my $ledgerCode = $args{ledgerCode};
  my $groupName = $args{groupName};
  my $sth;
  my @info;

  if ($number !~ /^\d+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = "Stock Sub Account = '$number' is invalid!<br>\n";
    return (0);
  }

  if (length $ledgerCode == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "ledgerCode = '$ledgerCode' is invalid!<br>\n";
    return (0);
  }

  if (length $groupName == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "groupName = '$groupName' is invalid!<br>\n";
    return (0);
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT ledger_code FROM stock_watch_tb WHERE sub_acct = ? AND ledger_code = ? AND group_name = ?",
                                    plug => [  $number, $ledgerCode, $groupName ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $self->{dbHandle}->errorMessage;
    return (0);
  }
  @info = $sth->fetchrow_array;
  if (!defined $info[0])
  {
    return (-1, $info[0]);  # The stock sub account is not being watched!
  }

  my $rc = $self->{dbHandle}->write(sql => "DELETE FROM stock_watch_tb WHERE sub_acct = ? AND ledger_code = ? AND group_name = ?",
                                    plug => [ $number, $ledgerCode, $groupName ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $self->{dbHandle}->errorMessage;
    return (0);
  }
  $self->{dbHandle}->commit;

  return (1);
}

# unWatchAllStocks
# requires: ledgerCode
# optional: groupName (if not specified, defaults to 'Default')
# returns: 1=OK, 0=Error
# summary: deletes all the watch this stock entry in the database for the specified ledger.
sub unWatchAllStocks
{
  my $self = shift;
  my %args = ( ledgerCode => "", groupName => "Default", @_ );
  my $ledgerCode = $args{ledgerCode};
  my $groupName = $args{groupName};

  if (length $ledgerCode == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "ledgerCode = '$ledgerCode' is invalid!<br>\n";
    return (0);
  }

  if (length $groupName == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "groupName = '$groupName' is invalid!<br>\n";
    return (0);
  }

  my $rc = $self->{dbHandle}->write(sql => "DELETE FROM stock_watch_tb WHERE ledger_code = ? AND group_name = ?",
                                    plug => [ $ledgerCode, $groupName ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $self->{dbHandle}->errorMessage;
    return (0);
  }
  $self->{dbHandle}->commit;

  return (1);
}

1;
__END__

=head1 Exported FUNCTIONS (non-Inline POD)

  scalar deleteSecurityEvent(eventRefNum)
    Retrieves the security event object from the database for the given
    event reference number.  Determines the type of event it represents
    and deletes any security_event_tb, journal_entry_tb, and
    debitcredit_tb table entries necessary.  It then recalculates the
    lot balances for all remaining events in the sec_event_tb table.

  scalar insertBuy(secobj [, subEntity][, counter])
    Inserts a buy into the sec_event_tb table and the amount of
    the purchase into the journal_entry_tb and debitcredit_tb tables
    and propagates the purchase through all existing future events.

  scalar insertSell(secobj [, subEntity][, counter])
    Inserts a sell into the sec_event_tb table and the amount of the
    sell along with any capital gain/loss into the journal_entry_tb
    and debitcredit_tb tables and propagates the sell through all
    existing future events.

  scalar insertStockSplit(secobj [, counter])
    Inserts a split into the sec_event_tb table and propagates the
    split through all future events to generate new lots.

  scalar insertStockDividend(secobj [, counter])
    Inserts a dividend into the sec_event_tb table and propagates the
    split through all future events to generate new lots.

  scalar insertStockSwap(secobj [, counter])
    Inserts a stock swap between to companies/entities into the
    sec_event_tb table and propagates the event into the future to
    generate new lots for both companies/entities.  Entries are
    made in the journal_entry_tb and debitcredit_tb tables in order
    to transfer cost.

  scalar insertAcquisition(secobj [, counter])
    Inserts an acquisition into the sec_event_tb table.  Entries are
    made in the journal_entry_tb and debitcredit_tb tables to
    transfer the entire cost from the old lots to the new lots.

  scalar insertSpinOff(secobj [, counter])
    Inserts a spinoff into the sec_event_tb table and makes entries
    in the journal_entry_tb and debitcredit_tb tables to transfer
    some cost from the original lots to the spun-off lots.

  scalar insertStockTransfer(secobj [, subEntity][, counter])
    Inserts a stock transfer into the sec_event_tb table and makes
    entries in the journal_entry_tb and debitcredit_tb tables to
    transfer cost and lots from one account to another.

  scalar insertMemo(secobj)
    Inserts a memo into the sec_event_tb table.

  scalar insertFee(secobj [, subEntity][, counter])
    Inserts the fee into the sec_event_tb and journal_entry_tb and
    debitcredit_tb tables.

  scalar insertReturnOfCapital(secobj [, subEntity][, counter])
    Inserts the return of capital event into the sec_event_tb and
    makes entries in the journal_entry_tb and debitcredit_tb tables.

  LotNodeObject getLotNode(nodeId)
    Returns the LotNodeObject or undef that represents the entry in
    the lot_node_tb with node_id = nodeId.

  %LotNodeObjects getLotTree(nodeId)
    This routine, starts with the Lot Node = nodeId and gathers all
    nodes that are children of it.  It creates their objects and
    stores them in the hash which is returned.  The nodeId for each
    node is the key in the hash.

  %LotNodeObjects getChildLotNodes(parentId)
    returns: hash of LotNodeObjects that represent all child nodes of
    the specified parent.  This is a potentially recursive function.

  @stockInfo getCurrentStockPrice(stock, date, year)
    required: stock, date, year
    optional:
    returns: array [ticker, price, date, stock]
    This routine gathers the current price for the specified stock
    (subaccount) along with the current ticker it is under.

    The year specifies the year to use if the date doesn't specify it.

  PortfolioObject getPortfolioData(accounts, date, stocks, year)
    required: accounts, date, year
    optional: stocks
    returns: PortfolioObject or undef if error.
    This routine takes the array of accounts, stocks and gathers all
    info for a Portfolio view upto the specified date.

    The year specifies the year to use if the date doesn't specify it.

  InvestmentJournalViewObject getSecurityEventsForDisplay(startIndex,
      numEntriesToDisplay, ledgerCode, showEvents, accounts, stocks,
      actions, startDate, endDate)
    requires: startIndex, numEntriesToDisplay, ledgerCode, showEvents
    optional: accounts, stocks, actions, startDate, endDate
    returns: InvestmentJournalViewObject or undef if error
    summary: Builds up the SQL statement to gather SecurityEvents from
    the database based upon the possible accounts, stocks, actions, and
    dates to search across.  accounts, stocks and actions are comma (,)
    separated strings of the numbers that represent the various items.
    If nothing is specified other than the ledgerCode, startIndex and
    numEntriesToDisplay, then all security event entries will be searched
    in the specified ledger.  The number of rows that would have been
    returned before limiting will be calculated and then the actual data
    will be gathered and stored in the InvestmentJournalViewObject.

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

PC & Web Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal::Accounting::Objects::SecurityEventObject(3), Portal::Accounting::DB::Transaction(3), Portal::Accounting::Objects::TransactionObject(3), Portal::Accounting::Objects::TransactionRowObject(3), Portal::Base(3)

=cut
